

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_FEATURESCHANNELCORNERS_H
#define ADORE_FEATURESCHANNELCORNERS_H

#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/edgesat.h>
#include <adrcore/imgprocessing/harrisat.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/util/calctype.h>

namespace adore
{

	namespace core
	{

		/**
		 * finds corners in an image by applying a local hough transformation. The strength are measured with a harris detector.
		 * @tparam T the image type
		 * @tparam gradT the type of the gradient images
		 * @tparam functionT the type of the callback functor that is called for each detected feature
		 *  
		*/
		template <class T, class gradT, class functionT>
		class ChannelCorners
		{
			private:
				typedef typename signedT< typename calcT<gradT>::type >::type harrisT;
			protected:
				T minEdgeStrength;	///the local hough threshold for edge strength
				float harrisThresh;	///the threshold an harris strength
				float harrisVals[8];	///harris values in a neighborhood for sub-pixel localization
				uint hood;		///a patch of size 2*hood+1 is examind by the local hough trafo
				int numAngleBins;	///the number of angles used in the angle histogram of the local hough
				int needed2nd;		///the minimum strenght of the second strongest edge in the local hough

				Channel<gradT> *gxx;
				Channel<gradT> *gyy;
				Channel<gradT> *gxy;
			public:
				/**
				 * Constructor
				 * @param strength a pixel is an edge pixel if its gradient value exceeds this value
				 * @param hStrenth the minimum result of the harris operator needed to be a valid corner candidate
				 * @param numEdges the needed number of edges that meet in a pixel to make that pixel a corner candidate
				 * @param strengthOfSecondStrongestEdge the minimum strength of the second strongest Edge
				 * @param _hood the half of the neighborhood that is examined (the region width is 2*hood+1)
				 * @param _needed the number of votes needed for an edge. This depends on the hood, of course
				 * @param _numAngleBins the number of angle bins used in the local hough trafo
				 */
				ChannelCorners( T strength = 0, harrisT hStrength = 0,
				                int strengthOfSecondStrongestEdge = 14,
				                uint _hood = 6, int _numAngleBins = 12 ):
						minEdgeStrength( strength ),
						harrisThresh( hStrength ),
						hood( _hood ),
						numAngleBins( _numAngleBins ),
						needed2nd( strengthOfSecondStrongestEdge )
				{
					;
				}

				/**
				 * Destructor
				 */
				~ChannelCorners()
				{
					;
				}

				/**
				 * finds the corner features and calls the callback functor
				 * @param img the image to find the corners in
				 * @param func the callback functor
				 */
				void operator()( const Channel<T> &img, functionT &func );

				/**
				 * @param theValue the minimum edge strength for a pixel to be regarded by the local hough
				 */
				void setMinEdgeStrength( const T& theValue )
				{
					minEdgeStrength = theValue;
				}

				/**
				 * @return the minimum edge strength for a pixel to be regarded by the local hough
				 */
				T getMinEdgeStrength() const
				{
					return minEdgeStrength;
				}

				/**
				 * @param harris the minimum required harris strength for a detected corner pixel
				 */
				void setHarrisThresh( const float harris )
				{
					harrisThresh = harris;
				}

				/**
				 * @return the minimum required harris strength for a detected corner pixel
				 */
				float gerHarrisThresh() const
				{
					return harrisThresh;
				}

				/**
				 * set the gradient images used by the harris detector
				 * @param _gxx channel containing gx*gx
				 * @param _gyy channel containing gy*gy
				 * @param _gxy channel containing gx*gy
				 */
				void setGradImages( Channel<gradT> *_gxx, Channel<gradT> *_gyy, Channel<gradT> *_gxy )
				{
					gxx = _gxx;
					gyy = _gyy;
					gxy = _gxy;
				}

		};

		template <class T, class gradT, class functionT>
		void ChannelCorners<T, gradT, functionT>::operator()( const Channel<T> &img, functionT &func )
		{
			assert( gxx != NULL );
			assert( gyy != NULL );
			assert( gxy != NULL );

			const float dxMax = 0.5F;
			const float dyMax = 0.5F;

			typename Channel<T>::Iterator2D begin = img.getIteratorBegin();
			typename Channel<T>::Iterator2D end = img.getIteratorEnd();

			EdgesAt<T> EdgeCounter( img, hood, 2, numAngleBins );
			EdgeCounter.setThresh( minEdgeStrength );

			HarrisAt<gradT> HarrisStrength( *gxx, *gyy, *gxy, 1 ); //1 because of best localization

			uint x, y;
			float subX, subY;
			for ( ; begin != end; ++begin ) //iterate over pixels
			{
				img.getCoors( begin.getPointer(), x, y );
				if ( x < hood || y < hood || x >= img.getWidth() - hood || y >= img.getHeight() - hood ) continue;

				float strength = HarrisStrength( x, y );
//				DBGOUT( DBGNORMAL, "HarrisStrength: %f (thresh: %f)\n", strength, harrisThresh );
				if ( strength > harrisThresh )
				{

					EdgeCounter( x, y );//TODO should returnvalue be used?

					//calculate second strongest edge strength
					int votes1st = 0;
					int votes2nd = 0;
					for ( typename EdgesAt<T>::iterator it = EdgeCounter.begin(); it != EdgeCounter.end(); ++it )
					{
						int val = *it;
						if ( val > votes1st )
						{
							votes2nd = votes1st;
							votes1st = val;
						}
						else if ( val > votes2nd )
						{
							votes2nd = val;
						}
					}
					DBGOUT( DBGNORMAL, "strongest: %d, 2nd strongest: %d\n", votes1st, votes2nd );

					/*if(x > 8 && x <12 && y > 8 && y < 12)
					{
					    DBGOUT(DBGNORMAL, "Numedges at (%d, %d): %d\n", x, y, numEdges);
					    for(typename EdgesAt<T>::iterator it = EdgeCounter.begin(); it != EdgeCounter.end(); ++it)
					    {
					        printf("%d  ", *it);
					    }
					    printf("\n");
					}*/

					if ( votes2nd > needed2nd )
					{
//						DBGOUT( DBGNORMAL, "(%d, %d) > numEdgesThresh\n", x, y );
						harrisVals[0] = HarrisStrength( x - 1, y - 1 ) - strength;
						harrisVals[1] = HarrisStrength( x    , y - 1 ) - strength;
						harrisVals[2] = HarrisStrength( x + 1, y - 1 ) - strength;

						harrisVals[3] = HarrisStrength( x - 1, y    ) - strength;
						harrisVals[4] = HarrisStrength( x + 1, y    ) - strength;

						harrisVals[5] = HarrisStrength( x - 1, y + 1 ) - strength;
						harrisVals[6] = HarrisStrength( x    , y + 1 ) - strength;
						harrisVals[7] = HarrisStrength( x + 1, y + 1 ) - strength;

						bool success = subCoorQuadratic( subX, subY, harrisVals );
//						printf( "subCoorQuadratic (%f, %f) (%s)\n", subX, subY, success ? "success" : "failed" );
						if ( success &&
						                fabs( subX ) < dxMax &&
						                fabs( subY ) < dyMax ) func( float( x ) + subX, float( y ) + subY, strength );
					}
				}
//				pressKey();
			}
		}


	}

}

#endif
