

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_FEATURESCHANNELEXTREMA_H
#define ADORE_FEATURESCHANNELEXTREMA_H

#include <adrcore/imgprocessing/channel.h>
#include <adrcore/types/featurepoint.h>

#include <functional>

namespace adore
{

	namespace core
	{

		/**
		 * Finds local extrema in an image and then calls a callback-function
		 * the extrema is compared to a threshold using a given comparator
		 * @tparam T image type
		 * @tparam functionT callback type
		 * @tparam Comparator determins what kind of extrema are reported
		 *  
		*/
		template <class T, class functionT, class Comparator = greater_equal<T> >
		class ChannelExtrema
		{
			public:
				/**
				 * Constructor
				 * @param t the threshold an extremum need to exceed
				 */
				ChannelExtrema ( T t = 0 ) :thresh ( t )
				{
					;
				}

				/**
				 * Destructor
				 */
				~ChannelExtrema()
				{
					;
				}

				/**
				 * identifies extrama in an image and calls the callback functor on each occurrence
				 * @param img the image to scan for extrema
				 * @param func the callback functor
				 */
				void operator() ( Channel<T> &img, functionT &func );

				/**
				 * @param theValue the threshold an extremum needs to exceed to get reported
				 */
				void setThresh ( const T& theValue )
				{
					thresh = theValue;
				}

				/**
				 * @return the threshold an extremum needs to exceed to get reported
				 */
				T getThresh() const
				{
					return thresh;
				}

			protected:

				T thresh;		///the threshold an extremum needs to exceed to get reported
				Comparator compare;	///a comparator for loacal extremum detection
		};

		template <class T, class functionT, class Comparator>
		void ChannelExtrema<T, functionT, Comparator>::operator() ( Channel<T> &img, functionT &func )
		{
			UsingROI r1 ( &img );
			img.minROIDistance ( 1 );

			typename Channel<T>::Iterator2D begin = img.getIteratorBegin();
			typename Channel<T>::Iterator2D end = img.getIteratorEnd();

//			cout << "ChannelExtrema: Thresh is " << thresh << endl;

			for ( ; begin != end; ++begin )
			{
				T val = *begin;

				if ( !compare ( val, thresh ) )
				{
//					cout << val << " is !compare to " << thresh << endl;
					continue;
				}

//				else
//				{
//					cout << "yeah "<<val << endl;
//					pressKey();
//				}

				bool extremum = true;

				for ( int c = 0; c < 8; ++c )   //iterate corona
				{
					if ( !compare ( val, *begin.getPointer ( c ) ) )
					{
						extremum = false;
						break;
					}
				}

				if ( extremum )
				{
					/*printf("%d - ", *begin.getPointer());

					for(int c = 0; c < 8; ++c)//iterate corona
					{
						printf("%d ", *begin.getPointer(c));
					}
					printf("\n");*/

					func ( img, begin.getPointer() );

//					++begin; if ( begin == end ) return;
				}
			}
		}


		template<bool> struct compile_time_assert_;
		template<> struct compile_time_assert_<true> {};

		/**
		 * Finds local extrema in an image and then calls a callback-function
		 * the extrema is compared to a threshold using a given comparator
		 * Skips right half of filter window if an extrema is found
		 * Performance is ok only for filter size > 3 and small threshold
		 * @tparam T image type
		 * @tparam Iterator1D 1D iterator to iterate in image
		 * @tparam functionT callback type
		 * @tparam filter_sz size of filter(min value is 3)
		 * @tparam Comparator determins what kind of extrema are reported
		 *  
		 */
		template <typename T, class Iterator1D, class functionT, int filter_sz = 3, class Comparator = std::greater_equal<T> >

		class Extrema
		{
			public:
				/**
				 * Constructor
				 * @param t the threshold an extremum need to exceed
				 */
				Extrema ( T t = 0 ) : thresh ( t ) { }

				~Extrema() { }

				/**
				 * @param theValue the threshold an extremum needs to exceed to get reported
				 */
				void setThresh ( const T& theValue )
				{
					thresh = theValue;
				}

				/**
				 * @return the threshold an extremum needs to exceed to get reported
				 */
				T getThresh() const
				{
					return thresh;
				}

				/**
				 * @param begin iterator pointing at beginnig of image
				 * @param width image width
				 * @param height image height
				 * @param lineStep width of one line of image in memory (can be >= width)
				 * @param func the callback functor
				 */
				void operator() ( Iterator1D begin, std::size_t width, std::size_t height,
				                  std::size_t lineStep, functionT& func ) const
				{
					using std::size_t;

					compile_time_assert_< 0 != filter_sz % 2 >();
					compile_time_assert_< 3 <= filter_sz >();
					size_t n = filter_sz / 2;
					size_t step = n + 1;

					if ( width < n || height < n ) return;

					for ( size_t i = n; i < height - n; i += step )
					{
						Iterator1D line = begin + i * lineStep;

						for ( size_t j = n; j < width - n; j += step )
						{
							size_t mi = i, mj = j;
							T extremum = * ( line + j );

							for ( size_t i2 = i; i2 <= i + n; ++i2 )
							{
								Iterator1D pt = begin + i2 * lineStep + j;

								for ( size_t j2 = j; j2 <= j + n; ++j2, ++pt )
								{
									if ( compare ( *pt, extremum ) )
									{
										extremum = *pt;
										mi = i2; mj = j2;
									}
								}
							}

							// TOP
							for ( size_t i2 = mi - n; i2 < i; ++i2 )
							{
								for ( size_t j2 = mj - n; j2 <= mj + n; ++j2 )
								{
									if ( compare ( * ( begin + i2 * lineStep + j2 ), extremum ) )
										goto not_extremum;
								}
							}

							// BOTTOM
							for ( size_t i2 = mi + n; i2 > i + n; --i2 )
							{
								for ( size_t j2 = mj - n; j2 <= mj + n; ++j2 )
								{
									if ( compare ( * ( begin + i2 * lineStep + j2 ), extremum ) )
										goto not_extremum;
								}
							}

							// LEFT
							for ( size_t i2 = i; i2 >= i + n; ++i2 )
							{
								for ( size_t j2 = mj - n; j2 < j; ++j2 )
								{
									if ( compare ( * ( begin + i2 * lineStep + j2 ), extremum ) )
										goto not_extremum;
								}
							}

							// RIGHT
							for ( size_t i2 = i; i2 >= i + n; ++i2 )
							{
								for ( size_t j2 = mj + n; j2 > j; --j2 )
								{
									if ( compare ( * ( begin + i2 * lineStep + j2 ), extremum ) )
										goto not_extremum;
								}
							}

							j += mj - j; // extrema found => all values at right are not extrema, skip them.

							if ( !compare ( extremum, thresh ) ) continue;

							func ( mj, mi ); // x, y

not_extremum:	;
						}
					}
				}

			protected:
				T thresh; ///< the threshold an extremum needs to exceed to get reported
				Comparator compare; ///< determins what kind of extrema are reported
		};


	} // core
} // adore

#endif
