

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_FEATURESCORNERFEATURES_H
#define ADORE_FEATURESCORNERFEATURES_H

#include <adrcore/features/featurefilter.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/features/channelcorners.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/search/grid2d.h>
#include <adrcore/util/recommendations.h>
//#include <adrcore/util/printimage.h>

namespace adore
{

	namespace core
	{

		/**
		 * takes an image in which edges are marked white and measures the
		 * number of lines that meet at each corner point. if the number exceed a predefined value,
		 * the harris corner strength is measured. If this also exceeds a predefined value, the point is accepted as a feature.
		 * @tparam T the image type of the image with edges
		 * @tparam gradT the gradient image type for harris (gxx, gyy, gxy)
		 *  
		*/
		template <class T, class gradT>
		class CornerFeatures
		{
			private:
				typedef typename signedT< typename calcT<gradT>::type >::type harrisT;
			protected:

				uint fid;		///feature id which is assigned to newly created features
				float minDistance;	///the minimum distance between features in the image plane
				float minimumFraction;	///the minimum required fraction of strength relative to the strongest feature
				int maxPoints;		///the maximum number of feature points allowed
				FeatureGrid candidates;	///candidate feature points
				ChannelCorners<T, gradT, CornerFeatures<T, gradT> > corners;///computes the corners
			public:
				/**
				 * Constructor
				 * @param ID the feature id newly constructed features shall have
				 * @param minDist the minimum distance of feature points
				 * @param minStrength the minimum feature strength
				 * @param minFraction the minimum fraction thr strength of a feature must have, relative to the strength of the strongest feature
				 * @param maxPts the maximum number of features returned. The strongest features are preferred.
				 * @param hStrength the required harris strength
				 * @param strengthOfSecondStrongestEdge the minimum strength of the second strongest edge in the local hough trafo
				 * @param hood a patch of size 2*hood+1 is examined
				 * @param numAngelBins the number of angles used by the local hough trafo
				 */
				CornerFeatures( uint ID = 0, float minDist = 10.0F, T minStrength = CFDefaults<T>::minThresh,
				                float minFraction = 0.00001, int maxPts = INT_MAX,
				                harrisT hStrength = 0,
				                int strengthOfSecondStrongestEdge = 14,
				                int hood = 6, int numAngleBins = 12 )
						:
						fid( ID ),
						minDistance( minDist ),
						minimumFraction( minFraction ),
						maxPoints( maxPts ),
						candidates( false ),//don't delete points
						corners( minStrength, hStrength, strengthOfSecondStrongestEdge, hood, numAngleBins )
				{
					;
				}


				/**
				 * Destructor
				 */
				~CornerFeatures()
				{
					;
				}

				/**
				 * set the minimum distance allowed for feature points.
				 * @param theValue new distance.
				 */
				void setMinDistance( float theValue )
				{
					minDistance = theValue;
				}

				/**
				 * get the current minimum distance allowed for feature points.
				 * @return current minimum distance.
				 */
				float getMinDistance() const
				{
					return minDistance;
				}

				/**
				 * callback for ChannelCorners
				 * This constructs the FeaturePoints and adds them to the local collection of candidates.
				 * @param x the feature's x co-ordinate
				 * @param y the feature's y co-ordinate
				 * @param strength the features strength
				 */
				inline void operator()( float x, float y, float strength )
				{
					FeaturePoint *pt = new FeaturePoint( x, y, fid );
					pt->setStrength( strength );
					candidates.add( pt );
				}

				/**
				 * computes the feature points and puts them into a vector.
				 * @param img the image features shall be computed from.
				 * @param f a vector into which the features are put.
				 */
				void getFeatures( Channel<T> &img, vector<FeaturePoint *> &f );

				/**
				 * set gradient images for the harris measure
				 * @param gxx an image containing gx*gx
				 * @param gyy an image containing gy*gy
				 * @param gxy an image containing gx*gy
				 */
				void setGradImages( Channel<gradT> *gxx, Channel<gradT> *gyy, Channel<gradT> *gxy )
				{
					corners.setGradImages( gxx, gyy, gxy );
				}

		};

		template <class T, class gradT>
		void CornerFeatures<T, gradT>::getFeatures( Channel<T> &img, vector<FeaturePoint *> &f )
		{
			candidates.init( 0, 0, img.getWidth(), img.getHeight() );

			corners( img, *this );

			DBGOUT( DBGNORMAL, "CornerFeatures: found %d maxima\n", candidates.getSize() );
#ifndef NDEBUG
			for ( typename FeatureGrid::Iterator it = candidates.getBegin(); it != candidates.getEnd(); ++it )
			{
				printf( "candidate (%f, %f)\n", ( **it )( 0 ), ( **it )( 1 ) );
			}
#endif

			filterFeatures( candidates, minDistance, minimumFraction, maxPoints );

			DBGOUT( DBGNORMAL, "CornerFeatures: after dist/str filtering %d maxima\n", candidates.getSize() );

			FeatureGrid::Iterator it = candidates.getBegin();
			FeatureGrid::Iterator endIt = candidates.getEnd();

			for ( ;it != endIt; ++it )
			{
				DBGOUT( DBGNORMAL, "CornerFeature: (%f, %f)\n", ( **it )( 0 ), ( **it )( 1 ) );
				f.push_back( *it );
			}
		}




	}

}

#endif
