

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGCHANNEL_H
#define ADORE_IMGPROCESSINGCHANNEL_H

#include <vector>
#include <limits.h>
#include <adrcore/util/debug.h>
#include <adrcore/imgprocessing/imageiterator.h>
#include <adrcore/imgprocessing/roi.h>
#include <adrcore/imgprocessing/imgconstants.h>
#include <adrcore/imgprocessing/imgfunctions.h>
#include <adrcore/imgprocessing/imgarray.h>
#include <adrcore/imgprocessing/channelset.h>

namespace adore
{

	namespace core
	{

		using namespace std;

		/**
		 * The Basis of a Channel, usually referencing to one channel of a multichannel image
		 *  
		*/
		class ChannelBase
		{
//			public:
//				typedef typename img_traits<T>::ImgArray ImgArray;

			public:
				/**
				 * Constructor
				 * @param _img an ImgArray which holds the pixel data
				 * @param roi a predefined ROI
				 */
				explicit ChannelBase( ImgArray *_img = NULL, const ROI &_roi = DefaultROI ):
						img( _img ),
						roi( _roi )
				{
				}

				/**
				 * Destructor
				 */
				virtual ~ChannelBase()
				{
					;
				}

				/**
				 * @return the width of the Channel
				 */
				uint getWidth() const
				{
					return getImgArrayWidth( img );
				}

				/**
				 * @return the height of the Channel
				 */
				uint getHeight() const
				{
					return getImgArrayHeight( img );
				}

				/**
				 * This function sets the height of the image to a value less than
				 * the current one. If you like to restore the old height, you
				 * have to store a backup beforehand
				 * @param h the new height
				 */
				void clipHeight( uint h )
				{
					assert( h < getHeight() && h > 0 );
					setImgArrayHeight( img, h );
				}

				/**
				 * modifies the ROI to ensure a minimum distance to the image border
				 * @param min the minimum distance the ROI shall have to the image border
				 */
				void minROIDistance( uint min )
				{
					ROI &roi = getROI();
					if ( roi.getX() < min )
					{
						roi.setWidth(roi.getWidth() - min + roi.getX());
						roi.setX( min );
					}
					if ( roi.getY() < min )
					{
						roi.setHeight(roi.getHeight() - min + roi.getY());
						roi.setY( min );
					}
					if ( roi.getX() + roi.getWidth() > getWidth() - min ) roi.setWidth( getWidth() - min - roi.getX() );
					if ( roi.getY() + roi.getHeight() > getHeight() - min ) roi.setHeight( getHeight() - min - roi.getY() );
				}

				/**
				 * @return the underlying ImgArray. May also be referenced by other Channels
				 */
				ImgArray *getImgArray() const
				{
					return img;
				}

				/**
				 * sets the Channel's ROI
				 * @param x the x co-ordinate of the upper left ROI corner
				 * @param y the y co-ordinate of the upper left ROI corner
				 * @param w the width of the ROI
				 * @param h the height of the ROI
				 */
				void setROI( int x, int y, int w, int h )
				{
//					printf( "stting a %d %d %d %d\n", x, y, w, h );
					roi.set( x, y, w, h );
				}

				/**
				 * sets the Channel's ROI
				 * @param r the new ROI
				 */
				void setROI( const ROI &r )
				{
//					printf( "stting b %d %d %d %d\n", r.getX(), r.getY(), r.getWidth(), r.getHeight() );
					roi = r;
				}

				/** 
				 * sets the Channel's ROI centered on the image.
				 * @param width the width of the ROI
				 * @param height the height of the ROI
				 */
				void setCenterdROI( int width, int height )
				{
					int x = ( getWidth() - width ) >> 1;
					int y = ( getHeight() - height ) >> 1;
					roi.set( x, y, width, height );
				}

				/**
				 * re-sets the Channel's ROI to cover the whole image
				 */
				void resetROI()
				{
//					printf( "%d %d\n", getWidth(), getHeight() );
					roi.set( 0, 0, getWidth(), getHeight() );
				}

				/**
				 * @return a reference to the Channel's ROI
				 */
				ROI &getROI()
				{
					return roi;
				}

				/**
				 * @return a const reference to the Channel's ROI
				 */
				const ROI &getROI() const
				{
					return roi;
				}

			protected:

				/**
				 * sets the ImgArray which contains the Channels pixel
				 * @param i the array containing the pixel data
				 */
				void setImgArray( ImgArray *i )
				{
					img = i;
				}

			private:

				ImgArray *img;	///the array containing the pixel data
				ROI roi;	///the current ROI
		};


		/**
		 * A collection of Channels.
		 * @tparam T the image type.
		 */
		template <class T> class ChannelSet;

		/** 
		 * A Channel, usually referencing a single channel of a multichannel image.
		 * @tparam T the image data type of the channel.
		 */
		template <class T>
		class Channel:public ChannelBase
		{

			public:

				typedef Image2DIterator<T> Iterator2D;
				typedef Image1DIterator<T> Iterator1D;
				typedef Iterator2D iterator;
				typedef T value_type;
				typedef ChannelSet<T> Set;

				/**
				 * Constructor.
				 * @param _img the image.
				 * @param _offSet the offSet to address the channel.
				 * @param _numBytes how may bytes represent one channel (e.g. uchar = 1).
				 * @param _incStep the Step to the next image value.
				 * @param _numChans the real number of channels encoded in _img.
				 */
				explicit Channel( ImgArray *_img = NULL, int _offSet = 0,
				                  int _numBytes = sizeof( T ), int _incStep = 1,
				                  int _wStep = 0, const ROI &_roi = DefaultROI );

				/**
				 * Destructor
				 */
				virtual ~Channel()
				{
					;
				}

				/**
				 * @return an iterator iterating over the pixel pointing to the begin of the channel data
				 */
				Image2DIterator<T> begin() const
				{
					return getIteratorBegin();
				}

				/**
				 * @return an iterator iterating over the pixel pointing after the last bit of the channel data
				 */
				Image2DIterator<T> end() const
				{
					return getIteratorEnd();
				}

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing to the begin of the channel data
				 */
				Image2DIterator<T> getIteratorBegin( int xStep = 1, int yStep = 1 ) const
				{
					return getIteratorBeginH( xStep, yStep );
				}

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing after the last bit of the channel data
				 */
				Image2DIterator<T> getIteratorEnd( int xStep = 1, int yStep = 1 ) const
				{
					return getIteratorEndH( xStep, yStep );
				}

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing to the begin of the channel data. Iterates the image columns first
				 */
				Image2DIterator<T> getIteratorBeginV( int xStep = 1, int yStep = 1 ) const;

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing after the last bit of the channel data. Iterates the image columns first
				 */
				Image2DIterator<T> getIteratorEndV( int xStep = 1, int yStep = 1 ) const;

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing to the begin of the channel data. Iterates the image rows first
				 */
				Image2DIterator<T> getIteratorBeginH( int xStep = 1, int yStep = 1 ) const;

				/**
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel pointing after the last bit of the channel data. Iterates the image rows first
				 */
				Image2DIterator<T> getIteratorEndH( int xStep = 1, int yStep = 1 ) const;

				/**
				 * A horizontally iterating iterator
				 * @param row the image row to start
				 * @param col the image col to start
				 * @param range the number of pixel after which to switch to the next row.
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel. Iterates the image rows first
				 */
				Image2DIterator<T> getIterator( int row, int col, int range, int xStep = 1, int yStep = 1 ) const
				{
					return getIteratorH( row, col, range, xStep, yStep );
				}

				/**
				 * A vertically iterating iterator
				 * @param row the image row to start
				 * @param col the image col to start
				 * @param range the number of pixel after which to switch to the next column.
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel. Iterates the image columns first
				 */
				Image2DIterator<T> getIteratorV( int row, int col, int range, int xStep = 1, int yStep = 1 ) const;

				/**
				 * A horizontally iterating iterator
				 * @param row the image row to start
				 * @param col the image col to start
				 * @param range the number of pixel after which to switch to the next row.
				 * @param xStep a value greater zero causes the iterator to skip pixel
				 * @param yStep a value greater zero causes the iterator to skip pixel
				 * @return an iterator iterating over the pixel. Iterates the image rows first
				 */
				Image2DIterator<T> getIteratorH( int row, int col, int range, int xStep = 1, int yStep = 1 ) const;


				/**
				 * Creates an Iterator which stays in its row.
				 * @param row the row to iterate over
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel. Starts at the beginning of the row.
				 */
				Image1DIterator<T> getXIteratorBegin( int row, int step = 1 ) const;

				/**
				 * Creates an Iterator which stays in its row.
				 * @param row the row to iterate over
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel. Starts after the end of the row.
				 */
				Image1DIterator<T> getXIteratorEnd( int row, int step = 1 ) const;

				/**
				 * Creates an Iterator which stays in its row.
				 * @param row the row to iterate over
				 * @param col the column to start the iteration at
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel.
				 */
				Image1DIterator<T> getXIterator( int row, int col, int step = 1 ) const;

				/**
				 * Creates an Iterator which stays in its column.
				 * @param col the column to iterate over
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel. Starts at the beginning of the column.
				 */
				Image1DIterator<T> getYIteratorBegin( int col, int step = 1 ) const;

				/**
				 * Creates an Iterator which stays in its column.
				 * @param col the column to iterate over
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel. Starts after the end of the column.
				 */
				Image1DIterator<T> getYIteratorEnd( int col, int step = 1 ) const;

				/**
				 * Creates an Iterator which stays in its column.
				 * @param row the row to start the iteration at
				 * @param col the column to iterator over
				 * @param step a value greater zero causes the iterator to skip pixel
				 * @return a 1D-iterator iterating over the pixel.
				 */
				Image1DIterator<T> getYIterator( int row, int col, int step = 1 ) const;

				/**
				 * deep copy of pixel values.
				 * @param ip the channel to copy. The ROIs should match in their size.
				 */
				void operator=( const Channel<T> &ip );

				/**
				 * copies data into the channel
				 * @param Iterator an Iterator type
				 * @param begin the begin of the data to be copied
				 * @param end the end of the data to be copied
				 */
				template<class Iterator>
				void copyFrom( Iterator begin, Iterator end );

				/**
				 * pixel access
				 * @param x the x co-ordinate of the pixel
				 * @param y the y co-ordinate of the pixel
				 * @return a reference to the pixel value
				 */
				inline T &operator()( uint x, uint y )
				{
					return *getPointer( x, y );
				}

				/**
				 * read only pixel access
				 * @param x the x co-ordinate of the pixel
				 * @param y the y co-ordinate of the pixel
				 * @return a const reference to the pixel value
				 */
				inline const T &operator()( uint x, uint y ) const
				{
					return *getPointer( x, y );
				}

				/**
				 * fills the whole channel with a constant value
				 * @param val the value
				 */
				void fill( const T &val );

				/** 
				 * get a pointer to the pixel position
				 * @param x the x co-ordinate of the pixel
				 * @param y the y co-ordinate of the pixel
				 * @return a pointer to the pixel position
				 */
				inline T *getPointer( const uint x, const uint y ) const;

				/** 
				 * get a pointer to a pixel position in the neighborhood of another pixel
				 * @param coronaIdx an index of the interval [0..7] identifying the neighbor pixel.
				 * @param ptr a pointer to the reference pixel position
				 * @return a pointer to the pixel position
				 */
				inline T *getPointerRel( int coronaIdx, T * ptr ) const;

				/** 
				 * get a pointer to a pixel position in the neighborhood of another pixel
				 * @param coronaIdx an index of the interval [0..7] identifying the neighbor pixel.
				 * @param x the x co-ordinate of the pixel
				 * @param y the y co-ordinate of the pixel
				 * @return a pointer to the pixel position
				 */
				inline T *getPointerRel( int coronaIdx, int x, int y ) const;

				/** 
				 * get a pointer to a pixel position relative of another pixel
				 * @param x the x coordinate relative to the reference pixel
				 * @param y the y coordinate relative to the reference pixel
				 * @param ptr a pointer to the reference pixel
				 * @return a pointer to the pixel position
				 */
				inline T *getPointerRel( int x, int y, T * ptr ) const;

				/** 
				 * get a pointer based on an offset to the begin of the channel data
				 * @param off the offset
				 * @return a pointer to the pixel position
				 */
				inline T *getPointer( uint off ) const;

				/** 
				 * get an offset from the begin of the channel data to the pixel position
				 * @param x the x co-ordinate of the pixel
				 * @param y the y co-ordinate of the pixel
				 * @return the offset
				 */
				inline int getOffset( const uint x, const uint y ) const;

				/** 
				 * get an offset from the begin of the channel data to the pixel position
				 * @param pixel a pointer to the pixel position
				 * @return the offset
				 */
				inline int getOffset( const T * const pixel ) const;

				/** 
				 * get an offset from the begin of the channel data to the pixel position
				 * @param x the x coordinate relative to the reference pixel
				 * @param y the y coordinate relative to the reference pixel
				 * @param off the offset to the reference pixel
				 * @return the offset
				 */
				inline int getOffsetRel( const int x, const int y, uint off ) const;

				/**
				 * get an offset from the begin of the channel data to the pixel position
				 * @param coronaIdx an index of the interval [0..7] identifying the neighbor pixel.
				 * @param off the offset to the reference pixel
				 * @return the offset
				 */
				inline int getOffsetRel( const int coronaIdx, uint off ) const;

				/**
				 * @return the base pointer all offset a computed from
				 */
				inline T *getBase() const;

				/**
				 * convert an offset into co-ordinates
				 * @param offset an offset
				 * @param x will contain the x co-ordinate belonging to the offset
				 * @param y will contain the y co-ordinate belonging to the offset
				 */
				inline void getCoors( int offset, uint &x, uint &y ) const;

				/**
				 * convert apointer into co-ordinates
				 * @param pointer a pointer
				 * @param x will contain the x co-ordinate belonging to the offset
				 * @param y will contain the y co-ordinate belonging to the offset
				 */
				inline void getCoors( const T* const pointer, uint &x, uint &y ) const;

				/**
				 * debug function, printing the co-ordinates of an offset to stdout
				 * @param offset an offset
				 */
				inline void printCoors( uint offset ) const;

				/**
				 * debug function, printing the co-ordinates of a pointer to stdout
				 * @param pointer a pointer
				 */
				inline void printCoors( const T* const pointer ) const;

				/**
				 * translates the offset of a pixel in srcChan in the corresponding offset
				 * for this channel.
				 * @param Tsrc the image data type of srcChan
				 * @param srcChan the channel the offset belongs to
				 * @param srcOff the offset in srcChan
				 * @return an offset in this channel that correspondes to srcOff in srcChan
				 */
				template <class Tsrc>
				inline int translateOffset( Channel<Tsrc> &srcChan, int srcOff )
				{
					div_t dv = div( srcOff, srcChan.getWStep() );
					int numChans = getImgArrayNChannels( this->getImgArray() );
					return dv.rem * numChans / srcChan.getNumChans() + getWStep() * dv.quot;
				}

				/**
				 * @return the width step to reach the same x coordinate in the next row.
				 */
				uint getWStep() const
				{
					return wStep;
				}

				/**
				 * This internal value is used by translateOffset()
				 * @return the number of Channels of the orig Image.
				 */
				uint getNumChans() const
				{
					return getImgArrayNChannels( this->getImgArray() );
				}

				/**
				 * This internal value is used by translateOffset()
				 * @return the number of Channels of the orig Image.
				 */
				uint getNumChannels() const
				{
					return getNumChans();
				}

				/**
				 * given a maximum range and a step size, the range is returned
				 * in which a integer number of steps fit without overflow
				 * @param maxRange the range to step through
				 * @param step the step size
				 * @return the number of steps one can savely take
				 */
				inline int getRange( int maxRange, int step ) const
				{
					int range = maxRange - 1;
					if ( step != 1 )
					{
						assert( step != 0 );
						div_t quot = div( range, step );
						range -= quot.rem;
					}
					range += step;
					return range;
				}

				/**
				 * pixel by pixel multiplication; modifies this channel to contain the result
				 * @param c2 the other channel
				 */
				template<class img2T> void operator*=( Channel<img2T> &c2 );

				/**
				 * pixel by pixel addition; modifies this channel to contain the result
				 * @param c2 the other channel
				 */
				template<class img2T> void operator+=( Channel<img2T> &c2 );
				
				/**
				 * pixel by pixel subtraction; modifies this channel to contain the result
				 * @param c2 the other channel
				 */
				template<class img2T> void operator-=( Channel<img2T> &c2 );
				
				/**
				 * pixel by pixel division; modifies this channel to contain the result
				 * @param c2 the other channel
				 */
				template<class img2T> void operator/=( Channel<img2T> &c2 );

				/**
				 * pixel by pixel addition; the values of the other channel are scaled first; modifies this channel to contain the result
				 * @param c2 the other channel
				 * @param scale the value that scales c2
				 */
				template<class img2T> void scaleAdd( Channel<img2T> &c2, float scale );

				/**
				 * replaces the values of this channel if the scaled values of another channel are greater
				 * this = max(factor * c2, this)
				 * @param c2 the other channel
				 * @param scale the value that scales c2
				 */
				template<class img2T> void max( Channel<img2T> &c2, float scale = 1.0 );

				/**
				 * Saves the image array into a file. Does not separate the channel!
				 * @param name the file name. Its extension determines the file type. PNG and JPG are perfectly sane choices.
				 */
				void save ( const char *name ) const
				{
					ImgSaver<T> saver;
					saver.saveImgArray ( name, getImgArray(), getROI() );
				}

			protected:

				uint offSet;	///the offset in the ImgArray at which the channel starts
				uint numBytes;	///how many bytes form a pixel of this channel?
				uint incStep;	///spacing between pixels of this channel

				/**
				 * the step to switch to the next image row in (sizeof(T))
				 * if the plane has 3 channels, the widthstep will be three times the number of pixels per row
				 */
				uint wStep;

// 				inline T* getImgArrayBase() const;
		};

		/**
		 * Sets the ROI of a channel and restores it upon destruction.
		 */
		class UsingROI
		{
			public:
				/**
				 * Constructor.
				 * Sets the ROI of a channel.
				 * @param i the channel.
				 * @param r the ROI.
				 */
				UsingROI( ChannelBase *i, const ROI &r ):
						roi( i->getROI() ),
						plane( i )
				{
					plane->setROI( r );
				}

				/**
				 * Constructor.
				 * Does not change the ROI, but copies the current ROI, which will be restored upon destruction.
				 * @param i the channel.
				 */
				UsingROI( ChannelBase *i ):
						roi( i->getROI() ),
						plane( i )
				{
				}

				/**
				 * Destructor.
				 * Restores the ROI.
				 */
				~UsingROI()
				{
					plane->setROI( roi );
				}
			private:
				ROI roi;
				ChannelBase *plane;
		};

		template<class T> Channel<T>::Channel( ImgArray *_img, int _offSet,
		                                       int _numBytes, int _incStep,
		                                       int _wStep, const ROI &_roi ):
				ChannelBase( _img, _roi ),
				offSet( _offSet ),
				numBytes( _numBytes ),
				incStep( _incStep ),
				wStep( _wStep )
		{
		}

		template<class T> template<class Iterator> void Channel<T>::copyFrom( Iterator begin, Iterator end )
		{
			Iterator2D dstIt = getIteratorBegin();
			Iterator2D dstEnd = getIteratorEnd();
			for ( ; ( begin != end ) && ( dstIt != dstEnd ); ++dstIt, ++begin )
			{
				*dstIt = *begin;
			}
		}

		template<class T>
		void Channel<T>::operator=( const Channel<T> &ip )
		{
			Iterator2D srcIt = ip.getIteratorBegin();
			Iterator2D dstIt = getIteratorBegin();
//			Iterator2D srcEnd = ip.getIteratorEnd();
			Iterator2D dstEnd = getIteratorEnd();
			for ( ; dstIt != dstEnd; ++srcIt, ++dstIt )
			{
				*dstIt = *srcIt;
			}
		}

		template<class T> Image2DIterator<T> Channel<T>::getIteratorV( int row, int col, int range, int xStep, int yStep ) const
		{
			Image2DIterator<T> it( getYIterator( row, col, yStep ),
			                       getYIterator( row + range, col, yStep ), xStep * incStep );
			return it;
		}


		template<class T> Image2DIterator<T> Channel<T>::getIteratorH( int row, int col, int range, int xStep, int yStep ) const
		{
			Image2DIterator<T> it( getXIterator( row, col, xStep ),
			                       getXIterator( row, col + range, xStep ), yStep * wStep );
			return it;
		}


		template<class T>     Image2DIterator<T> Channel<T>::getIteratorBeginV( int xStep, int yStep ) const
		{
			const ROI &roi = this->getROI();
			int range = getRange( roi.getHeight(), yStep );
			return getIteratorV( roi.getY(), roi.getX(), range, xStep, yStep );
		}

		template<class T>     Image2DIterator<T> Channel<T>::getIteratorBeginH( int xStep, int yStep ) const
		{
			const ROI &roi = this->getROI();
			int range = getRange( roi.getWidth(), xStep );
			return getIteratorH( roi.getY(), roi.getX(), range, xStep, yStep );
		}


		template<class T>     Image2DIterator<T> Channel<T>::getIteratorEndV( int xStep, int yStep ) const
		{
			const ROI &roi = this->getROI();
			int rangeX= getRange( roi.getWidth(), xStep );
			int rangeY= getRange( roi.getHeight(), yStep );
			return getIteratorV( roi.getY(), roi.getX() + rangeX, rangeY, xStep, yStep );
		}

		template<class T>     Image2DIterator<T> Channel<T>::getIteratorEndH( int xStep, int yStep ) const
		{
			const ROI &roi = this->getROI();
			int rangeX= getRange( roi.getWidth(), xStep );
			int rangeY= getRange( roi.getHeight(), yStep );
			return getIteratorH( roi.getY() + rangeY, roi.getX(), rangeX, xStep, yStep );
		}


		template<class T>     Image1DIterator<T> Channel<T>::getXIteratorBegin( int row, int step ) const
		{
			const ROI &roi = this->getROI();
			return getXIterator( row, roi.getX(), step );
		}

		template<class T>     Image1DIterator<T> Channel<T>::getXIteratorEnd( int row, int step ) const
		{
			const ROI &roi = this->getROI();
			int range= getRange( roi.getWidth(), step );
			return getXIterator( row, roi.getX() + range, step );
		}

		template<class T>     Image1DIterator<T> Channel<T>::getXIterator( int row, int col, int step ) const
		{
			T *startPtr = getBase() + getOffset( col, row );
			Image1DIterator<T> it( startPtr, incStep * step, numBytes );

			return it;
		}

		template<class T>     Image1DIterator<T> Channel<T>::getYIteratorBegin( int col, int step ) const
		{
			const ROI &roi = this->getROI();
			return  getYIterator( roi.getY(), col, step );
		}

		template<class T>     Image1DIterator<T> Channel<T>::getYIteratorEnd( int col, int step ) const
		{
			const ROI &roi = this->getROI();
			int range= getRange( roi.getHeight(), step );
			return  getYIterator( roi.getY() + range, col, step );
		}

		template<class T>     Image1DIterator<T> Channel<T>::getYIterator( int row, int col, int step ) const
		{
			T *startPtr = getBase() + getOffset( col, row );
			Image1DIterator<T> it( startPtr, wStep * step, numBytes );
			return  it;
		}

		template<class T> void Channel<T>::fill( const T &val )
		{
			Iterator2D it1 = getIteratorBegin();
			Iterator2D it2 = getIteratorEnd();
			for ( ; it1 != it2; ++it1 )
			{
				*it1 = val;
			}
		}

		template<class T> T *Channel<T>::getPointer( const uint x, const uint y ) const
		{
			assert_msg( x < this->getWidth(), "x = %d, width = %d\n", x, this->getWidth() );
			assert_msg( y < this->getHeight(), "y = %d, height = %d\n", y, this->getHeight() );
			return  getBase() + getOffset( x, y );
		}

		template<class T> T *Channel<T>::getPointerRel( int coronaIdx, T * ptr ) const
		{
			return  getPointerRel( CORONA[coronaIdx][0], CORONA[coronaIdx][1], ptr );
		}

		template<class T> T *Channel<T>::getPointerRel( int coronaIdx, int x, int y ) const
		{
			return  getPointer( x + CORONA[coronaIdx][0], y + CORONA[coronaIdx][1] );
		}

		template<class T> T *Channel<T>::getPointerRel( int x, int y, T * ptr ) const
		{
			return ptr + getOffset( x, y );
		}

		template<class T> T *Channel<T>::getPointer( uint off ) const
		{
			return getBase() + off;
		}

		template<class T> int Channel<T>::getOffset( const uint x, const uint y ) const
		{
			return x * getImgArrayNChannels( this->getImgArray() ) + y * wStep;
		}

		template<class T> void Channel<T>::getCoors( int offset, uint &x, uint &y ) const
		{
			div_t d = div( offset, wStep );
			y = d.quot;
			x = d.rem / getImgArrayNChannels( this->getImgArray() );
		}

		template<class T> void Channel<T>::printCoors( uint offset ) const
		{
			div_t d = div( int( offset ), int( wStep ) );
			printf( "Channel: Coors are (%d, %d)\n", d.rem, d.quot );
		}

		template<class T> void Channel<T>::getCoors( const T* const pointer, uint &x, uint &y ) const
		{
			getCoors( getOffset( pointer ), x, y );
		}

		template<class T> void Channel<T>::printCoors( const T* const pointer ) const
		{
			printCoors( getOffset( pointer ) );
		}


		template<class T> int Channel<T>::getOffset( const T * const pixel ) const
		{
			return pixel - getBase();
		}

		template<class T> int Channel<T>::getOffsetRel( const int x, const int y, uint off ) const
		{
			return off + getOffset( x, y );
		}

		template<class T> int Channel<T>::getOffsetRel( const int coronaIdx, uint off ) const
		{
			return off + getOffset( CORONA[coronaIdx][0], CORONA[coronaIdx][1] );
		}


		template<class T> T *Channel<T>::getBase() const
		{
			return getImgArrayData<T>( this->getImgArray() ) + offSet;
		}

		/**
			+=, -=, *=, /= operators
		*/
		template<class T>
		template<class img2T>
		void Channel<T>::operator*=( Channel<img2T> &c2 )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				*srcIt = *srcIt * *dstIt;
			}
		}

		template<class T>
		template<class img2T>
		void Channel<T>::operator+=( Channel<img2T> &c2 )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				*srcIt = *srcIt + *dstIt;
			}
		}

		template<class T>
		template<class img2T>
		void Channel<T>::scaleAdd( Channel<img2T> &c2, float factor )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				*srcIt = *srcIt + factor * *dstIt;
			}
		}


		template<class T>
		template<class img2T>
		void Channel<T>::operator-=( Channel<img2T> &c2 )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				*srcIt = *srcIt - *dstIt;
			}
		}

		template<class T>
		template<class img2T>
		void Channel<T>::operator/=( Channel<img2T> &c2 )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				*srcIt = *srcIt / *dstIt;
			}
		}

		template<class T>
		template<class img2T>
		void Channel<T>::max( Channel<img2T> &c2, float factor )
		{
			typename Channel<T>::Iterator2D srcIt = this->getIteratorBegin();
			typename Channel<img2T>::Iterator2D dstIt = c2.getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = this->getIteratorEnd();
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
				float val = factor * *dstIt;
				if ( val > *srcIt ) *srcIt = val;
			}
		}


	}
}

#endif
