

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGFLIPPER_H
#define ADORE_IMGPROCESSINGFLIPPER_H

#include <adrcore/types/aarray.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/channelfunctions.h>
#include <adrcore/util/calctype.h>

namespace adore
{

	namespace core
	{

		/**
		 * Flips an image, meaning a rotation about 90°, mirroring, etc.
		*/
		template <class inT>
		class Flipper
		{
			private:
				typedef Channel<inT> ChannelT;

			public:
				/**
				 * mirror vertically (exchange left and right)
				 */
				void mirrorV( const Channel<inT> *in, Channel<inT> *out );

				/**
				 * in-place version
				 */
				void mirrorV( Channel<inT> *in );

				/**
				 * mirror horizontally (exchange top and bottom)
				 */
				void mirrorH( const Channel<inT> *in, Channel<inT> *out );

				/**
				 * in-place version
				 */
				void mirrorH( Channel<inT> *in );

				/**
				 * rotates 90deg clockwise
				 */
				void rotate90cw( const Channel<inT> *in, Channel<inT> *out );

		};

		template <class inT>
		void Flipper<inT>::mirrorV( const Channel<inT> *in, Channel<inT> *out )
		{
			assert( in->getWidth() == out->getWidth() );
			int cols = in->getWidth();
			int rows = MIN( in->getHeight(), out->getHeight() );

			for ( int row = 0; row < rows; ++row )
			{
				typename Channel<inT>::Iterator1D iterIn  = in->getXIterator( row, 0 );
				typename Channel<inT>::Iterator1D iterOut = out->getXIterator( row, cols - 1 );

				for ( int col = 0; col < cols; ++col, ++iterIn, --iterOut )
				{
					*iterOut = *iterIn;
				}
			}
		}

		template <class inT>
		void Flipper<inT>::mirrorV( Channel<inT> *in )
		{
			int cols = in->getWidth();
			int rows = in->getHeight();

			for ( int row = 0; row < rows; ++row )
			{
				typename Channel<inT>::Iterator1D iterIn  = in->getXIterator( row, 0 );
				typename Channel<inT>::Iterator1D iterOut = in->getXIterator( row, cols - 1 );

				for ( int col = 0; col < cols >> 1; ++col, ++iterIn, --iterOut )
				{
					inT temp = *iterOut;
					*iterOut = *iterIn;
					*iterIn = temp;
				}
			}
		}

		template <class inT>
		void Flipper<inT>::mirrorH( const Channel<inT> *in, Channel<inT> *out )
		{
			assert( in->getHeight() == out->getHeight() );
			int cols = MIN( in->getWidth(), out->getWidth() );
			int rows = in->getHeight();

			for ( int col = 0; col < cols; ++col )
			{
				typename Channel<inT>::Iterator1D iterIn  = in->getYIterator( 0, col );
				typename Channel<inT>::Iterator1D iterOut = out->getYIterator( rows - 1, col );

				for ( int row = 0; row < rows; ++row, ++iterIn, --iterOut )
				{
					*iterOut = *iterIn;
				}
			}
		}

		template <class inT>
		void Flipper<inT>::mirrorH( Channel<inT> *in )
		{
			int cols = in->getWidth();
			int rows = in->getHeight();

			for ( int col = 0; col < cols; ++col )
			{
				typename Channel<inT>::Iterator1D iterIn  = in->getYIterator( 0, col );
				typename Channel<inT>::Iterator1D iterOut = in->getYIterator( rows - 1, col );

				for ( int row = 0; row < rows >> 1; ++row, ++iterIn, --iterOut )
				{
					inT temp = *iterOut;
					*iterOut = *iterIn;
					*iterIn = temp;
				}
			}
		}

		template <class inT>
		void Flipper<inT>::rotate90cw( const Channel<inT> *in, Channel<inT> *out )
		{
			//these to will be used on in's perspective
			int rows = MIN( in->getHeight(), out->getWidth() );
			int cols = MIN( in->getWidth(), out->getHeight() );

			for ( int row = 0; row < rows; ++row )
			{
				typename Channel<inT>::Iterator1D iterIn  = in->getXIterator( row, 0 );
				typename Channel<inT>::Iterator1D iterOut = out->getYIterator( 0, rows - row - 1 );

				for ( int col = 0; col < cols; ++col, ++iterIn, ++iterOut )
				{
//					printf( "(%d, %d)\n", row, col );
					*iterOut = *iterIn;
				}
			}
		}



	}

}

#endif
