

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGHARRISAT_H
#define ADORE_IMGPROCESSINGHARRISAT_H

#include <adrcore/types/aarray.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/gaussians.h>

namespace adore
{

	namespace core
	{

		/**
		 
		*/
		template <class inT>
		class HarrisAt
		{
				typedef typename signedT< typename calcT<inT>::type >::type harrisT;
				typedef AArray<double, 2> T_weight;

			public:
				HarrisAt( Channel<inT> &_gxx,
				          Channel<inT> &_gyy,
				          Channel<inT> &_gxy,
				          int _hood = 4,
				          T_weight *_weights = NULL,
				          float _k = 0.04 ):
						gxx( _gxx ),
						gyy( _gyy ),
						gxy( _gxy ),
						hood( _hood ),
						k( _k ),
						width( hood + hood + 1 ),
						weights( _weights ),
						deleteWeights( false )
				{
					if ( !weights )
					{
						weights = new T_weight( width, width );
						getGaussianWeights( *weights, hood );
						deleteWeights = true;
					}
				}

				~HarrisAt()
				{
					if ( deleteWeights ) delete weights;
				}

				harrisT operator()( int x, int y );

			private:

				HarrisAt()
				{
					;
				}

				Channel<inT> &gxx;
				Channel<inT> &gyy;
				Channel<inT> &gxy;
				const int hood;
				const float k;
				const int width;
				T_weight *weights;
				bool deleteWeights;
		};

		template <class inT>
		typename HarrisAt<inT>::harrisT HarrisAt<inT>::operator()( int x, int y )
		{
			float A = 0;
			float B = 0;
			float C = 0;

			UsingROI r1( &gxx );
			UsingROI r2( &gyy );
			UsingROI r3( &gxy );

			int hWidth = hood + hood + 1;
			int norm = hWidth * hWidth;

			gxx.setROI( x - hood, y - hood, hWidth, hWidth );
			gyy.setROI( x - hood, y - hood, hWidth, hWidth );
			gxy.setROI( x - hood, y - hood, hWidth, hWidth );

			typename Channel<inT>::Iterator2D gxxHIt1 = gxx.getIteratorBegin();
			typename Channel<inT>::Iterator2D gyyHIt  = gyy.getIteratorBegin();
			typename Channel<inT>::Iterator2D gxyHIt  = gxy.getIteratorBegin();
			typename Channel<inT>::Iterator2D gxxHIt2 = gxx.getIteratorEnd();
			typename T_weight::iterator wit = weights->begin();

			for ( ; gxxHIt1 != gxxHIt2; ++gxxHIt1, ++gyyHIt, ++gxyHIt, ++wit )
			{
				double w = *wit;
				A += w * *gxxHIt1;
				B += w * *gyyHIt;
				C += w * *gxyHIt;
//				printf( "%f  %f  %f  %f\n", w, A, B, C );
			}

//			printf( "%f  %f  %f\n", A, B, C );

			A /= norm;
			B /= norm;
			C /= norm;

//			printf( "%f  %f  %f\n", A, B, C );

			float trace = A + B;
			trace *= trace;
			float result = k * trace;

//			printf( "%f\n", A * B - C * C - result );

			return harrisT( A * B - C * C - result );
		}

	}

}

#endif
