

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREIMAGEARRANGER_H
#define ADORE_COREIMAGEARRANGER_H

#include <adrcore/search/kdtree3.h>
#include <adrcore/search/unionfind.h>
#include <adrcore/types/descriptorbackref.h>
#include <deque>

namespace adore
{

	namespace core
	{
		/**
		 * This class creates a relation between the members of a set of images that reflects their mutual similarity
		 * Its result can be used to group images that are good candidate matches
			 
		*/
		class ImageArranger
		{
			private:

				typedef FeatureDescriptor::DescVecT DescVecT;
				typedef DescVecT::scalar_type T_numtype;

				typedef DescriptorBackRef<int> DBackRef;//references to the view
				typedef DBackRef::Ruler BackRefRuler;
				typedef vector<DBackRef> CollectionT;
				typedef CollectionT::iterator CollIter;
				typedef KDTree3<CollIter, BackRefRuler> TreeT;

				typedef vector<ViewBase *> IdToView;//all views

			public:
				typedef AMatrixf AdMatrix;

			private:

				IdToView v;

			public:

				/**
				 * Note: Should preserve the enumeration applied in ViewRelations, if ViewRelations begin() is used as vit.
				 */
				template<class ViewIterator>
				AdMatrix arrange ( ViewIterator vit, int size, int featsId = 0, int descrId = 0 )
				{
					const int numNearest = MIN( 4, size );
					const float breakThresh = 0.9;

					//establish an indirect indexing in case of insane view ids
					v.resize( size );

					//count the number of features. this is to reserve the space the BackRefCollection may need
					ViewIterator it = vit;
					int numFeatures = 0;
					for ( int i = 0; i < size; ++i, ++it )
					{
						numFeatures += ( *it )->getNumFeaturePoints( featsId );

						v[i] = *it;
					}
					CollectionT backRefs;
					backRefs.reserve( numFeatures );
					const int bbfBins = MAX( 100, int( 0.2 * numFeatures ) );
					DBGOUT( DBGNORMAL, "ImageArranger using %d bbf bins for %d features\n", bbfBins, numFeatures );

					//initialize the square matrix relating the views
					AMatrixf viewScores( size, size );
					viewScores = 0;

					//create backrefs of all features in all views
					addBackRefIndices( &v[0], size, featsId, descrId, backRefs );

					//put all features into one big KD-Tree
					int dim = getDescriptorDim( *vit, featsId, descrId );
					BackRefRuler ruler( dim );
					TreeT tree( backRefs.begin(), backRefs.end(), dim, ruler );

					//for each feature score the views it most probably belongs to
					for ( CollIter bit = backRefs.begin(); bit != backRefs.end(); ++bit )
					{
						DBackRef search( *bit );

						TreeT::NearestCollectorT best( numNearest );
						//ResulT is <nearest, dist, numVisitedNodes>
						/*int visited = */
						tree.findNearestNeighbourBBF( search, bbfBins, best );
//						printf( "searched %d nodes, collected: %d\n", visited, best.size() );
						best.sort();
						//the best is the point itself, therefore the ++
						TreeT::NearestCollectorT::iterator it = ++best.begin();
						float dist = ( *it ).get<1>();
						int id1 = search.backRef;
						int id2 = ( *it ).get<0>()->backRef;
						if ( id1 != id2 )
						{
							float score = scoreFunc( dist );
							viewScores( id1, id2 ) += score;
							viewScores( id2, id1 ) += score;
							float lastDist = dist;
							for ( ; it != best.end(); ++it )
							{
								float dist = ( *it ).get<1>();
								assert( lastDist <= dist );
								if ( lastDist / dist < breakThresh ) break;
								int id1 = search.backRef;
								int id2 = ( *it ).get<0>()->backRef;
								if ( id1 == id2 ) break;
								float score = scoreFunc( dist );
								viewScores( id1, id2 ) += score;
								viewScores( id2, id1 ) += score;
							}
						}
					}

					cout << viewScores << endl;

					spanningTree( viewScores );

					cout << viewScores << endl;

					return viewScores;
				}

				ViewBase *getView( int idx )
				{
					return v[idx];
				}

			private:

				inline float scoreFunc( float dist )
				{
					return 1.0/( 0.0001 + dist );
				}

				void spanningTree( AMatrixf &scores )
				{
					typedef tuple<int, int, float> Edge;
					vector<Edge> edges;

					typedef UnionFind<int>::SetId SetId;
					UnionFind<int> vertexSets;

					for ( int row = 0; row < (int)scores.rows() - 1; ++row )
					{
						for ( int col = row + 1; col < (int)scores.cols(); ++col )
						{
							edges.push_back( make_tuple( row, col, scores( row, col ) ) );
						}
					}
					scores = 0;//set the matrix to zero
					TupleComparator<Edge, 2, greater<float> > compi;
					std::sort( edges.begin(), edges.end(), compi );
					for ( vector<Edge>::iterator it = edges.begin(); it != edges.end(); ++it )
					{
						Edge &e = *it;
						int row = e.get<0>();
						int col = e.get<1>();
						float benefit = e.get<2>();

//						check for circles
						SetId s1 = vertexSets.findSet( row );
						SetId s2 = vertexSets.findSet( col );
//						if s1 == s2 we are going to produce a circle
//						which is not what we want
						if ( s1 != NullSetId && s2 != NullSetId && ( s1 == s2 ) )
							{
								continue;
							}

//						no circle
						scores( row, col ) = benefit;
						scores( col, row ) = benefit;

//						add row and coll to the same set
						if ( s1 != NullSetId )
						{
							if ( s2 == NullSetId )
							{
								vertexSets.addSet( col );
							}
							else
							{
								vertexSets.unionSets( s1, col );
							}
						}
						else//s1 == NullSetId
						{
							if ( s2 == NullSetId )
							{
//								create new sets
								vertexSets.addSet( row );
								vertexSets.addSet( col );
								vertexSets.unionSets( row, col );
							}
							else
							{
								vertexSets.addSet( row );
							}
						}
					}
				}

		};

	}//namespace
}

#endif
