

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "imgfunctions.h"
#include <cmath>
#include <adrcore/util/debug.h>
#include <adrcore/util/printmat.h>


namespace adore
{

	namespace core
	{


		void getBicubicFactors( const float x, const float y,
		                        float Rx[4], float Ry[4] )
		{
			float dx = x - floor( x );
			float dy = y - floor( y );

			float dx1 = ( 1.0-dx ) * ( 1.0-dx ) * ( 1.0-dx );
			float dx2 = ( 2.0-dx ) * ( 2.0-dx ) * ( 2.0-dx );
			float dx3 = ( 3.0-dx ) * ( 3.0-dx ) * ( 3.0-dx );
			float dx4 = ( 4.0-dx ) * ( 4.0-dx ) * ( 4.0-dx );

			float dy1 = dy * dy * dy;
			float dy2 = ( dy + 1 ) * ( dy + 1 ) * ( dy + 1 );
			float dy3 = ( dy + 2 ) * ( dy + 2 ) * ( dy + 2 );
			float dy4 = ( dy + 3 ) * ( dy + 3 ) * ( dy + 3 );

			float factor = 1.0 / 6.0;

			Rx[0] = factor * dx1;
			Rx[1] = factor * ( dx2 - 4 * dx1 );
			Rx[2] = factor * ( dx3 - 4 * dx2 + 6 * dx1 );
			Rx[3] = factor * ( dx4 - 4 * dx3 + 6 * dx2 - 4 * dx1 );

			Ry[0] = factor * ( dy4 - 4 * dy3 + 6 * dy2 - 4 * dy1 );
			Ry[1] = factor * ( dy3 - 4 * dy2 + 6 * dy1 );
			Ry[2] = factor * ( dy2 - 4 * dy1 );
			Ry[3] = factor * dy1;
		}


		bool subCoorQuadratic( float &subX, float &subY, const float r[8] )
		{
			const float sixth = 1.0/6.0;

			double a =   0.1F * r[0] -  0.2F * r[1] +  0.1F  * r[2] +  0.3F  * r[3] +  0.3F  * r[4] +  0.1F  * r[5] -  0.2F  * r[6] + 0.1F  * r[7];
			double b =   0.1F * r[0] +  0.3F * r[1] +  0.1F  * r[2] -  0.2F  * r[3] -  0.2F  * r[4] +  0.1F  * r[5] +  0.3F  * r[6] + 0.1F  * r[7];
			double c =  0.25F * r[0] +  0.0F * r[1] -  0.25F * r[2] +  0.0F  * r[3] +  0.0F  * r[4] -  0.25F * r[5] +  0.0F  * r[6] + 0.25F * r[7];
			double d = -sixth * r[0] +  0.0F * r[1] +  sixth * r[2] -  sixth * r[3] +  sixth * r[4] -  sixth * r[5] +  0.0F  * r[6] + sixth * r[7];
			double e = -sixth * r[0] - sixth * r[1] -  sixth * r[2] +  0.0F  * r[3] +  0.0F  * r[4] +  sixth * r[5] +  sixth * r[6] + sixth * r[7];

			double divisor = c * c - 4 * a * b;
			if ( fabs( divisor ) <= FLT_EPSILON )
			{
//				DBGOUT( DBGALL, "degenerate Quadratic: a=%.15f, b=%.15f, c=%.15f, d=%.15f, e=%.15f\n", a, b, c, d, e );
//				getchar();
				return false;
			}

//			printf( "divsor: %.8f, denom1: %.8f, denom2: %.8f\n", divisor, ( 2 * b * d - c * e ), ( 2 * a * e - c * d ) );

			subX = ( 2 * b * d - c * e ) / divisor;
			subY = ( 2 * a * e - c * d ) / divisor;
//			printf( "subX: %.8f, subY: %.8f\n", subX, subY );
			return true;
		}

		bool subCoorQuadratic3D( float &subX, float &subY, const float r[26] )
		{
			const float c1 = 1.0F / 42.0F;
			const float c2 = 1.0F / 14.0F;
			const float c3 = 5.0F / 42.0F;
			const float c4 = 2.0F / 21.0F;
			const float c5 = 1.0F / 21.0F;
			const float c6 = 1.0F / 12.0F;
			const float c7 = 1.0F / 8.0F;
			const float c8 = 1.0F / 18.0F;

			float a = c1 *r[0] + c2 *r[1] + c1 *r[2] + c2 *r[3] + c3 *r[4] + c2 *r[5] + c1 *r[6] + c2 *r[7] + c1 *r[8] -c4 *r[9] -c5 *r[10] -c4 *r[11] -c5 *r[12] -c5 *r[13] -c4 *r[14] -c5 *r[15] -c4 *r[16] + c1 *r[17] + c2 *r[18] + c1 *r[19] + c2 *r[20] + c3 *r[21] + c2 *r[22] + c1 *r[23] + c2 *r[24] + c1 * r[25];

			float b = c1 *r[0] + c2 *r[1] + c1 *r[2] -c4 *r[3] -c5 *r[4] -c4 *r[5] + c1 *r[6] + c2 *r[7] + c1 *r[8] + c2 *r[9] + c3 *r[10] + c2 *r[11] -c5 *r[12] + c5 *r[13] + c2 *r[14] + c3 *r[15] + c2 *r[16] + c1 *r[17] + c2 *r[18] + c1 *r[19] -c4 *r[20] -c5 *r[21] -c4 *r[22] + c1 *r[23] + c2 *r[24] + c1 * r[25];

			float c = c1 *r[0] -c4 *r[1] + c1 *r[2] + c2 *r[3] -c5* r[4] + c2* r[5] + c1* r[6] -c4* r[7] + c1* r[8] + c2* r[9] -c5* r[10] + c2* r[11] + c3* r[12] + c3* r[13] + c2* r[14] -c5* r[15] + c2* r[16] + c1* r[17] -c4* r[18] + c1* r[19] + c2* r[20] -c5* r[21] + c2* r[22] + c1* r[23] -c4* r[24] + c1 * r[25];

			float d = c6* r[0] + c6* r[1] + c6* r[2] -c6* r[6] -c6* r[7] -c6* r[8] - c6* r[17] -c6* r[18] -c6* r[19] + c6* r[23] + c6* r[24] + c6 * r[25];

			float e = c6* r[0] -c6* r[2] + c6* r[3] - c6 * r[5] + c6* r[6] -c6* r[8] -c6* r[17] + c6* r[19] -c6* r[20] + c6* r[22] -c6* r[23] + c6 * r[25];

			float f = c6* r[0] -c6* r[2] -c6* r[6] + c6* r[8] + c6* r[9] -c6* r[11] -c6* r[14] + c6* r[16] + c6* r[17] -c6* r[19] -c6* r[23] + c6 * r[25];

			float g = c7* r[0] -c7* r[2] -c7* r[6] + c7* r[8] -c7* r[17] + c7* r[19] + c7* r[23] -c7 * r[25];

			float h = c8* r[0] + c8* r[1] + c8* r[2] + c8* r[3] + c8* r[4] + c8* r[5] + c8* r[6] + c8* r[7] + c8* r[8] -c8* r[17] -c8* r[18] -c8* r[19] -c8* r[20] -c8* r[21] -c8* r[22] -c8* r[23] -c8* r[24] -c8 * r[25];

			float i = c8* r[0] + c8* r[1] + c8* r[2] -c8* r[6] -c8* r[7] -c8* r[8] + c8* r[9] + c8* r[10] + c8* r[11] -c8* r[14] -c8* r[15] -c8* r[16] + c8* r[17] + c8* r[18] + c8* r[19] -c8* r[23] -c8* r[24] -c8 * r[25];

			float j = c8* r[0] -c8* r[2] + c8* r[3] -c8* r[5] + c8* r[6] -c8* r[8] + c8* r[9] -c8* r[11] + c8* r[12] -c8* r[13] + c8* r[14] -c8* r[16] + c8* r[17] -c8* r[19] + c8* r[20] -c8* r[22] + c8* r[23] -c8 * r[25];

			//minimization involves finding roots of a 5th degree polynom
			//... postponed

			ADORE_UNUSED( a );
			ADORE_UNUSED( b );
			ADORE_UNUSED( c );
			ADORE_UNUSED( d );
			ADORE_UNUSED( e );
			ADORE_UNUSED( f );
			ADORE_UNUSED( g );
			ADORE_UNUSED( h );
			ADORE_UNUSED( i );
			ADORE_UNUSED( j );

			return false;
		}

		void anglesCoveredByPixelAt( int x, int y, float &min, float &max )
		{
			if ( x > 0 && y > 0 )
			{
				min = atan2( y - 0.5, x + 0.5 );
				max = atan2( y + 0.5, x - 0.5 );
			}
			else if ( x < 0 && y > 0 )
			{
				min = atan2( y + 0.5, x + 0.5 );
				max = atan2( y - 0.5, x - 0.5 );
			}
			else if ( x > 0 && y < 0 )
			{
				min = atan2( y - 0.5, x - 0.5 );
				max = atan2( y + 0.5, x + 0.5 );
			}
			else if ( x < 0 && y < 0 )
			{
				min = atan2( y + 0.5, x - 0.5 );
				max = atan2( y - 0.5, x + 0.5 );
			}
			else if ( x == 0 && y > 0 )
			{
				min = atan2( y - 0.5, 0.5 );
				max = atan2( y - 0.5,-0.5 );
			}
			else if ( x == 0 && y < 0 )
			{
				min = atan2( y + 0.5, -0.5 );
				max = atan2( y + 0.5, 0.5 );
			}
			else if ( x > 0 && y == 0 )
			{
				min = atan2( -0.5, x - 0.5 );
				max = atan2( 0.5, x - 0.5 );
			}
			else if ( x < 0 && y == 0 )
			{
				min = atan2( 0.5, x + 0.5 );
				max = atan2( -0.5, x + 0.5 );
			}
			else if ( x == 0 && y == 0 )
			{
				min = -M_PI;
				max = M_PI;
			}
			else
			{
				assert( false );
			}
		}


	}
}
