

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESCALESPACEDOG_H
#define ADORE_CORESCALESPACEDOG_H

#include <adrcore/imgprocessing/imageplane.h>
#include <adrcore/imgprocessing/dizenzoat.h>
#include <adrcore/types/ringbuffer.h>
#include <adrcore/util/templatelimits.h>
#include <adrcore/util/printmat.h>

namespace adore
{

	namespace core
	{

		template<class imgT, class diffT> void calcDOG ( Channel<imgT> *oldI, Channel<imgT> *newI,
		                Channel<diffT> *diffI )
		{
			int step = 1;
			if ( oldI->getWidth() == newI->getWidth() )
			{
				assert ( diffI->getWidth() == oldI->getWidth() );
			}
			else
			{
				assert ( diffI->getWidth() == newI->getWidth() );
				step = 2;
			}

			typename Channel<imgT>::Iterator2D itoldI = oldI->getIteratorBegin ( step, step );
			typename Channel<imgT>::Iterator2D itnewI = newI->getIteratorBegin();
			typename Channel<imgT>::Iterator2D itnewIEnd = newI->getIteratorEnd();
			typename Channel<diffT>::Iterator2D itdiffI = diffI->getIteratorBegin();
			for ( ; itnewI != itnewIEnd; ++itnewI, ++itoldI, ++itdiffI )
			{
				*itdiffI = *itoldI - *itnewI;
			}
		}


		/**
		 * Given an image, its predecessor and successor in a scale space pyramid,
		 * extrema (minima and maxima simultaneously) are located.
		 */
		template<class imgT, class CallbackT>
		class DOGExtrema
		{
			public:
				typedef ImagePlane<imgT> ImageT;
				typedef typename signedT< typename calcT<imgT>::type >::type calcT;

			public:
				DOGExtrema ( CallbackT &cb, calcT minStrength, calcT minHarrisThresh ) :
						thresh ( minStrength ),
						hessian ( 3, 3 ),
						jacobian ( 3 ),
						offset ( 3 ),
						dzChans ( 1 ),
						edgeThresh ( minHarrisThresh ),
						wDiag ( 1.0 ),
						wOrtho ( 1.0 ),
						callback ( cb )
				{
					offset = 0,0,0;
				}

				bool isMaximum()
				{
					for ( imgT *ptr = cut; ptr != &cut[13]; ++ptr )
					{
						if ( cut[13] <= *ptr ) return false;
					}
					for ( imgT *ptr = &cut[14]; ptr != &cut[27]; ++ptr )
					{
						if ( cut[13] <= *ptr ) return false;
					}
					return true;
				}

				bool isMinimum()
				{
					for ( imgT *ptr = cut; ptr != &cut[13]; ++ptr )
					{
						if ( cut[13] >= *ptr ) return false;
					}
					for ( imgT *ptr = &cut[14]; ptr != &cut[27]; ++ptr )
					{
						if ( cut[13] >= *ptr ) return false;
					}
					return true;
				}

				/**
				 * finds points in candDOG that are maxima in (x, y, s)-direction
				 */
				void findExtrema ( Channel<imgT> *candDOG, Channel<imgT> *prevDOG, Channel<imgT> *nextDOG, int spacing )
				{
//					DBGOUT( DBGNORMAL, "finding dog maxima\n" );

					assert ( candDOG );
					assert ( thresh > 0 );

					uint x, y;//callback take coordinates
					resetCut();

					//first, check is the DOGs have an appropriate ROI
					//to prevent searching out of image boundaries
					UsingROI ur1 ( candDOG );
					if ( candDOG->getWidth() == candDOG->getROI().getWidth() ||
					                candDOG->getHeight() == candDOG->getROI().getHeight() )
					{
						candDOG->setROI ( 2, 2, candDOG->getWidth() - 4, candDOG->getHeight() - 4 );
					}

					typename Channel<imgT>::Iterator2D begin = candDOG->getIteratorBegin();
					typename Channel<imgT>::Iterator2D end = candDOG->getIteratorEnd();

					for ( ; begin != end; ++begin )
					{
						imgT val = *begin;

						if ( tabs<imgT> ( val ) < thresh )
						{
							continue;
						}

						fillCut ( begin.getPointer(), candDOG, prevDOG, nextDOG );
//						printCut();

						bool extremum = isMaximum() || isMinimum();
						if ( !extremum )
						{
							continue;
						}

//						candDOG->getCoors( begin.getPointer(), x, y );
//						printf( "(%d, %d) is an extremum\n", x, y );

						if ( isEdge() )
						{
//							printf( "failed\n\n\n" );
//							pressKey();
							continue;
						}
//						else
//						{
//							printf( "\n\n\npassed\n\n\n" );
//							pressKey();
//						}

//						printf( "(%d, %d) is an edge\n", x, y );

						//fill offset-array with offset
						float optValue = fitQuadratic(); //TODO: prevents simple squares from beeing matched
//						float optValue = *begin;

						if ( fabsf ( offset ( 2 ) ) < 0.5f && fabsf ( offset ( 1 ) ) < 0.5f && fabsf ( offset ( 0 ) ) < 0.5f )
						{
//							printf( "(%d, %d) is well located\n", x, y );
//							printf( "%d - %f (%f, %f, %f)\n", *begin, optValue, offset( 2 ), offset( 1 ), offset( 0 ) );
							candDOG->getCoors ( begin.getPointer(), x, y );
							callback ( spacing * x, spacing * y, offset ( 2 ), offset ( 1 ), offset ( 0 ), optValue );
//							pressKey();
						}
//						else
//						{
//							printf( "not stable: %d - %f (%f, %f, %f)\n", *begin, optValue, offset( 2 ), offset( 1 ), offset( 0 ) );
//						}
//						pressKey();
					}
				}

				/*
				    return the value of D at the accurate sub-pixel position. The Offset can be found via optOff
				*/
				float fitQuadratic()
				{
//					DBGOUT( DBGALL, "entered fitQuadratic\n" );

//					DBGOUT( DBGALL, "hessian row 1\n" );
					hessian ( 0, 0 ) = cut[22] - 2.0 * cut[13] + cut[4];
					hessian ( 0, 1 ) = 0.25 * ( cut[5] - cut[3] - cut[23] + cut[21] );
					hessian ( 0, 2 ) = 0.25 * ( cut[7] - cut[1] - cut[25] + cut[19] );

//					DBGOUT( DBGALL, "hessian row 2\n" );
					hessian ( 1, 0 ) = hessian ( 0, 1 );
					hessian ( 1, 1 ) = cut[12] - 2 * cut[13] + cut[14];
					hessian ( 1, 2 ) = 0.25 * ( cut[17] - cut[11] - cut[15] - cut[9] );

//					DBGOUT( DBGALL, "hessian row 3\n" );
					hessian ( 2, 0 ) = hessian ( 0, 2 );
					hessian ( 2, 1 ) = hessian ( 1, 2 );
					hessian ( 2, 2 ) = cut[10] - 2 * cut[13] + cut[16];

//					DBGOUT( DBGALL, "negative jacobian\n" );
                                        jacobian ( 0 ) = 0.5 * ( cut[4] - cut[22] );
                                        jacobian ( 1 ) = 0.5 * ( cut[12] - cut[14] );
                                        jacobian ( 2 ) = 0.5 * ( cut[10] - cut[16] );

//					DBGOUT( DBGALL, "hessian\n" );
//					printMat( hessian );
//					DBGOUT( DBGALL, "jacobian\n" );
//					printMat( jacobian );

//					hessian * x = -jacobian, solve this to get x
					              offset = jacobian;
					invert ( hessian );
//					DBGOUT( DBGALL, "inverse hessian\n" );
//					printMat( hessian );
					offset = matMul ( hessian, jacobian );
//					solveLU( hessian, offset );

					return cut[13] + 0.5 * dot ( jacobian, offset );
				}

				/**
				 * calculate harris strength of the cut-buffer
				 */
				bool isEdge()
				{
					const float k = 0.04;

					calcT gxx = cut[16] + cut[10] - 2 * cut[13];
					calcT gyy = cut[14] + cut[12] - 2 * cut[13];
					calcT gxy = calcT ( 0.25 * ( ( cut[17] - cut[15] ) -
					                             ( cut[11] - cut[9] ) ) );

					calcT trace = gxx + gyy;
					trace *= trace;
					calcT result = calcT ( k * trace );
					calcT harris = gxx * gyy - gxy * gxy - result;

//					printf( "%d / %d\n", harris, edgeThresh );
					return harris <= edgeThresh;
				}

				calcT getEdgeThresh() const
				{
					return edgeThresh;
				}

				imgT getThresh() const
				{
					return thresh;
				}


			private:
				imgT thresh; //DOG maximum must exceed this
				AArray<float, 2> hessian;
				AArray<float, 1> jacobian;
				AArray<float, 1> offset;//(scale, y, x)
				ChannelSet<imgT> dzChans;

				imgT cut[27];

				//for harris
				calcT edgeThresh;//edgeness must stay below this
				float wDiag, wOrtho; //weights of a gaussian

				CallbackT &callback;

			private:

				/**
				 * resets the array that holds a portion of the scalespace
				 */
				void resetCut()
				{
					memset ( cut, 0, 27 );
				}


				void fillCut ( int idx, int x, int y, Channel<imgT> *chan, Channel<imgT> *cand )
				{
					assert ( chan );
					if ( chan->getWidth() == cand->getWidth() )
					{
						cut[idx++] = *chan->getPointer ( x - 1, y - 1 );
						cut[idx++] = *chan->getPointer ( x - 1, y );
						cut[idx++] = *chan->getPointer ( x - 1, y + 1 );
						cut[idx++] = *chan->getPointer ( x, y - 1 );
						cut[idx++] = *chan->getPointer ( x, y );
						cut[idx++] = *chan->getPointer ( x, y + 1 );
						cut[idx++] = *chan->getPointer ( x + 1, y - 1 );
						cut[idx++] = *chan->getPointer ( x + 1, y );
						cut[idx++] = *chan->getPointer ( x + 1, y + 1 );
					}
					else if ( chan->getWidth() > cand->getWidth() )
					{
						x <<= 1; y <<= 1;
						cut[idx++] = *chan->getPointer ( x - 2, y - 2 );
						cut[idx++] = *chan->getPointer ( x - 2, y );
						cut[idx++] = *chan->getPointer ( x - 2, y + 2 );
						cut[idx++] = *chan->getPointer ( x, y - 2 );
						cut[idx++] = *chan->getPointer ( x, y );
						cut[idx++] = *chan->getPointer ( x, y + 2 );
						cut[idx++] = *chan->getPointer ( x + 2, y - 2 );
						cut[idx++] = *chan->getPointer ( x + 2, y );
						cut[idx++] = *chan->getPointer ( x + 2, y + 2 );
					}
					else
					{
						x >>= 1; y >>= 1;
						//interpolate
						if ( isEven ( x ) && isEven ( y ) )
						{
							imgT center = *chan->getPointer ( x, y );
							idx++;//diagonals are left blank
							cut[idx++] = imgT ( 0.5 * ( center + *chan->getPointer ( x - 1, y ) ) );
							idx++;
							cut[idx++] = imgT ( 0.5 * ( center + *chan->getPointer ( x, y - 1 ) ) );
							cut[idx++] = center;
							cut[idx++] = imgT ( 0.5 * ( center + *chan->getPointer ( x, y + 1 ) ) );
							idx++;
							cut[idx++] = imgT ( 0.5 * ( center + *chan->getPointer ( x + 1, y ) ) );
//							idx++;
						}
						else if ( isOdd ( x ) && isOdd ( y ) )
						{
							imgT i00 = *chan->getPointer ( x, y );
							imgT i01 = *chan->getPointer ( x, y + 1 );
							imgT i10 = *chan->getPointer ( x + 1, y );
							imgT i11 = *chan->getPointer ( x + 1, y + 1 );

							cut[idx++] = i00;
							cut[idx++] = imgT ( 0.5 * ( i00 + i01 ) );
							cut[idx++] = i01;
							cut[idx++] = imgT ( 0.5 * ( i00 + i10 ) );
							cut[idx++] = imgT ( 0.25 * ( i00 + i01 + i10 + i11 ) );
							cut[idx++] = imgT ( 0.5 * ( i01 + i11 ) );
							cut[idx++] = i10;
							cut[idx++] = imgT ( 0.5 * ( i10 + i11 ) );
							cut[idx++] = i11;
						}
						else if ( isEven ( x ) && isOdd ( y ) )
						{
							imgT a = *chan->getPointer ( x - 1, y );
							imgT b = *chan->getPointer ( x, y );
							imgT c = *chan->getPointer ( x + 1, y );
							imgT d = *chan->getPointer ( x - 1, y + 1 );
							imgT e = *chan->getPointer ( x, y + 1 );
							imgT f = *chan->getPointer ( x + 1, y + 1 );

							idx++;//diagonals are left blank
							cut[idx++] = imgT ( 0.25 * ( a + b + d + e ) );
							idx++;
							cut[idx++] = b;
							cut[idx++] = imgT ( 0.5 * ( b + e ) );
							cut[idx++] = e;
							idx++;
							cut[idx++] = imgT ( 0.25 * ( b + c + e + f ) );
						}
						else
						{
							imgT a = *chan->getPointer ( x, y - 1 );
							imgT b = *chan->getPointer ( x + 1, y - 1 );
							imgT c = *chan->getPointer ( x, y );
							imgT d = *chan->getPointer ( x + 1, y );
							imgT e = *chan->getPointer ( x, y + 1 );
							imgT f = *chan->getPointer ( x + 1, y + 1 );

							idx++;//diagonals are left blank
							cut[idx++] = c;
							idx++;
							cut[idx++] = imgT ( 0.25 * ( a + b + c + d ) );
							cut[idx++] = imgT ( 0.5 * ( c + d ) );
							cut[idx++] = imgT ( 0.25 * ( c + d + e + f ) );
							idx++;
							cut[idx++] = d;
						}
					}
				}

				void fillCut ( imgT *candPtr, Channel<imgT> *cand, Channel<imgT> *prev, Channel<imgT> *next )
				{
					assert ( cand );
					uint x, y;
					cand->getCoors ( candPtr, x, y );
					if ( prev )
					{
						fillCut ( 0, x, y, prev, cand );
					}
					cut[9] = *cand->getPointer ( x - 1, y - 1 );
					cut[10] = *cand->getPointer ( x - 1, y );
					cut[11] = *cand->getPointer ( x - 1, y + 1 );
					cut[12] = *cand->getPointer ( x, y - 1 );
					cut[13] = *cand->getPointer ( x, y );
					cut[14] = *cand->getPointer ( x, y + 1 );
					cut[15] = *cand->getPointer ( x + 1, y - 1 );
					cut[16] = *cand->getPointer ( x + 1, y );
					cut[17] = *cand->getPointer ( x + 1, y + 1 );
					if ( next )
					{
						fillCut ( 18, x, y, next, cand );
					}
				}

				void printCut()
				{
					printf ( "%d %d %d   %d %d %d   %d %d %d\n", cut[0], cut[3], cut[6], cut[9], cut[12], cut[15], cut[18], cut[21], cut[24] );
					printf ( "%d %d %d   %d %d %d   %d %d %d\n", cut[1], cut[4], cut[7], cut[10], cut[13], cut[16], cut[19], cut[22], cut[25] );
					printf ( "%d %d %d   %d %d %d   %d %d %d\n\n", cut[2], cut[5], cut[8], cut[11], cut[14], cut[17], cut[20], cut[23], cut[26] );
//					if ( cut[13] < 0 )
						pressKey();
				}

		};
	}
}

#endif

