

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGSCALESPACER_H
#define ADORE_IMGPROCESSINGSCALESPACER_H

#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/subsampler.h>
#include <adrcore/util/mathutil.h>

namespace adore
{

	namespace core
	{

		/**
		 * A Scalespacer that generates a mixed
		 * smoothing/subsampling pyramid
			 
		*/
		template<class ConvolverT, class PreConvolverT = ConvolverT>
		class ScaleSpacer
		{
			public:
				typedef typename ConvolverT::img_in_type img_in_type;
				typedef typename ConvolverT::img_in_type img_out_type;

				enum SCALE_ACTION
				{
					SCALE_PRESMOOTH,
					SCALE_SMOOTH,
					SCALE_SUBSAMPLE,
					SCALE_STOP
				};

			public:
				ScaleSpacer():convolver( NULL ), preConvolver( NULL )
				{}

				~ScaleSpacer()
				{}

				void process( Channel<img_in_type> *src, Channel<img_out_type> *dst, SCALE_ACTION action );

				ConvolverT *getConvolver()
				{
					return convolver;
				}

				void setConvolver( ConvolverT * c )
				{
					convolver = c;
				}

				PreConvolverT *getPreConvolver()
				{
					return preConvolver;
				}

				void setPreConvolver( PreConvolverT * c )
				{
					preConvolver = c;
				}

			private:

				ConvolverT *convolver;
				PreConvolverT *preConvolver;
				SubSamplerSmooth<img_in_type> subSampler;

		};

		template<class ConvolverT, class PreConvolverT>
		void ScaleSpacer<ConvolverT, PreConvolverT>::process( Channel<img_in_type> *src, Channel<img_out_type> *dst, SCALE_ACTION action )
		{
			assert( convolver );
			switch ( action )
			{
				case SCALE_PRESMOOTH:
					preConvolver->process( src, dst );
					break;
				case SCALE_SMOOTH:
					convolver->process( src, dst );
					break;
				case SCALE_SUBSAMPLE:
					subSampler.process( src, dst );
					break;
				case SCALE_STOP:
				default:
					return;
			}
		}

		/**
		 * returns the amount of necessary presmoothing to ensure a nice oversampled pyramid.
		 * the value is also a good starting scale
		 */
		inline float getPreSmoothingAmount( float deltaT = 0.5F )
		{
			return 0.3333333F * deltaT;
		}

		/**
		 * returns the Step in Scalespace dependent on the level of the Kernel K, the grid spacing of the image
		 * and deltaT, the amount of smoothing done by applying the kernel.
		 */
		inline float getScaleStep( int kernelLevel, int spacing, float deltaT = 0.5F )
		{
			return float( kernelLevel ) * float( spacing ) * deltaT;
		}

		/**
		 * returns the spacing at a given sigma (== sqrt(t)). Also needed is the
		 * subsampling rate rho.
		 * initially, the image has a spcing of 1.
		 * 2, 4, 8 ... follow.
		 */
		inline uint32_t getSpacingAtScale( float rho, float sigma )
		{
			float hmax = rho * sigma;
			if ( hmax < 1 ) return 1;
			uint32_t result = uint32_t( hmax );
			uint32_t hibit = getHiBit( result );
			uint32_t mask = 1 << hibit;

			assert( mask > 0 );
			if ( result == mask ) return mask;

			mask <<= 1;
			assert( mask > 0 );
			return mask;
		}

		inline float getTCycle( int n, int kernelLevel, float deltaT )
		{
			return float( n ) * float( kernelLevel ) * deltaT;
		}

		inline float getSubsamplingFactor( float tCycle )
		{
			return 1.732F /*sqrt(3)*/ / sqrt( tCycle );
		}

		/**
		 * n is the number of smooths
		 */
		inline float getSubSamplingDensity( int n )
		{
			return 2.0F / float( n );
		}

		inline std::pair<int, int> getCorrespondingWindowSizes( int minsize, float scale1, float scale2 )
		{
			std::pair<int, int> sizes;
			if ( scale1 == scale2 )
			{
				sizes.first = sizes.second = minsize;
			}
			else if ( scale1 < scale2 )
			{
				sizes.first = minsize;
				sizes.second = int( ( minsize * scale2 ) / scale1 + 0.5F );
			}
			else
			{
				sizes.second = minsize;
				sizes.first = int( ( minsize * scale1 ) / scale2 + 0.5F );
			}
			return sizes;
		}

	}
} //namespace

#endif
