

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGSIFTAT_H
#define ADORE_IMGPROCESSINGSIFTSAT_H

#include <adrcore/types/aarray.h>
#include <adrcore/types/featuredescriptor.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/channelfunctions.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/printimage.h>
#include <adrcore/util/printmat.h>

namespace adore
{

	namespace core
	{



		/**
		     * This class provides methods to compute rotational invariant feature vectors.
		     * It uses the main orientations in a neighborhood arount the feature point in question.
		     *  
		 */
		template <class imgT>
		class SiftAt
		{
			private:
				typedef float orienT;
				typedef typename signedT< typename calcT<imgT>::type >::type calcT;
				typedef typename FeatureDescriptor::DescVecT DescVecT;
				typedef vector<float> MagCollT;
				typedef vector<orienT> OrientCollT;
			public:
				typedef typename OrientCollT::iterator iterator;
				typedef typename MagCollT::iterator mIterator;
			public:
				/**
				    * Constructor
				    * @param numMainBins the number of histogram bins for the determination of the main direction
				    * @param numSubBins the number of histogram bins in the feature descriptor
				    * @param maxAngle the angles in the orientation images, that must be provided to some fuctions,
				    * are known to lie between [-maxAngle, maxAngle]
				*/
				SiftAt( Channel<imgT> &in, int numMainBins = 32, int _numSubBins = 8, float _maxAngle = 2.0 * M_PI ):
						srcX( 0 ), srcY( 0 ),
						chan( in ),
						maxAngle( _maxAngle ),
						hood( 8 ),
						width( hood + hood + 1 ),
						subHoodWidth( 4 ),
						numSubHoods( 2 * hood / subHoodWidth ),
						numSubBins( _numSubBins ),
						hist( numMainBins ),//for  [0, maxAngle].
						subHist( numSubBins ),
						patch( width, width, 1 ),
						orientPatch( width - 1, width - 1, 1 ),
						magPatch( width - 1, width - 1, 1 ),
						invAffine( 2, 2 ),
						trafo( 2, 2 )
				{
					orients.resize( ( numSubHoods + 2 ) * ( numSubHoods + 2 ) );
					mags.resize( ( numSubHoods + 2 ) * ( numSubHoods + 2 ) );
					invAffine.setIdentity();
					trafo.setIdentity();

//					float binIdx = toHistIdx( 0.01, hist.size() );
//					printf( "binIdx of 0: %f\n", binIdx );
//					binIdx = toHistIdx( 0.5*M_PI + .01, hist.size() );
//					printf( "binIdx of 0.5*M_PI: %f\n", binIdx );
//					binIdx = toHistIdx( M_PI + .01, hist.size() );
//					printf( "binIdx of M_PI: %f\n", binIdx );
//					binIdx = toHistIdx( 2.0*M_PI + .01, hist.size() );
//					printf( "binIdx of 2.0*M_PI: %f\n", binIdx );
				}

				~SiftAt()
				{
				}

				void prepare( float x, float y, AArray<float, 2> &invAffine )
				{
					srcX = x, srcY = y;
					this->invAffine = invAffine;
//					printf( "SiftAt: preparing at (%.2f, %.2f)\n", x, y );

//					invAffine /= ( hood >> 2 );
					warpAffine( chan, x, y, invAffine, patch.begin(), hood, hood );
//					invAffine *= ( hood >> 2 );

//					printChannelRegion( patch, 0, 0, patch.getWidth(), patch.getHeight() );
//					printf( "\n\n" );
					initMagAndOrient();
//					printChannelRegion( orientPatch, 0, 0, orientPatch.getWidth(), orientPatch.getHeight() );
//					printf( "\n\n" );
//					printChannelRegion( magPatch, 0, 0, magPatch.getWidth(), magPatch.getHeight() );
//					printf( "\n\n" );
					getMainOrientations();

//					ostringstream oss;
//					oss << "patch" << *begin() << ".png";
//					patch.save( oss.str().c_str() );

					//for(uint i = 0; i < hist.size(); ++i) printf("%.2f ", hist[i]);
					//printf("\n");
				}

				DescVecT operator()( orienT orientation, bool* ok = NULL )
				{
					//DBGOUT(DBGNORMAL, "SiftAt creates Feat of orientation %f\n", orientation);
					//you need indeed to apply the orientation and not its negative
					float c = cosf( orientation );
					float s = sinf( orientation );
					//multiply the invAffine with a rotation
					trafo( 0, 0 ) = invAffine( 0, 0 ) * c - invAffine( 1, 0 ) * s;
					trafo( 1, 0 ) = invAffine( 0, 0 ) * s + invAffine( 1, 0 ) * c;
					trafo( 0, 1 ) = invAffine( 0, 1 ) * c - invAffine( 1, 1 ) * s;
					trafo( 1, 1 ) = invAffine( 0, 1 ) * s + invAffine( 1, 1 ) * c;

					warpAffine( chan, srcX, srcY, trafo, patch.begin(), hood, hood );

					//ostringstream oss;
					//oss << "rotated" << orientation << ".png";
					//patch.save(oss.str().c_str());

					initMagAndOrient();

//					printChannelRegion( magPatch, 0, 0, magPatch.getWidth(), magPatch.getHeight() );
//					printf( "\n\n" );

					DescVecT result = getFeatureVector();
					hicapFeatureVector( result );
					normalizeFeatureVector( result, ok );
					return result;
				}

				iterator begin()
				{
					return orientations.begin();
				}

				iterator end()
				{
					return orientations.end();
				}

				uint getNumOrientations()
				{
					return orientations.size();
				}

				uint getDescriptorSize()
				{
					return numSubHoods * numSubHoods * numSubBins;
				}

				int getWidth()
				{
					return hood + hood + 1;
				}

				int getHood()
				{
					return hood;
				}

			private:

				/**
				     * This normalizes the feature vector to unit length.
				      * @param featureVector the vector to be normalized.
				     */
				void normalizeFeatureVector ( DescVecT &featureVector, bool* ok = NULL );

				/**
					 * Every entry of the feature vector that exceeds a certain value is set to that value.
					 * After applying this method the vector must be normalized again.
					 * @param featureVector the vector to be hi-capped.
					 * @param fvGradHicap the value not to be exceeded.
					 */
				void hicapFeatureVector ( DescVecT &featureVector, float fvGradHicap = 0.2 );

				//calculate orientation an magnitude
				//magnitude is multiplied with gaussian weights
				void initMagAndOrient()
				{
					float sigma = getSigmaForKernelOfSize( patch.getWidth() );

					uint hood = patch.getWidth() >> 1;
					for ( uint y = 0; y < patch.getHeight() - 1; ++y )
					{
						for ( uint x = 0; x < patch.getWidth() - 1; ++x )
						{
							float a = patch( x  , y );
							float b = patch( x+1, y );
							float c = patch( x  , y+1 );
							float d = patch( x+1, y+1 );
							float xx = 0.5f * ( b - a + d - c );
							float yy = 0.5f * ( c - a + d - b );

							float dx = float( x ) - float( hood ) + 0.5F;
							float dy = float( y ) - float( hood ) + 0.5F;
							float weight = expf( -0.5 * ( dx*dx+dy*dy )/sigma );
//							float weight = 1.0F;
//							printf( "%f  ", weight );

							magPatch( x, y ) = weight * sqrtf( xx * xx + yy * yy );
							orientPatch( x, y ) = atan2f( yy, xx );
						}
//						printf( "\n" );
					}
//					printChannelRegion( magPatch, 0, 0, magPatch.getWidth(), magPatch.getHeight() );printf( "\n" );
//					magPatch.save( "mag.png" );
//					printChannelRegion( orientPatch, 0, 0, orientPatch.getWidth(), orientPatch.getHeight() );printf( "\n" );
//					pressKey();
				}

				template <class OIterator, class MIterator>
				void getOrientationHistogramm( vector<float> &hist, OIterator oBegin,
				                               OIterator oEnd, MIterator mBegin )
				{
					for ( vector<float>::iterator it = hist.begin(); it != hist.end(); ++it ) *it = 0;

					for ( ; oBegin != oEnd; ++oBegin, ++mBegin )
					{
						float mag    = *mBegin;
						if ( mag <= FLT_EPSILON ) continue;

						float orient = *oBegin;

						//linear interpolating weights:
						float binIdx = toHistIdx( orient, hist.size() );
						assert( binIdx >= -10.0 * FLT_EPSILON );
						if ( binIdx < 0.0 ) binIdx = 0.0;
						int idx = ( int )floor( binIdx );
						float factor = binIdx - idx;

//						printf( "orient: %f, binIdx: %f\n", orient, binIdx );

						//if orientation is maxAngle, idx can be equal to binSize
						//in this cas, it has to be set to bin zero.
						assert( ( idx >= 0 && idx < ( int )hist.size() ) || ( idx == ( int )hist.size() && fabs( maxAngle - orient ) < 10 * FLT_EPSILON ) );

						hist[idx % hist.size()] += mag * ( 1.0 - factor );
						hist[( idx + 1 ) % hist.size()] += mag * factor;
					}

					//     for(uint i = 0; i < hist.size(); ++i) printf("%f ", hist[i]);
					//     printf("\n");
				}


				/**
				* calculates all peaks in a vector that are greater than percentage * maxpeak
				*/
				void getPeaks( vector<orienT> &hist, vector<int> &peaks, float percentage );

				/**
				* calcultes a histogram index from a radians
				*/
				inline float toHistIdx( float rad, int numBins )
				{
					float binIdx = rad * float( numBins ) / maxAngle;
					if ( binIdx < 0 ) binIdx += numBins;
					return binIdx;
				}

				/**
				* calcultes a radians from a histogram index
				*/
				inline float fromHistIdx( float idx, int numBins )
				{
					return maxAngle * ( idx ) / float( numBins );
				}

				void getMainOrientations( float fraction = 0.8 );

				/**
				* assumes a correctly rotated patch
				* mag and orient must be initialized
				*/
				DescVecT getFeatureVector()
				{
					int width = orientPatch.getWidth();
					assert( isEven( numSubHoods ) );
					assert( isEven( width ) );
					int dim = numSubHoods + 2;

					//first clear the remains of previous run
					for ( uint y = 1; y <= numSubHoods; ++y )
						for ( uint x = 1; x <= numSubHoods; ++x )
						{
							orients[x + y * dim].clear();
							mags[x + y * dim].clear();
						}

					//
					// for all subHoods: calculate histogram
					//
					DescVecT result( getDescriptorSize() );

					float off = 0.5F + 1.0F / float( subHoodWidth << 1 );
					for ( int y = 0; y < width; ++y )
					{
						for ( int x = 0; x < width; ++x )
						{
							float xFrac = float( x ) / float( subHoodWidth ) + off;
							float yFrac = float( y ) / float( subHoodWidth ) + off;

							int lx = int( floorf( xFrac ) );
							int ly = int( floorf( yFrac ) );
							int ux = lx + 1;
							int uy = ly + 1;
							float dx = xFrac - lx;
							float dy = yFrac - ly;

							/*
							a b
							c d
							*/
							float dPart = dx * dy;
							float aPart = 1.0F - dx - dy + dPart;
							float bPart = dx - dPart;
							float cPart = dy - dPart;

//							printf( "%d, %d -> %d, %d\n", x, y, lx, ly );

							orients[lx + ly * dim].push_back( orientPatch( x, y ) );
							orients[ux + ly * dim].push_back( orientPatch( x, y ) );
							orients[lx + uy * dim].push_back( orientPatch( x, y ) );
							orients[ux + uy * dim].push_back( orientPatch( x, y ) );

							mags[lx + ly * dim].push_back( aPart * magPatch( x, y ) );
							mags[ux + ly * dim].push_back( bPart * magPatch( x, y ) );
							mags[lx + uy * dim].push_back( cPart * magPatch( x, y ) );
							mags[ux + uy * dim].push_back( dPart * magPatch( x, y ) );
						}
					}

					int idx = 0;
					for ( uint y = 1; y <= numSubHoods; ++y )
						for ( uint x = 1; x <= numSubHoods; ++x )
						{
							getOrientationHistogramm( subHist, orients[x + y * dim].begin(),
							                          orients[x + y * dim].end(), mags[x + y * dim].begin() );

							for ( uint i = 0; i < subHist.size(); ++i ) result( idx++ ) = subHist[i];
						}

					return result;
				}

				uint getPatchWidth()
				{
					return hood + hood + 1;
				}

			private:
				float srcX, srcY;
				Channel<imgT> &chan;
				float maxAngle;
				const uint hood;
				const uint width;
				uint subHoodWidth;
				uint numSubHoods;//==hood/subBinWidth
				uint numSubBins;//the angle bins in the subs
				vector<orienT> hist;
				vector<orienT> subHist;
				OrientCollT orientations;//the main ones
				MagCollT magnitudes;//the main ones
				ImagePlane<imgT> patch;
				ImagePlane<float> orientPatch;
				ImagePlane<float> magPatch;
				vector< vector<float> > orients;
				vector< vector<float> > mags;
				AMatrixf invAffine;
				AMatrixf trafo;
		};


		template <class imgT>
		void SiftAt<imgT>::getMainOrientations( float fraction )
		{
			for ( uint i = 0; i < hist.size(); ++i ) hist[i] = 0.0;
			getOrientationHistogramm<ImagePlane<float>::Iterator2D, ImagePlane<float>::Iterator2D>
			( hist, orientPatch.begin(), orientPatch.end(), magPatch.begin() );

//			printf( "histogram is:\n" );
//			for ( uint i = 0; i < hist.size(); ++i ) printf( "%.20f \n", hist[i] );
//			printf( "\n" );

//			printf( "finding peaks\n" );
			vector<int> peaks;
			peaks.reserve( 10 );
			getPeaks( hist, peaks, fraction );

			//this can't be an extremum, if peaks[0] too small
//			assert( peaks[0] > 0.001 );

//			printf( "interpolating peaks\n" );
			//interpolate peaks
			double orientInc, realPeak;
			orientations.clear();
			magnitudes.clear();
			for ( vector<int>::iterator it = peaks.begin(); it != peaks.end(); ++it )
			{
				int idx = *it;
				int leftI = ( idx + hist.size() - 1 ) % hist.size();
				int rightI = ( idx + 1 ) % hist.size();
				if ( fitParabola ( hist[leftI], hist[idx], hist[rightI], orientInc, realPeak ) )
				{
					float orient = fromHistIdx( float( idx ) + orientInc, hist.size() );
//					DBGOUT( DBGNORMAL, "peak %f at orientation %f\n", realPeak, orient );
					orientations.push_back( orient );
					magnitudes.push_back( realPeak );
				}
				else
				{
//					printf( "parabola not fit\n" );
				}
			}
		}

		template <class imgT>
		void SiftAt<imgT>::getPeaks( vector<orienT> &hist, vector<int> &peaks, float percentage )
		{
			int maxBin = 0;
			for ( uint i = 1; i < hist.size(); ++i )
			{
				if ( hist[i] > hist[maxBin] )
				{
					maxBin = i;
				}
			}
			peaks.push_back( maxBin );
			orienT thresh = percentage * hist[maxBin];

			for ( int i = 0; i < maxBin; ++i ) if ( hist[i] > thresh )
				{
					int leftI = ( i + hist.size() - 1 ) % hist.size();
					int rightI = ( i + 1 ) % hist.size();
					if ( hist[leftI] < hist[i] && hist[rightI] < hist[i] ) peaks.push_back( i ); //local maximum
				}
			for ( uint i = maxBin + 1; i < hist.size(); ++i ) if ( hist[i] > thresh )
				{
					int leftI = ( i + hist.size() - 1 ) % hist.size();
					int rightI = ( i + 1 ) % hist.size();
					if ( hist[leftI] < hist[i] && hist[rightI] < hist[i] ) peaks.push_back( i ); //local maximum
				}
		}

		template <class imgT>
		void SiftAt<imgT>::normalizeFeatureVector ( DescVecT &featureVector, bool* ok )
		{
			float norm = 0.0;
			for ( int n = 0 ; n < (int)featureVector.size(); ++n ) norm += featureVector( n ) * featureVector( n );

			norm = sqrt( norm );//TODO check value range
			if ( 0 == norm )
			{
				if ( ok ) *ok = false;
				else assert( false );
			}
			featureVector /= norm;
		}

		template <class imgT>
		void SiftAt<imgT>::hicapFeatureVector ( DescVecT &featureVector, float fvGradHicap )
		{
			for ( int n = 0 ; n < (int)featureVector.size(); ++n )
			{
				if ( featureVector( n ) > fvGradHicap )
				{
					featureVector( n ) = fvGradHicap;
				}
			}
		}


	}
}

#endif
