

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESUBSAMPLER_H
#define ADORE_CORESUBSAMPLER_H

#include <adrcore/imgprocessing/channelprocessor.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>

namespace adore
{

	namespace core
	{

		/**
		 * A function to calculate the size of a dimension (e.g. width or height) after subsampling
		 */
		inline uint getSubSampleDim ( uint oldD )
		{
			return ( oldD >> 1 ) + ( oldD & 1 );
		}

		inline uint getSubSampleDim ( uint oldD, int steps )
		{
			if((oldD & -oldD) >> steps == 0) return (oldD >> steps) + 1;
			else return oldD >> steps;
		}

		/**
			returns the closest coordintate after smooth downsampling steps times
		*/
		inline int down ( uint coor, uint steps )
		{
			return coor >> steps;
		}

		/**
			 
		    subsampling of an image
			works also in place
		*/
		template<class imgT>
		class SubSampler : public ChannelProcessor<1, 1, imgT, imgT>
		{
			public:
				SubSampler() :ChannelProcessor<1, 1, imgT, imgT>()
				{
					;
				}

				~SubSampler()
				{
					;
				}

				void process ( Channel<imgT> *in, Channel<imgT> *out )
				{
					assert ( in );
					assert ( out );
//					assert( out->getROI().getWidth() == getSubSampleDim( in->getROI().getWidth() ) );
//					assert( out->getROI().getHeight() == getSubSampleDim( in->getROI().getHeight() ) );

					typename Channel<imgT>::Iterator2D itIn = in->getIteratorBegin ( 2, 2 );
					typename Channel<imgT>::Iterator2D itOut = out->getIteratorBegin();
					typename Channel<imgT>::Iterator2D itOutE = out->getIteratorEnd();
					for ( ; itOut != itOutE; ++itOut, ++itIn )
					{
						*itOut = *itIn;
					}
				}

		};


		/**
		 * Unlike Subsampler, this one doesn't simple discard every second line/column,
		 * but replaces 4 pixels with their average
		 * works also in place
		 */
		template<class imgT>
		class SubSamplerSmooth : public ChannelProcessor<1, 1, imgT, imgT>
		{
			public:
				SubSamplerSmooth() :ChannelProcessor<1, 1, imgT, imgT>()
				{
					;
				}

				~SubSamplerSmooth()
				{
					;
				}

				void process ( Channel<imgT> *in, Channel<imgT> *out )
				{
					assert ( in );
					assert ( out );
//					assert( out->getROI().getWidth() == getSubSampleDim( in->getROI().getWidth() ) );
//					assert( out->getROI().getHeight() == getSubSampleDim( in->getROI().getHeight() ) );

					assert ( in->getROI().getWidth() > 1 );
					assert ( in->getROI().getHeight() > 1 );

					int x = in->getROI().getX();
					int y = in->getROI().getY();
					int inwidth = in->getROI().getWidth();
					int inheight = in->getROI().getHeight();

					//ensure it's even
					int width = ( inwidth >> 1 ) << 1;
					int height = ( inheight >> 1 ) << 1;

					int outx = out->getROI().getX();
					int outy = out->getROI().getY();
					int outwidth = width >> 1;
					int outheight = height >> 1;

					typename Channel<imgT>::Iterator2D it1    = in->getIteratorH ( y, x, width, 2, 2 );
					typename Channel<imgT>::Iterator2D it2    = in->getIteratorH ( y, x + 1, width, 2, 2 );
					typename Channel<imgT>::Iterator2D it3    = in->getIteratorH ( y + 1, x, width, 2, 2 );
					typename Channel<imgT>::Iterator2D it4    = in->getIteratorH ( y + 1, x + 1, width, 2, 2 );

					typename Channel<imgT>::Iterator2D itOut  = out->getIteratorH ( outy, outx, outwidth );
					typename Channel<imgT>::Iterator2D itOutE = out->getIteratorH ( outy + outheight, outx, outwidth );
					for ( ; itOut != itOutE; ++itOut, ++it1, ++it2, ++it3, ++it4 )
					{
						*itOut = ( ( *it1 ) + ( *it2 ) + ( *it3 ) + ( *it4 ) ) / 4;
					}

					//now it's possible that the last row and col are not processed
					//copy them

					//bottom row
					if ( height != inheight )
					{
//						DBGOUT( DBGNORMAL, "bottom row\n" );
						typename Channel<imgT>::Iterator1D itH1 = in->getXIterator ( y + height, x, 2 );
						typename Channel<imgT>::Iterator1D itH2 = in->getXIterator ( y + height, x + 1, 2 );

						//bottom - 1
						typename Channel<imgT>::Iterator1D itH3 = in->getXIterator ( y + height - 1, x, 2 );
						typename Channel<imgT>::Iterator1D itH4 = in->getXIterator ( y + height - 1, x + 1, 2 );


						typename Channel<imgT>::Iterator1D itOut  = out->getXIterator ( outy + outheight, outx );
						typename Channel<imgT>::Iterator1D itOutE = out->getXIterator ( outy + outheight, outx + outwidth );

						for ( ; itOut != itOutE; ++itOut, ++itH1, ++itH2, ++itH3, ++itH4 )
						{
							*itOut = ( ( *itH1 ) + ( *itH2 ) + ( *itH3 ) + ( *itH4 ) ) / 4;
						}
					}

					//right col
					if ( width != inwidth )
					{
//						DBGOUT( DBGNORMAL, "right col\n" );
						typename Channel<imgT>::Iterator1D itV1 = in->getYIterator ( y, x + width, 2 );
						typename Channel<imgT>::Iterator1D itV2 = in->getYIterator ( y + 1, x + width, 2 );

						//right  - 1
						typename Channel<imgT>::Iterator1D itV3 = in->getYIterator ( y, x + width - 1, 2 );
						typename Channel<imgT>::Iterator1D itV4 = in->getYIterator ( y + 1, x + width - 1, 2 );

						typename Channel<imgT>::Iterator1D itOut  = out->getYIterator ( outy, outx + outwidth );
						typename Channel<imgT>::Iterator1D itOutE = out->getYIterator ( outy + outheight, outx + outwidth );

						for ( ; itOut != itOutE; ++itOut, ++itV1, ++itV2, ++itV3, ++itV4 )
						{
							*itOut = ( ( *itV1 ) + ( *itV2 ) + ( *itV3 ) + ( *itV4 ) ) / 4;
						}
					}

					//bottom right pixel
					if ( height != inheight && width != inwidth )
					{
//						DBGOUT( DBGNORMAL, "bottom right pixel\n" );
						( *out ) ( outx + outwidth, outy + outheight ) = ( *in ) ( width, height ) + ( *in ) ( width - 1, height ) + ( *in ) ( width, height - 1 ) + ( *in ) ( width - 1, height - 1 );
					}
				}

		};

		/**
		 * Like SubSamplerSmooth, but only horizontally
		 * works also in place
		 */
		template<class imgT>
		class SubSamplerHorSmooth : public ChannelProcessor<1, 1, imgT, imgT>
		{
			public:
				SubSamplerHorSmooth() :ChannelProcessor<1, 1, imgT, imgT>()
				{
					;
				}

				~SubSamplerHorSmooth()
				{
					;
				}

				void process ( Channel<imgT> *in, Channel<imgT> *out )
				{
					assert ( in );
					assert ( out );
//					assert( out->getROI().getWidth() == getSubSampleDim( in->getROI().getWidth() ) );
//					assert( out->getROI().getHeight() == in->getROI().getHeight() );

					int x = in->getROI().getX();
					int y = in->getROI().getY();
					int inwidth = in->getROI().getWidth();
					int inheight = in->getROI().getHeight();

					//ensure it's even
					int width = ( inwidth >> 1 ) << 1;
					int height = inheight;

					int outx = out->getROI().getX();
					int outy = out->getROI().getY();
					int outwidth = width >> 1;
					int outheight = height;

//					printf( "%d, %d, %d, %d\n", outx, outy, outwidth, outheight );

					typename Channel<imgT>::Iterator2D it1    = in->getIteratorH ( y, x, width, 2, 1 );
					typename Channel<imgT>::Iterator2D it2    = in->getIteratorH ( y, x + 1, width, 2, 1 );

					typename Channel<imgT>::Iterator2D itOut  = out->getIteratorH ( outy, outx, outwidth );
					typename Channel<imgT>::Iterator2D itOutE = out->getIteratorH ( outy + outheight, outx, outwidth );
					for ( ; itOut != itOutE; ++itOut, ++it1, ++it2 )
					{
//						printf( "%d, %d -> ", *it1, *it2 );
						*itOut = ( ( *it1 ) + ( *it2 ) ) / 2;
//						printf( "%d\n", *itOut );
					}

					//now it's possible that the last row and col are not processed
					//copy them

					//right col
					if ( width != inwidth )
					{
//						DBGOUT( DBGNORMAL, "right col\n" );
						typename Channel<imgT>::Iterator1D itV1 = in->getYIterator ( y, x + width );
						typename Channel<imgT>::Iterator1D itV2 = in->getYIterator ( y, x + width - 1 );

						typename Channel<imgT>::Iterator1D itOut  = out->getYIterator ( outy, outx + outwidth );
						typename Channel<imgT>::Iterator1D itOutE = out->getYIterator ( outy + outheight, outx + outwidth );

						for ( ; itOut != itOutE; ++itOut, ++itV1, ++itV2 )
						{
							*itOut = ( ( *itV1 ) + ( *itV2 ) ) / 2;
						}
					}

				}
		};


		/**
		 * use it, if the new size is not too different from the old one (< times 2)
		 */
		template<class imgT>
		class ReSampler : public ChannelProcessor<1, 1, imgT, imgT>
		{
			public:
				ReSampler() :ChannelProcessor<1, 1, imgT, imgT>()
				{
					;
				}

				~ReSampler()
				{
					;
				}

				void process ( Channel<imgT> *in, Channel<imgT> *out )
				{
					assert ( in );
					assert ( out );
					assert ( in != out );

					int inW = in->getROI().getWidth();
					int inH = in->getROI().getHeight();
					int outW = out->getROI().getWidth();
					int outH = out->getROI().getHeight();

					float scaleX = float( inW ) / float( outW );
					float scaleY = float( inH ) / float( outH );

					BilinearInterpolator<imgT> sampler;

					for ( int y = 0; y < outH; ++y )
					{
						for ( int x = 0; x < outW; ++x )
						{
							float srcX = scaleX * float( x );
							float srcY = scaleY * float( y );

							assert( srcX >= 0 && srcX <= inW - 1 && srcY >= 0 && srcY <= inH - 1 );

							sampler.init( srcX, srcY );
							( *out )( x, y ) = sampler.valueOf( *in );
						}
					}
				}
		};


	}

}

#endif
