

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORETHUMBNAILER_H
#define ADORE_CORETHUMBNAILER_H

#include <adrcore/imgprocessing/subsampler.h>
#include <adrcore/types/view.h>
#include <adrcore/util/unptr.h>

namespace adore
{

	namespace core
	{

		/**
		 * shrinks an image until its smaller extend equals a given value
			 
		*/
		template<class imgT>
		class Thumbnailer
		{
			private:
				typedef ImagePlane<imgT> ImageT;
				typedef ChannelSet<imgT> CSetT;

				int maxExt;

			public:
				/**
				 * @param mExt the maximum extension the shorter image side is allowed to have
				 */
				Thumbnailer( int mExt ) : maxExt( mExt )
				{
					;
				}

				/**
					return false, if image is already small enough
				*/
				bool getDims( int w, int h, int &tw, int &th ) const
				{
					int e = MIN( w, h );
					if ( e <= maxExt ) return false;
					float scale = float( maxExt ) / float( e );
					tw = roundf( scale * w );
					th = roundf( scale * h );
					return true;
				}

				/**
				 * creates a thumbnail through iterative subsampling
				 * if the image was already small enough, NULL is returned
				 */
				ImageT* create( ImageT *img )
				{
					int w = img->getROI().getWidth();
					int h = img->getROI().getHeight();
					int tw, th;
					bool okay = getDims( w, h, tw, th );
					if ( !okay ) return NULL; //already small enough
					ImageT *result = new ImageT( tw, th, img->getNumChannels() );
					create( img, result );
					return result;
				}

				void create( ImageT *img, ImageT *result )
				{
					assert( result );
					int w = img->getROI().getWidth();
					int h = img->getROI().getHeight();
					int tw = result->getWidth();
					int e = MIN( w, h );
					float fsteps = log2f( float( e ) / float( maxExt ) );
					int steps = floorf(fsteps);

					ImageT intermediate( *img, true/*copy*/ );
					CSetT set1( intermediate, true/*own*/ );
					CSetT set2( intermediate, true/*own*/ );
					uint smallW = w;//img->getWidth();
					uint smallH = h;//img->getHeight();
					SubSamplerSmooth<imgT> sub;

					for ( int s = 0; s < steps; ++s )
					{
						uint bigW = smallW;
						uint bigH = smallH;
						smallW = getSubSampleDim ( bigW );
						smallH = getSubSampleDim ( bigH );
						for ( uint i = 0; i < img->getNumChannels(); ++i )
						{
							set1[i]->setROI( 0, 0, bigW, bigH );
							set2[i]->setROI( 0, 0, smallW, smallH );
							sub.process( set1[i], set2[i] );
						}
					}
					intermediate.setROI( 0, 0, smallW, smallH );

					if((int)smallW == tw) //done, just copy
					{
						*result = intermediate;
					}
					else
					{
						assert_msg( float( smallW ) / tw <= 2.0F, "smallW: %d, tw: %d\n", smallW, tw );
						assert_msg( float( smallW ) / tw > 1.0F, "smallW: %d, tw: %d\n", smallW, tw );

						ReSampler<imgT> resample;
						for ( uint i = 0; i < img->getNumChannels(); ++i )
						{
							intermediate.setChannelMode( i );
							result->setChannelMode( i );
							resample.process( &intermediate, result );
						}
						result->setChannelMode( CHANNEL_SPLIT );
					}
				}

		};

		/**
		 * creates new views that contain a thumnail version of the original views' image
		 *
		 */
		template<class ViewIteratorIn, class ViewIteratorOut>
		int createThumbViews( ViewIteratorIn in, int numViews, int maxSize, ViewIteratorOut out )
		{
			typedef typename UnPtr<typename ViewIteratorOut::value_type>::type ViewT;
			typedef typename ViewT::imgT imgT;
			typedef ImagePlane<imgT> ImageT;
			Thumbnailer<imgT> thumb( maxSize );

			int numDone = 0;
			for ( int i = 0; i < numViews; ++i, ++in, ++out )
			{
				ViewT *v = dynamic_cast<ViewT *>( *in );
				ImageT *i = thumb.create( v->getImage() );
				if ( i )
				{
					*out = new ViewT( i, true/*own*/ );
					++numDone;
				}
				else
				{
					*out = NULL;
				}
			}
			return numDone;
		}

	}

}

#endif
