

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin
/// Author: Klaus Haeming


#include "klttracker.h"

#include <cassert>
#include <set>
#include <stack>

#include <adrcore/types/nview.h>
#include <adrcore/types/nviewmap.h>
#include <adrcore/types/viewrelations.h>
#include <adrcore/imgprocessing/rgbtogray.h>
#include <memory>

namespace adore
{
namespace core
{

KltTracker::KltTracker(int mindist, int w_width, int w_height,
	bool sm, bool li, int check, int minval, uint maxpts, uint id, int sRange)
	: minDist(mindist),
	winWidth(w_width),
	winHeight(w_height),
	smoothing(sm),
	lighting(li),
	checking(check),
	minValue(minval),
	searchRange(sRange),
	maxPoints(maxpts),
	fid(id)
{
	assert(-2 < checking && 3 > checking);
}

void KltTracker::setupKlt(KLT_TrackingContext & context) const {
	context->mindist = minDist;
	context->window_width = winWidth;
	context->window_height = winHeight;
	context->smoothBeforeSelecting = smoothing;
	context->lighting_insensitive = lighting;
	context->min_eigenvalue = minValue;
	context->affineConsistencyCheck = checking;
	KLTChangeTCPyramid(context, searchRange);
	KLTUpdateTCBorder(context);
}

void KltTracker::trackPairs(NViewMap & map) {
	ViewRelations& rel = map.getViewRelations();
	size_t nFrames = rel.getNumViews();
	if(!nFrames) {
		cout << "KltTracker::trackPairs: no Views found!" << endl;
		return;
	}
	uint begin = maxNeighbours(rel);

	std::map< uint, vector<int> > values;

	std::map<ViewT*, imgT*> images;
	makeImages(map, images);

	ViewT* v = dynamic_cast<ViewT *>(rel.getView(begin));
	assert(v);
	ImagePlane<imgT>* img = v->getImage(); assert(img);
	uint height = img->getHeight(), width = img->getWidth();

	KLT_TrackingContext context(KLTCreateTrackingContext());
	setupKlt(context);
	KLT_FeatureList f_list(KLTCreateFeatureList(maxPoints));
	KLT_FeatureTable f_table(KLTCreateFeatureTable(nFrames, maxPoints));

	bool selected = false;

	std::map<uint, std::vector<FeaturePoint*> > fpoints;


	std::set<uint> tracked;
	std::stack<uint> toTrack;
	toTrack.push(begin);

	while(!toTrack.empty()) {
		uint id = toTrack.top(); toTrack.pop();

		ViewT* v1 = dynamic_cast<ViewT *>(rel.getView(id));
		assert(v1);
		std::map< ViewT *, imgT* >::const_iterator iit = images.find(v1);
		assert(iit != images.end());
		imgT* img1 = (*iit).second; assert(img1);
		if(!selected) {
			KLTSelectGoodFeatures(context, img1, width, height, f_list);
			KLTStoreFeatureList(f_list, f_table, id);
			fillValues(f_list, values[id]);
			selected = true;
		}

		tracked.insert(id);

		ViewRelations::IdIterator idi = rel.begin(id);
		assert(rel.begin(id) != rel.end(id));
		while(idi != rel.end(id)) {
			KLTExtractFeatureList(f_list, f_table, id);
			uint nei = *idi;
			if(std::find(tracked.begin(), tracked.end(), nei) == tracked.end()) {
				ViewT* v2 = dynamic_cast<ViewT *>(rel.getView(nei));
				assert(v2);
				iit = images.find(v2);
				assert(iit != images.end());
				imgT* img2 = (*iit).second; assert(img2);

				KLTTrackFeatures(context, img1, img2, width, height, f_list);
				KLTReplaceLostFeatures(context, img2, width, height, f_list);
				KLTStoreFeatureList(f_list, f_table, nei);
				fillValues(f_list, values[nei]);
				KLTExtractFeatureList(f_list, f_table, id);
				fillValues(f_list, values[id], values[nei]);
				insertCorrs(id, nei, map, f_table, fpoints, values[nei]);
				toTrack.push(nei);
				tracked.insert(nei);
			}
			++idi;
		}
	}

	KLTFreeFeatureList(f_list);
	KLTFreeFeatureTable(f_table);
	KLTFreeTrackingContext(context);
	releaseImages(images);
}

void KltTracker::insertCorrs(int id1, int id2, NViewMap & map,
	KLT_FeatureTable & f_table, std::map<uint, std::vector<FeaturePoint*> >& fpoints,
	const std::vector<int>& values)
{
	ViewRelations& rel = map.getViewRelations();
	ViewT* v1 = dynamic_cast<ViewT *>(rel.getView(id1));
	ViewT* v2 = dynamic_cast<ViewT *>(rel.getView(id2));
	assert(v1 && v2);

	KLT_FeatureList f_list1(KLTCreateFeatureList(maxPoints));
	KLTExtractFeatureList(f_list1, f_table, id1);
	KLT_FeatureList f_list2(KLTCreateFeatureList(maxPoints));
	KLTExtractFeatureList(f_list2, f_table, id2);

	bool resizing = false;
	if(fpoints.find(id1) == fpoints.end()) resizing = true;
	vector<vector<FeaturePoint*> > pts;
	vector<FeaturePoint*>& fpv1 = fpoints[id1];
	if(resizing) {
		fpv1.resize(maxPoints, 0);
		resizing = false;
	}
	pts.push_back(fpv1);

	if(fpoints.find(id2) == fpoints.end()) resizing = true;
	vector<FeaturePoint*>& fpv2 = fpoints[id2];
	if(resizing) {
		fpv2.resize(maxPoints, 0);
	}
	pts.push_back(fpv2);

	for(uint i = 0; i < maxPoints; ++i) {
		if(KLT_TRACKED != f_list2->feature[i]->val)  {
			continue;
		}

		FeaturePoint* fp1 = fpv1[i];
		bool added = false;
		if(!fp1) {
			fp1 = new FeaturePoint(f_list1->feature[i]->x, f_list1->feature[i]->y, fid);
			fp1->setView(v1);
			if(v1->add(fp1) != GRID_SUCCESS) {
				delete fp1;
				continue;
			}
			added = true;
		}

		FeaturePoint* fp2 = new FeaturePoint(f_list2->feature[i]->x, f_list2->feature[i]->y, fid);
		fp2->setView(v2);
		if(v2->add(fp2) != GRID_SUCCESS) {
			delete fp2;
			if(added) v1->remove(fp1);
			continue;
		}
		if(added) fpv1[i] = fp1;
		fpv2[i] = fp2;
	}


	NView& nv = map.getMasterNView();
	for(std::size_t i = 0; i < fpv1.size(); ++i) {
		if(fpv1[i] && fpv2[i]) {
			vector<FeaturePoint*> vec;
			vec.push_back(fpv1[i]);
			vec.push_back(fpv2[i]);
			nv.addPointCorr(vec.begin(), vec.end(), 1.0f / values[i]);
		}
	}

	KLTFreeFeatureList(f_list2);
	KLTFreeFeatureList(f_list1);
}

uint adore::core::KltTracker::maxNeighbours(const ViewRelations & relations) const
{
	uint num = relations.getNumViews();
	uint max = 0;
	for(uint i = 0; i < num; ++i)
		if(max < relations.getNumNeighbors(i))
			max = i;
	return max;
}

void KltTracker::fillValues(const KLT_FeatureList & f_list, std::vector< int > & values) const
{
	values.resize(f_list->nFeatures, -1);
	for(int i = 0; i < f_list->nFeatures; ++i)
		values[i] = f_list->feature[i]->val;
}

void KltTracker::fillValues(const KLT_FeatureList & f_list,
	const std::vector< int > & old_values, std::vector< int > & values) const
{
	values.resize(f_list->nFeatures, -1);
	for(int i = 0; i < f_list->nFeatures; ++i) {
		if(KLT_TRACKED != values[i])  {
			continue;
		}
		values[i] = old_values[i];
	}
}

void KltTracker::makeImages(const NViewMap & map, std::map< ViewT *, imgT * > & images)
{
	NViewMap::ViewIterator vit = map.getAllViewsBegin();
	while(vit != map.getAllViewsEnd()) {
		ViewT* view = dynamic_cast<ViewT *>(*vit);
		assert(view && !images.count(view));
		images[view] = toGray(view);
		++vit;
	}
}

void adore::core::KltTracker::releaseImages(std::map< ViewT *, imgT * > & images)
{
	std::map<ViewT*, imgT*>::iterator iit = images.begin();
	while(iit != images.end()) {
		delete [] (*iit).second;
		++iit;
	}
}

uint8_t * KltTracker::toGray(ViewT * v) {
	ImagePlane<imgT>* img = v->getImage();
	assert ( img );

	ImagePlane<imgT> * gray = 0;
	if (img->getNumChannels() == 3) {
		gray = new ImagePlane<imgT>((int) v->getWidth(), (int) v->getHeight(), 1);
		RGBtoGray<imgT> converter;
		Channel<imgT>::Set in(true);
		img->createChannels(in);
		converter.process(in, gray);
		img = gray;
	} else if(1 != img->getNumChannels()) return 0;

	auto_ptr< ImagePlane<imgT> > grayDeletor(gray); //does nothing if gray is NULL

	imgT* data = new imgT[(int) v->getWidth() * (int) v->getHeight()];
	assert(data);

	uint height = img->getHeight(), width = img->getWidth();
	for(uint y = 0; y < height; ++y) {
		imgT* scanLine = data + y * width;
		for(uint x = 0; x < width; ++x) {
			imgT* ptr = img->getPointer(x, y);
			*(scanLine + x) = *ptr;
		}
	}

	return data;
}

} // namespace core
} // namespace adore
