

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "filterransac.h"
#include <limits.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/divisiondistortion.h>
#include <adrcore/random/ransac2.h>
#include "filterutil.h"

namespace adore
{

	namespace core
	{

		template<class Solution>
		FilterRANSAC<Solution>::FilterRANSAC ( const double t, const double iF, const int s ) :
				helper ( this, getOuterSampleSize<Solution>() ),
				inlierThresh ( t * t ), //value from HZ, p. 291
				incFactor ( iF ),
				innerSteps ( s ),
				hasBest ( false ),
				optMethod ( INNER_RANSAC )
		{
			assert ( incFactor >= 1.0 );
			assert ( t < 1.0F );//because of mandatory normalization
		}

		template<class Solution>
		FilterRANSAC<Solution>::FilterRANSAC ( const double t, RANSAC_OPT_METHOD method ) :
				helper ( this, getOuterSampleSize<Solution>() ),
				inlierThresh ( t * t ), //value from HZ, p. 291
				incFactor ( 1.0 ),
				innerSteps ( 1 ),
				hasBest ( false ),
				optMethod ( method )
		{
			assert ( t < 1.0F );//because of mandatory normalization
		}

		template<class Solution>
		FilterRANSAC<Solution>::FilterRANSAC ( const FilterRANSAC &fr ) :NVIRANSACClient ( fr ),
				helper ( this, getOuterSampleSize<Solution>() ),
				inlierThresh ( fr.inlierThresh ), //value from HZ, p. 291
				incFactor ( fr.incFactor ),
				innerSteps ( fr.innerSteps ),
				hasBest ( fr.hasBest ),
				optMethod ( fr.optMethod )
		{
		}

		template<class Solution>
		bool FilterRANSAC<Solution>::applySample ( const vector<int> &sample )
		{
			assert ( ( int ) sample.size() == helper.numSamples );

			PointCorr *sam[helper.numSamples];

			for ( int i = 0; i < helper.numSamples; ++i )
			{
				sam[i] = helper.getPointCorr ( sample[i] );
			}

			///calculate model
			int numResults = solCalc.calcSloppy ( &sam[0], helper.numSamples, vito );

			DBGOUT ( DBGALL, "%d solutions\n", numResults );

			if ( numResults < 1 )
			{
				return false;
			}

			///for all results, find the best and compute its inliers
			bool success = false;

			for ( int i = 0; i < numResults; ++i )
			{
				DBGOUT ( DBGALL, "result %d\n", i );
				Solution sol = solCalc.getSolutionSloppy ( i );
				success |= helper.updateSolution ( sol, sample, inlierThresh, false );
			}

			return success;
		}

		template<class Solution>
		bool FilterRANSAC<Solution>::optimize()
		{
			hasBest = true;
			bool success = false;

			switch ( optMethod )
			{
				case NONE:
					success = true;
					break;
				case ITERATIVE:
					success = optimize1();
					break;
				case INNER_RANSAC:
					success = optimize2();
					break;
				default:
					return false;
			}

			return success;
		}

		template<class Solution>
		bool FilterRANSAC<Solution>::optimize1()
		{
			bool success = false;

			const int numIter=innerSteps;
			float thresh = inlierThresh * incFactor;
			const float minThresh = inlierThresh;
			const float step = ( thresh - minThresh ) /float ( numIter - 1 );
			const int minSampleSize = getInnerSampleSize<Solution>();

			for ( int epoch = numIter; epoch > 0; --epoch, thresh -= step )
			{
				const int sampleSize = this->getNumIn();

				if ( sampleSize < minSampleSize ) return success;

				DBGOUT ( DBGNORMAL, "THRESH: %f, samplesize: %d, error: %f\n", thresh, sampleSize, this->getError() );

				PointCorr *sam[sampleSize];

				for ( int idx = 0; idx < sampleSize; ++idx )
				{
					sam[idx] = helper.getPointCorr ( helper.inliers[idx] );
				}

				int numResults = solCalc.calc ( &sam[0], sampleSize, vito );

				DBGOUT ( DBGNORMAL, "numResults: %d\n", numResults );

				if ( numResults < 1 ) return success;

				for ( int i = 0; i < numResults; ++i )
				{
					Solution sol = solCalc.getSolution ( i );
					success |= helper.updateSolution ( sol, helper.inliers, thresh, true );
					DBGOUT ( DBGNORMAL, "inliers: %zd, error %f\n", helper.candInliers.size(), helper.DBGErr );
				}
			}

			return success;
		}

		template<class Solution>
		bool FilterRANSAC<Solution>::optimize2()
		{
			DBGOUT ( DBGALL, "---------------------------------------\n" );
			bool success = false;

			const int numIter=innerSteps;
			float thresh = inlierThresh * incFactor;
			const float minThresh = inlierThresh;
			const float step = ( thresh - minThresh ) /float ( MAX ( numIter - 1, 1 ) );
//			ADRSAC ransac;
			RANSAC2 ransac;
			Inner ranser ( this );

			const int minSampleSize = getInnerSampleSize<Solution>();

			for ( int epoch = numIter; epoch > 0; --epoch, thresh -= step )
			{
				const int sampleSize = this->getNumIn();

				if ( sampleSize < minSampleSize )
				{
					DBGOUT ( DBGNORMAL, "not enough inliers for inner ransac loop\n" );
					return success;
				}

				DBGOUT ( DBGALL, "THRESH: %f, samplesize: %d, error: %f\n", thresh, sampleSize, this->getError() );

				ocorrs.resize ( sampleSize );

				for ( int i = 0; i < sampleSize; ++i )
				{
					ocorrs[i] = helper.getPointCorr ( helper.inliers[i] );
				}

				ranser.hasBest = false;

				ranser.thresh = thresh;
				ranser.helper.init ( ocorrs.begin(), ocorrs.size(), vito );
				assert ( ranser.isOkay() );
				ransac ( &ranser, RansacParams ( 400 ) );

				if ( ranser.success() )
				{
					success |= helper.updateSolution ( ranser.helper.bestSolution, helper.inliers, thresh, true );
					DBGOUT ( DBGALL, "inliers: %zd, error %f\n", helper.candInliers.size(), helper.DBGErr );
				}
				else
				{
					DBGOUT ( DBGNORMAL, "Inner RANSAC failed\n" );
					break;
				}
			}

			DBGOUT ( DBGALL, "---------------------------------------\n" );

			return success;
		}

		template<class Solution>
		NViewInfo *FilterRANSAC<Solution>::createNViewInfo()
		{
			NViewInfo *info = helper.bestSolution.createNViewInfo();
			info->setOkay ( hasBest );
			return info;
		}

//		instantiation

		template class FilterRANSAC<Helper2F>;
		template class FilterRANSAC<FilterRANSAC2ViewHelper<FMatSolution, SampsonF> >;

		template class FilterRANSAC<Helper2RDF>;

		template class FilterRANSAC<HelperTT>;
		template class FilterRANSAC<HelperTTSampson>;

		template class FilterRANSAC< Helper2H >;
		template class FilterRANSAC< Helper2HC >;

		template class FilterRANSAC< Helper2HSampson >;
		template class FilterRANSAC< Helper2HCSampson >;

	}
} //namespace
