

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_AUTOCALIBRATIONAUTOCALIBRATOR2_H
#define ADORE_AUTOCALIBRATIONAUTOCALIBRATOR2_H

#include <adrcore/types/camera.h>
#include <adrcore/types/nview.h>
#include <adrcore/nview/nviewoperator.h>
#include <adrcore/minimization/lm2.h>
#include <adrcore/minimization/lmfunctionjacobi2.h>
#include <adrcore/nview/autocalibrationminfunc2.h>
#include <adrcore/nview/autocalibratorcommon.h>
#include <adrcore/nview/linearqcalculatordummy.h>
#include <adrcore/parameterization/calibrationparamversatile.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/solvesvd.h>

namespace adore
{
	namespace core
	{
		template<class LinQCalc> inline bool hasLinearQCalculator()
		{
			return true;
		}
		template<> inline bool hasLinearQCalculator<LinearQCalculatorDummy>()
		{
			return false;
		}

		/**
		 
		*/
		template<class CalibrationParamT = CalibrationParamVersatile, class LinearQInitializer = LinearQCalculatorDummy>
		class Autocalibrator2 : public NViewOperator
		{
			protected:

				typedef AutocalibrationMinFunc2<CalibrationParamT> MinFuncT;
				typedef LMFunctionJacobi2<MinFuncT > JacobiT;
				typedef ZeroErrorFunc2<double> ErrFuncT;
				typedef LM2<MinFuncT, JacobiT, ErrFuncT, InvCovarUnused<double>::type > lmT;

			public:

				Autocalibrator2() :
						param(), params()
				{}

				Autocalibrator2 ( const Autocalibrator2 &ac ) :
						param ( ac.param ), params()
				{}

				~Autocalibrator2()
				{
					;
				}

				NViewOperator *clone()
				{
					return new Autocalibrator2 ( *this );
				}

				void modify ( NView &nview );

				CalibrationParamT &getCalibrationParam()
				{
					return param;
				}

				const CalibrationParamT &getCalibrationParam() const
				{
					return param;
				}

				/**
				 * if initialize is true, Q is used as a starting point, not only as the result
				 */
				int calcDualQuadric ( NView::CameraIterator camIt, const int numCams, AMatrix &Q, vector<AMatrix> &wi, bool initialize );

			private:

				AutocalibratorCommon ac;

				MinFuncT minFunc;
				ErrFuncT errFunc;
				lmT lm;

				CalibrationParamT param;
				AVector params;

		};

		template<class CalibrationParamT, class LinearQInitializer>
		int Autocalibrator2<CalibrationParamT, LinearQInitializer>::calcDualQuadric ( NView::CameraIterator camIt, const int numCams, AMatrix &Q, vector<AMatrix> &wi, const bool initialize )
		{
			DBGOUT ( DBGALL, "numParams: %d\n", param.getNumParams() );
			if ( (int) params.rows() != param.getNumParams() ) params.resize ( param.getNumParams() );

			if ( initialize )
			{
				DBGOUT ( DBGNORMAL, "Testing the paramtrization of wi:\n" );
//				DBGOUT( DBGNORMAL, "wi before:\n" );
//				for ( uint i = 0; i < wi->size(); ++i ) printMat( ( *wi )[i] );

				param.getInitializedParams ( params, 0, &Q, &wi );

//				param.calculateQuadricAndConics( params, 0, Q, *wi );//debug, too
//				DBGOUT( DBGNORMAL, "wi after:\n" );
//				for ( uint i = 0; i < wi->size(); ++i ) printMat( ( *wi )[i] );
			}
			else param.getInitializedParams ( params, 0 );

//			DBGOUT( DBGNORMAL, "autocalibration is using %d parameter\n", params.rows() );
//			printMat( params );


//			DBGOUT( DBGNORMAL, "cameras in autocalibration:\n" );
//			for ( uint i = 0; i < cams.size(); ++i ) printMat( cams[i]->getProjMat() );

			minFunc.setCalibrationParam ( &param );
			minFunc.setCameras ( camIt, numCams );
			int domainDim = params.rows();
			int rangeDim = minFunc.getRangeDim();

			JacobiT jacobi ( minFunc, domainDim, rangeDim );
			DBGOUT ( DBGNORMAL, "param.getNumParams(): %d, params.rows(): %zd, rangeDim: %d\n", param.getNumParams(), params.rows(), rangeDim );
			lm.solve ( params, minFunc, rangeDim, jacobi, errFunc, lmT::DefaultInvCovar, DBL_EPSILON, 2000 );

			DBGOUT ( DBGNORMAL, "Autocalib: Residual: %.20f - %.20f after %d iterations\n", lm.getInitialError(), lm.getLastError(), lm.getLastIterations() );

			param.calculateQuadricAndConics ( params, 0, Q, wi );
			return 0;
		}

		/**
		 *
		 * @param nviewset
		 */
		template<class CalibrationParamT, class LinearQInitializer>
		void Autocalibrator2<CalibrationParamT, LinearQInitializer>::modify ( NView &nview )
		{
//			assert_msg( param, "You need to provide a CalibrationParamT\n" );
			if ( nview.getNumCameras() <= 0 )
			{
				DBGOUT ( DBGCRITICAL, "No Cameras to calibrate\n" );
				return;
			}

			//got Cameras and 3D-points
			DBGOUT ( DBGNORMAL, "autocalibrating %d cams, %d points\n", nview.getNumCameras(), nview.getNumPoints3D() );

			//allocate the DIACs and Quadric for autocalibration
			vector<AMatrix> wi ( nview.getNumCameras() );
			for ( uint i = 0; i < wi.size(); ++i )
			{
				wi[i].resize ( 3, 3 );
			}
			AMatrix Q ( 4, 4 );
			calculateQ<LinearQInitializer>( nview, Q, ac );

			//since calculateQ does not initialize the ws, we need to do this now.
			//this is, because calcDualQuadric insists on initialized ws also, if we tell it
			//we have initialized data
			if ( hasLinearQCalculator<LinearQInitializer>() )
			{
				ac.enforceRank3SVD ( Q );
				NView::CameraIterator cb = nview.getCamerasBegin();
				for ( uint i = 0; i < wi.size(); ++i, ++cb )
				{
					Camera *cam = *cb;
					ac.computePQPT( Q, cam->getProjMat(), wi[i] );
				}
			}

			calcDualQuadric ( nview.getCamerasBegin(), nview.getNumCameras(), Q, wi, hasLinearQCalculator<LinearQInitializer>() );

			ac.computeMetricReconstruction ( Q, nview.getCamerasBegin(), nview.getNumCameras(), nview.getPoints3DBegin(), nview.getNumPoints3D() );
		}

	}

}

#endif
