

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREBADPOINTREMOVER_H
#define ADORE_COREBADPOINTREMOVER_H

#include <adrcore/nview/nviewoperator.h>
#include <adrcore/util/c++0x.h>

namespace adore
{

	namespace core
	{

		/**
			takes an NView an sorts the correspondences according to their repro error
			from smallest to largest
		*/
		class ReproErrorSorter
		{
			public:
				typedef tuple<PointCorr *, double> ReproT;
				typedef vector<ReproT> ReproColl;
				typedef ReproColl::iterator iterator;

			private:
				class ReproComp
				{
					public:
						bool operator() ( const ReproT &r1, const ReproT &r2 )
						{
							return r1.get<1>() < r2.get<1>();
						}
				};

				ReproColl repros;

			public:
				void init( NView &nview );

				//remove the corr with largest reproerr from list
				//does not delete the corr
				void pop_back()
				{
					repros.pop_back();
				}

				ReproT back()
				{
					return repros.back();
				}

				iterator begin()
				{
					return repros.begin();
				}
				iterator end()
				{
					return repros.end();
				}
		};

		/**
		 * this filters correspondences that posses exceptionally large reprojection errors
		 * The reprojection error of each correspondence is calculated. According to this they
		 * get sorted. the largest gap in this sequence of errors is taken as a candidate
		 * position to remove those correspondence who have larger errors
		 *
			 
		*/
		class BadPointRemover : public NViewOperator
		{
			private:
				ReproErrorSorter repros;

				typedef ReproErrorSorter::iterator ReproIter;
				typedef tuple<ReproIter, double> JumpT;//jump in errorvalue

				const double goodRatio;

			public:

				BadPointRemover ( double ratio = 0.5 ) :goodRatio ( ratio ) {}

				void modify ( NView &nview );
		};

		/**
		 * removes all points whose repro err exceeds a certain amount
		 */
		class BadPointRemover2 : public NViewOperator
		{
			private:
				double reproThresh;
				int removed;//for reporting purposes

			public:

				BadPointRemover2 ( double thresh = DBL_MAX ):NViewOperator(),reproThresh ( thresh ) {}

				BadPointRemover2( const BadPointRemover2& bpr2 ):NViewOperator(), reproThresh( bpr2.reproThresh ) {}

				NViewOperator *clone()
				{
					return new BadPointRemover2( *this );
				}

				void modify ( NView &nview );

				void setThresh( const double thresh )
				{
					reproThresh = thresh;
				}

				double getThresh() const
				{
					return reproThresh;
				}

				int getRemoved() const
				{
					return removed;
				}

		};

		/**
		 * removes all points not in front of all cameras
		 */
		class BadPointRemover3 : public NViewOperator
		{
			private:
				int removed;//for reporting purposes
				bool pretend;

			public:

				/**
					@param p if true, a message is printed how many points would have been removed
				*/
				BadPointRemover3( bool p = false ):NViewOperator(), removed( 0 ), pretend( p ) {}

				BadPointRemover3( const BadPointRemover3& b ):NViewOperator(), removed( 0 ), pretend( b.pretend ) {}

				NViewOperator *clone()
				{
					return new BadPointRemover3( *this );
				}

				void modify ( NView &nview );

				int getRemoved() const
				{
					return removed;
				}

		};

	}

}

#endif
