

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWCAMERACALCULATOR_H
#define ADORE_NVIEWCAMERACALCULATOR_H

#include <adrcore/types/nview.h>
#include <adrcore/types/camera.h>
#include <adrcore/nview/cameraminfunc.h>
#include <adrcore/nview/cameraminjacobi.h>
#include <adrcore/nview/nviewoperator.h>
#include <adrcore/util/debug.h>
#include <vector>
#include <adrcore/util/solvesvd.h>

namespace adore
{
	namespace core
	{

		class TTensorCalculator;
		class FMatrixCalculator;

		using namespace std;

		/**
		 * The objective of this class is to calculate Camera matrices from 3D-Point/2D-Point correspondences.
		 *  
		*/
		class CameraCalculator : public NViewOperator
		{
			public:
				/**
				 * Constructor.
				 */
				CameraCalculator();

				CameraCalculator( const CameraCalculator &cc );

				/**
				 * Destructor.
				 */
				virtual ~CameraCalculator();

				NViewOperator *clone()
				{
					return new CameraCalculator( *this );
				}

				void modify( NView &nview );

				void modify( NView &nview, vector<ViewBase *> &views );

				void modify2( NView &nview, FMatrixCalculator *fcalc = NULL );

				void modify3( NView &nview, TTensorCalculator *tcalc = NULL );
				/**
				 * Calculates a camera matrix from a set of 3D/2D correspondences.
				 * @param pts3D homogeneous column vectors.
				 * @param pts2D homogeneous row or column vectors (the case is detected by inspection).
				 * @param cam Space to hold the result.
				 */
				int calculateCamera( const vector<AVector3D *> & pts3D, const vector<AVector2D *> & pts2D, AMatrix & cam );
				inline int calculateCamera( vector<AVector3D> & pts3D, vector<AVector2D> & pts2D, AMatrix & cam )
				{
					assert( pts2D.size() == pts3D.size() );
					vector<AVector2D*> p2dpv(pts3D.size());
					vector<AVector3D*> p3dpv(pts3D.size());
					for(size_t i = 0; i < pts2D.size(); ++i)
					{
						p2dpv[i] = &pts2D[i];
						p3dpv[i] = &pts3D[i];
					}
					return calculateCamera(p3dpv, p2dpv, cam);
				}

				/**
				 * Calculate a camera from an already set (2n x 12)-matrix of the form
				 *    0    0    0    0 -wX0 -wX1 -wX2 -wX3  yX0  yX1  yX2  yX3
				 * -wX0 -wX1 -wX2 -wX3    0    0    0    0 -xX0 -xX1 -xX2 -xX3
				 * ...
				 * Where the n 2D-Points are (x, y, w) and 3D-Points are (X0, X1, X2, X3)
				 * @param camEq the camera equation.
				 * @param cam Space to hold the result.
				 */
				int calculateCamera( const AMatrix &camEq, AMatrix &cam );

				/**
				 * Minimize the geometric error of a least squares solution obtained by one of the other methods.
				 */
				void minimizeError( const vector<AVector3D *> & pts3D, const vector<AVector2D *> & pts2D, AMatrix & cam );
				/**
				 * calculates all four cameras that belong to an essential matrix
				 * make sure the CamIterator points to valid space
				 */
				template<class CamIterator>
				static void createEssentialCameras( const AMatrix &emat, CamIterator cit )
				{
					SVDStoraged store( emat, SVD_U | SVD_VT );
					AVector &sing = store.D;
					AMatrix &U = store.U;
					AMatrix &VT = store.VT;

					svdUVT( emat, store );
					AVector u = U( Range::all(), 2 );

					cout << sing << endl;

					AMatrix camMat( 3, 4 );
					AMatrix M = camMat( Range::all(), Range( 0, 2 ) );
					AVector t = camMat( Range::all(), 3 );
					AMatrix temp = mulW( VT, false );
					matMul2( U, temp, M );
					t = u;
					Camera *cam = new Camera( camMat );
					*cit = cam; ++cit;

					t = -u;
					cam = new Camera( camMat );
					*cit = cam; ++cit;

					temp = mulW( VT, true );
					matMul2( U, temp, M );
					t = u;
					cam = new Camera( camMat );
					*cit = cam; ++cit;

					t = -u;
					cam = new Camera( camMat );
					*cit = cam;
				}


			private:

				SVDStoraged store;
				CameraMinFunc cameraMinFunc;

				/**
				 * a helper function for createEssentialCameras
				 * multiplicate a matrix with
				 * 0 -1  0
				 * 1  0  0
				 * 0  0  1
				 * or the tranpose
				 */
				static AMatrix mulW( const AMatrix & V, bool trans );

		};




	}

}

#endif
