

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWCAMERAPOINTDUALITY_H
#define ADORE_NVIEWCAMERAPOINTDUALITY_H

#include <adrcore/util/mathutil.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/types/camera.h>

namespace adore
{
	namespace core
	{

		/**
		 * In this class functions related to the Carlsson-Weinshall duality can be found.
		 *
		 *  
		*/
		class CameraPointDuality
		{
			public:
				/**
				 * Constructor.
				 */
				CameraPointDuality();

				/**
				 * Destructor.
				 */
				~CameraPointDuality();

				/**
				 * Retrieve the dual point from a camera.
				 * @param cam the camera.
				 * @param pt the resulting dual.
				 */
				AVector cameraToPoint( const AMatrix &cam ) const;

				/**
				 * Retrieve the dual camera from a point.
				 * @param pt the point
				 * @param cam the resulting dual camera.
				 */
				AMatrix pointToCamera( const AVector &pt ) const;

				/**
				 * The canonical X1 of algorithm 20.1 (p.511) of the HZ-Book
				 * @return X1
				 */
				const AVector & getCanonicalX1() const
				{
					return canonicalX1;
				}
				/**
				 * The canonical X3 of algorithm 20.1 (p.511) of the HZ-Book
				 * @return X3
				 */
				const AVector & getCanonicalX2() const
				{
					return canonicalX2;
				}
				/**
				 * The canonical X4 of algorithm 20.1 (p.511) of the HZ-Book
				 * @return X4
				 */
				const AVector & getCanonicalX3() const
				{
					return canonicalX3;
				}
				/**
				 * The canonical X5 of algorithm 20.1 (p.511) of the HZ-Book
				 * @return X5
				 */
				const AVector & getCanonicalX4() const
				{
					return canonicalX4;
				}
				/**
				 * The canonical X6 of algorithm 20.1 (p.511) of the HZ-Book
				 * @return X6
				 */
				const AVector & getCanonicalX5() const
				{
					return canonicalX5;
				}

				/**
				 * Here's the funtion you need to compute X2. Is does exactly what it's name tells you.
				 * @param fmatrix a reduced fundametal matrix.
				 * @return dual the resulting dual camera.
				 */
				AVector calculateReducedCameraDualFromReducedFMatrix( const AMatrix &fmatrix );

				int reconstructFrom6PointCorrespondences( const std::vector<std::vector<AVector2D> > &points, AVector &X6_1, AVector &X6_2, AVector &X6_3 );

			protected:

				AVector canonicalX1;
				AVector canonicalX2;
				AVector canonicalX3;
				AVector canonicalX4;
				AVector canonicalX5;

				SVDStoraged store3;
				SVDStoraged store4;
				SVDStoraged store10_12;
				SVDStoraged store3_5;

				AMatrix fmatrixT;
				AMatrix fshuffle;
				AMatrix ratioMat;

				//6Point-Algo
				AMatrix M;
				AMatrix Q;
				AVector tmpPt;
				AMatrix tmpCam;
				AMatrix tmpA4x4;
				AMatrix tmpB4x4;
				AMatrix cam1;
				AMatrix cam2;
				AMatrix w;
				AMatrix designMat;

				void initFirst8ColumnsOfM();
				void initLast4ColumnsOfM( double x1, double y1,
				                          double x2, double y2,
				                          double x3, double y3,
				                          double x4, double y4,
				                          double x5, double y5 );
				void calculateQ( const AMatrix &cam1, const AMatrix &cam2, const double x, const double y );
				void wFrom6Points( double x1, double y1,
				                   double x2, double y2,
				                   double x3, double y3,
				                   double x4, double y4,
				                   double x5, double y5,
				                   double x6, double y6,
				                   AMatrix &w, int rowOfW
				                 );
				void convertLineCam( const AVector &lineCam, AMatrix &cam );

				int solveForAlpha( const AMatrix &V_T5, double &alpha1, double &alpha2, double &alpha3 );

				void computeX6( double a, double b, double c, double d, double e, AVector &X6 );

				double getSummand3( double x[2], double y[2], double z[2] )
				{
					return ( x[0]-x[1] ) * ( y[0] - y[1] ) * ( z[0] - z[1] );
				}

				double getSummand2( double x[2], double y[2], double z[2] )
				{
					return ( x[0]-x[1] ) * ( y[0] - y[1] ) * z[1] +
					       ( x[0]-x[1] ) * y[1] * ( z[0] - z[1] ) +
					       x[1] * ( y[0] - y[1] ) * ( z[0] - z[1] );
				}

				double getSummand1( double x[2], double y[2], double z[2] )
				{
					return x[1] * ( y[0] - y[1] ) * z[1] +
					       ( x[0]-x[1] ) * y[1] * z[1] +
					       x[1] * y[1] * ( z[0] - z[1] );
				}
		};

	}

}

#endif
