

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_VIEWMATCHCORRESPONDENCECORRECTOR_H
#define ADORE_VIEWMATCHCORRESPONDENCECORRECTOR_H

#include <adrcore/nview/fmatrixcalculator.h>
#include <adrcore/util/mathutil.h>

namespace adore
{
	namespace core
	{

		/**
		 * Given a fundamental matrix and two corresponding points with inaccurate coordinates, this class
		 * corrects the coordintes of the points to their maximum likelihood position.
		 * This class is a Functor using Alg. 12.1, page 318 of Hartley-Zisserman.
		 *
		 *  
		 */
		class CorrespondenceCorrector
		{
			public:
				/**
				 * Simple Constructor.
				 */
				CorrespondenceCorrector();

				/**
				 * Simple Destructor.
				 */
				~CorrespondenceCorrector();

				/**
				 * One incarnation of the main function of this class. The pair of points are corrected according to the fundamental matrix.
				 * Because their values are altered, you may have to copy them before in order to save them.
				 * Please notice, that the coordinates must not be homogeneous.
				 * @param p1 The first of the corresponding points.
				 * @param p2 The second of the corresponding points.
				 * @param fmat The fundamental matrix.
				 * @return true, if succeeded, false otherwise
				 */
				bool operator()( AVector2D &p1, AVector2D &p2, AMatrix &fmat );

				/**
				 * One incarnation of the main function of this class. The coordinates are corrected according to the fundamental matrix.
				 * Because their values are altered, you may have to copy them before in order to save them.
				 * Please notice, that the coordinates must not be homogeneous.
				 * @param p1x The x-coordinte of the first of the corresponding points.
				 * @param p1y The y-coordinte of the first of the corresponding points.
				 * @param p2x The x-coordinte of the second of the corresponding points.
				 * @param p2y The y-coordinte of the second of the corresponding points.
				 * @param fmat The fundamental matrix.
				 * @return true, if succeeded, false otherwise
				 */
				bool operator()( double &p1x, double &p1y, double &p2x, double &p2y, AMatrix &fmat );

			private:

				FMatrixCalculator fcalc;

				AMatrix T1;
				AMatrix T2;
				AMatrix T1inv;
				AMatrix T2inv;
				AMatrix R1;
				AMatrix R2;
				double rootsR[6];
				double rootsI[6];

				AVector x1, x2, e1, e2;//optimal points
				AMatrix Ftemp1, Ftemp1T, Ftemp2;

				//vars of g(t)
				double f1, f2, a, b, c, d;
                                std::vector<double> abcd;
				//coefficients of t^0 ... t^6
				float t[7];

				/**
				 * Calculates the coefficients of g(t), the derivation of s(t). (Alg. 12.1, page 318)
				 */
				void prepareGt();

				/**
				 * Calculates the roots of g(t), the derivation of s(t). (Alg. 12.1, page 318)
				 * @param rootsR the real part of the roots is stored here.
				 * @param rootsI the imaginary part of the roots is stored here.
				 */
				bool getRootsOfGt( double rootsR[6], double rootsI[6] );

				/**
				 * Evaluates s(t). (Alg. 12.1, page 318)
				 * @param t s(t) is evaluated.
				 */
				double getStAt( double t );

				/**
				 * Evaluates g(t). For debugging.
				 * @param t g(t) is evaluated.
				 */
				double getGtAt( double t );

				/**
				 * there are many multiplications it it may happen that numbers reach infty
				 * in this case the also return true (also if nan)
				 *
				 */
				bool reachedInfty()
				{
					for ( int i = 0; i < 7; ++i )
					{
						if ( isinff( t[i] ) || isnanf( t[i] ) ) return true;
					}
					return false;
				}

		};

	}

}

#endif
