

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_VIEWMATCHCORRESPONDENCECORRECTORP_H
#define ADORE_VIEWMATCHCORRESPONDENCECORRECTORP_H

#include <adrcore/nview/projectivitycalculator.h>
#include <adrcore/util/mathutil.h>

	namespace adore
	{
		namespace core
		{

			/**
			 * Given a projectivity and two corresponding points with inaccurate coordinates, this class
			 * corrects the coordintes of the points to their maximum likelihood position.
			 *
			 *  
			 */
			class CorrespondenceCorrectorP
			{
				public:
					/**
					 * Simple Constructor.
					 */
					CorrespondenceCorrectorP();

					/**
					 * One incarnation of the main function of this class. The pair of points are corrected according to the fundamental matrix.
					 * Because their values are altered, you may have to copy them before in order to save them.
					 * Please notice, that the coordinates must not be homogeneous.
					 * @param p1 The first of the corresponding points.
					 * @param p2 The second of the corresponding points.
					 * @param fmat The fundamental matrix.
					 * @return true, if succeeded, false otherwise
					 */
					bool operator()( AVector2D &p1, AVector2D &p2, const AMatrix &H );

					bool correct( AVector2D &p1, AVector2D &p2, const AMatrix &H );

					/**
					 * One incarnation of the main function of this class. The coordinates are corrected according to the fundamental matrix.
					 * Because their values are altered, you may have to copy them before in order to save them.
					 * Please notice, that the coordinates must not be homogeneous.
					 * @param p1x The x-coordinte of the first of the corresponding points.
					 * @param p1y The y-coordinte of the first of the corresponding points.
					 * @param p2x The x-coordinte of the second of the corresponding points.
					 * @param p2y The y-coordinte of the second of the corresponding points.
					 * @param fmat The fundamental matrix.
					 * @return true, if succeeded, false otherwise
					 */
					bool operator()( double &p1x, double &p1y, double &p2x, double &p2y, const AMatrix &H );

					void prepare( const AMatrix &H );

					bool correct( double &p1x, double &p1y, double &p2x, double &p2y, const AMatrix &H );

				private:

					ProjectivityCalculator fcalc;

					AMatrix B;
					AMatrix C;
					double cosA, sinA;
					double rootsR[8];
					double rootsI[8];

					//coefficients of t^0 ... t^8
					float p[9];
					//the parameters to determin the coefficients
					double q1, q2, q3, q5, q6, q7, q9;
					//and their squares
					double q1_2, q2_2, q3_2, q5_2, q6_2, q7_2, q9_2;

					/**
					 * Calculates the coefficients of g(t), the derivation of s(t). (Alg. 12.1, page 318)
					 */
					void prepareGt();

					/**
					 * Calculates the roots of g(t), the derivation of s(t). (Alg. 12.1, page 318)
					 * @param rootsR the real part of the roots is stored here.
					 * @param rootsI the imaginary part of the roots is stored here.
					 */
					bool getRootsOfGt( double rootsR[8], double rootsI[8] );

					double getY( double x );

					double getError( double x, double y );

					/**
					 * there are many multiplications it it may happen that numbers reach infty
					 * in this case the also return true (also if nan)
					 *
					 */
					bool reachedInfty()
					{
						for ( int i = 0; i < 9; ++i )
						{
							if ( isinff( p[i] ) || isnanf( p[i] ) ) return true;
						}
						return false;
					}

			};

		}

	}

#endif
