

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREDEFAULTOPCOLLECTION_H
#define ADORE_COREDEFAULTOPCOLLECTION_H

#include <adrcore/nview/nviewoperator.h>
#include <adrcore/nview/quasiaffinator.h>
#include <adrcore/nview/point3dcalculator.h>
#include <adrcore/nview/baprojective.h>
#include <adrcore/nview/autocalibratorlinear.h>
#include <adrcore/nview/autocalibrator2.h>
#include <adrcore/nview/linearqcalculatorfl.h>
#include <adrcore/nview/badpointremover.h>
#include <adrcore/viewmatch/epilinematcher.h>
#include <adrcore/nview/tensorupdater.h>

namespace adore
{

	namespace core
	{

		/**
			recalculates the fmatix from cameras
		*/
		class FMatRecalc : public NViewOperator
		{
			public:
				FMatRecalc() {}
				FMatRecalc ( const FMatRecalc & ) {}
				NViewOperator *clone()
				{
					return new FMatRecalc ( *this );
				}

				void modify ( NView &nv )
				{
					NView::ViewIterator vit = nv.getViewsBegin();
					Camera *cam1 = nv.getCamera ( *vit ); ++vit;
					assert ( cam1 );
					Camera *cam2 = nv.getCamera ( *vit );
					assert ( cam2 );
					AMatrix & fmat = nv.getInfo()->getFMatrix();
					FMatrixCalculator::calculateFundamentalMatrix2Cams ( cam1->getProjMat(), cam2->getProjMat(), fmat );
				}
		};

		/**
			combines the basic projective reconstruction steps
		*/
		class ProjectiveStep : public NViewOperator
		{
			private:
				TensorUpdater tup;
				BAProjectiveSimple BA;
				CameraCalculator calcCam;
				Point3DCalculator calc3D;

			public:

				ProjectiveStep()
				{
					init();
				}
				ProjectiveStep ( const ProjectiveStep & )
				{
					init();
				}
				NViewOperator *clone()
				{
					return new ProjectiveStep ( *this );
				}

				void init()
				{
					tup.setMinAlgebraic ( true );
					calc3D.setMinimizeError ( true );
					BA.setThresh ( 0.1 * FLT_EPSILON );
				}

				void modify ( NView &nv )
				{
					tup.modify ( nv );
					calcCam.modify ( nv );
					calc3D.modify ( nv );
					BA.modify ( nv );
				}
		};

		/**
			This one captures a good procedure for inital (pre-merge, post-filter) reconstruction
			 
		*/
		class DefaultOpCollection : public NViewOperator
		{
			private:
				AutocalibratorLinear<LinearQCalculatorFL> lincalibrator;
				Autocalibrator2<CalibrationParamVersatile> calibrator;
				BAProjectiveSimple BA;
				CameraCalculator calcCam;
				Point3DCalculator calc3D;
				BadPointRemover2 bpr2;
				ReproErrorSorter sorter;
				TensorUpdater tup;
				bool initRepro;
				bool calibrate;
				bool caliTries;

			public:
				DefaultOpCollection ( float normalizedThresh ) :bpr2 ( normalizedThresh ), initRepro ( true ), calibrate ( true ), caliTries ( 10 )
				{
					init();
				}

				DefaultOpCollection ( const DefaultOpCollection &doc ) :bpr2 ( doc.bpr2.getThresh() ),initRepro ( true ), calibrate ( doc.calibrate ), caliTries ( doc.caliTries )
				{
					init();
				}

				NViewOperator *clone()
				{
					return new DefaultOpCollection ( *this );
				}

				void projRecon ( NView &nv )
				{
					tup.modify ( nv );
					/*					ATensor3 t(3, 3, 3);
										t = nv.getInfo()->getTTensor();
										normalize(t, 1.0);
										nv.getInfo()->setTTensor(t);*/
					calcCam.modify ( nv );
					calc3D.modify ( nv );
					BA.modify ( nv );
				}

				void modify ( NView &nv )
				{
					projRecon ( nv );
					bpr2.modify ( nv );
					projRecon ( nv );
					bpr2.modify ( nv );

					if ( calibrate )
					{
						initRepro = true;
						int invisible = 0;
						int tries = 0;
						do
						{
							BA.modify ( nv );
							//lincalibrator.modify(nv);
							calibrator.modify ( nv );
							invisible = QuasiAffinator::countInvisible ( nv );
							if ( invisible > 0 )
							{
								DBGOUT ( DBGNORMAL, "DefaultOpCollection: incoherent reconstruction (%d invis), trying to improve\n", invisible );
								removeMostEvil ( nv );
							}
						}
						while ( invisible > 0 && tries++ < 10 );
					}
				}

				void setCalibrate ( bool theValue )
				{
					calibrate = theValue;
				}

				bool setCalibrate() const
				{
					return calibrate;
				}

				void setNumCalibrationTries ( bool theValue )
				{
					caliTries = theValue;
				}


				bool getNumCalibrationTries() const
				{
					return caliTries;
				}


			private:
				void removeMostEvil ( NView &nv )
				{
					if ( initRepro )
					{
						initRepro = false;
						sorter.init ( nv );
					}
					ReproErrorSorter::ReproT worst = sorter.back();
					nv.removePointCorrAnd3D ( worst.get<0>() );
					sorter.pop_back();
				}

				void init()
				{
					BA.setThresh ( 0.01 * FLT_EPSILON );
					tup.setMinAlgebraic ( true );
					calc3D.setMinimizeError ( true );

					CalibrationParamVersatile &calibParam = calibrator.getCalibrationParam();
					SoftConstraintSet &constSoftConstraints = calibParam.getCONSTSoftConstraints();
					CamParamChoice &metricChoice = calibParam.getChoice();

					metricChoice.setFLyMode ( FREE_CONST );
					metricChoice.setARMode ( FREE_CONST );
					metricChoice.setPPxMode ( FREE_CONST );
					metricChoice.setPPyMode ( FREE_CONST );
					metricChoice.setSkewMode ( HARD_CONST );
					metricChoice.setSkew ( 0.0 );

					constSoftConstraints.addConstraint ( SoftConstraint ( FL, 1.0, 3.0 ), 0 );
					constSoftConstraints.addConstraint ( SoftConstraint ( AR, 1.0, 0.1 ), 0 );
					constSoftConstraints.addConstraint ( SoftConstraint ( PPX, 0.0, 0.1 ), 0 );
					constSoftConstraints.addConstraint ( SoftConstraint ( PPY, 0.0, 0.1 ), 0 );
				}
		};

	}

}

#endif
