

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <adrcore/nview/relativeorientationcalculator.h>

#include "homographycalculator.h"

#include <adrcore/util/mathutil.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/types/camera.h>

#include <iostream>
using namespace std;

namespace adore {
	namespace core {

	AMatrix HomographyCalculator::calculateHomographyAlgebraic ( const AMatrix & cam1, const AMatrix & cam2, const vector< AVector* > & pts3D1, const vector< AVector* > & pts3D2 )
	{
		assert( pts3D1.size() == pts3D2.size() );
		assert( pts3D1.size() >= 2 );

		AMatrix eq( 3 * pts3D1.size(), 4 );
		AVector c( 3 * pts3D1.size() );

		AMatrix pseudo = pseudoInvertDeficient( cam1 );
		AMatrix H(4, 4);
		matMul2( pseudo, cam2, H );

		//calculate camera center
		AVector center1 = getVectorOfMinors3x4( cam1 );
//		cout << cam1 << center1 << endl;
//		AMatrix ccc = cam1.copy();
//		normalizeExponents( ccc.begin(), ccc.end() );
//		center1 = getVectorOfMinors3x4( ccc );
//		cout << cam1 << center1 << endl;

		eq = 0;
		int eqIdx = 0;
		for ( uint p = 0; p < pts3D1.size(); ++p )
		{
			AVector &pt1 = *pts3D1[p];
			AVector &pt2 = *pts3D2[p];
			AVector a = matMul( H, pt2 );
			double b;

			for ( int i = 0; i < 3; ++i )
			{
				b = center1( i ) * pt1( 3 ) - center1( 3 ) * pt1( i );
				c( eqIdx ) = pt1( i ) * a( 3 ) - pt1( 3 ) * a( i );
				eq( eqIdx, 0 ) =  b * pt2( 0 );
				eq( eqIdx, 1 ) =  b * pt2( 1 );
				eq( eqIdx, 2 ) =  b * pt2( 2 );
				eq( eqIdx, 3 ) =  b * pt2( 3 );
				++eqIdx;
			}
		}

		AVector v = solveSVD( eq, c );

		for ( int row = 0; row < 4; ++row )
			for ( int col = 0; col < 4; ++col )
			{
				H( row, col ) = H( row, col ) + center1( row ) * v( col );
			}

		return H;
	}

	void mul(const AVector& a, const AVector& b, AMatrix& A)
	{
		assert(a.rows() == b.rows());
		assert(A.rows() == A.cols() && A.rows() == a.rows());
                for(unsigned i = 0; i < a.rows(); ++i)
		{
			for(unsigned j = 0; j < a.rows(); ++j)
			{
				A(i, j) = a(i) * b(j);
			}
		}
	}

	bool get_scale(const vector<AVector *>& a, const vector<AVector *>& b, double& scale)
	{
		double asum = 0, bsum = 0;
		for(uint i = 0; i < a.size(); ++i)
		{
			const AVector& pt1 = *a[i];
			const AVector& pt2 = *b[i];
			asum += pt1(0) * pt1(0) + pt1(1) * pt1(1) + pt1(2) * pt1(2);
			bsum += pt2(0) * pt2(0) + pt2(1) * pt2(1) + pt2(2) * pt2(2);
		}

		if(asum <= FLT_EPSILON || bsum <= FLT_EPSILON)
			return false;

		scale = sqrt(asum / bsum);
		return true;
	}

	bool HomographyCalculator::calculateAssumingMetricUsePoints (const AMatrix &cam1, const AMatrix &cam2, const vector<AVector *> &pts1, const vector<AVector *> &pts2, AMatrix& H)
	{
		assert(pts1.size() == pts2.size());

		Camera cam_1(cam1);
		cam_1.decompose();
		const AMatrix& R1 = cam_1.getRotationMatrix();
		const AVector& t1 = cam_1.getLocation();

		AVector tmp3(3);
		matMul2 ( R1, t1, tmp3, -1.0 );

		AMatrix RT(4, 4);
		RT =
				R1(0, 0), R1(0, 1), R1(0, 2), tmp3(0),
				R1(1, 0), R1(1, 1), R1(1, 2), tmp3(1),
				R1(2, 0), R1(2, 1), R1(2, 2), tmp3(2),
				0, 0, 0, 1;
		invert(RT);

		AMatrix cam44(4, 4);
		cam44(Range(0, 2), Range(0, 4)) = cam1;
		cam44(3, 0) = cam44(3, 1) = cam44(3, 2) = 0;
		cam44(3, 3) = 1.0;

		AMatrix T1(4, 4);
		matMul2(RT, cam44, T1);

		Camera cam_2(cam2);
		cam_2.decompose();

		const AMatrix& R2 = cam_2.getRotationMatrix();
		const AVector& t2 = cam_2.getLocation();

		matMul2 ( R2, t2, tmp3, -1.0 );

		RT =
				R2(0, 0), R2(0, 1), R2(0, 2), tmp3(0),
				R2(1, 0), R2(1, 1), R2(1, 2), tmp3(1),
				R2(2, 0), R2(2, 1), R2(2, 2), tmp3(2),
				0, 0, 0, 1;
		invert(RT);

		cam44(Range(0, 2), Range(0, 4)) = cam2;

		AMatrix T2(4, 4);
		matMul2(RT, cam44, T2);

		vector<AVector*> points1(pts1.size());
		vector<AVector*> points2(pts2.size());
		AVector tmp4(4), tmp4_1(4);
		for(uint i = 0; i < pts1.size(); ++i)
		{
			const AVector & pt1 = *pts1[i];
			tmp4 = pt1(0) / pt1(3), pt1(1) / pt1(3), pt1(2) / pt1(3), 1.0;
			matMul2(T1, tmp4, tmp4_1);

			AVector * v1 = new AVectord(3);
			(*v1) = tmp4_1(0) / tmp4_1(3), tmp4_1(1) / tmp4_1(3), tmp4_1(2) / tmp4_1(3);
			points1[i] = v1;

			const AVector & pt2 = *pts2[i];
			tmp4 = pt2(0) / pt2(3), pt2(1) / pt2(3), pt2(2) / pt2(3), 1.0;
			matMul2(T2, tmp4, tmp4_1);

			AVector * v2 = new AVectord(3);
			(*v2) = tmp4_1(0) / tmp4_1(3), tmp4_1(1) / tmp4_1(3), tmp4_1(2) / tmp4_1(3);
			points2[i] = v2;
		}

		RelativeOrientationCalculator calc;
		bool ok = calc.calculate(points1, points2);

		for(uint i = 0; i < points1.size(); ++i)
		{
			delete points1[i];
			delete points2[i];
		}

		if(!ok) return false;

		const AMatrix & R = calc.rotation();
		const AVector & t = calc.translation();
		const double s = calc.scale();

		AMatrix temp1(4, 4);
		temp1 =
				s, 0, 0, t(0),
				0, s, 0, t(1),
				0, 0, s, t(2),
				0, 0, 0, 1;

		AMatrix temp2(4, 4);
		temp2 = 0;
		temp2(Range(0, 2), Range(0, 2)) = R;
		temp2(3, 3) = 1;

		matMul2(temp1, temp2, H);

		invert(T1);
		matMul2(T1, H, temp1);
		matMul2(temp1, T2, H);

		return true;
	}

	bool HomographyCalculator::calculateAssumingMetricUseAll (const AMatrix &cam1, const AMatrix &cam2, const vector<AVector *> &pts1, const vector<AVector *> &pts2, AMatrix& H)
	{
		assert(pts1.size() == pts2.size());

		Camera cam_1(cam1);
		cam_1.decompose();
		const AMatrix& R1 = cam_1.getRotationMatrix();
		const AVector& t1 = cam_1.getLocation();

		AVector tmp3(3);
		matMul2 ( R1, t1, tmp3, -1.0 );

		AMatrix RT(4, 4);
		RT =
				R1(0, 0), R1(0, 1), R1(0, 2), tmp3(0),
				R1(1, 0), R1(1, 1), R1(1, 2), tmp3(1),
				R1(2, 0), R1(2, 1), R1(2, 2), tmp3(2),
				0, 0, 0, 1;
		invert(RT);

		AMatrix cam44(4, 4);
		cam44(Range(0, 2), Range(0, 4)) = cam1;
		cam44(3, 0) = cam44(3, 1) = cam44(3, 2) = 0;
		cam44(3, 3) = 1.0;

		AMatrix T1(4, 4);
		matMul2(RT, cam44, T1);

		Camera cam_2(cam2);
		cam_2.decompose();

		const AMatrix& R2 = cam_2.getRotationMatrix();
		const AVector& t2 = cam_2.getLocation();

		matMul2 ( R2, t2, tmp3, -1.0 );

		RT =
				R2(0, 0), R2(0, 1), R2(0, 2), tmp3(0),
				R2(1, 0), R2(1, 1), R2(1, 2), tmp3(1),
				R2(2, 0), R2(2, 1), R2(2, 2), tmp3(2),
				0, 0, 0, 1;
		invert(RT);

		cam44(Range(0, 2), Range(0, 4)) = cam2;

		AMatrix T2(4, 4);
		matMul2(RT, cam44, T2);

		vector<AVector*> points1(pts1.size() + 1);
		vector<AVector*> points2(pts2.size() + 1);
		AVector tmp4(4), tmp4_1(4);
		AVector dummy_p1(3), dummy_p2(3);
		dummy_p1 = 0; dummy_p2 = 0;

		for(uint i = 0; i < pts1.size(); ++i)
		{
			const AVector & pt1 = *pts1[i];
			tmp4 = pt1(0) / pt1(3), pt1(1) / pt1(3), pt1(2) / pt1(3), 1.0;
			matMul2(T1, tmp4, tmp4_1);

			AVector * v1 = new AVectord(3);
			(*v1) = tmp4_1(0) / tmp4_1(3), tmp4_1(1) / tmp4_1(3), tmp4_1(2) / tmp4_1(3);
			points1[i] = v1;
			dummy_p1 = dummy_p1 + *v1;

			const AVector & pt2 = *pts2[i];
			tmp4 = pt2(0) / pt2(3), pt2(1) / pt2(3), pt2(2) / pt2(3), 1.0;
			matMul2(T2, tmp4, tmp4_1);

			AVector * v2 = new AVectord(3);
			(*v2) = tmp4_1(0) / tmp4_1(3), tmp4_1(1) / tmp4_1(3), tmp4_1(2) / tmp4_1(3);
			points2[i] = v2;
			dummy_p2 = dummy_p2 + *v2;
		}

//		dummy_p1 = dummy_p1 / double(pts1.size() + 1);
                dummy_p1 /= double(pts1.size() + 1);
//		dummy_p2 = dummy_p2 / double(pts1.size() + 1);
		dummy_p2 /= double(pts1.size() + 1);
		dummy_p1 = t1(0) - dummy_p1(0), t1(1) - dummy_p1(1), t1(2) - dummy_p1(2);
		dummy_p2 = t2(0) - dummy_p2(0), t2(1) - dummy_p2(1), t2(2) - dummy_p2(2);
		dummy_p1 = dummy_p1 * double(pts1.size() + 1);
		dummy_p2 = dummy_p2 * double(pts1.size() + 1);
//		points1.insert(points1.begin() + pts1.size(), pts1.size(), const_cast<AVector*>(&t1));
//		points2.insert(points2.begin() + pts2.size(), pts2.size(), const_cast<AVector*>(&t2));
		points1[pts1.size()] = const_cast<AVector*>(&dummy_p1); // TODO
		points2[pts1.size()] = const_cast<AVector*>(&dummy_p2);
//		cout << "center1: " << t1 << endl;
//		cout << "center2: " << t2 << endl;
//		for(uint i = 0; i < points1.size(); ++i)
//		{
//			cout << "pt1: " << *points1[i] << endl;
//			cout << "pt2: " << *points2[i] << endl;
//		}

		RelativeOrientationCalculator calc;
		bool ok = calc.calculate(points1, points2);
		if(!ok) return false;

//		for(uint i = 0; i < points1.size() - 1; ++i)
//		{
//			delete points1[i];
//			delete points2[i];
//		}


		const AMatrix & R = calc.rotation();
		const AVector & t = calc.translation();
		const double s = calc.scale();

//		AVector temp(3);
//		for(uint i = 0; i < points1.size(); ++i)
//		{
//			const AVector & pt1 = *points1[i];
//			const AVector & pt2 = *points2[i];
//			cout << "pt1: " << pt1 << endl;
//			cout << "pt2: " << pt2 << endl;
//			cout << "distance: " << sqrt((pt1(0) - pt2(0))*(pt1(0) - pt2(0)) + (pt1(1) - pt2(1))*(pt1(1) - pt2(1)) +
//										 (pt1(2) - pt2(2))*(pt1(2) - pt2(2))) << endl;
//			matMul2(R, pt2, temp);
//			temp = temp * s;
//			temp = temp + t;
//			cout << "rotated p2: " << temp << endl;
//			cout << "distance: " << sqrt((pt1(0) - temp(0))*(pt1(0) - temp(0)) + (pt1(1) - temp(1))*(pt1(1) - temp(1)) +
//										 (pt1(2) - temp(2))*(pt1(2) - temp(2))) << endl;
//		}

		AMatrix temp1(4, 4);
		temp1 =
				s, 0, 0, t(0),
				0, s, 0, t(1),
				0, 0, s, t(2),
				0, 0, 0, 1;
//		cout << "S.T: " << temp1;
		AMatrix temp2(4, 4);
		temp2 = 0;
		temp2(Range(0, 2), Range(0, 2)) = R;
		temp2(3, 3) = 1;
//		cout << "R: " << temp2;

		matMul2(temp1, temp2, H);
//		cout << "H: " << H;


//		cout << T1;
		invert(T1);
		matMul2(T1, H, temp1);
		matMul2(temp1, T2, temp2);

//		for(uint i = 0; i < pts1.size(); ++i)
//		{
//			const AVector & pt1 = *pts1[i];
//			const AVector & pt2 = *pts2[i];
////			cout << "pt1: " << pt1 << endl;
////			cout << "pt2: " << pt2 << endl;

//			AVector pt11(4), pt22(4);
//			pt11 = pt1(0) / pt1(3), pt1(1) / pt1(3), pt1(2) / pt1(3), 1.0;
//			pt22 = pt2(0) / pt2(3), pt2(1) / pt2(3), pt2(2) / pt2(3), 1.0;
////			cout << "pt1: " << pt11 << endl;
////			cout << "pt2: " << pt22 << endl;

//			matMul2(temp2, pt22, tmp4);
////			cout << "rotated p2: " << tmp4 << endl;
////			cout << "distance: " << sqrt((pt11(0) - tmp4(0))*(pt11(0) - tmp4(0)) + (pt11(1) - tmp4(1))*(pt11(1) - tmp4(1)) +
////										 (pt11(2) - tmp4(2))*(pt11(2) - tmp4(2))) << endl;
//		}

		H = temp2;

		return true;
	}

	AMatrix HomographyCalculator::calculateFromPointsMetric ( const vector< AVector* > & pts1, const vector< AVector* > & pts2)
	{
		assert(pts1.size() == pts2.size());

		vector<AVector*> points1(pts1.size());
		vector<AVector*> points2(pts2.size());
		for(uint i = 0; i < pts1.size(); ++i)
		{
			AVector * pt = pts1[i];
			AVector * v = new AVectord(3);
			(*v)(0) = (*pt)(0) / (*pt)(3);
			(*v)(1) = (*pt)(1) / (*pt)(3);
			(*v)(2) = (*pt)(2) / (*pt)(3);
			points1[i] = v;

			pt = pts2[i];
			v = new AVectord(3);
			(*v)(0) = (*pt)(0) / (*pt)(3);
			(*v)(1) = (*pt)(1) / (*pt)(3);
			(*v)(2) = (*pt)(2) / (*pt)(3);
			points2[i] = v;
		}



		for(uint i = 0; i < pts1.size(); ++i)
		{
			delete points1[i];
			delete points2[i];
		}


		AMatrix H(4, 4);
		return H;
	}

	} // core
} // adore
