

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "linearqcalculatorfl.h"
#include <adrcore/util/printmat.h>

namespace adore
{

	namespace core
	{

		int LinearQCalculatorFL::calcDualQuadric ( NView::CameraIterator camIt, const int numCams, AMatrix &Q1, AMatrix &Q2 )
		{
			NView::CameraIterator cit = camIt;
			for ( int i = 0; i < numCams; ++i, ++cit )
			{
				normalize( ( *cit )->getProjMat(), 1.0 );
			}

			//Q is
			//x1, x2, x3, x4
			//x2, x5, x6, x7
			//x3, x6,  1, x8
			//x4, x7, x8, x9

			getLinearEquation ( camIt, numCams, left, right );

//			printMat( left );
//			printMat( right );

			svdUVT ( left, store.store );
			solveSVD2_DUVTb ( store, right, x );
			Q1 ( 0, 0 ) = x ( 0 );
			Q1 ( 0, 1 ) = x ( 1 );
			Q1 ( 0, 2 ) = x ( 2 );
			Q1 ( 0, 3 ) = x ( 3 );
			Q1 ( 1, 0 ) = x ( 1 );
			Q1 ( 1, 1 ) = x ( 4 );
			Q1 ( 1, 2 ) = x ( 5 );
			Q1 ( 1, 3 ) = x ( 6 );
			Q1 ( 2, 0 ) = x ( 2 );
			Q1 ( 2, 1 ) = x ( 5 );
			Q1 ( 2, 2 ) = 1.0;
			Q1 ( 2, 3 ) = x ( 7 );
			Q1 ( 3, 0 ) = x ( 3 );
			Q1 ( 3, 1 ) = x ( 6 );
			Q1 ( 3, 2 ) = x ( 7 );
			Q1 ( 3, 3 ) = x ( 8 );

//			printf( "Q1:\n" );
//			printMat( Q1 );

			if ( store.store.D ( cols - 1 ) < DBL_EPSILON ) //DBL_EPSILON is also used in solveSVD2_DUVTb
			{
//				printf( "singular values:\n" );
//				printMat( store.store.D );

				const AMatrix &VT = store.store.VT;
				const int lr = VT.rows() - 1;
				Q2 ( 0, 0 ) = VT ( lr, 0 );
				Q2 ( 0, 1 ) = VT ( lr, 1 );
				Q2 ( 0, 2 ) = VT ( lr, 2 );
				Q2 ( 0, 3 ) = VT ( lr, 3 );
				Q2 ( 1, 0 ) = VT ( lr, 1 );
				Q2 ( 1, 1 ) = VT ( lr, 4 );
				Q2 ( 1, 2 ) = VT ( lr, 5 );
				Q2 ( 1, 3 ) = VT ( lr, 6 );
				Q2 ( 2, 0 ) = VT ( lr, 2 );
				Q2 ( 2, 1 ) = VT ( lr, 5 );
				Q2 ( 2, 2 ) = 1.0;
				Q2 ( 2, 3 ) = VT ( lr, 7 );
				Q2 ( 3, 0 ) = VT ( lr, 3 );
				Q2 ( 3, 1 ) = VT ( lr, 6 );
				Q2 ( 3, 2 ) = VT ( lr, 7 );
				Q2 ( 3, 3 ) = VT ( lr, 8 );

//				printf( "Q2:\n" );
//				printMat( Q2 );

				return CALIB_Q_PENCIL;
			}
			else if ( numCams == 2 )
			{
				DBGOUT ( DBGCRITICAL, "AutocalibratorLinear: 2 Views -> solution should be a pencil, but it's not.\n" );
				assert ( false );
				return CALIB_Q_ERROR;
			}

			return CALIB_Q_UNIQUE;
		}


		/**
		 * fills in the values for a linear equation to solve autocalibration
		 * this is for P1=general camera
		 */
		void LinearQCalculatorFL::getLinearEquation ( NView::CameraIterator camIt, const int numCams, AMatrix &left, AVector &right )
		{
			int cols = left.cols();
			assert ( cols == 9 );
			assert ( (int) left.rows() == numCams * 4 );
			ADORE_UNUSED( cols );

			//K is
			//f 0 0
			//0 f 0
			//0 0 1

			//Q is
			//x1, x2, x3, x4
			//x2, x5, x6, x7
			//x3, x6,  1, x8
			//x4, x7, x8, x9

			//matrix cols:
			// 0   1   2   3   4   5   6   7,  8
			//x1, x2, x3, x4, x5, x6, x7, x8, x9
			int row = 0;
			for ( int k = 0; k < numCams; ++k, ++camIt )
			{
				Camera *cam = *camIt;
				assert ( cam );
				AMatrix &camMat = cam->getProjMat();
				const double &a = camMat ( 0, 0 );
				const double &b = camMat ( 0, 1 );
				const double &c = camMat ( 0, 2 );
				const double &d = camMat ( 0, 3 );
				const double &e = camMat ( 1, 0 );
				const double &f = camMat ( 1, 1 );
				const double &g = camMat ( 1, 2 );
				const double &h = camMat ( 1, 3 );
				const double &i = camMat ( 2, 0 );
				const double &j = camMat ( 2, 1 );
				const double &k = camMat ( 2, 2 );
				const double &l = camMat ( 2, 3 );

				//equation K(0, 0) == K(1, 1)
				//-g^2 + c^2
				right ( row ) = c * c - g * g;
				//(-e^2 + a^2) * x1
				left ( row, 0 ) = a * a - e * e;
				//(-2*e*f + 2*a*b) * x2
				left ( row, 1 ) = 2 * a * b - 2 * e * f;
				//(-2*e*g + 2*a*c) * x3
				left ( row, 2 ) = 2 * a * c - 2 * e * g;
				//(-2*e*h + 2*a*d) * x4
				left ( row, 3 ) = 2 * a * d - 2 * e * h;
				//(-f^2 + b^2) * x5
				left ( row, 4 ) = b * b - f * f;
				//(-2*f*g + 2*b*c) * x6
				left ( row, 5 ) = 2 * b * c - 2 * f * g;
				//(-2*f*h + 2*b*d) * x7
				left ( row, 6 ) = 2 * b * d - 2 * f * h;
				//(-2*g*h + 2*c*d) * x8
				left ( row, 7 ) = 2 * c * d - 2 * g * h;
				//(-h^2 + d^2) * x9
				left ( row, 8 ) = d * d - h * h;
				++row;

				//equation K(0, 1) == 0
				//c*g
				right ( row ) = c * g;
				//a*e*x1
				left ( row, 0 ) = a * e;
				//a*f*x2+b*e*x2
				left ( row, 1 ) = a * f + b * e;
				//a*g*x3+c*e*x3
				left ( row, 2 ) = a * g + c * e;
				//a*h*x4+d*e*x4
				left ( row, 3 ) = a * h + d * e;
				//b*f*x5
				left ( row, 4 ) = b * f;
				//b*g*x6+c*f*x6
				left ( row, 5 ) = b * g + c * f;
				//b*h*x7+d*f*x7
				left ( row, 6 ) = b * h + d * f;
				//c*h*x8+d*g*x8
				left ( row, 7 ) = c * h + d * g;
				//d*h*x9
				left ( row, 8 ) = d * h;
				++row;

				//equation K(0, 2) == 0
				//c*k
				right ( row ) = c * k;
				//a*i*x1
				left ( row, 0 ) = a * i;
				//a*j*x2+b*i*x2
				left ( row, 1 ) = a * j + b * i;
				//a*k*x3+c*i*x3
				left ( row, 2 ) = a * k + c * i;
				//a*l*x4+d*i*x4
				left ( row, 3 ) = a * l + d * i;
				//b*j*x5
				left ( row, 4 ) = b * j;
				//b*k*x6+c*j*x6
				left ( row, 5 ) = b * k + c * j;
				//b*l*x7+d*j*x7
				left ( row, 6 ) = b * l + d * j;
				//c*l*x8+d*k*x8
				left ( row, 7 ) = c * l + d * k;
				//d*l*x9
				left ( row, 8 ) = d * l;
				++row;

				//equation K(1, 2) == 0
				//g*k
				right ( row ) = g * k;
				//e*i*x1
				left ( row, 0 ) = e * i;
				//e*j*x2+f*i*x2
				left ( row, 1 ) = e * j + f * i;
				//e*k*x3+g*i*x3
				left ( row, 2 ) = e * k + g * i;
				//e*l*x4+h*i*x4
				left ( row, 3 ) = e * l + h * i;
				//f*j*x5
				left ( row, 4 ) = f * j;
				//f*k*x6+g*j*x6
				left ( row, 5 ) = f * k + g * j;
				//f*l*x7+h*j*x7
				left ( row, 6 ) = f * l + h * j;
				//g*l*x8+h*k*x8
				left ( row, 7 ) = g * l + h * k;
				//h*l*x9
				left ( row, 8 ) = h * l;
				++row;
			}
		}

		int LinearQCalculatorFL::getDet0Solutions ( AMatrix &Q1, AMatrix &Q2, double l[4] )
		{
			//Q form is
			//x1, x2, x3, x4
			//x2, x5, x6, x7
			//x3, x6,  1, x8
			//x4, x7, x8, x9

			const double &x1 = Q1 ( 0, 0 );
			const double &x2 = Q1 ( 0, 1 );
			const double &x3 = Q1 ( 0, 2 );
			const double &x4 = Q1 ( 0, 3 );
			const double &x5 = Q1 ( 1, 1 );
			const double &x6 = Q1 ( 1, 2 );
			const double &x7 = Q1 ( 1, 3 );
			const double &x8 = Q1 ( 2, 3 );
			const double &x9 = Q1 ( 3, 3 );

			const double &x2_2 = x2 * x2;
			const double &x3_2 = x3 * x3;
			const double &x4_2 = x4 * x4;
			const double &x6_2 = x6 * x6;
			const double &x7_2 = x7 * x7;
			const double &x8_2 = x8 * x8;

			const double &y1 = Q2 ( 0, 0 );
			const double &y2 = Q2 ( 0, 1 );
			const double &y3 = Q2 ( 0, 2 );
			const double &y4 = Q2 ( 0, 3 );
			const double &y5 = Q2 ( 1, 1 );
			const double &y6 = Q2 ( 1, 2 );
			const double &y7 = Q2 ( 1, 3 );
			const double &y8 = Q2 ( 2, 3 );
			const double &y9 = Q2 ( 3, 3 );

			const double &y2_2 = y2 * y2;
			const double &y3_2 = y3 * y3;
			const double &y4_2 = y4 * y4;
			const double &y6_2 = y6 * y6;
			const double &y7_2 = y7 * y7;
			const double &y8_2 = y8 * y8;


			//coefficients
			float c[5];
			double zeror[4];
			double zeroi[4];

			//das ergebnis von maxima (von l^4 bis l^0):
			//-y1*y6^2*y9*l^4+2*y2*y3*y6*y9*l^4-y3^2*y5*y9*l^4+y1*y5*y9*l^4-y2^2*y9*l^4-y1*y5*y8^2*l^4+y2^2*y8^2*l^4+2*y1*y6*y7*y8*l^4-2*y2*y3*y7*y8*l^4-2*y2*y4*y6*y8*l^4+2*y3*y4*y5*y8*l^4+y3^2*y7^2*l^4-y1*y7^2*l^4-2*y3*y4*y6*y7*l^4+2*y2*y4*y7*l^4+y4^2*y6^2*l^4-y4^2*y5*l^4

			c[0] = ( -y1*y6_2 + 2*y2*y3*y6 + ( y1-y3_2 ) *y5-y2_2 ) *y9 +
			       ( y2_2-y1*y5 ) *y8_2 +
			       ( ( 2*y1*y6 - 2*y2*y3 ) *y7- 2*y2*y4*y6 + 2*y3*y4*y5 ) *y8 +
			       ( y3_2-y1 ) *y7_2 +
			       ( 2*y2*y4 - 2*y3*y4*y6 ) *y7 +
			       y4_2*y6_2-
			       y4_2*y5;

			//-x1*y6^2*y9*l^3+2*x2*y3*y6*y9*l^3+2*x3*y2*y6*y9*l^3-2*x6*y1*y6*y9*l^3-2*x3*y3*y5*y9*l^3+y1*y5*y9*l^3+x1*y5*y9*l^3-x5*y3^2*y9*l^3+2*x6*y2*y3*y9*l^3-y2^2*y9*l^3-2*x2*y2*y9*l^3+x5*y1*y9*l^3-x1*y5*y8^2*l^3+2*x2*y2*y8^2*l^3-x5*y1*y8^2*l^3+2*x1*y6*y7*y8*l^3-2*x2*y3*y7*y8*l^3-2*x3*y2*y7*y8*l^3+2*x6*y1*y7*y8*l^3-2*x2*y4*y6*y8*l^3-2*x4*y2*y6*y8*l^3+2*x7*y1*y6*y8*l^3+2*x3*y4*y5*y8*l^3+2*x4*y3*y5*y8*l^3-2*x8*y1*y5*y8*l^3+2*x5*y3*y4*y8*l^3-2*x6*y2*y4*y8*l^3-2*x7*y2*y3*y8*l^3+2*x8*y2^2*y8*l^3+2*x3*y3*y7^2*l^3-y1*y7^2*l^3-x1*y7^2*l^3-2*x3*y4*y6*y7*l^3-2*x4*y3*y6*y7*l^3+2*x8*y1*y6*y7*l^3-2*x6*y3*y4*y7*l^3+2*y2*y4*y7*l^3+2*x2*y4*y7*l^3+2*x7*y3^2*y7*l^3-2*x8*y2*y3*y7*l^3+2*x4*y2*y7*l^3-2*x7*y1*y7*l^3+2*x4*y4*y6^2*l^3-x9*y1*y6^2*l^3+2*x6*y4^2*y6*l^3-2*x7*y3*y4*y6*l^3-2*x8*y2*y4*y6*l^3+2*x9*y2*y3*y6*l^3-y4^2*y5*l^3+2*x8*y3*y4*y5*l^3-2*x4*y4*y5*l^3-x9*y3^2*y5*l^3+x9*y1*y5*l^3-x5*y4^2*l^3+2*x7*y2*y4*l^3-x9*y2^2*l^3

			c[1] = ( -x1*y6_2+ ( 2*x2*y3 + 2*x3*y2 - 2*x6*y1 ) *y6 + ( -2*x3*y3 + y1+x1 ) *y5	-x5*y3_2 + 2*x6*y2*y3 - y2_2 - 2*x2*y2 + x5*y1 ) * y9
			       + ( -x1*y5 + 2*x2*y2 - x5*y1 ) *y8_2
			       + ( ( 2*x1*y6 - 2*x2*y3 - 2*x3*y2 + 2*x6*y1 ) *y7 + ( -2*x2*y4 - 2*x4*y2 + 2*x7*y1 ) * y6	+ ( 2*x3*y4 + 2*x4*y3 - 2*x8*y1 ) * y5 + ( 2*x5*y3 - 2*x6*y2 ) *y4 - 2*x7*y2*y3 + 2*x8*y2_2 ) *y8
			       + ( 2*x3*y3 - y1-x1 ) *y7_2
			       + ( ( -2*x3*y4 - 2*x4*y3 + 2*x8*y1 ) *y6+ ( -2*x6*y3 + 2*y2 + 2*x2 ) *y4 + 2*x7*y3_2 - 2*x8*y2*y3 + 2*x4*y2 - 2*x7*y1 ) *y7
			       + ( 2*x4*y4 - x9*y1 ) *y6_2
			       + ( 2*x6*y4_2+ ( -2*x7*y3 - 2*x8*y2 ) *y4 + 2*x9*y2*y3 ) *y6
			       + ( -y4_2+ ( 2*x8*y3 - 2*x4 ) *y4 - x9*y3_2 + x9*y1 ) *y5
			       -x5*y4_2
			       +2*x7*y2*y4
			       -x9*y2_2;


			//-2*x1*x6*y6*y9*l^2+2*x2*x3*y6*y9*l^2-x3^2*y5*y9*l^2+x1*y5*y9*l^2+2*x2*x6*y3*y9*l^2-2*x3*x5*y3*y9*l^2+2*x3*x6*y2*y9*l^2-2*x2*y2*y9*l^2-x6^2*y1*y9*l^2+x5*y1*y9*l^2+x1*x5*y9*l^2-x2^2*y9*l^2-x1*x5*y8^2*l^2+x2^2*y8^2*l^2+2*x1*x6*y7*y8*l^2-2*x2*x3*y7*y8*l^2+2*x1*x7*y6*y8*l^2-2*x2*x4*y6*y8*l^2-2*x1*x8*y5*y8*l^2+2*x3*x4*y5*y8*l^2-2*x2*x6*y4*y8*l^2+2*x3*x5*y4*y8*l^2-2*x2*x7*y3*y8*l^2+2*x4*x5*y3*y8*l^2+4*x2*x8*y2*y8*l^2-2*x3*x7*y2*y8*l^2-2*x4*x6*y2*y8*l^2-2*x5*x8*y1*y8*l^2+2*x6*x7*y1*y8*l^2+x3^2*y7^2*l^2-x1*y7^2*l^2+2*x1*x8*y6*y7*l^2-2*x3*x4*y6*y7*l^2-2*x3*x6*y4*y7*l^2+2*x2*y4*y7*l^2-2*x2*x8*y3*y7*l^2+4*x3*x7*y3*y7*l^2-2*x4*x6*y3*y7*l^2-2*x3*x8*y2*y7*l^2+2*x4*y2*y7*l^2+2*x6*x8*y1*y7*l^2-2*x7*y1*y7*l^2-2*x1*x7*y7*l^2+2*x2*x4*y7*l^2-x1*x9*y6^2*l^2+x4^2*y6^2*l^2-2*x2*x8*y4*y6*l^2-2*x3*x7*y4*y6*l^2+4*x4*x6*y4*y6*l^2+2*x2*x9*y3*y6*l^2-2*x4*x7*y3*y6*l^2+2*x3*x9*y2*y6*l^2-2*x4*x8*y2*y6*l^2-2*x6*x9*y1*y6*l^2+2*x7*x8*y1*y6*l^2+2*x3*x8*y4*y5*l^2-2*x4*y4*y5*l^2-2*x3*x9*y3*y5*l^2+2*x4*x8*y3*y5*l^2+x9*y1*y5*l^2-x8^2*y1*y5*l^2+x1*x9*y5*l^2-x4^2*y5*l^2+x6^2*y4^2*l^2-x5*y4^2*l^2+2*x5*x8*y3*y4*l^2-2*x6*x7*y3*y4*l^2-2*x6*x8*y2*y4*l^2+2*x7*y2*y4*l^2+2*x2*x7*y4*l^2-2*x4*x5*y4*l^2-x5*x9*y3^2*l^2+x7^2*y3^2*l^2+2*x6*x9*y2*y3*l^2-2*x7*x8*y2*y3*l^2-x9*y2^2*l^2+x8^2*y2^2*l^2-2*x2*x9*y2*l^2+2*x4*x7*y2*l^2+x5*x9*y1*l^2-x7^2*y1*l^2

			c[2] =	( ( 2*x2*x3-2*x1*x6 ) *y6+ ( x1-x3_2 ) *y5+ ( 2*x2*x6-2*x3*x5 ) *y3+ ( 2*x3*x6-2*x2 ) *y2+ ( x5-x6_2 ) *y1+x1*x5-x2_2 ) *y9
			       + ( x2_2-x1*x5 ) *y8_2
			       + ( ( 2*x1*x6-2*x2*x3 ) *y7+ ( 2*x1*x7-2*x2*x4 ) *y6+ ( 2*x3*x4-2*x1*x8 ) *y5+ ( 2*x3*x5-2*x2*x6 ) *y4+ ( 2*x4*x5-2*x2*x7 ) *y3+ ( 4*x2*x8-2*x3*x7-2*x4*x6 ) *y2+ ( 2*x6*x7-2*x5*x8 ) *y1 ) *y8
			       + ( x3_2-x1 ) *y7_2
			       + ( ( 2*x1*x8-2*x3*x4 ) *y6+ ( 2*x2-2*x3*x6 ) *y4+ ( -2*x2*x8+4*x3*x7-2*x4*x6 ) *y3+ ( 2*x4-2*x3*x8 ) *y2+ ( 2*x6*x8-2*x7 ) *y1-2*x1*x7+2*x2*x4 ) *y7
			       + ( x4_2-x1*x9 ) *y6_2
			       + ( ( -2*x2*x8-2*x3*x7+4*x4*x6 ) *y4+ ( 2*x2*x9-2*x4*x7 ) *y3+ ( 2*x3*x9-2*x4*x8 ) *y2+ ( 2*x7*x8-2*x6*x9 ) *y1 ) *y6
			       + ( ( 2*x3*x8-2*x4 ) *y4+ ( 2*x4*x8-2*x3*x9 ) *y3+ ( x9-x8_2 ) *y1+x1*x9-x4_2 ) *y5
			       + ( x6_2-x5 ) *y4_2
			       + ( ( 2*x5*x8-2*x6*x7 ) *y3+ ( 2*x7-2*x6*x8 ) *y2+2*x2*x7-2*x4*x5 ) *y4
			       + ( x7_2-x5*x9 ) *y3_2
			       + ( 2*x6*x9-2*x7*x8 ) *y2*y3
			       + ( x8_2-x9 ) *y2_2
			       + ( 2*x4*x7-2*x2*x9 ) *y2
			       + ( x5*x9-x7_2 ) *y1;

			//-x1*x6^2*y9*l+2*x2*x3*x6*y9*l-x3^2*x5*y9*l+x1*x5*y9*l-x2^2*y9*l-2*x1*x5*x8*y8*l+2*x2^2*x8*y8*l+2*x1*x6*x7*y8*l-2*x2*x3*x7*y8*l-2*x2*x4*x6*y8*l+2*x3*x4*x5*y8*l+2*x1*x6*x8*y7*l-2*x2*x3*x8*y7*l+2*x3^2*x7*y7*l-2*x1*x7*y7*l-2*x3*x4*x6*y7*l+2*x2*x4*y7*l-2*x1*x6*x9*y6*l+2*x2*x3*x9*y6*l+2*x1*x7*x8*y6*l-2*x2*x4*x8*y6*l-2*x3*x4*x7*y6*l+2*x4^2*x6*y6*l-x3^2*x9*y5*l+x1*x9*y5*l-x1*x8^2*y5*l+2*x3*x4*x8*y5*l-x4^2*y5*l-2*x2*x6*x8*y4*l+2*x3*x5*x8*y4*l-2*x3*x6*x7*y4*l+2*x2*x7*y4*l+2*x4*x6^2*y4*l-2*x4*x5*y4*l+2*x2*x6*x9*y3*l-2*x3*x5*x9*y3*l-2*x2*x7*x8*y3*l+2*x4*x5*x8*y3*l+2*x3*x7^2*y3*l-2*x4*x6*x7*y3*l+2*x3*x6*x9*y2*l-2*x2*x9*y2*l+2*x2*x8^2*y2*l-2*x3*x7*x8*y2*l-2*x4*x6*x8*y2*l+2*x4*x7*y2*l-x6^2*x9*y1*l+x5*x9*y1*l-x5*x8^2*y1*l+2*x6*x7*x8*y1*l-x7^2*y1*l+x1*x5*x9*l-x2^2*x9*l-x1*x7^2*l+2*x2*x4*x7*l-x4^2*x5*l

			c[3] =	( -x1*x6_2+2*x2*x3*x6+ ( x1-x3_2 ) *x5-x2_2 ) *y9
			       + ( ( 2*x2_2-2*x1*x5 ) *x8+ ( 2*x1*x6-2*x2*x3 ) *x7-2*x2*x4*x6+2*x3*x4*x5 ) *y8
			       + ( ( 2*x1*x6-2*x2*x3 ) *x8+ ( 2*x3_2-2*x1 ) *x7-2*x3*x4*x6+2*x2*x4 ) *y7
			       + ( ( 2*x2*x3-2*x1*x6 ) *x9+ ( 2*x1*x7-2*x2*x4 ) *x8-2*x3*x4*x7+2*x4_2*x6 ) *y6
			       + ( ( x1-x3_2 ) *x9-x1*x8_2+2*x3*x4*x8-x4_2 ) *y5+ ( ( 2*x3*x5-2*x2*x6 ) *x8+ ( 2*x2-2*x3*x6 ) *x7+2*x4*x6_2-2*x4*x5 ) *y4+ ( ( 2*x2*x6-2*x3*x5 ) *x9+ ( 2*x4*x5-2*x2*x7 ) *x8+2*x3*x7_2-2*x4*x6*x7 ) *y3+ ( ( 2*x3*x6-2*x2 ) *x9+2*x2*x8_2+ ( -2*x3*x7-2*x4*x6 ) *x8+2*x4*x7 ) *y2+
			       ( ( x5-x6_2 ) *x9-x5*x8_2+2*x6*x7*x8-x7_2 ) *y1+ ( x1*x5-x2_2 ) *x9-x1*x7_2+2*x2*x4*x7-x4_2*x5;

			//-x1*x6^2*x9+2*x2*x3*x6*x9-x3^2*x5*x9+x1*x5*x9-x2^2*x9-x1*x5*x8^2+x2^2*x8^2+2*x1*x6*x7*x8-2*x2*x3*x7*x8-2*x2*x4*x6*x8+2*x3*x4*x5*x8+x3^2*x7^2-x1*x7^2-2*x3*x4*x6*x7+2*x2*x4*x7+x4^2*x6^2-x4^2*x5

			c[4] = ( -x1*x6_2+2*x2*x3*x6+ ( x1-x3_2 ) *x5-x2_2 ) *x9+ ( x2_2-x1*x5 ) *x8_2+ ( ( 2*x1*x6-2*x2*x3 ) *x7-2*x2*x4*x6+2*x3*x4*x5 ) *x8+ ( x3_2-x1 ) *x7_2+ ( 2*x2*x4-2*x3*x4*x6 ) *x7+x4_2*x6_2-
			       x4_2*x5;

			printf( "COEFFS: %f, %f, %f, %f, %f\n", c[0], c[1], c[2], c[3], c[4] );

			int dim = 4;
			bool success = rpoly ( c, dim, zeror, zeroi );
			assert ( success );
			ADORE_UNUSED( success );

			//filter imaginary results
			int numSolutions = 0;
			for ( int i = 0; i < 4; ++i )
			{
				if ( zeroi[i] == 0 )
				{
					l[numSolutions] = zeror[i];
					++numSolutions;
				}
				else
				{
					DBGOUT ( DBGNORMAL, "AutocalibratorLinear: imaginary solution (%f, %f)\n", zeror[i], zeroi[i] );
				}
			}
			return numSolutions;
		}


	}

}
