

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWOPRCALCULATOR_H
#define ADORE_NVIEWOPRCALCULATOR_H

#include <adrcore/types/camera.h>
#include <adrcore/nview/nviewoperator.h>
#include <vector>

namespace adore
{

	namespace core
	{

		using namespace std;

		/**
		 * This class calculate a quasi affine reconstruction
		 * Is solves the cheiriality inequalities using a "monte carlo" approach.
		 * In fact, it is not realy a monte carlo approach, because only 8 different combinations are tested:
		 * The fourth row of H is +/-1, +/-1, +/-1, 1, the rest is an identity matrix.
		 * However, it is much better than the QuasiAffinator.
		 *
		 * Additionally, for every equation PX = wx, positive w and right handed cameras are
		 * as far as possible established. If metricOnly is set, than this is the only modification done.
		 * If you intend to paint the nview using opengl, you should call this Modifier with the
		 * metricOnly option set. This ensures that a decomposition of the rotation matrix to angle axis
		 * succeeds.
		 *
		 *  
		*/
		class OPRCalculator : public NViewOperator
		{
			public:
				/**
				 * Constructor.
				 */
				OPRCalculator();

				OPRCalculator( const OPRCalculator &opr );

				/**
				 * Destructor.
				 */
				~OPRCalculator();

				NViewOperator *clone()
				{
					return new OPRCalculator ( *this );
				}

				/**
				 * overloaded from NViewOperator. This carries out the main functionality.
				 * An oriented projective reconstrution is computed and/or the modififations
				 * concerning handedness are carried out.
				 * @param nview a NView
				 */
				void modify ( NView &nview );

				/**
				 * tests, whether all points lie on the same side of the plain at infinity.
				 * @param pts3D the points
				 * @param cams the cameras
				 * @return the answer to the question "Is this Quasi Affine?"
				 */
				bool isQuasiAffine ( vector<AVector3D *> &pts3D, vector<Camera *> &cams );

				bool isQuasiAffine ( NView &nview );

				/**
				 * calculates a transformation that will upgrade the reconstruction to quasi-affine.
				 * it checks the return value of isQuasiAffine before doing anything.
				 * @param cams the cameras
				 * @param pts3D the points
				 * @param H a 4x4 matrix that will hold the result.
				 */
				bool findTrafo ( vector<Camera *> &cams, vector<AVector3D *> &pts3D, AMatrix &H );

				/**
				 * applies the transformation to the cameras
				 * @param H the transformation
				 * @param cams the cameras
				 */
				void apply ( AMatrix &H, vector<Camera *> &cams );
				void apply ( AMatrix &Hinv, vector<AVector3D *> &pts3D );

				/**
				 * a test whether or not all points are in front of all cameras.
				 * @param nview the Nview to test
				 * @return true, if alle points are in front of all cameras, false otherwise
				 */
				static bool pointsInFront ( NView &nview );

				/**
				 * if set to "true", all cameras will be upgraded to be right-handed and
				 * PX=wx with x=(a, b, 1) will hold for positive w.
				 * @param theValue the new value
				 */
				void setMetricOnly ( bool theValue )
				{
					metricOnly = theValue;
				}

				/**
				 * is the metric-only-option set or not?
				 * @return true, if only modifications are made, that don't affect the metric frame
				 */
				bool getMetricOnly() const
				{
					return metricOnly;
				}

			private:

				bool metricOnly;

				void findInitialD ( vector<AVector3D *> &pts3D, vector<double> &z, vector<int> &d, AMatrix &H4 );
				bool goodTransformation ( vector<AVector3D *> &pts3D, vector<double> &z, vector<int> &d, AMatrix &H4 );
				void fillZ ( vector<Camera *> &cams, vector<AVector3D *> &pts3D, vector<double> &z );

				/**
				 * This method's name is not quite right.
				 * In fact, it ensures that all cameras are right handed.
				 */
				void makeScaleFactorsPositive ( NView &nview );

				void fixCams ( set<AVector3D *> &pts3Ddone, set<Camera *> &camstodo,
				               set<Camera *> &camsdone, NView &nview );
				void fixPoints ( set<Camera *> &camsdone, set<AVector3D *> &pts3Dtodo,
				                 set<AVector3D *> &pts3Ddone, NView &nview );

				/**
				 * Tests, if all points are in front of all cameras.
				 * If not, the homography
				 * 1 0 0 0
				 * 0 1 0 0
				 * 0 0 1 0
				 * 0 0 0-1
				 * is applied
				 */
				void putPointsInFrontMetric ( NView &nview );
		};

	}
}

#endif
