

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#ifndef RELATIVEORIENTATIONCALCULATOR_H
#define RELATIVEORIENTATIONCALCULATOR_H

#include <adrcore/types/aarray.h>
#include <adrcore/util/solvesvd.h>

#include <vector>

namespace adore {
namespace core {

	/**
	 * This class calculates the relative orientation(R, s and t of pts1 = s*R.pts2 + t) for two sets of points.
	 * Calculation fails, if points are coincindent, collinear or very noisy.
	 * Literature used:
	 * "Least-Squares Fitting of Two 3-D Point Sets." K. S. Arun, T. S. Huang, S. D. Blostein
	 * "Relative Orientation" BERTHOLD K.P. HORN
	 *  
	*/
	class RelativeOrientationCalculator
	{
	public:
		typedef std::vector<AVector*> PointColl;

	private:
		AMatrix m_R;
		AVector m_t;
		double m_scale;

		AVector m_c1, m_c2;
		AMatrix m_temp33;
		SVDSolveStorage<double, 1> m_store;

		PointColl m_pts1, m_pts2;

	public:

		RelativeOrientationCalculator() : m_R(3, 3), m_t(3), m_scale(1.0), m_c1(3), m_c2(3), m_temp33(3, 3), m_store(3, 3) {}

		~RelativeOrientationCalculator()
		{
			cleanup(m_pts1);
			cleanup(m_pts2);
		}

		/** find R, s and t of pts1 = s*R.pts2 + t
		  * The number of points must be same for pts1 and pts2 .
		  * The minimum number of points is 3.
		  * @returns true on success.
		*/
		bool calculate(const PointColl& pts1, const PointColl& pts2);

		/**
		  * @returns R of pts = s*R.pts2 + t (if previous calculate-call was successfull)
		  */
		const AMatrix& rotation() const { return m_R; }

		/**
		  * @returns t of pts = s*R.pts2 + t (if previous calculate-call was successfull)
		  */
		const AVector& translation() const { return m_t; }

		/**
		  * @returns s of pts = s*R.pts2 + t (if previous calculate-call was successfull)
		  */
		const double scale() const { return m_scale; }

	private:
		void get_centroid(const PointColl& pts, AVector& c);

		void get_translated_points(const PointColl& a, const AVector& t, PointColl& b);

		void cleanup(PointColl& pts);

		bool get_rotation(const PointColl& a, const PointColl& b, AMatrix& R);

		void get_M(const PointColl& a, const PointColl& b, AMatrix& M);

		bool get_scale(const PointColl& a, const PointColl& b, double& scale);

		void get_translation(const AVector& c1, const AVector& c2, const AMatrix& R, const double s, AVector& t);
	};

} // core
} // adore

#endif // RELATIVEORIENTATIONCALCULATOR_H
