

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWTTENSORCALCULATORPPP_H
#define ADORE_NVIEWTTENSORCALCULATORPPP_H

#include <adrcore/nview/ttensorcalculator.h>
#include <adrcore/test/progressobserver.h>
#include <adrcore/nview/lineartriangulator.h>
#include <adrcore/minimization/bundleadjuster.h>
#include <adrcore/minimization/baminfunction.h>
#include <adrcore/minimization/sparsedefaulterrorfunc.h>

namespace adore
{
	namespace core
	{

		/**
		 * A specialization of the TTensorCalculator that takes Point-Point-Point-Correspondences and carries out (de-)normalization.
		 * Additionally, the Gold-Standard-Algorithm is implemented.
		 *  
		*/
		class TTensorCalculatorPPP : public TTensorCalculator
		{
			private:
				//these cameras are used in algebraic minimization
				Camera *cam2;
				Camera *cam3;
				bool minAlgebraic;

			public:
				/**
				 * Constructor.
				 * If you like to retrieve the cameras that the algebraic minimizer may compute, pass two pointers
				 */
				TTensorCalculatorPPP( Camera *c2 = NULL, Camera *c3 = NULL ) :TTensorCalculator(), cam2( c2 ), cam3( c3 ), minAlgebraic ( true ) {}

				TTensorCalculatorPPP( const TTensorCalculatorPPP& ):TTensorCalculator(), minAlgebraic ( true )
				{
					//the cam pointers would lead to currupted memory
					assert( cam2 == NULL );
					assert( cam3 == NULL );
				}

				/**
				 * intializes the TTensorCalculatorPPP
				 * @param it1 an iterator pointing to the points of the first View.
				 * @param it2 an iterator pointing to the points of the second View.
				 * @param it3 an iterator pointing to the points of the third View.
				 * @param numpts the number of point correspondences
				 */
				template <class CorrIterator, class ViewIterator >
				int calculateTrifocalTensor ( CorrIterator it, int size, ViewIterator vit )
				{
					ViewBase *v1 = *vit; ++vit;
					ViewBase *v2 = *vit; ++vit;
					ViewBase *v3 = *vit; --vit; --vit;//vit is used later
					CorrIterator cit = it;

					TTensorCalculator::init ( size );
					int row = 0;
					for ( int i = 0; i < size; ++i, ++cit )
					{
						PointCorr *corr = *cit;
						if ( !corr ) continue;
						FeaturePoint *fp1 = corr->getFeature ( v1 );
						if ( !fp1 ) continue;
						FeaturePoint *fp2 = corr->getFeature ( v2 );
						if ( !fp2 ) continue;
						FeaturePoint *fp3 = corr->getFeature ( v3 );
						if ( !fp3 ) continue;

						assert ( ( *fp1 ) ( 2 ) == 1 );
						assert ( ( *fp2 ) ( 2 ) == 1 );
						assert ( ( *fp3 ) ( 2 ) == 1 );

						if ( size == 6 )
						{
							setPPP ( row, ( *fp1 ) ( 0 ), ( *fp1 ) ( 1 ), ( *fp2 ) ( 0 ), ( *fp2 ) ( 1 ), ( *fp3 ) ( 0 ), ( *fp3 ) ( 1 ) );
						}
						else
						{
							setPPP ( row, ( *fp1 ) ( 0 ), ( *fp1 ) ( 1 ), 1.0, ( *fp2 ) ( 0 ), ( *fp2 ) ( 1 ), ( *fp3 ) ( 0 ), ( *fp3 ) ( 1 ) );
						}
						++row;
					}

					if ( row < 6 )
					{
						DBGOUT( DBGCRITICAL, "TTensorCalculatorPPP: too few corrs\n" );
						return -1;
					}

					TTensorCalculator::init ( row );//resize
					assert_msg ( row * 4 == (int) linearEq.rows(), "%d <> %zd\n", row * 4 ,linearEq.rows() );

					if ( row == 6 )
					{
						return calculateTrifocalTensorBase6Point();
					}
					else if ( minAlgebraic )
					{
						DBGOUT ( DBGNORMAL, "minimizing algebraic\n" );
						if ( cam2 && cam3 )
						{
							calculateTrifocalTensorBaseMinAlgebraicError ( true, &cam2->getProjMat(), &cam3->getProjMat() );
						}
						else
						{
							calculateTrifocalTensorBaseMinAlgebraicError ( true, NULL, NULL );
						}
//						if ( bundleAdjust ) bundleAdjustTTensor ( it, size, vit, lineTensor, cam2, cam3 );
					}
					else
					{
						calculateTrifocalTensorBaseLinear();
//						if ( bundleAdjust ) bundleAdjustTTensor ( it, size, vit, lineTensor );
					}

					tensor1 = buildTensorFromLineTensor ( lineTensor );

#ifndef NDEBUG
//					for ( ATensor3::iterator it = tensor1.begin(); it != tensor1.end(); ++it )
//					{
//						if ( fabs ( *it ) < FLT_EPSILON ) *it = 0;
//					}
//					COUT( DBGALL, "-----------\n" << tensor1 << "\n-----------\n" );
#endif

					return 1;
				}

				void setMinAlgebraic ( bool theValue )
				{
					minAlgebraic = theValue;
				}

				bool getMinAlgebraic() const
				{
					return minAlgebraic;
				}
		};

	}
}

#endif
