

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "angleaxis.h"
#include <adrcore/util/mathutil.h>
#include <adrcore/util/printmat.h>

namespace adore
{

	namespace core
	{

		AngleAxis::AngleAxis( double midVal ):
				mat( 3, 3 ),
				store3( 3, 3, SVD_VT ), n1( 3 ), n2( 3 )
		{
			setIntervalCentre( midVal );
		}

		AVector2D AngleAxis::toParam( const AMatrix &R )
		{
			AVector2D t( 3 );
			toParam( R, t );
			return t;
		}

		void AngleAxis::toParam( const AMatrix &R, AVector2D &param )
		{
			double angle;
			matToAngleAxis( R, param, angle );
			param( 0 ) = angle * param( 0 );
			param( 1 ) = angle * param( 1 );
			param( 2 ) = angle * param( 2 );
		}

//		angle is normalized to lie between -M_PI and +M_PI
		AVector2D AngleAxis::matToAngleAxis( const AMatrix &R, double &angle )
		{
			AVector2D t( 3 );
			matToAngleAxis( R, t, angle );
			return t;
		}

		void AngleAxis::matToAngleAxis( const AMatrix &R, AVector2D &t, double &angle )
		{
			assert( t.rows() == 3 );

                        mat = R/*.copy()*/;
			mat( 0, 0 ) = mat( 0, 0 ) - 1.0;
			mat( 1, 1 ) = mat( 1, 1 ) - 1.0;
			mat( 2, 2 ) = mat( 2, 2 ) - 1.0;

			svdVT( mat, store3 );

#ifndef NDEBUG
			if ( fabs( store3.D( 2 ) ) >= 0.001 )
			{
				printf( "Rotation matrix\n" );
				printMat( R );
				printf( "Rotation matrix - I\n" );
				printMat( mat );
				printf( "Singular values\n" );
				printMat( store3.D );
			}
#endif
			assert_msg( fabs( store3.D( 2 ) ) < 0.001, "ANGLEAXIS - There's still a reflection in the rotation matrix!\n" );

			t = store3.VT( 2, Range::all() );

			double cosT = 0.5 * ( R( 0, 0 ) + R( 1, 1 ) + R( 2, 2 ) - 1.0 );
			store3.D( 0 ) = R( 2, 1 ) - R( 1, 2 );
			store3.D( 1 ) = R( 0, 2 ) - R( 2, 0 );
			store3.D( 2 ) = R( 1, 0 ) - R( 0, 1 );

            double sinT = 0.5 * ( store3.D( 0 ) *  t( 0 ) +
                                  store3.D( 1 ) *  t( 1 ) +
                                  store3.D( 2 ) *  t( 2 ) );

			angle = atan2( sinT, cosT );//result is between -PI and PI
			normAngleToInterval( angle );
		}

		AMatrix AngleAxis::fromParam( const AVector &t )
		{
			AMatrix R( 3, 3 );
			fromParam( t, R );
			return R;
		}

		void AngleAxis::fromParam( const AVector &t, AMatrix &R )
		{
			getCrossMat( t, R );

			double angle2 = getSqrNorm( t );
			if ( angle2 == 0.0 )
			{
				setIdentity( R );
				return;
			}
			double angle = sqrt( angle2 );

			valueToInterval( angle, -M_PI, M_PI );
			double cosT = cos( angle );
			double sincT = sinc( angle );
			double factor = ( 1.0 - cosT ) / angle2;

            R( 0, 0 ) = sincT * R( 0, 0 ) + factor * t( 0 ) * t( 0 ) + cosT;
            R( 0, 1 ) = sincT * R( 0, 1 ) + factor * t( 0 ) * t( 1 );
            R( 0, 2 ) = sincT * R( 0, 2 ) + factor * t( 0 ) * t( 2 );
            R( 1, 0 ) = sincT * R( 1, 0 ) + factor * t( 1 ) * t( 0 );
            R( 1, 1 ) = sincT * R( 1, 1 ) + factor * t( 1 ) * t( 1 ) + cosT;
            R( 1, 2 ) = sincT * R( 1, 2 ) + factor * t( 1 ) * t( 2 );
            R( 2, 0 ) = sincT * R( 2, 0 ) + factor * t( 2 ) * t( 0 );
            R( 2, 1 ) = sincT * R( 2, 1 ) + factor * t( 2 ) * t( 1 );
            R( 2, 2 ) = sincT * R( 2, 2 ) + factor * t( 2 ) * t( 2 ) + cosT;

#ifndef NDEBUG
			if ( !isRightHand( R ) )
			{
				printf( "Rotation matrix\n" );
				printMat( R );
			}
#endif
			assert_msg( isRightHand( R ), "R is not right hand after creation form AngleAxis\n" );
		}

		AMatrix AngleAxis::angleAxisToMat( const double t1, const double t2, const double t3 )
		{
            store3.D( 0 ) =  t1;
            store3.D( 1 ) =  t2;
            store3.D( 2 ) =  t3;
			return fromParam( store3.D );
		}

		void AngleAxis::angleAxisToMat( const double t1, const double t2, const double t3, AMatrix &R )
		{
            store3.D( 0 ) =  t1;
            store3.D( 1 ) =  t2;
            store3.D( 2 ) =  t3;
			fromParam( store3.D, R );
		}

		bool AngleAxis::normAngleToInterval( double &angle )
		{
			double newAngle = valueToInterval( angle, interval[0], interval[1] );
			if ( newAngle == angle ) return false;
			angle = newAngle;
			return true;
		}

		bool AngleAxis::normLengthToInterval( AVector &r )
		{
			double angle = sqrt( getSqrNorm( r ) );
			double old = angle;

			//printf("angle is %f\n", angle);
			if ( normAngleToInterval( angle ) )
			{
				double factor = angle / old;
                                r( 0 ) = factor * r( 0 );
                                r( 1 ) = factor * r( 1 );
                                r( 2 ) = factor * r( 2 );
				return true;
			}
			return false;
		}

		AVector AngleAxis::vectorsToAngleAxis( AVector &vec1, AVector &vec2, double &angle )
		{
			AVector anax( 3 );
			vectorsToAngleAxis( vec1, vec2, angle, anax );
			return anax;
		}

		void AngleAxis::vectorsToAngleAxis( AVector &vec1, AVector &vec2, double &angle, AVector &result )
		{
			assert( result.rows() == 3 );
			AVector &anax = result;
			anax = 0,0,0;
			n1 = vec1;//copy;
			n2 = vec2;//copy;
			normalize( n1, 1.0 );
			normalize( n2, 1.0 );

			angle = acos( dot( n1, n2 ) );

			if ( fabs( angle ) < 0.0001 ) return;

			if ( fabs( angle-M_PI ) < 0.0001 )
			{
				anax( 0 ) = -vec1( 1 );
				anax( 1 ) = vec1( 0 );
				anax( 2 ) = vec1( 2 );
				return;
			}

			anax = cross( vec1, vec2 );
			normalize( anax,angle );
		}

	}
}
