

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#ifndef ADORE_PARAMETERIZATIONQUATERNIONCAMPARAM_H
#define ADORE_PARAMETERIZATIONQUATERNIONCAMPARAM_H

#include <adrcore/types/camera.h>
#include <adrcore/parameterization/angleaxis.h>
#include "quaternion.h"

namespace adore
{

	namespace core
	{

		/**
		  * converts camera matrix (K.[R|t]) in parameter vector (and v.v.):
		  * < qw, qx, qy, qz, tx, ty, tz, f1, f2, s, x0, y0 >
		  * K :=
		  * [ f1, s,  x0 ]
		  * [ 0,  f2, y0 ]
		  * [ 0,  0,  1  ]
		  * quaternion is build from matrix < invert(K).R >
		  * < tx, ty, tz > := < invert(K).t >
		  *  
		*/
		class QuaternionCamParam
		{
			mutable AMatrix K, R, temp33;
			mutable AVector t, t2;
			mutable Quaternion quaternion;

			public:
				QuaternionCamParam() : K(3, 3), R(3, 3), temp33(3, 3), t(3), t2(3) {}

				int getSize() { return 12; }

				void fromParam( const vector<AVector> & param, vector<Camera *> & cams ) const
				{
					for ( uint i = 0; i < param.size(); ++i )
					{
						fromParam( param[i], *cams[i] );
					}
				}

				void toParam( const vector<Camera *> & cams, vector<AVector> & param ) const
				{
					for ( uint i = 0; i < param.size(); ++i )
					{
						toParam( *cams[i], param[i] );
					}
				}

				void fromParam( const AVector & param, Camera & cam ) const
				{
					fromParam( param, cam.getProjMat() );
				}

				void fromParam( const AVector & vec, AMatrix & cam ) const
				{
					quaternion(0) = vec(1);
					quaternion(1) = vec(2);
					quaternion(2) = vec(3);
					quaternion(3) = vec(0);

					double n = sqrt(quaternion(0) * quaternion(0) + quaternion(1) * quaternion(1) + quaternion(2) * quaternion(2) + quaternion(3) * quaternion(3));
					assert(n > FLT_EPSILON);

					quaternion(0) = quaternion(0) / n;
					quaternion(1) = quaternion(1) / n;
					quaternion(2) = quaternion(2) / n;
					quaternion(3) = quaternion(3) / n;
					quaternion.toRotationMatrix(R);

					K =
							vec(7), vec(9), vec(10),
							0,      vec(8), vec(11),
							0,      0,      1;
					matMul2(K, R, temp33);
					cam(Range::all(), Range(0, 2)) = temp33;

					t = vec(4), vec(5), vec(6);
					matMul2(K, t, t2);
					cam(Range::all(), 3) = t2;
				}

				void toParam( const Camera & cam, AVector & vec ) const
				{
					toParam( cam.getProjMat(), vec );
				}

				void toParam( const AMatrix & cam, AVector & vec ) const
				{
					temp33 = cam ( Range::all(), Range ( 0, 2 ) );
					rq ( temp33, K, R );
					makeDiagOfROfRQPosisitve ( K, R );

					double div = K ( 2, 2 );
					assert ( fabs( div ) > FLT_EPSILON );
					div = 1.0 / div;
					vec(7) = K(0, 0) * div;
					vec(8) = K(1, 1) * div;
					vec(9) = K(0, 1) * div;
					vec(10) = K(0, 2) * div;
					vec(11) = K(1, 2) * div;

					quaternion.fromRotationMatrix(R);

					vec(0) = quaternion(3);
					vec(1) = quaternion(0);
					vec(2) = quaternion(1);
					vec(3) = quaternion(2);

					invertUpperTriangular3x3Matrix ( K );

					t = cam ( Range::all(), 3 );
					matMul2 ( K, t, t2 );

					vec(4) = t2(0);
					vec(5) = t2(1);
					vec(6) = t2(2);
				}

		};

		/**
		  * converts camera matrix (K.[R|t]) in parameter vector (and v.v.):
		  * < qw, qx, qy, qz, tx, ty, tz, f, x0, y0 >
		  * K :=
		  * [ f, 0, 0 ]
		  * [ 0, f, 0 ]
		  * [ 0, 0, 1 ]
		  * quaternion is build from matrix < invert(K).R >
		  * < tx, ty, tz > := < invert(K).t >
		  *  
		*/
		class QuaternionCamParam_F
		{
			mutable AMatrix K, R, temp33;
			mutable AVector t, t2;
			mutable Quaternion quaternion;

			public:
				QuaternionCamParam_F() : K(3, 3), R(3, 3), temp33(3, 3), t(3), t2(3) {}

				int getSize() { return 8; }

				void fromParam( const vector<AVector> & param, vector<Camera *> & cams ) const
				{
					for ( uint i = 0; i < param.size(); ++i )
					{
						fromParam( param[i], *cams[i] );
					}
				}

				void toParam( const vector<Camera *> & cams, vector<AVector> & param ) const
				{
					for ( uint i = 0; i < param.size(); ++i )
					{
						toParam( *cams[i], param[i] );
					}
				}

				void fromParam( const AVector & param, Camera & cam ) const
				{
					fromParam( param, cam.getProjMat() );
				}

				void fromParam( const AVector & vec, AMatrix & cam ) const
				{
					quaternion(0) = vec(1);
					quaternion(1) = vec(2);
					quaternion(2) = vec(3);
					quaternion(3) = vec(0);

					double n = sqrt(quaternion(0) * quaternion(0) + quaternion(1) * quaternion(1) + quaternion(2) * quaternion(2) + quaternion(3) * quaternion(3));
					assert(n > FLT_EPSILON);

					quaternion(0) = quaternion(0) / n;
					quaternion(1) = quaternion(1) / n;
					quaternion(2) = quaternion(2) / n;
					quaternion(3) = quaternion(3) / n;
					quaternion.toRotationMatrix(R);

					K =
							vec(7), 0,      0,
							0,      vec(7), 0,
							0,      0,      1;
					matMul2(K, R, temp33);
					cam(Range::all(), Range(0, 2)) = temp33;

					t = vec(4), vec(5), vec(6);
					matMul2(K, t, t2);
					cam(Range::all(), 3) = t2;
				}

				void toParam( const Camera & cam, AVector & vec ) const
				{
					toParam( cam.getProjMat(), vec );
				}

				void toParam( const AMatrix & cam, AVector & vec ) const
				{
					temp33 = cam ( Range::all(), Range ( 0, 2 ) );
					rq ( temp33, K, R );
					makeDiagOfROfRQPosisitve ( K, R );

					double div = K ( 2, 2 );
					assert ( fabs( div ) > FLT_EPSILON );
					vec(7) = (0.5 * (K(0, 0) + K(1, 1))) / div;

					quaternion.fromRotationMatrix(R);

					vec(0) = quaternion(3);
					vec(1) = quaternion(0);
					vec(2) = quaternion(1);
					vec(3) = quaternion(2);

					invertUpperTriangular3x3Matrix ( K );

					t = cam ( Range::all(), 3 );
					matMul2 ( K, t, t2 );

					vec(4) = t2(0);
					vec(5) = t2(1);
					vec(6) = t2(2);
				}

		};

	}

}

#endif // ADORE_PARAMETERIZATIONQUATERNIONCAMPARAM_H

