

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "prosac.h"
#include <adrcore/util/avgvar.h>

namespace adore
{

	namespace core
	{

		void PROSAC::operator() ( RANSACClient *client, const RansacParams params )
		{
			int maxIterations = params.getNumIterations();

			if ( maxIterations < 1 ) return;

			//if the samples contain muiltiple candidates, then
			const int numCandidates = client->getNumCandidates();

			const int sampleSize    = client->getSampleSize();

			if ( !client->isOkay() )
			{
				DBGOUT ( DBGCRITICAL, "!!! PROSAC: Ransac Client was not okay !!!\n" );
				return;
			}

			assert ( numCandidates > 0 );

			assert ( sampleSize > 0 );

			//initialize the parameters controlling the ransac interation
			float TN = 1000 * numCandidates;
			float Tn = TN;
			float Tnn = 0;

			for ( int i = 0; i < sampleSize; ++i )
			{
				Tn *= ( sampleSize - i ) /float ( numCandidates - i );
			}

			int N = INT_MAX;

			int scount = 0;//the t
			double minErr = DBL_MAX; //the minimum current error
			int mostIn = 0;
			double errRatio;
			const double prob = params.getProbability(); //the probability for at least one sample consisting exclusivly of inliers
			const double logdivident = log ( 1.0-prob );

			vector<int>sample ( sampleSize );

			NumberStore pool ( numCandidates, sampleSize );
			pool.init ( 0, numCandidates );

			//the ransac algorithm
			int Tin = 1;
			int n = sampleSize;//n is the number of all samples taken into account. In PROSAC, this gets increased gradually

			while ( N > scount )
			{
				assert ( scount <= Tin );

				if ( ( scount == Tin ) && ( n < numCandidates ) )
				{
					//take one more sample into account.
					createChoice ( sample.begin(), sampleSize - 1, n, pool );
					sample.back() = n;
					++n;
					Tnn = ( n + 1 ) / float ( n + 1 - sampleSize ) * Tn;
					float add = Tnn - Tn;
					assert ( add > 0 );
					Tn = Tnn;
					Tin += ( int ) ceil ( add );

				}
				else
				{
					createChoice ( sample.begin(), sampleSize, n, pool );
				}

				//print the sample --- debug
//				for ( int sss = 0; sss < sample.size(); ++sss )
//				{
//					printf( "%d ", sample[sss] );
//				}
//				printf( "\n" ); /*pressKey();*/

				bool success = client->applySample ( sample );

				if ( success )
				{
					//compute the number of inliers for every result and determin the maximum number of inliers.
//					printf( "getting inliers of %d results\n", numResults );
					double err = client->getError();
					int in  = client->getNumIn();

//					assert( in >= sampleSize );//may fail when not using minimum sample size

					DBGOUT ( DBGNORMAL, "numIn: %d, mostIn: %d, iterations-max: %d\n", in, mostIn, N );
//					if ( in > mostIn || ( in == mostIn && err <= minErr ) )

						if ( err < minErr )
						{
							bool optimized = client->optimize();

							if ( optimized )   //LO-step, modifies numIn and curErr
							{
								double oerr = client->getError();
								int oin  = client->getNumIn();
								//after optimization, there may be fewer inliers.
								//if this happens, then we fake the number of
								//inliers to be the greater one
//								err = MIN ( err, minErr );
//								in  = MAX ( in, mostIn );

								if ( oerr < err )
								{
									err = oerr;
									in = oin;
								}
							}

							mostIn = in;

							minErr = err;

							//recalculate N
							DBGOUT ( DBGNORMAL, "mostIn: %d/%d, minErr: %f, done: %d/%d\n", mostIn, numCandidates, minErr, scount, N );
							errRatio = mostIn / double ( numCandidates );

							if ( errRatio == 1.0 )
							{
								N = 0; //stops immediately.
							}
							else
							{
								double logdivisor = log ( 1.0-intPow ( errRatio, sampleSize ) );

								if ( -logdivisor > FLT_EPSILON )
								{
									N = ( int ) ( logdivident/logdivisor + 0.5 );

									if ( N < 0 ) N = INT_MAX;
								}

								DBGOUT ( DBGNORMAL, "New N! - done: %d/%d,  errRatio: %f\n", scount, N, errRatio );
							}
						}
				}
				else
				{
					DBGOUT ( DBGALL, "PROSAC: no result\n" );
				}

				if ( scount >= maxIterations )   //if no improvement occurrs, check for iteration limit
				{
					N = 0; //stops immediately.
				}

				++scount;
			} // end of ransac algorithm.

			client->ransacDone();
		}

	}
}
