

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_RANDOMRANSAC_H
#define ADORE_RANDOMRANSAC_H

#include <vector>
#include <set>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/debug.h>
#include <adrcore/random/integerrandomset.h>
#include <adrcore/random/ransacinterface.h>

namespace adore
{
	namespace core
	{

		using namespace std;

		/**
		 * A RANSAC filter.
		 * This is an abstract class. If you want to filter point correspondences using a RANSAC technique, feel free to derive.
		 * There are three methods you will have to implement (see below).
		 * Do not override the filter function!
		 *  
		*/
		class RANSAC
		{
			public:
				/**
				 * Constructor.
				 * @param numviews the supported number of views.
				 * @param numResults A sample can lead to more than one possible result. Tell how many results are there at most (e.g. 3 for both fundamental matrix and trifocal tensor).
				 * @param sample the number of samples used to calculate the results (e.g. 7 and 6 for fundmental matrix and trifocal tensor, respectivly).
				 * @param maxIts maximum allowed number of iterations.
				 */
				RANSAC( int numResults, int samples, int maxIts = 1500 );

				/**
				 * Destructor.
				 */
				virtual ~RANSAC();

				/**
				 * Main function. Fills the internal vectors.
				 * @param indices A vector of the vectors that should be filtered.
				 */
				void operator()( vector< int > &indices );


				void setClient( RANSACInterface* theValue )
				{
					user = theValue;
				}

				RANSACInterface* getClient() const
				{
					return user;
				}

				const vector<int> &getBestSample() const
				{
					return bestSample;
				}

				int getBestResult() const
				{
					return bestResult;
				}

			private:

				RANSACInterface *user;

				int maxIterations;
				int maxNumResults;
				uint sampleSize;
				IntegerRandomSet randSet;
				vector<int>sample;
				vector<int>bestSample;
				int bestResult;//of all results bestSample leads to, bestResult is the best.

				vector<int> outMask;
				vector<int> inMask;

				/**
				 * Creates a random subset for RANSAC
				 * @param indices the Correspondences.
				 * @param size the number of correspondences that should be taken into account.
				 */
				void createChoice( int size );

				/**
				 * Resolves ties between the results if they produce the same number of inliers. The variance of the inliers is here
				 * the criterium.
				 * @param ties an array of indices naming the tieing correspondences.
				 * @param numTies the number of ties.
				 * @param inlierFlags an array of vectors masking out the inliers.
				 * @param distvals the distance of every inlier.
				 */
				int resolveTies( int *ties, int numTies, vector<int> *inlierFlags, vector<double> *distVals );

				bool testChoice( int val );
		};

	}
}

#endif
