

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_SEARCHKDTREE_H
#define ADORE_SEARCHKDTREE_H

#include <cstdlib>
#include <cfloat>
//#include <queue>
#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/comparators.h>
#include <adrcore/types/astack.h>
#include <functional>
#include <cstring>
#include "collectors.h"

namespace adore
{
	namespace core
	{

		using namespace std;

		/**
		This is an implemetation of a kd-tree. A kd-tree is a partition of the k-space, that supports
		the search of the nearest neighbour in O(log n).
		The complexity of the construction is O(n log n).

		the template class T represents the type of data that is sought, e.g. in k-space, this might be vector&lt;double&gt;, each vector holding k elements.
		 *
		 * The Ruler measures the distance between two elements. Is must have the method
		 * sqrDist(const T &a, const T &b)
		 *
		 *
		 
		*/
		template<	class Iterator,
		class Ruler,
		class Indexer = ElemAtSB<typename Iterator::value_type>
		>
		class KDTree3
		{
			private:

				typedef typename Iterator::value_type T;
				typedef typename Ruler::distance_type DistT;
				typedef typename Indexer::value_type ElemT;

				class KDNode
				{
					public:
						T* sample;

						DistT *minCoors; //hyperrectangle
						DistT *maxCoors; //hyperrectangle
						float *vars;

						int split;
						KDNode* left;
						KDNode* right;

						const int dims;

					public:

						KDNode( int d ):sample( NULL ), split( 0 ), left( NULL ), right( NULL ), dims( d )
						{
							minCoors = new DistT[dims];
							maxCoors = new DistT[dims];
							vars = new float[dims];
						}

						/**
							does not really copy anything, but uses templ as a template for how much dimensions	to allocate
						*/
						KDNode( const KDNode &templ ):sample( NULL ), split( 0 ), left( NULL ), right( NULL ), dims( templ.dims )
						{
							minCoors = new DistT[dims];
							maxCoors = new DistT[dims];
							vars = new float[dims];
						}

						~KDNode()
						{
							delete [] minCoors;
							delete [] maxCoors;
							delete [] vars;
						}

						/**
						* @return true, if leaf.
						*  
						*/
						bool isLeaf() const
						{
							return ( !left && !right );
						}
				};

				/**
				     * A nested class essentially coupling a subtree with the information about the distance
				     * of its hyperrectangle to the current search point.
				     * This class is needed by the Best Bin Fit Algorithm.
				     *  
				     */
				class FurtherInfo
				{
					public:

						KDNode* node;///The tree node.
						DistT dist; ///The distance.

					public:

						/**
						    		Constructor. Initializes the subtree and his distance with the given parameters.
						    		 
						    	 */
						FurtherInfo( KDNode *f, DistT d ):node( f ), dist( d ) {;}

						/**
						    		This class is used in a priority_queue. That means, the &lt; operator would result in the biggest distance beeing first.
						    		 
						    	 */
						bool operator<( const FurtherInfo f ) const
						{
							return dist > f.dist;
						}
				};

			public:

				typedef KDTupleComparator<T, DistT> ComparatorT;
				typedef SingleBest<T, DistT> SingleBestT;
				typedef RangeCollector<T, DistT> RangeCollectorT;
				typedef NearestCollector<T, DistT> NearestCollectorT;

			private:
				const KDNode templ;//for vector creation
				vector<KDNode> nodes;
				const int dims;
				const Ruler ruler; ///measures distances
				int nextIdx;
				int depth;
				const Indexer idxAt;

			public:

				/**
					Enum to choose the type of pivot calculation for the kd-tree.
					Currently, only MEDIAN is supported. That means, the median of all values in the
					current dimension is used to define the hyperplane splitting the space into hyperrectangles.
				*/
				enum PIVOT_TYPE
				{
					MEDIAN
				};

				/**
				* Constructs a KDTree3. Side effect: Depending on the pivot search, the element positions between begin and end are not preserved.
				* Any previous sorting etc. has to be done again.
				* @param begin pointer to the first element in the sample vector
				* @param end pointer BEHIND the last entry in the sample-vector (The usual stl-convention).
				* @param k number of dimensions
				* @param splitDim the dimension that is split creating the two hyperrectangles.
				* @param p method for choosing the pivot that splits the current hyperrectangle. Defaults to MEDIAN.
				* @param min the minimum co-ordinates of the initial hyperrectangle. Leaving it NULL is perfectly legal. Else, the entries are copied.
				* @param max the maximum co-ordinates of the initial hyperrectangle. Leaving it NULL is perfectly legal. Else, the entries are copied.
				*
				*  
				*/
				KDTree3( Iterator begin,
				         Iterator end,
				         int numDims,
				         Ruler &r,
				         PIVOT_TYPE p = MEDIAN ):
						templ( numDims ),
						nodes( distance( begin, end ), templ ),
						dims( numDims ),
						ruler( r ), nextIdx( 0 ), depth( 0 ), idxAt()
				{
					assert( dims > 1 );//else some loops may break

					typedef tuple<Iterator, Iterator, int, bool, int> TodoElem;//interval + parent idx + left? + depth

					if ( begin == end )
					{
						return;
					}

					//first node is special
					//add node
					KDNode &node = nodes[nextIdx];
					computeVariances( begin, end, node.vars );
					node.split = getSplitDim( node.vars );
					Iterator sampleIdx = getPivotAndSort( begin, end, node.split, p );
					node.sample = &( *sampleIdx );
					assert( node.sample );
					for ( int i = dims - 1; i >= 0; --i )
					{
						node.minCoors[i] = minLimit<DistT>();
						node.maxCoors[i] = maxLimit<DistT>();
					}

					AStack< TodoElem > todo( MAX( 10, static_cast<int>( distance( begin, end ) ) ) );

					if ( begin != sampleIdx )
					{
						todo.push( TodoElem( begin, sampleIdx, nextIdx, true, 1 ) );
					}
					if ( sampleIdx + 1 != end )
					{
						todo.push( TodoElem( sampleIdx + 1, end, nextIdx, false, 1 ) );
					}
					++nextIdx;

					//other nodes
					while ( !todo.empty() )
					{
//						printf( "stack size: %d, index: %d\n", todo.size(), nextIdx );
						TodoElem interval = todo.top();
						todo.pop();

						Iterator &start = get<0>( interval );
						Iterator &stop = get<1>( interval );
						int parentIdx = get<2>( interval );
						bool left = get<3>( interval );
						int currentDepth = get<4>( interval ) + 1;

						KDNode &parent = nodes[parentIdx];

//						printf( "interval: %d - %d\n", distance( begin, start ), distance( begin, stop ) );

						KDNode &node = nodes[nextIdx];
						memcpy( node.vars, parent.vars, dims * sizeof( float ) );

						assert( ensurePartition( start, stop, parent, left ) );

						node.vars[parent.split] = computeVariance( start, stop, parent.split );
//						printVars( node.vars );
						node.split = getSplitDim( node.vars );
//						printf( "split idx = %d\n", node.split );

						Iterator sampleIdx = getPivotAndSort( start, stop, node.split, p );

						node.sample = &( *sampleIdx );
						assert( node.sample );
						memcpy( node.minCoors, parent.minCoors, dims * sizeof( DistT ) );
						memcpy( node.maxCoors, parent.maxCoors, dims * sizeof( DistT ) );
						if ( left )
						{
							node.maxCoors[parent.split] = idxAt( *parent.sample, parent.split );
							parent.left = &node;
						}
						else
						{
							node.minCoors[parent.split] = idxAt( *parent.sample, parent.split );
							parent.right = &node;
						}

						if ( start != sampleIdx )
						{
							todo.push( TodoElem( start, sampleIdx, nextIdx, true, currentDepth ) );
						}
						if ( sampleIdx + 1 != stop )
						{
							todo.push( TodoElem( sampleIdx + 1, stop, nextIdx, false, currentDepth ) );
						}

//						depth = MAX( depth, ( int )todo.size() );
						depth = MAX( depth, currentDepth );

						++nextIdx;
					}
					assert_msg( ( int )nodes.size() == nextIdx, "not all nodes put into tree (%d <-> %zd)\n", nextIdx, nodes.size() );
					assert_msg( ( int )nodes.size() == distance( begin, end ), "not all nodes put into tree (%d <-> %zd)\n", nextIdx, nodes.size() );
					DBGOUT( DBGNORMAL, "KD-Tree constructed with depth %d\n", depth );
				}

				/**
				* Simple destructor
				*  
				*/
				~KDTree3() {}

				bool isEmpty()
				{
					return nextIdx == 0;
				}

				/**
					returns number of visited nodes
				 */
				int findNearestNeighbour( const T& sought )
				{
					FakeCollector DummyCollector;
					return findNearestNeighbour<FakeCollector>( sought, DummyCollector );
				}

				/**
					returns number of visited nodes
				*/
				template<class Collector>
				int findNearestNeighbour( const T& sought, Collector &collector )
				{
					typedef tuple< KDNode*, KDNode*> PathT;//node, further

					if ( isEmpty() ) return 0;

					int visited = 1;

					AStack< PathT > path( ( depth << 2 ) + 1 );

					KDNode *node = NULL;
					KDNode *nextNode = &nodes[0];//get root
					assert( nextNode->sample );
					while ( nextNode )
					{
						++visited;
						node = nextNode;
						assert( node->sample );
						assert( node->split < dims );
						ElemT sVal = idxAt( sought, node->split );
						ElemT nVal = idxAt( *node->sample, node->split );
						if ( sVal <= nVal )
						{
							path.push( PathT( node, node->right ) );
							nextNode = node->left;
						}
						else
						{
							path.push( PathT( node, node->left ) );
							nextNode = node->right;
						}
					}

					//now check the other leafes that lie in range of distSqr
					while ( !path.empty() )
					{
						PathT toppy = path.top();
						path.pop(); ++visited;
						KDNode *node = toppy.get<0>();

						DistT candDist = ruler.sqrDist( sought, *node->sample );
//						printf( "candDist : %f\n", candDist );
						collector.insert( node->sample, candDist );
						DistT distSqr = collector.getDist();
						if ( distSqr == 0 ) return visited;

						KDNode *further = toppy.get<1>();
						if ( further && furtherIsWithinReach( distSqr, sought, further ) )
						{
//							printf( "searching further\n" );
							node = further;
							while ( node )
							{
								if ( idxAt( sought, node->split ) < idxAt( *node->sample, node->split ) )
								{
									path.push( PathT( node, node->right ) );
									node = node->left;
								}
								else
								{
									path.push( PathT( node, node->left ) );
									node = node->right;
								}
							}
						}
					}

					return visited;
				}

				int findNearestNeighbourBBF( const T& sought, int maxNumNodes )
				{
					FakeCollector DummyCollector;
					return findNearestNeighbourBBF<FakeCollector>( sought, maxNumNodes, DummyCollector );
				}

				template<class Collector>
				int findNearestNeighbourBBF( const T& sought, int maxNumNodes, Collector &collector )
				{
					if ( isEmpty() ) return 0;

					int visited = 1;

					FixedPQ<FurtherInfo> bestFurther( ( depth << 2 ) + 1 );
					//set initial minimum distance, usually a huge value
					DistT distSqr = collector.getDist();

					bestFurther.push( FurtherInfo( &nodes[0], 0 ) );

					while ( !bestFurther.empty() )
					{
						FurtherInfo fi = bestFurther.top();
						bestFurther.pop();
						if ( fi.dist < distSqr )
						{
							KDNode *node = NULL;
							KDNode *nextNode = fi.node;
							assert( nextNode->sample );
							while ( nextNode )
							{
								node = nextNode;
								assert( node->sample );
								assert( node->split < dims );
								++visited;

								distSqr = ruler.sqrDist( sought, *node->sample );
								collector.insert( node->sample, distSqr );
								distSqr = collector.getDist();
								if ( distSqr == 0.0F || visited >= maxNumNodes ) return visited;

								KDNode *nearer = node->right;
								KDNode *further = node->left;
								ElemT sVal = idxAt( sought, node->split );
								ElemT nVal = idxAt( *node->sample, node->split );
								if ( sVal <= nVal )
								{
									nearer = node->left;
									further = node->right;
								}
								nextNode = nearer;

								if ( further )
								{
									DistT dist = getDistance( sought, further );
									if ( dist < distSqr ) bestFurther.push( FurtherInfo( further, dist ) );
								}
							}
						}
					}

					return visited;
				}

				/**
					debug. tests whether the tree was constructed correctly
				*/
				bool isValid()
				{
					if ( isEmpty() ) return true;
					KDNode *node = &nodes[0];
					assert( node );
					if ( node->isLeaf() ) return true;

					AStack<KDNode *> todo( depth + 2 );
					todo.push( node );

					while ( !todo.empty() )
					{
						node = todo.top(); todo.pop();
						assert( node );
						KDNode *left = node->left;
						KDNode *right = node->right;
						if ( left )
						{
							if ( idxAt( *left->sample, node->split ) > idxAt( *node->sample, node->split ) ) return false;
							todo.push( left );
						}
						if ( right )
						{
							if ( idxAt( *right->sample, node->split ) <= idxAt( *node->sample, node->split ) ) return false;
							todo.push( right );
						}
					}
					return true;
				}

			private:
				/**
					a debug function ensuring the median property
				*/
				bool ensurePartition( Iterator start, Iterator stop, KDNode &parent, bool left )
				{
					for ( ; start != stop; ++start )
					{
						if ( left && idxAt( *parent.sample, parent.split ) < idxAt( *start, parent.split ) )
						{
							cout << "ensurePartition error left: " << idxAt( *parent.sample, parent.split ) << " < " << idxAt( *start, parent.split ) << endl;
							return false;
						}

						if ( !left && idxAt( *parent.sample, parent.split ) >= idxAt( *start, parent.split ) )
						{
							cout << "ensurePartition error right: " << idxAt( *parent.sample, parent.split ) << " >= " << idxAt( *start, parent.split ) << endl;
							return false;
						}
					}
					return true;
				}

				void computeVariances( Iterator begin, Iterator end, float *vars )
				{
					for ( int s = dims - 1; s >= 0; --s )
					{
						vars[s] = computeVariance( begin, end, s );
					}
				}

				float computeVariance( Iterator begin, Iterator end, int s )
				{
					float sum = 0.0F;
					int size = distance( begin, end );

					Iterator start = begin;
					for ( int i = size - 1 ; i >= 0; --i, ++start )
					{
						sum += idxAt( *start, s );
					}
					float avg = sum / float( size );

					sum = 0.0F;
					for ( int i = size - 1 ; i >= 0; --i, ++begin )
					{
						float diff = avg - idxAt( *begin, s );
						sum += diff * diff;
					}
					return sum / float( size );
				}

				int getSplitDim( float *vars )
				{
					int maxIdx = 0;
					float max = vars[dims - 1];
					for ( int i = dims - 2; i >= 0; --i )
					{
						if ( vars[i] > max )
						{
							max = vars[i];
							maxIdx = i;
						}
					}
					return maxIdx;
				}

				/**
					debug function
				*/
				void printVars( float *vars )
				{
					printf( "variances:\n" );
					for ( int i = 0; i < dims; ++i )
					{
						printf( "%.2f ", vars[i] );
					}
					printf( "\n" );
				}

				/**
				*
				* @param begin Begin of the collection of the canditates for the pivot choice
				* @param end End of the collection of the canditates for the pivot choice
				* @param p The method to choose the pivot element
				* @return Iterator pointing to the chosen pivot element.
				*/
				Iterator getPivotAndSort( Iterator begin, Iterator end, int split, PIVOT_TYPE p )
				{
					using std::swap;

					Iterator med;
					IndexComparator3< T, Indexer > ic( split );

					assert( begin != end );

					switch ( p )
					{
						case MEDIAN:
						{
							int len = distance( begin, end );
							med = begin + len / 2;
							nth_element( begin, med, end, ic );

							//in the remaining right part, elements of the same
							//value may exist. The pivot has to be moved to
							//ensure these are left of it.
							ElemT ref = idxAt( *med, split );

							Iterator swappy = med + 1;
							while ( ( swappy != end ) && ( idxAt( *swappy, split ) == ref ) ) ++swappy;
							if ( swappy == end )
							{
#ifndef NDEBUG
								if ( len > 2 ) { DBGOUT( DBGALL, "KD-Tree3: Degenerate case (len = %d)\n", len );}
#endif
								return end - 1;//degenerate case
							}

							Iterator runner = swappy + 1;
							for ( ;runner != end; ++runner )
							{
								if ( idxAt( *runner, split ) == ref )
								{
									swap( *runner, *swappy );
									++swappy;
								}
							}
							assert( swappy != end );//degenerate case should have been caught before
							med = swappy - 1;

							return med;
						}
						default:
							error( "Unsupported pivot type" );
							break;
					}
					error( "Unsupported pivot type" );
					return med;
				}

				bool furtherIsWithinReach( DistT reach, const T& sought, KDNode*further )
				{
					DistT d = getDistance( sought, further );
//					printf( "further dist was %f\n", d );
					return d < reach;
				}

				/**
					Straight forward computatin of the distance between "sought" and the hyperrectangle of "further"
				*/
				DistT getDistance( const T& sought, KDNode*further )
				{
					assert( further );

					DistT *min = further->minCoors;
					DistT *max = further->maxCoors;
					DistT dist = ( DistT )0;

					for ( int i = dims - 1; i >= 0; --i )
					{
						if ( idxAt( sought, i ) < min[i] )
						{
							dist += SQR( idxAt( sought, i ) - min[i] );
						}
						else if ( idxAt( sought, i ) > max[i] )
						{
							dist += SQR( idxAt( sought, i ) - max[i] );
						}
						//else 0 is added
					}

					return dist;
				}

		};

	}
}

#endif
