

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef _ADORE_GEOMETRY3_
#define _ADORE_GEOMETRY3_

#include <cassert>
#include <cmath>
#include <list>
#include <set>
#include <algorithm>

namespace adore
{
	namespace core
	{
		/**	3D point
		 *
		 *	A standard mathematical 3 dimensional point with
		 *	double precision coordinates.
		 *
		 *	 
		 */
		class Point3D
		{
				double v[3];			///< point coordinates

			public:
				/**	Default constructor
				 */
				Point3D()
				{
				};

				/**	Constructor with initialization
				 */
				Point3D( double x, double y, double z )
				{
					v[0] = x;
					v[1] = y;
					v[2] = z;
				};

				/**	Access point components as an array
				 */
				double& operator[]( int i )
				{
					return v[i];
				};

				/**	Access point components as an array
				 */
				const double operator[]( int i ) const
				{
					return v[i];
				};

				/**	Point metric (squared distance to the origin)
				 */
				double operator()() const
				{
					return v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
				};

				/** Get mirror point about the origin
				 */
				Point3D operator-() const
				{
					return Point3D( -v[0],-v[1],-v[2] );
				};

				/**	Vector-like dot product
				 */
				double operator*( const Point3D& a ) const
				{
					return v[0]*a.v[0]+v[1]*a.v[1]+v[2]*a[2];
				};

				/**	Vector-like cross product
				 */
				Point3D operator%( const Point3D& a ) const
				{
					return Point3D( v[1]*a[2] - v[2]*a[1], v[2]*a[0] - v[0]*a[2], v[0]*a[1] - v[1]*a[0] );
				};

				/**	Vector-like point-scalar product
				 */
				Point3D operator*( double a )
				{
					return Point3D( v[0]*a,v[1]*a,v[2]*a );
				};

				/**	Vector-like point-inverse scalar product
				 */
				inline Point3D operator/( const double a ) const
				{
					double inv_a = 1.0 / a;
					return Point3D( v[0]*inv_a,v[1]*inv_a,v[2]*inv_a );
				};

				/**	Vector-like point addition
				 */
				Point3D operator+( const Point3D& a ) const
				{
					return Point3D( v[0]+a.v[0],v[1]+a.v[1],v[2]+a[2] );
				};

				/**	Vector-like point subtraction
				 */
				Point3D operator-( const Point3D& a ) const
				{
					return Point3D( v[0]-a.v[0],v[1]-a.v[1],v[2]-a[2] );
				};

				/**	Point equality operator
				 */
				bool operator==( const Point3D& a ) const
				{
					return v[0]==a.v[0] && v[1]==a.v[1] && v[2]==a.v[2];
				};

				/**	Point inequality operator
				 */
				bool operator!=( const Point3D& a ) const
				{
					return v[0]!=a.v[0] || v[1]!=a.v[1] || v[2]!=a[2];
				};

				/**	Distance to the origin
				 */
				double length()
				{
					return std::sqrt( v[0]*v[0]+v[1]*v[1]+v[2]*v[2] );
				};
		};

		/**	Vector-like point-scalar product
		 */
		inline Point3D operator*( const double a, const Point3D b )
		{
			return Point3D( b[0]*a,b[1]*a,b[2]*a );
		}

		/**	3D face
		 *
		 *	A standard mathematical 3 dimensional face, which is triangle
		 *	in 3D space.
		 *
		 *	@warning Vertex indices must not excceed the capacity of 16 bits
		 *
		 *	 
		 */
		class Face3D
		{
				unsigned v[3];				///< face vertices

				/**	Pack vertex indices in one 32 bit word
				 */
				unsigned pack() const
				{
					unsigned src[3];
					sort( src );
					return ( v[0] & 0x0000FFFF ) | ( ( v[1]<<16 ) ^ ( ( v[2] & 0x00FFFFFFFF )<<8 ) );
				};

				/**	sort vertices
				 */
				void sort( unsigned out[3] ) const
				{
					if ( v[0] > v[1] && v[0] > v[2] )
					{
						out[0] = v[0];
						if ( v[1] > v[2] )
						{
							out[1] = v[1];
							out[2] = v[2];
						}
						else
						{
							out[1] = v[2];
							out[2] = v[1];
						}
					}
					else if ( v[1] > v[2] )
					{
						out[0] = v[1];
						if ( v[2] > v[0] )
						{
							out[1] = v[2];
							out[2] = v[0];
						}
						else
						{
							out[1] = v[0];
							out[2] = v[2];
						}
					}
					else
					{
						out[0] = v[2];
						if ( v[1] > v[0] )
						{
							out[1] = v[1];
							out[2] = v[0];
						}
						else
						{
							out[1] = v[0];
							out[2] = v[1];
						}
					}
				};

			public:
				/**	Default constructor
				 */
				Face3D()
				{
				};

				/**	Constructor with initialization
				 */
				Face3D( unsigned a, unsigned b, unsigned c )
				{
					v[0] = a;
					v[1] = b;
					v[2] = c;
				};

				/**	Access face vertices as an array
				 */
				unsigned& operator[]( int i )
				{
					return v[i];
				};

				/**	Access face vertices as an array
				 */
				const unsigned operator[]( int i ) const
				{
					return v[i];
				};

				/**	Face equality operator
				 */
				bool operator==( const Face3D& rv ) const
				{
					unsigned a[3];
					unsigned b[3];

					sort( a );
					rv.sort( b );

					return a[0]==b[0] && a[1]==b[1] && a[2]==b[2];
				};

				/**	Face equality operator
				 */
				bool operator!=( const Face3D& rv ) const
				{
					unsigned a[3];
					unsigned b[3];

					sort( a );
					rv.sort( b );

					return a[0]!=b[0] || a[1]!=b[1] || a[2]!=b[2];
				};

				/**	Face comparison operator
				 */
				bool operator<( const Face3D& rv ) const
				{
					unsigned a[3];
					unsigned b[3];

					sort( a );
					rv.sort( b );

					if ( a[0] < b[0] )
						return true;
					else if ( a[1] < b[1] )
						return true;
					else if ( a[2] < b[2] )
						return true;
					else
						return false;
				};

				/**	Face packing for hash operation
				 */
				operator size_t() const
				{
					return static_cast<size_t>( pack() );
				};
		};

		/**	Determine the signed triangle area, depending on the triangle winding.
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return Signed triangle area
		 */
		inline double area( Point3D* src, unsigned a, unsigned b, unsigned c )
		{
			Point3D u = src[b] - src[a];
			Point3D v = src[c] - src[a];
			Point3D s = u % v;
			return 0.5 * std::sqrt( s*s );
		}

		/**	Determine the signed tetrahedra volume, depending on the vertex order.
		 *	@param src Source array of points
		 *	@param a Index of the first tetrahedra vertex in the source index
		 *	@param b Index of the second tetrahedra vertex in the source index
		 *	@param c Index of the third tetrahedra vertex in the source index
		 *	@param d Index of the fourth tetrahedra vertex in the source index
		 *	@return Signed triangle area
		 */
		inline double volume( Point3D* src, unsigned a, unsigned b, unsigned c, unsigned d )
		{
			Point3D u = src[b] - src[a];
			Point3D v = src[c] - src[a];
			Point3D w = src[d] - src[a];
			double det = u[0]*v[1]*w[2] + v[0]*w[1]*u[2] + w[0]*u[1]*v[2]
			             - w[0]*v[1]*u[2] - v[0]*u[1]*w[2] - u[0]*w[1]*v[2];
			return 0.5 * det;
		}

		/**	Determine the halfspace of a point relative to the face
		 *	@param src Source array of points
		 *	@param f Face with indices indicating points in the source array
		 *	@param a Point for which to determine the halfspace
		 *	@return -1 for negative, +1 for positiove halfspace, 0 if the point lies on line of the edge
		 */
		inline int halfspace( Point3D* src, Face3D f, Point3D a, double epsilon )
		{
			Point3D u = src[f[1]] - src[f[0]];
			Point3D v = src[f[2]] - src[f[0]];
			Point3D s = u % v;
			Point3D p = a - src[f[0]];
			double alpha = p*s;

			if ( alpha > epsilon )
				return -1;
			else if ( alpha < epsilon )
				return 1;
			else
				return 0;
		}

		/**	3D simplex
		 *
		 *	A standard mathematical 2 dimensional simplex, which is a
		 *	tetrahedron in 3D space.
		 *
		 *	 
		 */
		class Simplex3D
		{
				unsigned v[4];

			public:
				/**	Default constructor
				 */
				Simplex3D()
				{
				};

				/**	Constructor with initialization
				 */
				Simplex3D( unsigned a, unsigned b, unsigned c, unsigned d )
				{
					v[0] = a;
					v[1] = b;
					v[2] = c;
					v[3] = d;
				};

				/** Make null simplex
				 */
				static Simplex3D zero()
				{
					return Simplex3D( 0,0,0,0 );
				};

				/**	Access simplex vertices as an array
				*/
				unsigned& operator[]( int i )
				{
					return v[i];
				};

				/**	Access simplex vertices as an array
				 */
				const unsigned operator[]( int i ) const
				{
					return v[i];
				};

				/**	Test if the simplex is null.
				 *	The simplex is null if any of it's two vertices are identical.
				 */
				bool nullSimplex()
				{
					return v[0]==v[1] || v[0]==v[2] || v[0]==v[3] ||
					       v[1]==v[2] || v[1]==v[3] || v[2]==v[3];
				};

				Face3D face( unsigned i )
				{
					switch ( i )
					{
						case 0:
							return Face3D( v[0],v[1],v[2] );
						case 1:
							return Face3D( v[0],v[2],v[3] );
						case 2:
							return Face3D( v[0],v[3],v[1] );
						case 3:
							return Face3D( v[3],v[2],v[1] );
						default:
							return Face3D( 0,0,0 );
					};
				};
		};

		/**	3D circle
		 *
		 *	 
		 */
		struct Circle3D
		{
			Point3D center;					///< circle center
			double radius;					///< circle radius
		};

		/**	3D Plane
		 *
		 * 	 
		 */
		struct Plane3D
		{
			Point3D centroid;				///< Point on plane
			Point3D normal;					///< Normal to plane
		};

		/**	Determine the radius of the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumradius
		 */
		inline double circumradius( Point3D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			double dca = ( src[c]-src[a] ) * ( src[b]-src[a] );
			double dba = ( src[c]-src[b] ) * ( src[a]-src[b] );
			double dcb = ( src[a]-src[c] ) * ( src[b]-src[c] );

			double radius = 0.5 * std::sqrt( ( dca+dba )*( dba+dcb )*( dcb+dca ) / ( dba*dcb+dcb*dca+dca*dba ) );

			return radius;
		}

		/**	Determine the center of the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumcenter
		 */
		inline Point3D circumcenter( Point3D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			double dca = ( src[c]-src[a] )*( src[b]-src[a] );
			double dba = ( src[c]-src[b] )*( src[a]-src[b] );
			double dcb = ( src[a]-src[c] )*( src[b]-src[c] );

			double n1 = dba * dcb;
			double n2 = dcb * dca;
			double n3 = dca * dba;

			Point3D center = 0.5 * ( ( n2+n3 )*src[a]+( n3+n1 )*src[b]+( n1+n2 )*src[c] )/( n1+n2+n3 );

			return center;
		}

		/**	Determine the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumcirlce
		 */
		inline Circle3D circumcircle( Point3D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			Circle3D result;

			double dca = ( src[c]-src[a] ) * ( src[b]-src[a] );
			double dba = ( src[c]-src[b] ) * ( src[a]-src[b] );
			double dcb = ( src[a]-src[c] ) * ( src[b]-src[c] );

			double n1 = dba * dcb;
			double n2 = dcb * dca;
			double n3 = dca * dba;

			result.center = 0.5 * ( ( n2+n3 )*src[a]+( n3+n1 )*src[b]+( n1+n2 )*src[c] )/( n1+n2+n3 );
			result.radius = 0.5 * std::sqrt( ( dca+dba )*( dba+dcb )*( dcb+dca ) / ( dba*dcb+dcb*dca+dca*dba ) );

			return result;
		}


		/**	Determine the radius of the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumradius
		 */
		inline double circumradius( Point3D* src, unsigned a, unsigned b, unsigned c, unsigned d )
		{
			assert( a!=b && a!=c && a!=d && b!=c && b!=d && c!=d );

			Point3D ba = src[b] - src[a];
			Point3D ca = src[c] - src[a];
			Point3D da = src[d] - src[a];

			Point3D bxc = ba % ca;
			Point3D dxb = da % ba;
			Point3D cxd = ca % da;

			Point3D num = ( da*da )*bxc+( ca*ca )*dxb+( ba*ba )*cxd;
			double denom = 2.0*( ba*cxd );
			double radius = std::sqrt( num*num ) / denom;

			return std::abs( radius );
		}

		/**	Determine the center of the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumcenter
		 */
		inline Point3D circumcenter( Point3D* src, unsigned a, unsigned b, unsigned c, unsigned d )
		{
			assert( a!=b && a!=c && a!=d && b!=c && b!=d && c!=d );

			Point3D ba = src[b] - src[a];
			Point3D ca = src[c] - src[a];
			Point3D da = src[d] - src[a];

			Point3D bxc = ba % ca;
			Point3D dxb = da % ba;
			Point3D cxd = ca % da;

			Point3D num = ( da*da )*bxc+( ca*ca )*dxb+( ba*ba )*cxd;
			double denom = 2*( ba*cxd );
			Point3D center = src[a] + num / denom;

			return center;
		}

		/**	Determine the circumcirlce of triangle
		 *	@param src Source array of points
		 *	@param a Index of the first triangle vertex in the source index
		 *	@param b Index of the second triangle vertex in the source index
		 *	@param c Index of the third triangle vertex in the source index
		 *	@return circumcirlce
		 */
		inline Circle3D circumcircle( Point3D* src, unsigned a, unsigned b, unsigned c, unsigned d )
		{
			assert( a!=b && a!=c && a!=d && b!=c && b!=d && c!=d );

			Circle3D result;

			Point3D ba = src[b] - src[a];
			Point3D ca = src[c] - src[a];
			Point3D da = src[d] - src[a];

			Point3D bxc = ba % ca;
			Point3D dxb = da % ba;
			Point3D cxd = ca % da;

			Point3D num = ( da*da )*bxc+( ca*ca )*dxb+( ba*ba )*cxd;
			double denom = 2*( ba*cxd );
			result.center = src[a] + num / denom;
			result.radius = std::abs( std::sqrt( num*num ) / denom );
			return result;
		}

		/**	Face set that stores its members in an STL-type set.
		 *	Set operations are made explicitly.
		 *
		 *	 
		 */
		template <class T = std::set<Face3D> > class ActiveFaceList3D
		{
				T storage;						///< face storing container
				typename T::iterator it;		///< container iterator

			public:
				/**	Set container iterator to the beginning of the container
				 */
				void start()
				{
					it = storage.begin();
				};

				/**	Increment container iterator and return if the end of the container is not reached yet
				 *	@param f current iterated element, if the end is not reached
				 */
				bool next( Face3D& f )
				{
					if ( it==storage.end() )
						return false;
					else
					{
						f = *it;
						++it;
						return true;
					}
				};

				/**	Insert face into the set
				 *	@param f Face to insert
				 */
				void insert( Face3D f )
				{
					storage.insert( f );
				};

				/**	Remove and return the first element
				 */
				Face3D extract()
				{
					typename T::iterator iter = storage.begin();
					Face3D result = *iter;
					storage.erase( iter );
					return result;
				};

				/**	Remove face from the set
				*	@param f Face to remove
				*/
				void erase( Face3D f )
				{
					storage.erase( f );
				};

				/**	Test if the face is already in the set
				 *	@param f Face to test
				 */
				bool member( Face3D f )
				{
					return storage.find( f )!=storage.end();
				};

				/**	Update face in the set.
				 *	The face is inserted into the set if it is not in the set
				 *	The face is removed from the set if it is already in the set
				 *	@param f Face to update
				 */
				void update( Face3D f )
				{
					typename T::iterator iter = storage.find( f );
					if ( iter==storage.end() )
						storage.insert( f );
					else
						storage.erase( iter );
				};

				/**	Get number of elements
				 */
				unsigned size()
				{
					return static_cast<unsigned>( storage.size() );
				};

				/**	Test if the set is empty
				 */
				bool empty()
				{
					return storage.empty();
				};
		};

		/**	Face set that stores its members in an STL list.
		 *	Set operations are made explicitly.
		 *
		 *	 
		 */
		class FaceListSet3D
		{
				std::list<Face3D> storage;				///< face storing container
				std::list<Face3D>::iterator it;			///< container iterator

			public:
				/**	Set container iterator to the beginning of the container
				 */
				void start()
				{
					it = storage.begin();
				};

				/**	Increment container iterator and return if the end of the container is not reached yet
				 *	@param f current iterated element, if the end is not reached
				 */
				bool next( Face3D& f )
				{
					if ( it==storage.end() )
						return false;
					else
					{
						f = *it;
						++it;
						return true;
					}
				};

				/**	Insert face into the set
				 *	@param f Face to insert
				 */
				void insert( Face3D f )
				{
					std::list<Face3D>::iterator eli = storage.begin();
					while ( eli!=storage.end() )
					{
						if ( *eli==f )
							return;
						++eli;
					}
					storage.push_back( f );
				};

				/**	Remove and return the first element
				 */
				Face3D extract()
				{
					std::list<Face3D>::iterator eli = storage.begin();
					Face3D result = *eli;
					storage.erase( eli );
					return result;
				};

				/**	Remove face from the set
				 *	@param f Face to remove
				 */
				void erase( Face3D f )
				{
					std::list<Face3D>::iterator eli = storage.begin();
					while ( eli!=storage.end() )
					{
						if ( *eli==f )
						{
							eli = storage.erase( eli );
							return;
						}
						++eli;
					}
				};

				/**	Test if the face is already in the set
				 *	@param f Face to test
				 */
				bool member( Face3D f )
				{
					std::list<Face3D>::iterator eli = storage.begin();
					while ( eli!=storage.end() )
					{
						if ( *eli==f )
							return true;
						++eli;
					}
					return false;
				};

				/**	Update face in the set.
				 *	The face is inserted into the set if it is not in the set
				 *	The face is removed from the set if it is already in the set
				 *	@param f Face to update
				 */
				void update( Face3D f )
				{
					if ( member( f ) )
						erase( f );
					else
						insert( f );
				};

				/**	Get number of elements
				 */
				unsigned size()
				{
					return static_cast<unsigned>( storage.size() );
				};

				/**	Test if the set is empty
				 */
				bool empty()
				{
					return storage.empty();
				}
		};

		/**	Simplex list that stores its members in STL-type sequence container
		 *
		 *	 
		 */
		template<class T> class SimplexStore3D
		{
				T storage;					///< face storing container
				typename T::iterator it;	///< container iterator

			public:
				/**	Set container iterator to the beginning of the container
				 */
				void start()
				{
					it = storage.begin();
				};

				/**	Increment container iterator and return if the end of the container is not reached yet
				 *	@param f current iterated element, if the end is not reached
				 */
				bool next( Simplex3D& t )
				{
					if ( it==storage.end() )
						return false;
					else
					{
						t = *it;
						++it;
						return true;
					}
				};

				/**	Insert the simplex into the list
				 *	@param t Simplex to add
				 */
				void insert( Simplex3D& t )
				{
					storage.push_back( t );
				};
		};

		/**	Simplex list that outputs its members directly to the receiver
		 *
		 *	 
		 */
		class SimplexCallback3D
		{
				typedef void ( *callback )( unsigned* v, void* extra );	///< Triangle callback function prototype
				callback receiver;					///< Triangle callback function
				void* eparam;						///Extra parameter for callback function

			public:
				/**	Constructor
				 *	@param cb Triangle callback function
				 *	@param extra Extra parameter for callback function
				 */
				SimplexCallback3D( callback cb, void* extra )
						:receiver( cb ),eparam( extra ) {};

				/**	Insert the simplex into the list
				 *	@param t Simplex to add
				 */
				void insert( Simplex3D& t )
				{
					receiver( ( unsigned* )&t,eparam );
				};
		};

		/**	Type definitions
		 */
		typedef FaceListSet3D FaceList3D;
		typedef SimplexStore3D< std::list<Face3D> > SimplexStoreList3D;
		typedef SimplexCallback3D SimplexList3D;
	}
}

#endif	//_ADORE_GEOMETRY3_
