

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <stdint.h>
#include <adrcore/types/aarray.h>
#include <adrcore/fortran/fortran_switch.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/lapackutil.h>
#include <adrcore/util/storageorder.h>
#include <utility>

namespace adore
{
	namespace core
	{

		//returns the permutation info. row i was interchenged with row return(i)
		//A is changed
		AArray<int, 1> lu( AArray<double, 2> &A )
		{
			integer rowsA = A.rows();
			integer colsA = A.cols();

			double cmArray[rowsA * colsA];
			checkStorageOrder( A );
			rowToColumnMajor( A.begin(), cmArray, rowsA, colsA );

			integer LDA = rowsA;

			integer returnVal = 0;
			//these "new" are necessary to prevent hitting the stack limit
			AArray<int, 1> IPIV( min( rowsA, colsA ) );

			dgetrf_( &rowsA, &colsA,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(), &returnVal );

			assert( returnVal >= 0 );

			columnToRowMajor( cmArray, A.begin(), rowsA, colsA );

			return IPIV;//OK6
		}

		AArray<int, 1> lu( AArray<float, 2> &A )
		{
			integer rowsA = A.rows();
			integer colsA = A.cols();

			float cmArray[rowsA * colsA];
			checkStorageOrder( A );
			rowToColumnMajor( A.begin(), cmArray, rowsA, colsA );

			integer LDA = rowsA;

			integer returnVal = 0;
			//these "new" are necessary to prevent hitting the stack limit
			AArray<int, 1> IPIV( min( rowsA, colsA ) );

			sgetrf_( &rowsA, &colsA,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(), &returnVal );

			assert( returnVal >= 0 );

			columnToRowMajor( cmArray, A.begin(), rowsA, colsA );

			return IPIV;
		}

		double detN( const AArray<double, 2> &A )
		{
                        AArray<double, 2> LU = A;
			AArray<int, 1> perm = lu( LU );

			assert( perm.rows() == LU.rows() );
			assert( LU.cols() == LU.rows() );

			double ret = 1.0;
			double sign = 1.0;
			for ( unsigned i = 0; i < perm.rows(); ++i )
			{
				if ( perm( i ) != (int)i + 1 ) //i+1, because fortran is 0-based
				{
					sign *= -1.0;
				}
				ret *= LU( i, i );
			}
			return sign * ret;
		}

		bool invert( AArray<double, 2> &LU, AArray<int, 1> &IPIV )
		{
			assert( LU.rows() == LU.cols() );

			integer order = LU.rows();
			integer LDA = order;

			double cmArray[order * order];
			checkStorageOrder( LU );
			rowToColumnMajor( LU.begin(), cmArray, order, order );

			integer returnVal = 0;
			AArray<double, 1> WORK( 1 );
			integer LWORK = -1;

			//first, a workspace query
			dgetri_( &order,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(),
			         WORK.data(), &LWORK,
			         &returnVal );

			if ( returnVal != 0 ) return false;

			LWORK = ( int )WORK( 0 );
			WORK.resize( LWORK );

			dgetri_( &order,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(),
			         WORK.data(), &LWORK,
			         &returnVal );

			if ( returnVal != 0 ) return false;

			columnToRowMajor( cmArray, LU.begin(), order, order );

			return true;
		}

		bool invert( AArray<float, 2> &LU, AArray<int, 1> &IPIV )
		{
			assert( LU.rows() == LU.cols() );

			integer order = LU.rows();
			integer LDA = order;

			float cmArray[order * order];
			checkStorageOrder( LU );
			rowToColumnMajor( LU.begin(), cmArray, order, order );

			integer returnVal = 0;
			AArray<float, 1> WORK( 1 );
			integer LWORK = -1;

			//first, a workspace query
			sgetri_( &order,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(),
			         WORK.data(), &LWORK,
			         &returnVal );

			if ( returnVal != 0 ) return false;

			LWORK = ( int )WORK( 0 );
			WORK.resize( LWORK );

			sgetri_( &order,
			         cmArray, &LDA,
			         ( integer * )IPIV.data(),
			         WORK.data(), &LWORK,
			         &returnVal );

			if ( returnVal != 0 ) return false;

			columnToRowMajor( cmArray, LU.begin(), order, order );

			return true;
		}

		bool invert( AArray<double, 2> &A )
		{
			AArray<int, 1> IPIV = lu( A );
			return invert( A, IPIV );
		}

		bool invert( AArray<float, 2> &A )
		{
			AArray<int, 1> IPIV = lu( A );
			return invert( A, IPIV );
		}

		int cholesky( const AArray<double, 2> &A, AArray<double, 2> &result )
		{
                        integer rowsA = A.rows();
                        integer colsA = A.cols();
                        assert( (int)result.rows() == rowsA );
                        assert( (int)result.cols() == colsA );

			double K[rowsA * colsA];
			checkStorageOrder( A );
			rowToColumnMajor( A.begin(), K, rowsA, colsA );

			char UPLO = 'U';
			integer LDA = rowsA;
			integer returnVal = 0;

			dpotrf_( &UPLO, &rowsA,
			         K, &LDA,
			         &returnVal );

			assert_msg( returnVal >= 0, "cholesky returnval is %d\n", int(returnVal) );

			columnToRowMajor( K, result.begin(), rowsA, colsA );

			//set the lower triangle to zero
			for ( int row = 1; row < rowsA; ++row )
			{
				for ( int col = 0; col < row; ++col )
				{
					result( row, col ) = 0;
				}
			}

			return returnVal;
		}

		std::pair<AArray<double, 2>, int> cholesky( AArray<double, 2> &A )
		{
			AArray<double, 2> C( A.rows(), A.cols() );
			int retVal = cholesky( A, C );
			std::pair<AArray<double, 2>, int> p( C, retVal );
			return p;
		}

		int cholesky( const AArray<float, 2> &A, AArray<float, 2> &result )
		{
			integer rowsA = A.rows();
			integer colsA = A.cols();
                        assert( result.rows() == (unsigned)rowsA );
                        assert( result.cols() == (unsigned)colsA );

			float K[rowsA * colsA];
			checkStorageOrder( A );
			rowToColumnMajor( A.begin(), K, rowsA, colsA );

			char UPLO = 'U';
			integer LDA = rowsA;

			integer returnVal = 0;

			spotrf_( &UPLO, &rowsA,
			         K, &LDA,
			         &returnVal );

			assert_msg( returnVal >= 0, "cholesky returnval is %d\n", int(returnVal) );

			columnToRowMajor( K, result.begin(), rowsA, colsA );

			//set the lower triangle to zero
			for ( int row = 1; row < rowsA; ++row )
			{
				for ( int col = 0; col < row; ++col )
				{
					result( row, col ) = 0;
				}
			}

			return returnVal;
		}

		std::pair<AArray<float, 2>, int> cholesky( AArray<float, 2> &A )
		{
			AArray<float, 2> C( A.rows(), A.cols() );
			int retVal = cholesky( A, C );
			std::pair<AArray<float, 2>, int> p( C, retVal );
			return p;
		}

		/*	void solveLUInPlace(AArray<float, 2> &a, AArray<float, 1> &b)
			{
				//  assert(false);//this one is broken TODO

				integer rowsA = a.rows();
				integer colsA = a.cols();
				integer colsB = 1;
				assert(rowsA == colsA);
				checkStorageOrder(a);

				real A[rowsA * colsA];
				rowToColumnMajor(a.begin(), A, rowsA, colsA);

				AArray<int, 1> IPIV(min(rowsA, colsA));

				integer returnVal = 0;
				sgesv_(&rowsA, &colsB, A, &rowsA, (integer*)IPIV.data(), b.data(), &rowsA, &returnVal);

				assert(returnVal == 0);
			}*/

		template< typename RealT>
		const AArray<RealT, 1> solveLU( AArray<RealT, 2> &a, AArray<RealT, 1> &b )
		{
			AArray<RealT,1> x( b.rows() );
			x = b;
			solveLUinPlace( a, x );
			return x;
		}
		template
		const AArray<float, 1> solveLU( AArray<float, 2> &a, AArray<float, 1> &b );
		template
		const AArray<double, 1> solveLU( AArray<double, 2> &a, AArray<double, 1> &b );

		template< typename RealT>
		inline integer luHelper( integer, integer, RealT*, integer*, RealT* );

		template<>
		inline integer luHelper( integer rA, integer cB, double* A, integer* IPIV, double* B )
		{
			integer returnVal = 0;
			dgesv_( &rA, &cB, A, &rA, IPIV, B, &rA, &returnVal );
			return returnVal;
		}

		template<>
		inline integer luHelper( integer rA, integer cB, float* A, integer* IPIV, float* B )
		{
			integer returnVal = 0;
			sgesv_( &rA, &cB, A, &rA, IPIV, B, &rA, &returnVal );
			return returnVal;
		}

		template< typename RealT>
		bool solveLUinPlace( const AArray< RealT, 2 > & a, AArray< RealT, 2 > & b_x )
		{
			integer rowsA = a.rows();
			integer colsA = a.cols();
			integer rowsB = b_x.rows();
			integer colsB = b_x.cols();
			assert( rowsA == colsA );
			checkStorageOrder( a );
			checkStorageOrder( b_x );

			RealT A[rowsA * colsA];
			rowToColumnMajor( a.begin(), A, rowsA, colsA );
			RealT B[rowsB * colsB];
			rowToColumnMajor( b_x.begin(), B, rowsB, colsB );

			AArray<int, 1> IPIV( min( rowsA, colsA ) );

			integer returnVal =
			        luHelper( rowsA, colsB, ( RealT* )A, ( integer* )IPIV.data(), ( RealT* )B );
			columnToRowMajor( B, b_x.begin(), rowsB, colsB );

			return returnVal == 0;
		}

		template< typename RealT>
		bool solveLUinPlace( const AArray< RealT, 2 > & a, AArray< RealT, 1 > & b_x )
		{
			integer rowsA = a.rows();
			integer colsA = a.cols();
			integer rowsB = b_x.rows();
			integer colsB = 1;
			assert( rowsA == colsA );
			checkStorageOrder( a );

			RealT A[rowsA * colsA];
			rowToColumnMajor( a.begin(), A, rowsA, colsA );
			RealT B[rowsB * colsB];
			rowToColumnMajor( b_x.begin(), B, rowsB, colsB );

			AArray<int, 1> IPIV( min( rowsA, colsA ) );

			integer returnVal =
			        luHelper( rowsA, colsB, ( RealT* )A, ( integer* )IPIV.data(), ( RealT* )B );
			columnToRowMajor( B, b_x.begin(), rowsB, colsB );

			return returnVal == 0;
		}

		template
		bool solveLUinPlace<float>( const AArray< float, 2 > & a, AArray< float, 2 > & b_x );
		template
		bool solveLUinPlace<double>( const AArray< double, 2 > & a, AArray< double, 2 > & b_x );
		template
		bool solveLUinPlace<float>( const AArray< float, 2 > & a, AArray< float, 1 > & b_x );
		template
		bool solveLUinPlace<double>( const AArray< double, 2 > & a, AArray< double, 1 > & b_x );
	}
}


