

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <stdint.h>
#include <adrcore/types/aarray.h>
#include <adrcore/util/matmul2.h>
#include <adrcore/fortran/fortran_switch.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/lapackutil.h>
#include <adrcore/util/storageorder.h>

namespace adore
{

	namespace core
	{

		template <class T> inline T MIN ( const T &a, const T &b ) {return a < b ? a : b;}

		/**
			creates a linearization fo the matrizes of a
			D1 * f + l * D2 * f + l^2 * D3 * f = 0
			problem.
			The reuslting matrizes can be used to put into a generalized
			eigenvalue solver
		*/
		void createQEPLinearization ( const AArray<double,2>& D1,
									  const AArray<double,2>& D2,
									  const AArray<double,2>& D3,
									  AArray<double,2>& A,
									  AArray<double,2>& B )
		{
                        unsigned size = D1.cols();
                        unsigned entries = D1.rows();
                        unsigned lsize = size << 1;
			assert ( size == D1.cols() );
			assert ( entries == D2.rows() );
			assert ( size == D2.cols() );
			assert ( entries == D3.rows() );
			assert ( size == D3.cols() );
			assert ( lsize == A.rows() );
			assert ( lsize == A.cols() );
			assert ( lsize == B.rows() );
			assert ( lsize == B.cols() );

			ADORE_UNUSED( entries );
			ADORE_UNUSED( lsize );

			A = 0;
			B = 0;

			AMatrix temp(size - 1, size - 1);
			matMul2_AT ( D1, D1, temp );
			A ( Range ( 0, size - 1 ), Range ( 0, size - 1 ) ) = -temp;//-matMul_AT ( D1, D1 );
			A ( Range ( size, 2 * size - 1 ), Range ( size, 2 * size - 1 ) ).setIdentity();

			matMul2_AT ( D1, D2, temp );
			B ( Range ( 0, size - 1 ),Range ( 0, size - 1 ) ) = temp;//matMul_AT ( D1, D2 );

			AMatrix temp2(size - 1, 2 * size - 1);
			matMul2_AT ( D1, D3, temp2 );
			B ( Range ( 0, size - 1 ),Range ( size, 2 * size - 1 ) ) = temp2;//matMul_AT ( D1, D3 );
			B ( Range ( size, 2 * size - 1 ),Range ( 0, size - 1 ) ).setIdentity();
		}

		/**
			solves generalized eigenvalue problem
			(A - l*B)*f = 0
			imaginary, infinite, null and abs-value-greater-thresh eigenvalues and their eigenvectors are discarded
			returns number of valid results
		*/
		int solveGEV (	const AArray<double,2>& A,
					   const AArray<double,2>& B,
					   AArray<double,1> & lambdas,
					   AArray<double,2> & eigenvecs,
		               const float minThresh,
		               const float maxThresh )
		{
			typedef double T_double;
			typedef AArray<T_double,2> Arr2D;
			typedef AArray<T_double,1> Arr1D;

                        unsigned size = A.rows();
			assert ( A.cols() == size );
			assert ( B.cols() == size );
			assert ( B.rows() == size );

//			0. Teste korrekte Dimension
			assert ( size == lambdas.rows() );
			assert ( size == eigenvecs.cols() );

//			2. Berechne DGGEV von Lapack

			integer N=size;

			T_double A_[N*N];
			T_double B_[N*N];
			checkStorageOrder ( A );
			checkStorageOrder ( B );
			rowToColumnMajor ( A.begin(), A_, N, N );
			rowToColumnMajor ( B.begin(), B_, N, N );

			char JOBVL,JOBVR;
			JOBVL='N'; JOBVR='V';
			integer LDA=N;
			integer LDB=N;
			T_double alphar[N];
			T_double alphai[N];
			T_double beta[N];
			T_double VL[N*N];
			integer LDVL=N;
			T_double VR[N*N];
			integer LDVR=N;
			integer LWORK = 8 * max ( 1, 8*N );
			T_double work[LWORK];
			integer INFO = 0;

			dggev_ ( &JOBVL, &JOBVR, &N, A_, &LDA, B_, &LDB, alphar, alphai,
			         beta, VL, &LDVL, VR, &LDVR, work, &LWORK, &INFO );

			assert ( INFO == 0 );

//			AArray<double, 2> E( N, N );
//			columnToRowMajor( VR, E.begin(), N, N );

			int validcounter = 0;
			int minAbsA = -1;
			double minEig = DBL_MAX;
			for ( int a = 0; a < N; ++a )
			{
//				printf( "(%f, %f)/%f\n", alphar[a], alphai[a], beta[a] );
				if ( alphai[a] == 0 )
				{
					bool betaokay  = abs ( beta[a] )   > FLT_EPSILON;
					bool alphaokay = abs ( alphar[a] ) > FLT_EPSILON;

					double eigval = 0;//if |alphar[a]| < eps
					if ( betaokay ) eigval = alphar[a] / beta[a];

					if ( betaokay || !alphaokay )
					{
						if ( eigval >= minThresh && eigval <= maxThresh )
						{
							// 				printf("adding eigval %f\n", eigval);
							lambdas ( validcounter ) = eigval;
                                                        for ( unsigned c = 0; c < size; ++c )
							{
								eigenvecs ( c, validcounter ) = VR[N * a + c];//E(c, a);//
							}
							++validcounter;
						}
						else
						{
							if ( fabs ( eigval ) < fabs ( minEig ) )
							{
								minEig = eigval;
								minAbsA = a;
							}
						}
					}
				}
			}

			//if no solution inside the valid interval has been found, return
			//one solution with the min eigval so far
			if ( validcounter == 0 && minAbsA != -1 )
			{
				lambdas ( 0 ) = minEig;
                                for ( unsigned c = 0; c < size; ++c )
				{
					eigenvecs ( c, 0 ) = VR[N * minAbsA + c];//E(c, a);//
				}
				validcounter = 1;
			}

			return validcounter;
		}


	}

}
