

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "solvesvd.h"
#include <adrcore/types/aarray.h>
#include <adrcore/fortran/fortran_switch.h>
#include <adrcore/util/lapackutil.h>
#include <adrcore/util/storageorder.h>
#include <adrcore/util/debug.h>

namespace adore
{

	namespace core
	{


		typedef float T_float;
		typedef double T_double;

		void checkReturn( int returnVal )
		{
#ifndef NDEBUG
			if ( returnVal == 0 )
			{
				DBGOUT( DBGALL, "*gesvd behaved well\n" );
			}
			else if ( returnVal < 0 )
			{
				DBGOUT( DBGCRITICAL, "*gesvd discovered an illegal value at position %d\n", -returnVal );
			}
			else if ( returnVal > 0 )
			{
				DBGOUT( DBGCRITICAL, "*gesvd: %d superdiagonals of DBDSQR dis not converge\n", returnVal );
			}
#endif
			ADORE_UNUSED( returnVal );
		}

		////float

		int svd ( const AArray<T_float, 2> &_A, SVDStoragef &store )
		{
			char jobU = 'N';
			char jobVT = 'N';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_float U = 0, V_T = 0;
			integer u = 1, v_t = 1;

			assert ( rowsA > 0 );

			sgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          &U, &u,
			          &V_T, &v_t,
			          store.work, &LWORK,
			          &returnVal );

			checkStorageOrder ( store.D ); //may fail if S contains NAN

			return returnVal;

		}

		int svdU ( const AArray<T_float, 2> &_A, SVDStoragef &store )
		{
			char jobU = 'S';
			char jobVT = 'N';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );
//			assert ( store.U.isStorageContiguous() );
			assert( store.U.cols() <= store.U.rows() );//because of the 'S'

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_float V_T = 0;
			integer v_t = 1;

			assert ( rowsA > 0 );

			sgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          store.Uwork, &rowsA,
			          &V_T, &v_t,
			          store.work, &LWORK,
			          &returnVal );

			columnToRowMajor( store.Uwork, store.U.begin(), store.getURows(), store.getUCols() );
			//transInPlace ( store.U );

			checkStorageOrder ( store.D ); //may fail if S contains NAN
			checkStorageOrder ( store.U );

			return returnVal;

		}

		int svdUVT ( const AArray<T_float, 2> &_A, SVDStoragef &store )
		{
			char jobU = 'S';
			char jobVT = 'A';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );
//			assert ( store.U.isStorageContiguous() );

//			assert ( store.VT.isStorageContiguous() );
			assert( store.U.cols() <= store.U.rows() );//because of the 'S'

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			assert ( rowsA > 0 );

			sgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          store.Uwork, &rowsA,
			          store.VT.data(), &colsA,
			          store.work, &LWORK,
			          &returnVal );

			columnToRowMajor( store.Uwork, store.U.begin(), store.getURows(), store.getUCols() );
			//transInPlace ( store.U );
			transInPlace ( store.VT );

			checkStorageOrder ( store.D ); //may fail if S contains NAN
			checkStorageOrder ( store.U );
			checkStorageOrder ( store.VT );

			return returnVal;

		}

		int svdVT ( const AArray<T_float, 2> &_A, SVDStoragef &store )
		{
			char jobU = 'N';
			char jobVT = 'A';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );
//			assert ( store.VT.isStorageContiguous() );

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_float U = 0;
			integer u = 1;

			assert ( rowsA > 0 );

			sgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          &U, &u,
			          store.VT.data(), &colsA,
			          store.work, &LWORK,
			          &returnVal );

			transInPlace ( store.VT );

			checkStorageOrder ( store.D ); //may fail if S contains NAN
			checkStorageOrder ( store.VT );

			return returnVal;

		}

		////double

		int svd ( const AArray<T_double, 2> &_A, SVDStoraged &store )
		{
			char jobU = 'N';
			char jobVT = 'N';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_double U = 0, V_T = 0;
			integer u = 1, v_t = 1;

			assert ( rowsA > 0 );

			dgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          &U, &u,
			          &V_T, &v_t,
			          store.work, &LWORK,
			          &returnVal );

			checkStorageOrder ( store.D ); //may fail if S contains NAN

			return returnVal;

		}

		int svdU ( const AArray<T_double, 2> &_A, SVDStoraged &store )
		{
			char jobU = 'S';
			char jobVT = 'N';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );
//			assert ( store.U.isStorageContiguous() );
			assert( store.U.cols() <= store.U.rows() );//because of the 'S'

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_double V_T = 0;
			integer v_t = 1;

			assert ( rowsA > 0 );

			dgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          store.Uwork, &rowsA,
			          &V_T, &v_t,
			          store.work, &LWORK,
			          &returnVal );

			columnToRowMajor( store.Uwork, store.U.begin(), store.getURows(), store.getUCols() );
			//transInPlace ( store.U );

			checkStorageOrder ( store.D ); //may fail if S contains NAN
			checkStorageOrder ( store.U );

			return returnVal;

		}

		int svdUVT ( const AArray<T_double, 2> &_A, SVDStoraged &store )
		{
			char jobU = 'S';
			char jobVT = 'A';

			assert ( store.D.rows() == min ( _A.rows(), _A.cols() ) );
//			assert ( store.U.isStorageContiguous() );

//			assert ( store.VT.isStorageContiguous() );
			assert( store.U.cols() <= store.U.rows() );//because of the 'S'

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

//			printf( "storeorder A\n" );
			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			assert ( rowsA > 0 );
//			for ( int h = 0; h < rowsA * colsA; ++h ) printf( "%f, ", store.Awork[h] );
//			printf( "\n" );
//			for ( int h = 0; h < store.U.rows() * store.U.cols(); ++h ) store.Uwork[h] = 0;

			dgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          store.Uwork, &rowsA,
			          store.VT.data(), &colsA,
			          store.work, &LWORK,
			          &returnVal );

			checkReturn( returnVal );

//			for ( int h = 0; h < store.U.rows() * store.U.cols(); ++h ) printf( "%f, ", store.Uwork[h] );
//			printf( "\n" );

			columnToRowMajor( store.Uwork, store.U.begin(), store.getURows(), store.getUCols() );
			//transInPlace ( store.U );
			transInPlace ( store.VT );

//			printf( "storeorder D\n" );
//			cout << store.D;
			checkStorageOrder ( store.D ); //may fail if S contains NAN
//			printf( "storeorder U\n" );
//			cout << store.U;
			checkStorageOrder ( store.U );
//			printf( "storeorder VT\n" );
//			cout << store.VT;
			checkStorageOrder ( store.VT );

			return returnVal;

		}

		int svdVT ( const AArray<T_double, 2> &_A, SVDStoraged &store )
		{
			char jobU = 'N';
			char jobVT = 'A';

                        assert_msg ( store.D.rows() == min ( _A.rows(), _A.cols() ), "store.D.rows() = %zd, while A.rows() = %zd and A.cols() = %zd\n", store.D.rows(), _A.rows(), _A.cols() );
//			assert ( store.VT.isStorageContiguous() );

			integer rowsA = _A.rows();
			integer colsA = _A.cols();

			checkStorageOrder ( _A );
			rowToColumnMajor ( _A.begin(), store.Awork, rowsA, colsA );

			integer LWORK = store.worksize;
			integer returnVal = 0;

			T_double U = 0;
			integer u = 1;

			assert ( rowsA > 0 );

			dgesvd_ ( &jobU, &jobVT, &rowsA, &colsA,
			          store.Awork, &rowsA,
			          store.D.data(),
			          &U, &u,
			          store.VT.data(), &colsA,
			          store.work, &LWORK,
			          &returnVal );

			transInPlace ( store.VT );

			checkStorageOrder ( store.D ); //may fail if S contains NAN
			checkStorageOrder ( store.VT );

			return returnVal;

		}

	}

}
