

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESOLVESVD_H
#define ADORE_CORESOLVESVD_H

#include <adrcore/util/mathutil.h>

namespace adore
{

	namespace core
	{

		enum
		{
			//SVD_D = 1,
			SVD_U = 2,
                        SVD_VT = 4
		};

		/**
		 * A helper class that provides matrices used by solveSVD
			 
		*/
		template<class T_numtype>
		class SVDStorage
		{
			public:
				int rows, cols, largestRows;//the dims of the matrix to be svd'd
				AArray<T_numtype, 1> D;
				AArray<T_numtype, 2> U;
				AArray<T_numtype, 2> VT;
				int worksize;
				T_numtype *work;
				T_numtype *Awork;
				T_numtype *Uwork;

			private:
				static inline int getWorksize( const int rows, const int cols ) {return 10 * MAX ( 3*MIN ( rows, cols ) +MAX ( rows, cols ), 5*MIN ( rows, cols ) );}

				void applyFlags( const int flags = 0 )
				{
					if ( flags & SVD_U )
					{
						U.resize ( getURows(), getUCols() );
						Uwork = new T_numtype[getURows() * getUCols()];
					}
					if ( flags & SVD_VT ) VT.resize ( cols, cols );
				}

			public:

				SVDStorage ():rows( 0 ), cols( 0 ), largestRows( 0 ), worksize( 0 ), work( NULL ), Awork( NULL ), Uwork( NULL ) {;}

				SVDStorage ( const AArray<T_numtype, 2> &left, const int flags = 0 ) : rows ( left.rows() ), cols ( left.cols() ), largestRows( rows ),
						D ( MIN ( rows, cols ) ),
						worksize ( getWorksize( rows, cols ) ),
						work ( new T_numtype[worksize] ),
						Awork ( new T_numtype[ rows * cols ] ),
						Uwork( NULL )
				{
					applyFlags( flags );
				}

				/**
				 * if you dont need U, then leftRows can be set to an arbitrary value
				 * */
				SVDStorage ( const int leftRows, const int leftCols, const int flags = 0 ) : rows ( leftRows ), cols ( leftCols ), largestRows( rows ),
						D ( MIN ( rows, cols ) ),
						worksize ( getWorksize( rows, cols ) ),
						work ( new T_numtype[worksize] ),
						Awork ( new T_numtype[ rows * cols ] ),
						Uwork( NULL )
				{
					applyFlags( flags );
				}

				~SVDStorage()
				{
					delete [] work;
					delete [] Awork;
					delete [] Uwork;
				}

				/**
				 * re-initializes and re-allocates the storage
				 */
				void init( const int leftRows, const int leftCols, const int flags = 0 )
				{
					rows = leftRows;
					cols = leftCols;
					largestRows = rows;
					D.resize ( MIN ( rows, cols ) );
					worksize = getWorksize( rows, cols );
					delete [] work;  work = new T_numtype[worksize];
					delete [] Awork; Awork = new T_numtype[ rows * cols ];
					delete [] Uwork; Uwork = NULL;
					applyFlags( flags );
				}

				/**
				 * adapts the workspace to a new number of rows. If you use the U-Matrix, this is not sufficient
				 * never shrinks
				 * remember: D must have size MIN(rows, cols) -> see repairD
				 * */
				void changeRowsWorkspaceOnly ( int r )
				{
					rows = r;
					if ( rows <= largestRows ) return;

					assert( isNullArray( U ) );
					delete [] work;
					delete [] Awork;
					worksize  =  10 * MAX ( 3*MIN ( rows, cols ) +MAX ( rows, cols ), 5*MIN ( rows, cols ) );
					work = new T_numtype[worksize];
					Awork = new T_numtype[ rows * cols ];
					largestRows = rows;
				}

				inline void repairD()
				{
					D.resize ( MIN ( rows, cols ) );
				}

				int getRows()const {return rows;}
				int getCols()const {return cols;}

				int getURows()const {return rows;}
				int getUCols()const {return D.size();}

			private:
				SVDStorage ( const SVDStorage & ) {;}
		};

		/**
		 * A helper class that provides matrices used by solveSVD, and a temporary vector
		 
		 */

                template<class T_numtype, int rank> struct SVDSolveStorage_helper
                { void operator()(AArray<T_numtype, rank>&, int, int) { BOOST_STATIC_ASSERT( rank == 1 || rank == 2 ); } };
                template<class T_numtype> struct SVDSolveStorage_helper<T_numtype, 1>
                { void operator()(AArray<T_numtype, 1>& a, int i, int) { a.resize(i); } };
                template<class T_numtype> struct SVDSolveStorage_helper<T_numtype, 2>
                { void operator()(AArray<T_numtype, 2>& a, int i, int j) { a.resize(i, j); } };

                template<class T_numtype, int rank>
		class SVDSolveStorage
		{
			public:
				SVDStorage<T_numtype> store;
				AArray<T_numtype, rank> temp;//for linear equation solving

				SVDSolveStorage () {;}

                                SVDSolveStorage ( const AArray<T_numtype, 2> &left, const AArray<T_numtype, 1> &/*right*/ ) : store ( left, SVD_U | SVD_VT ), temp ( left.cols() )
                                {;}

                                SVDSolveStorage ( const AArray<T_numtype, 2> &left, const AArray<T_numtype, 2> &right ) : store ( left, SVD_U | SVD_VT ), temp ( left.cols(), right.cols() )
				{;}

                                SVDSolveStorage ( const int leftRows, const int leftCols, const int rightCols = 1 ) : store ( leftRows, leftCols, SVD_U | SVD_VT )
                                { SVDSolveStorage_helper<T_numtype, rank> h; h( temp, leftCols, rightCols ); }

				void init( const int leftRows, const int leftCols, const int rightCols = 1 )
				{
					store.init( leftRows, leftCols, SVD_U | SVD_VT );

                                        SVDSolveStorage_helper<T_numtype, rank> h;
                                        h( temp, leftCols, rightCols );
				}
		};

		typedef SVDStorage<double> SVDStoraged;
		typedef SVDStorage<float> SVDStoragef;
		typedef SVDSolveStorage<double, 2> SVDSolveStoraged2;
		typedef SVDSolveStorage<double, 1> SVDSolveStoraged1;
		typedef SVDSolveStorage<float, 2> SVDSolveStoragef2;
		typedef SVDSolveStorage<float, 1> SVDSolveStoragef1;

		int svd ( const AArray<float, 2> &A, SVDStoragef &store );
		int svdU ( const AArray<float, 2> &A, SVDStoragef &store );
		int svdUVT ( const AArray<float, 2> &A, SVDStoragef &store );
		int svdVT ( const AArray<float, 2> &A, SVDStoragef &store );

		int svd ( const AArray<double, 2> &A, SVDStoraged &store );
		int svdU ( const AArray<double, 2> &A, SVDStoraged &store );
		int svdUVT ( const AArray<double, 2> &A, SVDStoraged &store );
		int svdVT ( const AArray<double, 2> &A, SVDStoraged &store );

		template<class T_numtype>
		void solveSVD2_DUVTb ( SVDSolveStorage<T_numtype, 2> &store, const AArray<T_numtype, 2> &b, AArray<T_numtype, 2> &result )
		{
			const AArray<T_numtype, 2> &U = store.store.U;
			const AArray<T_numtype, 1> &D = store.store.D;
			const AArray<T_numtype, 2> &VT = store.store.VT;
			AArray<T_numtype, 2> &temp = store.temp;

			matMul2_AT ( U, b, temp );
			for ( int i = 0; i < temp.rows(); ++i )
			{
				const T_numtype d = D ( i );
				for ( int j = 0; j < temp.cols(); ++j )
				{
					if ( d > DBL_EPSILON ) temp ( i, j ) = temp ( i, j ) / d;
					else temp ( i, j ) = 0;
				}
			}
			matMul2_AT ( VT, temp, result );
		}

		template<class T_numtype>
		void solveSVD2_DUVTb ( SVDSolveStorage<T_numtype, 1> &store, const AArray<T_numtype, 1> &b, AArray<T_numtype, 1> &result )
		{
			const AArray<T_numtype, 2> &U = store.store.U;
			const AArray<T_numtype, 1> &D = store.store.D;
			const AArray<T_numtype, 2> &VT = store.store.VT;
			AArray<T_numtype, 1> &temp = store.temp;

			matMul2_AT ( U, b, temp );
                        for ( unsigned i = 0; i < D.rows(); ++i )
			{
				T_numtype d = D ( i );
				if ( d > DBL_EPSILON ) temp ( i ) = temp ( i ) / d;
				else temp ( i ) = 0;
			}
                        for ( unsigned i = D.rows(); i < temp.rows(); ++i )
			{
				temp ( i ) = 0;
			}
			matMul2_AT ( VT, temp, result );
		}


		template<class T_numtype>
		const AArray<T_numtype, 2> solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 2> &b )
		{
			SVDSolveStorage<T_numtype, 2> store ( A, b );
			AArray<T_numtype, 2> x ( A.cols(), b.cols() );

			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, x );
			return x;
		}

		template<class T_numtype>
		const AArray<T_numtype, 1> solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 1> &b )
		{
			SVDSolveStorage<T_numtype, 1> store ( A, b );
			AArray<T_numtype, 1> x ( A.cols() );

			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, x );
			return x;
		}

		template<class T_numtype>
		void solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 2> &b, AArray<T_numtype, 2> &result )
		{
			SVDSolveStorage<T_numtype, 2> store ( A, b );
			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, result );
		}

		template<class T_numtype>
		void solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 1> &b, AArray<T_numtype, 1> &result )
		{
			SVDSolveStorage<T_numtype, 1> store ( A, b );
			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, result );
		}

		template<class T_numtype>
		void solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 2> &b, AArray<T_numtype, 2> &result, SVDSolveStorage<T_numtype, 2> &store )
		{
			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, result );
		}

		template<class T_numtype>
		void solveSVD ( const AArray<T_numtype, 2> &A, const AArray<T_numtype, 1> &b, AArray<T_numtype, 1> &result, SVDSolveStorage<T_numtype, 1> &store )
		{
			svdUVT ( A, store.store );
			solveSVD2_DUVTb ( store, b, result );
		}

		/**
		 * does what it pretends to do (hopefully)
		 */
		template<class T> inline int getRank ( const AArray<T, 2> &m, double thresh = FLT_EPSILON )
		{
			SVDStorage<T> store ( m );
			svd ( m, store );
			for ( int i = store.D.rows() -1; i >= 0; --i )
			{
				if ( store.D ( i ) > thresh ) { return ( i+1 ); }
			}
			return 0;
		}

		template<class T> inline void printSingularValues ( const AArray<T, 2> &m )
		{
#ifndef NDEBUG
			SVDStorage<T> store ( m );
			svd ( m, store );
			for ( int i = store.D.rows() -1; i >= 0; --i )
			{
				printf( "%f\n", store.D ( i ) );
			}
#endif
		}

	}

}

#endif
