

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "epilinematcher2.h"

#include <adrcore/features/linedistfilter.h>
#include <adrcore/viewmatch/matchsorter.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/clock.h>

namespace adore
{

	namespace core
	{

        static void getEpi2( AMatrix & fmat, AVector & epipole, SVDStoraged & store3 )
		{
			transInPlace( fmat );
			FMatrixCalculator::calculateEpipolarPoint( fmat, epipole, store3 );
			transInPlace( fmat );
			inhomogenize<2>( epipole );
		}

		template <typename Tree, typename Assessor>
		void EpiLineMatcher2<Tree, Assessor>::match ( ViewBase *v1, ViewBase *v2, NView *set )
		{
			if ( isNullArray( fmat ) )
			{
				if ( set->getNumViews() != 2 )
				{
					DBGOUT( DBGNORMAL, "EpiLineMatcher2: wrong number of views\n" );
					return;
				}
				if ( !set->isNViewSet() )
				{
					DBGOUT( DBGNORMAL, "EpiLineMatcher2: not an NViewSet\n" );
					return;
				}
				NViewInfo *info = set->getInfo();
				if ( isNullArray( info->getFMatrix() ) )
				{
					DBGOUT( DBGNORMAL, "EpiLineMatcher2: No fundamental matrix\n" );
					return;
				}
				fmat.resize( 3, 3 );
				fmat = info->getFMatrix();
			}
			if ( isNullArray( fmat ) )
			{
				DBGOUT( DBGNORMAL, "EpiLineMatcher2: Still No fundamental matrix\n" );
				return;
			}
			DBGOUT( DBGNORMAL, "EpiLineMatcher2: Everything is fine\n" );
			match ( v1, v2, set, fmat );
		}

		template <typename Tree, typename Assessor>
		typename EpiLineMatcher2<Tree, Assessor>::PairT
		EpiLineMatcher2<Tree, Assessor>::findSuitable(
				FeaturePoint &fp, ViewBase *v2, const AMatrix &fmat, const AVector &epi2, Assessor & assessor )
		{
			PairT best( NULL, FLT_MAX );
			PairT good( NULL, FLT_MAX );

			AVector line = matMul( fmat, fp );
			LineDistance ldist( line( 0 ), line( 1 ), line( 2 ) );
			if ( !ldist.isOkay() )
			{
				DBGOUT( DBGNORMAL, "EpiLineMatcher2: insane line occurred" );
				printMat( DBGNORMAL, fmat );
				printMat( DBGNORMAL, fp );
				return best;
			}

			double w = v2->getWidth();
			double h = v2->getHeight();
			AVector p1(3), p2(3);
			p1 = 0, 0, 1.0;
			p2 = 0, 0, 1.0;
			bool ok = pointAsLine2( line( 0 ), line( 1 ), line( 2 ), p1(0), p1(1), p2(0), p2(1), w, h );
			if(!ok) {
				return best;
			}

			const vector<FeaturePoint*>& points = assessor.assess(p1, p2, lineDistThresh);

			int count = 0;
			FeatureDescriptor *fd1 = fp.getDescriptor( descrID );
			if ( fd1 )
			{
				for ( uint k = 0; k < fd1->size(); ++k )
				{
					for ( vector<FeaturePoint*>::const_iterator it = points.begin(); it != points.end(); ++it )
					{
						FeaturePoint &cand = **it;
						assert( cand( 2 ) == 1.0 );
						++count;

						bool ori = FMatrixCalculator::isOriented ( fmat, epi2, fp, cand );
						if ( ori )
						{
							FeatureDescriptor *fd2 = ( *it )->getDescriptor( descrID );
							if ( fd2 )
							{
								for ( uint j = 0; j < fd2->size(); ++j )
								{
									float dist = getDistance( ( *fd1 )[k], ( *fd2 )[j] );
									if ( dist < featDistThresh )
									{
										if ( dist < best.second )
										{
											good = best;
											best.first = *it;
											best.second = dist;
										}
										else if ( dist < good.second )
										{
											good.first = *it;
											good.second = dist;
										}
									}
								}
							}
						}
						else
						{
							DBGOUT( DBGALL, "Epilinematcher featpair was not oriented\n" );
						}
					}
				}
			}
			DBGOUT( DBGALL, "EpiLineMatcher2: ckecked %d candidates\n", count );

			if ( best.second / good.second < ratioThresh )	return best;
			return PairT( NULL, FLT_MAX );
		}

		template <typename Tree, typename Assessor>
		void EpiLineMatcher2<Tree, Assessor>::match( ViewBase *v1, ViewBase *v2, NView *set, AMatrix &fmat )
		{
			DBGOUT( DBGALL, "matching %d with %d\n", v1->getNumFeaturePoints( matchID ), v2->getNumFeaturePoints( matchID ) );

			getEpi2 ( fmat, epi2, store3 );

			//re-orient fmt in case is has been optimized and lost its sign
			for ( NView::CorrIterator cit = set->getPointCorrsBegin(); cit != set->getPointCorrsEnd(); ++cit )
			{
				PointCorr *pc = *cit;
				FeaturePoint *fp1 = pc->getFeature( v1 );
				FeaturePoint *fp2 = pc->getFeature( v2 );
				bool io1 = FMatrixCalculator::isOriented ( fmat, epi2, *fp1, *fp2 );
				bool io2 = FMatrixCalculator::isOriented ( -fmat, epi2, *fp1, *fp2 );
				if ( io1 != io2 )
				{
					if ( io2 ) fmat *= -1;
					break;
				}
			}

			/// prepare KDTree
			vector<FeaturePoint*> pointsInView2;
			pointsInView2.reserve(v2->getNumAllFeaturePoints());
			for ( ViewT::PointsIterator it = v2->getPointsBegin( matchID ); it != v2->getPointsEnd( matchID ); ++it )
			{
				FeaturePoint* fp = *it;
				if(NULL == fp) continue;
				pointsInView2.push_back(fp);
			}
			shared_ptr<helper::Clock> clock(new helper::Clock("building tree"));
			tree.build(pointsInView2);
			clock.reset();
			Assessor assessor(tree, pointsInView2.size());

			MatchSorter mSort( 100/*reserve*/ );
			for ( ViewT::PointsIterator it = v1->getPointsBegin( matchID ); it != v1->getPointsEnd( matchID ); ++it )
			{
				FeaturePoint *fp1 = *it;
				PairT fp2 = findSuitable( *fp1, v2, fmat, epi2, assessor );

				if ( fp2.first )
				{
					mSort.add( fp1, fp2.first, fp2.second );
				}
				else
				{
					DBGOUT( DBGALL, "EpiLineMatcher2 wasn't able to find a match.\n" );
//					printMats( DBGNORMAL, set->getCamera( v1 )->getProjMat(), set->getCamera( v2 )->getProjMat() );
//					pressKey();
				}
			}
			mSort.sort();

			int count = 0;
			for ( MatchSorter::iterator it = mSort.begin();
			                ( it != mSort.end() ) && ( count < countThresh ); ++it )
			{
//				DBGOUT( DBGALL, "try adding match with badness %f\n", ( *it ).badness );
//				if ( ( *it ).badness > thresh ) break;

				DBGOUT( DBGALL, "adding match with badness %f\n", ( *it ).badness );
				FeaturePoint *fp1 = ( *it ).fp1;
				FeaturePoint *fp2 = ( *it ).fp2;
//				DBGOUT( DBGNORMAL, "%p --- %p\n", fp1, fp2 );

				count += this->addPairToNview( fp2 /*certain*/, fp1 /*doubtful*/, set, ( *it ).badness ) ? 1 : 0;
			}

			DBGOUT( DBGNORMAL, "EpiLineMatcher2 added %d corrs\n", count );
		}


		template class EpiLineMatcher2<KDTree2D, KDTreeAssessorLine2d<KDTree2D> >;
		template class EpiLineMatcher2<QuadTree, QuadTreeLineAssessor>;
	}
}
