

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "epilinesegmentmatcher.h"
#include <adrcore/imgprocessing/channelfunctions.h>
#include <adrcore/imgprocessing/templatematching.h>
#include <adrcore/imgprocessing/drawing.h>
#include <adrcore/imgprocessing/rgbtogray.h>
#include <adrcore/types/corr1ton.h>
#include <memory>

namespace adore
{
	namespace core
	{
		typedef uint8_t imgT;
		typedef ImagePlane<imgT> ImageT;
		typedef View<imgT> ViewT;
		const double eps = numeric_limits<double>::epsilon();

		double distance( const AVector& p1, const AVector& p2 )
		{
			assert( p1( 2 ) == 1.0 && p2( 2 ) == 1.0 );
			AVector v( 3 ); v = p2 - p1;
			return sqrt( v( 0 ) * v( 0 ) + v( 1 ) * v( 1 ) );
		}

		void EpiLineSegmentMatcher::matchDerived( ViewT * v1, ViewT * v2, NView * set )
		{
			NViewInfo *info = set->getInfo();
			if ( isNullArray( info->getFMatrix() ) )
			{
				DBGOUT( DBGNORMAL, "EpiLineSegmentMatcher: No fundamental matrix\n" );
				return;
			}
			bool normalized = set->isNormalized();
			if ( normalized )
			{
				set->denormalize();
				v1->denormalize();
				v2->denormalize();
			}
			AMatrix fmat( 3, 3 );
			fmat = info->getFMatrix();
			if ( isNullArray( fmat ) )
			{
				DBGOUT( DBGNORMAL, "EpiLineSegmentMatcher: Still No fundamental matrix\n" );
				return;
			}
			DBGOUT( DBGNORMAL, "EpiLineSegmentMatcher: Everything is fine\n" );

			candMapT candidates;
			candOrigMapT candOrig;
			findCandidates( v1, v2, fmat, candidates, candOrig );

			ImageT *i1 = static_cast<ViewT*>( v1 )->getImage();
			ImageT *i2 = static_cast<ViewT*>( v2 )->getImage();
			ImageT *img = combineImages( i1, i2 );
			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );

			std::auto_ptr< ImagePlane<imgT> > grayDeletor1;
			ImagePlane<imgT> * gray = NULL;
			if ( i1->getNumChannels() == 3 )
			{
				gray = new ImagePlane<imgT> ( w1, h1, 1 );
				RGBtoGray<imgT> converter;
				Channel<imgT>::Set in ( true );
				i1->createChannels ( in );
				converter.process ( in, gray );
				i1 = gray;
			};
			grayDeletor1.reset( gray );
			std::auto_ptr< ImagePlane<imgT> > grayDeletor2;
			gray = NULL;
			if ( i2->getNumChannels() == 3 )
			{
				gray = new ImagePlane<imgT> ( w2, h2, 1 );
				RGBtoGray<imgT> converter;
				Channel<imgT>::Set in ( true );
				i2->createChannels ( in );
				converter.process ( in, gray );
				i2 = gray;
			};
			grayDeletor2.reset( gray );

			candMapT::iterator it = candidates.begin();
			candOrigMapT::iterator itO = candOrig.begin();
			const double scale1 = 2.0, scale2 = 2.0;
			const double thresh = 0.98;
			candListT lll;
			AMatrix R( 2, 2 ), S( 2, 2 ), A( 2, 2 );
			for ( ; it != candidates.end(); ++it, ++itO )
			{
				candListT& cl = it->second;
				candOrigListT& clO = itO->second;
				uint num2 = cl.size();
				map<uint, double> scores;
				for ( uint i = 0; i < num2; ++i )
				{
					const AVector& p1_1 = cl[i].first.first;
					const AVector& p1_2 = cl[i].first.second;
					double x1 = p1_2( 0 ) - p1_1( 0 );
					double y1 = p1_2( 1 ) - p1_1( 1 );
					double l1 = sqrt( x1 * x1 + y1 * y1 );
					if ( l1 < eps ) continue;
					const AVector& p2_1 = cl[i].second.first;
					const AVector& p2_2 = cl[i].second.second;
					double x2 = p2_2( 0 ) - p2_1( 0 );
					double y2 = p2_2( 1 ) - p2_1( 1 );
					double l2 = sqrt( x2 * x2 + y2 * y2 );
					if ( l2 < eps ) continue;

					double w1 = scale1 * 3, w2 = scale2 * 3;
					int width = std::max( round( w1 ), round( w2 ) );
					int length = std::max( round( l1 ), round( l2 ) );
					if ( 15 > length ) continue;
					int w_width = width / 2;
					if ( w_width % 2 == 0 )
					{
						++w_width;
						width = w_width + w_width;
					}

					x1 /= l1; y1 /= l1;
					double phi = atan2( y1, x1 ); // inverted y
					phi -= M_PI * 0.5;
					R = y1, x1, -x1, y1;
					S = w1 / width, 0, 0, l1 / length;
					matMul2( R, S, A );
					ImagePlane<double> patch1( width, length, 1 );
					double midX = ( p1_1( 0 ) + p1_2( 0 ) ) * 0.5;
					double midY = ( p1_1( 1 ) + p1_2( 1 ) ) * 0.5;
					warpAffine2( *i1, midX, midY, A, patch1.begin(), width, length );

					x2 /= l2; y2 /= l2;
					phi = atan2( y2, x2 ); // inverted y
					phi -= M_PI * .5;
					R = y2, x2, -x2, y2;
					S = w2 / width, 0, 0, l2 / length;
					matMul2( R, S, A );
					ImagePlane<double> patch2( width, length, 1 );
					midX = ( p2_1( 0 ) + p2_2( 0 ) ) * 0.5;
					midY = ( p2_1( 1 ) + p2_2( 1 ) ) * 0.5;
					warpAffine2( *i2, midX, midY, A, patch2.begin(), width, length );

					ImagePlane<double> patch1_l( w_width, length, 1 );
					ImagePlane<double> patch1_r( w_width, length, 1 );
					ImagePlane<double> patch2_l( w_width, length, 1 );
					ImagePlane<double> patch2_r( w_width, length, 1 );
					for ( int y = 0; y < length; ++y )
					{
						for ( int x = 0; x < w_width; ++x )
						{
							*patch1_l.getPointer( x, y ) = *patch1.getPointer( x, y );
							*patch1_r.getPointer( x, y ) = *patch1.getPointer( x + w_width, y );
							*patch2_l.getPointer( x, y ) = *patch2.getPointer( x, y );
							*patch2_r.getPointer( x, y ) = *patch2.getPointer( x + w_width, y );
						}
					}
					subtractMean( patch1_l.begin(), w_width * length );
					subtractMean( patch1_r.begin(), w_width * length );
					subtractMean( patch2_l.begin(), w_width * length );
					subtractMean( patch2_r.begin(), w_width * length );

					typedef CCORRNormedFunctor<double> CCORRNormedFunctorT;
					double scorel = getFastCorrelation<double, ImagePlane<double>::Iterator2D, CCORRNormedFunctorT > ( patch2_l, w_width / 2, length / 2,
					                patch1_l.begin(), w_width / 2, length / 2 );
					double scorer = getFastCorrelation<double, ImagePlane<double>::Iterator2D, CCORRNormedFunctorT > ( patch2_r, w_width / 2, length / 2,
					                patch1_r.begin(), w_width / 2, length / 2 );
					double sum = scorel + scorer;
					if ( thresh < sum )
						scores[i] = sum;
				}
				uint bestIdx = 0;
				double bestScore = 0.0;
				for ( map<uint, double>::iterator it = scores.begin(); it != scores.end(); ++it )
				{
					if ( bestScore < it->second )
					{
						bestScore = it->second;
						bestIdx = it->first;
					}
				}
				if ( 0.0 == bestScore ) continue;

				const AVector& p1_1 = cl[bestIdx].first.first;
				const AVector& p1_2 = cl[bestIdx].first.second;
				img->setROI( 0, 0, w1, h1 );
				drawLine( *img, int( p1_1( 0 ) ), int( p1_1( 1 ) ), int( p1_2( 0 ) ), int( p1_2( 1 ) ), ( imgT )( maxLimit<imgT>() ) );
				const AVector& p2_1 = cl[bestIdx].second.first;
				const AVector& p2_2 = cl[bestIdx].second.second;
				img->setROI( w1, 0, w2, h2 );
				drawLine( *img, int( p2_1( 0 ) ), int( p2_1( 1 ) ), int( p2_2( 0 ) ), int( p2_2( 1 ) ), ( imgT )( maxLimit<imgT>() ) );

				vector < pair<Feature*, float> > matched;
				if ( mode == SingleBest )
					matched.push_back( make_pair( clO[bestIdx].second.get(), bestScore ) );
				else
				{
					for ( map<uint, double>::iterator it = scores.begin(); it != scores.end(); ++it )
					{
						matched.push_back( make_pair( clO[it->first].second.get(), 1.0 / it->second ) );
					}
				}
				set->addMultiLineNoCheck( clO[bestIdx].first.get(), matched.begin(), matched.end() );
			}
			img->resetROI();
			img->save( "segments.png" );

			if ( normalized )
			{
				v1->normalize();
				v2->normalize();
			}
		}

		EpiLineSegmentMatcher::Relation EpiLineSegmentMatcher::testLinePoint(const AVector& line, const AVector& pt)
		{
			double temp = dot2(line, pt);
			if(abs(temp) < eps)
				return Collinear;
			return (temp < 0) ? Negative : Positive;
		}

		EpiLineSegmentMatcher::PointsInfo EpiLineSegmentMatcher::testLine2Points(const AVector& line, const AVector& pt1, const AVector& pt2)
		{
			return make_pair(testLinePoint(line, pt1), testLinePoint(line, pt2));
		}

		bool EpiLineSegmentMatcher::evaluateIntersection(const AVector& line, const AVector& pt1, const AVector& pt2, AVector& pt, PointsInfo& x)
		{
			x = testLine2Points(line, pt1, pt2);
			if(x.first == Collinear || x.second == Collinear) {
				if(x.first == x.second)
					return false;
				if(x.first == Collinear) pt = pt1;
				else pt = pt2;
				return true;
			} else if(x.first != x.second) {
				AVector temp(3);
				cross(pt1, pt2, temp);
				cross(line, temp, pt);
				pt /= pt(2);
				return true;
			}
			return false;
		}

		void EpiLineSegmentMatcher::findCandidates( ViewBase* v1, ViewBase* v2, const AMatrix& fmat, candMapT& candidates, candOrigMapT& candOrig )
		{
			candidates.clear();
			candOrig.clear();
			uint num1 = v1->getNumSegments();
			uint num2 = v2->getNumSegments();
			for ( uint i = 0; i < num1; ++i )
			{
				candListT& cand_list = candidates[i];
				candOrigListT& cand_listOr = candOrig[i];
				cand_list.reserve( num2 );
				cand_listOr.reserve( num2 );
				ViewBase::segmentT seg1 = v1->getSegment( i );
				const AVector& pt1 = seg1->getPt1();
				const AVector& pt2 = seg1->getPt2();

				AVector epiline_1(3), epiline_2(3);
				epiline_1 = matMul(fmat, pt1);
				epiline_2 = matMul(fmat, pt2);
				PointsInfo rel1, rel2;
				AVector pt1_1(3), pt1_2(3), pt2_1(3), pt2_2(3);

				for(uint j = 0; j < num2; ++j)
				{
					ViewBase::segmentT seg2 = v2->getSegment(j);
					const AVector& pt3 = seg2->getPt1();
					const AVector& pt4 = seg2->getPt2();

					bool havePt1 = evaluateIntersection(epiline_1, pt3, pt4, pt2_1, rel1);
					bool havePt2 = evaluateIntersection(epiline_2, pt3, pt4, pt2_2, rel2);
					if(havePt1 || havePt2)
					{
						if(havePt1) pt1_1 = pt1;
						if(havePt2) pt1_2 = pt2;

						AVector pt_in(3), epiline(3);
						if(havePt1 != havePt2) {
							if(rel1.first != rel2.first)
								pt_in = pt3;
							else
								pt_in = pt4;
							matMul2_AT(fmat, pt_in, epiline);
							if(!havePt1) {
								if(!evaluateIntersection(epiline, pt1, pt2, pt1_1, rel1))
									continue;
								pt2_1 = pt_in;
							} else {
								if(!evaluateIntersection(epiline, pt1, pt2, pt1_2, rel1))
									continue;
								pt2_2 = pt_in;
							}
						}
					}
					else
					{
						if(rel1.first == rel2.first)
							continue;
						AVector epiline(3);
						matMul2_AT(fmat, pt3, epiline);
						if(!evaluateIntersection(epiline, pt1, pt2, pt1_1, rel1))
							continue;
						pt2_1 = pt3;
						matMul2_AT(fmat, pt4, epiline);
						if(!evaluateIntersection(epiline, pt1, pt2, pt1_2, rel1))
							continue;
						pt2_2 = pt4;
					}

					double len2_2 = dist2(pt2_1, pt2_2);
					if(len2_2 < minLength2)
						continue;
					double len2_1 = dist2(pt1_1, pt1_2);
					if(len2_1 < minLength2)
						continue;
					if(maxRatio2 < std::max(len2_1, len2_2) / std::min(len2_1, len2_2))
						continue;

					cand_list.push_back(make_pair( make_pair(pt1_1/*.copy()*/, pt1_2/*.copy()*/), make_pair(pt2_1/*.copy()*/, pt2_2/*.copy()*/)));
					cand_listOr.push_back( make_pair( seg1, seg2 ) );
				}
			}
		}
	}
}
