

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <flann/flann.hpp>

#include "flannmatcher.h"

#include <adrcore/util/debug.h>

#include <cassert>

#include <vector>

#include <boost/shared_array.hpp>

using std::vector;

using boost::shared_array;

namespace adore
{
	namespace core
	{

		template<typename Tuple, int n>
		struct Sorter_N {
			bool operator()(const Tuple& a, const Tuple& b)
			{
				return a.get<n>() < b.get<n>();
			}
		};

		FlannMatcher::FlannMatcher(float threshold, uint count_threshold) :
			m_threshold(threshold),
			m_count_threshold(count_threshold)
		{
		}

		FlannMatcher::FlannMatcher(const FlannMatcher& rhs) :
			m_threshold(rhs.m_threshold),
			m_count_threshold(rhs.m_count_threshold)
		{
		}

		shared_array<float> FlannMatcher::copy_data(ViewBase * view, vector<FeaturePoint*>& fps, uint& columns)
		{
			shared_array<float> result(NULL);

			fps.clear();

			if(view->getNumFeaturePoints() == 0) return result;

			fps.reserve(view->getNumFeaturePoints() * 2);
			vector<uint> descriptor_indices;
			descriptor_indices.reserve(view->getNumFeaturePoints() * 2);
			columns = 0;
			for(ViewBase::PointsIterator it = view->getPointsBegin(); it != view->getPointsEnd(); ++it)
			{
				FeaturePoint * fp = *it;
				FeatureDescriptor * fd = fp->getDescriptor(0);
				if(fd == NULL || fd->size() < 1)  continue;

				FeatureDescriptor::DescVecT & descr = (*fd)(0);
				if(columns == 0) {
					columns = descr.rows(); // initialize
					if(columns == 0) return result;
				}

				for(uint i = 0; i < fd->size(); ++i) {
					FeatureDescriptor::DescVecT & d = (*fd)(i);
					if(columns == (uint)d.rows()) {
						fps.push_back(fp);
						descriptor_indices.push_back(i);
					}
				}
			}

			if(fps.empty()) return result;

			result.reset(new float[fps.size() * columns]);

			float * ptr = result.get();
			for(uint i = 0; i < fps.size(); ++i)
			{
				FeaturePoint * fp = fps[i];
				uint descr_index = descriptor_indices[i];
				FeatureDescriptor * fd = fp->getDescriptor(0);
				FeatureDescriptor::DescVecT & descr = (*fd)(descr_index);

				for(uint j = 0; j < columns; ++j, ++ptr)
					*ptr = descr(j);
			}

			return result;
		}

		void FlannMatcher::match ( ViewBase * v1, ViewBase * v2, NView * set )
		{
			set->addView( v1 ); // just in case ...
			set->addView( v2 );

			vector<FeaturePoint*> fps1, fps2;
			uint columns1, columns2;
			shared_array<float> descriptors1 = copy_data(v1, fps1, columns1);
			if(!descriptors1) return;
			shared_array<float> descriptors2 = copy_data(v2, fps2, columns2);
			if(!descriptors2 || columns1 != columns2) return;

			DBGOUT( DBGNORMAL, "FlannMatcher: matching set of size %zd and %zd\n", fps1.size(), fps2.size() );

			const uint nn = 1;
			const uint qrows = fps2.size();

			flann::Matrix<float> dataset(descriptors1.get(), fps1.size(), columns1);
			flann::Matrix<float> query(descriptors2.get(), fps2.size(), columns1);

			shared_array<int> indices_deleter(new int[qrows * nn]);
			flann::Matrix<int> indices(indices_deleter.get(), qrows, nn);
			shared_array<float> dists_deleter(new float[qrows * nn]);
			flann::Matrix<float> dists(dists_deleter.get(), qrows, nn);

			// construct an randomized kd-tree index using 1 kd-tree
			flann::Index<flann::L2<float> > index(dataset, flann::KDTreeIndexParams(1));
			index.buildIndex();

			// do a knn search, using unlimited checks
			index.knnSearch(query, indices, dists, nn, flann::SearchParams(flann::FLANN_CHECKS_UNLIMITED));

			// sort matches
			vector<tuple<int, int, float> > matches(qrows);
			for(uint i = 0; i < qrows; ++i)
				matches[i] = make_tuple(indices[i][0], int(i), dists[i][0]);

			Sorter_N<tuple<int, int, float>, 2> sorter;
			std::sort(matches.begin(), matches.end(), sorter);

			uint num_added = 0;
			for(uint i = 0; i < qrows && i < m_count_threshold; ++i)
			{
				float badness = matches[i].get<2>();
				if(badness > m_threshold) break;

				FeaturePoint * fp1 = fps1[matches[i].get<0>()];
				FeaturePoint * fp2 = fps2[matches[i].get<1>()];
				num_added += addPairToNview(fp1, fp2, set, badness) ? 1 : 0;
			}

			DBGOUT( DBGNORMAL, "FlannMatcher: added %d correspondences\n", num_added );
		}

	} // core
} // adore
