

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESCALEHARRISFINDER_H
#define ADORE_CORESCALEHARRISFINDER_H

#include <adrcore/imgprocessing/stretchnskewremover.h>
#include <adrcore/viewmatch/featurefinder.h>
//#include <adrcore/imgprocessing/scalespacedog.h>
//#include <adrcore/imgprocessing/imagering.h>
#include <adrcore/imgprocessing/derivatives.h>
#include <adrcore/viewmatch/descriptorcreator.h>
#include <adrcore/features/extremafeatures.h>
#include <adrcore/types/view.h>
#include <adrcore/features/channelextrema.h>
//#include <adrcore/imgprocessing/harrisfast.h>
#include <adrcore/imgprocessing/harrisfasthood.h>
#include <adrcore/imgprocessing/channelfunctions.h>
#include <adrcore/imgprocessing/scalehelpers.h>

#ifndef NDEBUG
//uncomment this
#define NSCALEDEBUG
#else
#define NSCALEDEBUG
#endif

#define DBGSHF DBGALL

namespace adore
{

	namespace core
	{

		/**
		 *
			 
		*/
		class ScaleHarrisFinder : public FeatureFinder
		{
			public:
				typedef signedT< imgCalcT<imgT>::type >::type diffT; //uint8_t -> int16_t
				typedef signedT< calcT<imgT>::type >::type calcT; //uint8_t -> int32_t
				typedef harrisOutT<imgT>::ouT harrisT;

			private:
				uint fid;//feature id

				ImagePlane< harrisT > harrisImg;
				ImagePlane< imgT > patchImg;//for StretchNSkew-Removal
				ChannelExtrema< harrisT , ScaleHarrisFinder> extrema;
				FeatureGrid candidates;

				int numSmooths;
				int numImages;
				float scale;
				int spacing;
				DescriptorCreator *descriptorcreator;
				bool ownCreator;

				int maxPts;
				float minDist;
				float minFraction;

				ViewT currentView;//a View that wraps the images to please the descriptorcreators
				int patchFactor;//this times scaleVal gets the patchsize for SnS-Revoe and Decriptor-Affine-Det

				bool snsRemove;
				AMatrixf snsMat;
				StretchNSkewRemover<imgT> snsrm;

			public:
				/**
				 * @param numSmooths number of smooths between subsamplings
				 * @param numImages number of smoothed images computed
				 */
				ScaleHarrisFinder( uint ID, int numSmooths, int numImages, calcT minHarrisThresh, int maxPoints = INT_MAX, float minDistance = 3.0F, float minFract = 0.000001 );

				ScaleHarrisFinder( const ScaleHarrisFinder &shf );

				~ScaleHarrisFinder();

				FeatureFinder *clone() {return new ScaleHarrisFinder( *this );}

				inline void operator()( std::size_t x, std::size_t y )
				{
					operator ( ) ( harrisImg, harrisImg.getPointer( x, y ) );
				}

				inline void operator()( const Channel<harrisOutT<imgT>::ouT> &img, harrisOutT<imgT>::ouT *localMaximum )
				{
//					DBGOUT( DBGSHF, "ScaleHarris: extrema callback\n" );
					const float dxMax = 0.5F, dyMax = 0.5F;

					uint x = 0, y = 0;
					float dx = 0, dy = 0;
					img.getCoors( localMaximum, x, y );

//					DBGOUT( DBGNORMAL, "incoming feat (%d, %d)\n", x, y );

					bool success = subCoorQuadratic( img.getIterator( y - 1, x - 1, 3 ), dx, dy );
					if ( success && fabs( dx ) < dxMax && fabs( dy ) < dyMax )
					{
						dx += x;
						dy += y;


						float scaleVal = max( 1.0F, scale / spacing );
						int hood = patchFactor * int( scaleVal + 0.5F );
						int width = 2 * hood + 1;

						bool success = true;
						if ( snsRemove )
						{
							snsrm.setPatchSize( width );
							UsingROI ur1( currentView.getImage(), ROI( x - hood, y - hood, width, width ) );
							UsingROI ur2( &patchImg, ROI( 0, 0, width, width ) );
							success = snsrm.process( currentView.getImage(), dx, dy, &patchImg );
							snsMat = snsrm.getMatrix();
							if ( !success )
							{
								DBGOUT( DBGNORMAL, "sns-remove failed for (%f, %f)\n", dx, dy );
							}
						}

						if ( success )
						{
							//dont's scale the coordinate here
							//there are used inside the descriptorcreators which
							//access the also scaled image
							FeaturePoint *pt = new FeaturePoint( dx, dy, fid );
							pt->setStrength( double ( *localMaximum ) );//TODO - this is approximately
							pt->setScale( scale );
#ifndef NSCALEDEBUG
							pt->getInvAffineMat() = scaleVal * snsMat;

							Rect<float> r = getAreaOfTrafo( pt->getInvAffineMat(), 16.0F, 16.0F );
							DBGOUT( DBGSHF, "(%.2f)-(%.2f)\n", r.getWidth(), r.getHeight() );

							ImagePlane<uint8_t> *p = new ImagePlane<uint8_t>( 17, 17, 1 );
							warpAffine( *currentView.getImage(), dx, dy, pt->getInvAffineMat(), p->begin(), 8, 8 );
							pt->patch = p;
#else
							pt->getInvAffineMat() = scaleVal * snsMat;
#endif

							if ( descriptorcreator )
							{
//								DBGOUT( DBGSHF, "ScaleHarris: currentView add\n" );
								if ( GRID_SUCCESS != currentView.add( pt ) ) error( "Grid in trouble\n" );
							}
							else
							{
								//scale now
								( *pt )( 0 ) = getPixelPosition( dx, spacing );
								( *pt )( 1 ) = getPixelPosition( dy, spacing );
//								DBGOUT( DBGSHF, "ScaleHarris: candidates add\n" );
								candidates.add( pt );
							}
						}
					}
				}

				void setDescriptorCreator ( DescriptorCreator* theValue, bool own = false )
				{
					ownCreator = false;
					descriptorcreator = theValue;
				}

				DescriptorCreator* getDescriptorCreator() const
				{
					return descriptorcreator;
				}

				void setSnsRemove ( bool theValue )
				{
					snsRemove = theValue;
				}

				bool getSnsRemove() const
				{
					return snsRemove;
				}

			protected:
				unsigned int findFeaturesDerived( ViewT *v );


		};

	}

}

#endif
