

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLNAVIGATOR_H
#define ADORE_GLNAVIGATOR_H

typedef unsigned int uint;
namespace adore
{
	namespace visual
	{
		/**	Base class for scene manipulation
		 *
		 *	 
		 */
		class GLNavigator
		{
		public:
			/**	Destructor
			 */
			virtual ~GLNavigator() {};

			/**	Initialization, must be called whenever the geometric data in NView is changed
			 *
			 *	GLVisualizer is designed as a renderer for static geometry.
			 *	Therefor GLVisualizer::init() must be called after every change
			 *	in the geometry of the NView to reflect all changes.
			 */
			virtual void init() = 0;

			/**	Must be called before painting sequence
			 *
			 *	This function sets up presentation parametres which
			 *	may differ per Visualizer but stay constant each
			 *	successive paint operation.
			 */
			virtual void focus() = 0;

			/**	Must be called when another Visualizer gains focus
			 */
			virtual void unfocus() = 0;

			/**	Paint scene
			 */
			virtual void paint() = 0;
			
			/**	Resize viewport
			 *
			 *	@param width Viewport witdth in pixels
			 *	@param height Viewport height in pixels
			 */
			virtual void resize(int width, int height) = 0;

			/**	Determine the object at position (x,y) and notify the listeners
			 *
			 *	@param x Coordinate x of the picking cursor
			 *	@param y Coordinate y of the picking cursor
			 */
			virtual void pick(int x, int y) = 0;

			/**	Set up perspective projection
			 *
			 *	@param fov Field of view
			 */
			virtual void perspective(float fov) = 0;

			/**	Set up the camera position. The vector pointing up becomes
			 *	the spine of the virtual observer. Turning and tilting function will
			 *	consider the spine rotating the camera.
			 *
			 *	@param eye An eye location in world space
			 *	@param look An object location in world space
			 *	@param up A vector pointing up in world space
			 */
			virtual void lookAt(const float eye[3], const float aim[3], const float up[3]) = 0;

			/**	Roll the camera
			 *
			 *	@param angle Roll angle in radians
			 */
			virtual void roll(float angle) = 0;

			/**	Yaw the camera
			 *
			 *	@param angle Yaw angle in radians
			 */
			virtual void yaw(float angle) = 0;

			/**	Increase the pitch of the camera
			 *
			 *	@param angle Pitch angle in radians
			 */
			virtual void pitch(float angle) = 0;

			/**	Slide the camera
			 *
			 *	@param delta Distance to slide in camera space
			 */
			virtual void slide(const float delta[3]) = 0;

			/**	Turn the camera around the spine keeping the up vector collinear to it
			 *
			 *	@param angle turning angle in radians
			 */
			virtual void turn(float angle) = 0;

			/**	Turn the camera around the y-axis keeping the minimum angle
			 *	between the up-vector and the spine no less than gap
			 *
			 *	@param angle tilting angle in radians
			 *	@param gap minimum gap value
			 */
			virtual void tilt(float angle, float gap) = 0;

			/**	Translate the scene
			 *
			 *	@param delta Distance to translate in world space
			 *	@param absolute Absolute or relative translation
			 */
			virtual void translate(const float delta[3], bool absolute) = 0;

			/**	Rotate the scene
			 *
			 *	@param axis Rotation axis in world space
			 *	@param angle rotation angle in radians
			 *	@param absolute Absolute or relative rotation
			 */
			virtual void rotate(const float axis[3], float angle, bool absolute) = 0;
			
			virtual void turnTable(float angle, bool absolute) = 0;
			
			virtual void setTurnTableSpine(const float axis[3]) = 0;

			/**	Scale the scene
			 *
			 *	@param scale Scaling factor for 3 axes in object space
			 *	@param absolute Absolute or relative scaling
			 */
			virtual void scale(const float factor[3], bool absolute) = 0;

			/** Reset all scene transformations to defaut values
			 */
			virtual void identity() = 0;

			virtual void getTranslation(float (& v3)[3], uint idx) = 0;
		};
	}
}

#endif
