

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_OPENGLSEGGRAPHDISPLAYCLIENT_H
#define ADORE_OPENGLSEGGRAPHDISPLAYCLIENT_H

#include "displayclient.h"
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/channelfunctions.h>
#include <boost/graph/undirected_dfs.hpp>
#include <boost/graph/adjacency_list.hpp>

namespace adore {
namespace visual {

    using namespace adore::core;
    using namespace boost;
    
/**
	 
    Consider BGL-Visitors are passed by value!
*/
template<class SegGraphT>
        class SegGraphDisplayClient : public DisplayClient, public boost::dfs_visitor<>
{
    typedef typename SegGraphT::SegmentT SegmentT;
    typedef typename SegGraphT::SegmentVertex SegVertexT;
    typedef typename SegGraphT::SegmentEdge SegEdgeT;
    
    public:
        /**
         * Parameters are the width and height of the image from which 
         * the graph was computed. This is needed for a proper overlay.
         */
        SegGraphDisplayClient(SegGraphT *g, int chanWidth, int chanHeight):
        graph(g),
        width(chanWidth),
        height(chanHeight)
        {
        }
    
        ~SegGraphDisplayClient(){;}
        
        //below are the visitor functions
        /**
            is invoked on every vertex of the graph before the start of the graph search
        */
        template<class Vertex_T, class Graph_T>
        void initialize_vertex(Vertex_T s, Graph_T &g)
        {
            SegmentT *seg = g[s].seg;
            assert(seg);
//             printf("initialize vertex: id: %d, ptr: %p\n", seg->getID(), seg);
            seg->calculateCenterOfGravity();
        }
        
        /**
         *  is invoked when a vertex is encountered for the first time
         */
        template<class Vertex_T, class Graph_T>
        void discover_vertex(Vertex_T u, Graph_T &g)
        {
            draw_vertex(g[u].seg);
        }
        
        /**
         * is invoked on every out-edge of each vertex after it is discovered
         */
        template<class Edge_T, class Graph_T>
        void tree_edge(Edge_T e, Graph_T &g)
        {
            draw_edge(g[source(e, g)].seg, g[target(e, g)].seg);
        }
        
        template<class Edge_T, class Graph_T>
        void back_edge(Edge_T e, Graph_T &g)
        {
            draw_edge(g[source(e, g)].seg, g[target(e, g)].seg);
        }
        
        void draw_vertex(SegmentT *seg)
        {
            assert(seg);
//             printf("discover vertex: id: %d, ptr: %p\n", seg->getID(), seg);
    
            float x, y;
            seg->getCenterOfGravity(x, y);
            imageToWindow(x, y);
    
            glColor4f(0.0, 1.0, 0.0, 1.0);
            drawDot(x, y);
        }
        
        void draw_edge(SegmentT *seg1, SegmentT *seg2)
        {
//             printf("examine edge: (%d - %d)\n", seg1->getID(), seg2->getID());
            
            assert(seg1);
            assert(seg2);
            
            float x1, y1, x2, y2;
            seg1->getCenterOfGravity(x1, y1);
            seg2->getCenterOfGravity(x2, y2);
            imageToWindow(x1, y1);
            imageToWindow(x2, y2);
            
            glBegin(GL_LINES);
            glColor4f(0.0, 0.0, 1.0, 1.0);
            glVertex2f(x1, y1);
            glVertex2f(x2, y2);
            glEnd();
        }
    
    private:
        
        void drawDot(float x = 0.0f, float y = 0.0f, float rad = 3.0) {
            float d = rad / sqrt(2.0);
            glBegin(GL_TRIANGLE_FAN);
            glVertex3f( x      ,  y      , 0.0f);
            glVertex3f( x + rad,  y      , 0.0f);
            glVertex3f( x +   d,  y +   d, 0.0f);
            glVertex3f( x      ,  y + rad, 0.0f);
            glVertex3f( x -   d,  y +   d, 0.0f);
            glVertex3f( x - rad,  y      , 0.0f);
            glVertex3f( x -   d,  y -   d, 0.0f);
            glVertex3f( x      ,  y - rad, 0.0f);
            glVertex3f( x +   d,  y -   d, 0.0f);
            glVertex3f( x + rad,  y      , 0.0f);
            glEnd();            
        }

        void drawSceneImpl()
        {
            prepareRemap(width, height);
//             printf("drawSceneImpl - num_vertices:%d\n", num_vertices(*graph));
            
            enable2D();
            undirected_dfs(*graph, *this, get(&SegVertexProps<SegmentT>::color, *graph), get(&SegEdgeProps::color, *graph));
//             dumbDraw();
            disable2D();
        }
        
        void dumbDraw()
        {
            typename graph_traits<SegGraphT>::vertex_iterator vi, vi_end;
            
            for (tie(vi, vi_end) = vertices(*graph); vi != vi_end; ++vi)
            {
                SegmentT *seg = (*graph)[*vi].seg;
                seg->calculateCenterOfGravity();
            }
            
            for (tie(vi, vi_end) = vertices(*graph); vi != vi_end; ++vi)
            {
                SegmentT *seg1 = (*graph)[*vi].seg;
                draw_vertex(seg1);
                typename graph_traits<SegGraphT>::adjacency_iterator avi, avi_end;
                for (tie(avi, avi_end) = adjacent_vertices(*vi, *graph); avi != avi_end; ++avi)
                {
                    SegmentT *seg2 = (*graph)[*avi].seg;
                    draw_edge(seg1, seg2);
                }
            }
        }

        
        SegGraphT *graph;
        int width, height;
        
};

}}



#endif
