

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "featuresscene.h"
#include "pixmapplane.h"
#include "featureslayer.h"
#include "corrgl2d.h"
#include "corrlayeritem.h"

#include <adrcore/types/pointcorr.h>
#include <adrcore/types/nview.h>
#include <adrcore/nview/ttensorcalculator.h>

#include <cassert>
#include <set>

using namespace std;
using namespace adore::core;

FeaturesScene::FeaturesScene( QObject* p ) :
		parent( p ),
		rect( 0, 0, 0, 0 ),
		featsShown( true ),
		linesShown( true ),
		pCorrsShown( true ),
		lCorrsShown( true ),
		lastNView( NULL ),
		firstView( NULL ),
		secondView( NULL )

{
	assert( p != NULL );
}

FeaturesScene::~FeaturesScene()
{
	reset();
}

void FeaturesScene::reset()
{
	removeEpipolarLines();

	imagePlanes.clear();
	orderedImages.clear();
	lastFocusedImage.reset();
	lastHighlightedImage.reset();

	featureLayers.clear();
	corrLayers.clear();
	lastNView = NULL;
	lastSelectedCorr.reset();

	rect.setCoords( 0, 0, 0, 0 );
}

void FeaturesScene::resetPreservingViews()
{
	removeEpipolarLines();

	lastFocusedImage.reset();
	lastHighlightedImage.reset();

	featureLayers.clear();
	corrLayers.clear();
	lastNView = NULL;
	lastSelectedCorr.reset();
}

void FeaturesScene::addView( ViewT* view )
{
	ImagePtr image( new PixmapPlane( view ) );
	imagePlanes[view] = image;
	orderedImages.push_back(image);
	QPointF pos = nextViewPos( image );
	image->setPosition( pos );
	calcSceneRect();

	QObject::connect ( image.get(), SIGNAL ( signalSelected ( const ViewT* const, bool ) ),
					   parent, SLOT ( slotSelected ( const ViewT* const, bool ) ) );
	QObject::connect ( image.get(), SIGNAL ( signalFocused ( const ViewT* const ) ),
					   parent, SLOT ( slotFocused ( const ViewT* const ) ) );

	FeaturesPtr fLayer( new FeaturesLayer( image->getView(), true ) );
	fLayer->setPosition( image->getPosition() );
	fLayer->setFeaturesVisible( featsShown );
	fLayer->setLinesVisible( linesShown );
	featureLayers[image->getView()] = fLayer;
}

QPointF FeaturesScene::nextViewPos( ImagePtr image ) const
{
	double x = rect.right() + imagePadding;
	double y = rect.top() + ( rect.height() - image->height() ) / 2;
	if ( 0 == rect.height() )
		y = 0;
	if ( 0 == rect.width() )
		x = 0;
	return QPointF( x, y );
}

void FeaturesScene::calcSceneRect()
{
	QRectF r;
	map<const ViewBase*, ImagePtr >::iterator it = imagePlanes.begin();
	for ( ; it != imagePlanes.end(); ++it )
	{
		ImagePtr img = it->second;
		r |= img->rect();
	}
	rect = r;
}

void FeaturesScene::paint( const QRectF& r )
{
	I_iterator it = imagePlanes.begin();
	for ( ; it != imagePlanes.end(); ++it )
	{
		ImagePtr img = it->second;
		if ( r.intersects( img->rect() ) )
		{
			img->paint();
			FL_iterator fit = featureLayers.find( it->first );
			if ( fit != featureLayers.end() )
				fit->second->paint();
		}
	}
	if ( lastNView != NULL )
	{
		CL_iterator cit = corrLayers.find( lastNView );
		if ( cit != corrLayers.end() )
			cit->second->paint();
	}

	if ( lastSelectedCorr )
		lastSelectedCorr->paint();

	glFlush();
}

void FeaturesScene::getSelectedViews( std::vector<ViewT*>& views ) const
{
	OrderedImagesT::const_iterator it = orderedImages.begin();
	while(it != orderedImages.end()) {
		if( (*it)->isSelected() )
			views.push_back( (*it)->getView() );
		++it;
	}
}

ViewT* FeaturesScene::getFocusedView() const
{
	ViewT* result = NULL;
	OrderedImagesT::const_iterator it = orderedImages.begin();
	while(it != orderedImages.end()) {
		if( (*it)->isFocused() ) {
			result = (*it)->getView();
			break;
		}
		++it;
	}
	return result;
}

void FeaturesScene::highlightView ( const ViewBase * v )
{
	if ( lastHighlightedImage )
		lastHighlightedImage->setHighlighted ( false );
	lastHighlightedImage.reset();

	if ( !v ) return;

	if ( imagePlanes.find( v ) == imagePlanes.end() )
		return;
	ImagePtr item = imagePlanes[v];
	if ( !item ) return;

	item->setHighlighted ( true );
	lastHighlightedImage = item;
}

void FeaturesScene::focusView ( const ViewBase* v )
{
	if ( lastFocusedImage )
		lastFocusedImage->setFocused ( false );
	lastFocusedImage.reset();

	if ( !v ) return;

	if ( imagePlanes.find( v ) == imagePlanes.end() )
		return;
	ImagePtr item = imagePlanes[v];
	if ( !item ) return;

	item->setFocused ( true );
	lastFocusedImage = item;
}

void FeaturesScene::selectView ( const ViewBase* const v, bool on )
{
	if ( imagePlanes.find( v ) == imagePlanes.end() )
		return;
	ImagePtr item = imagePlanes[v];
	if ( item ) item->setSelected ( on );
}

void FeaturesScene::mouseClicked( const QPointF& pos, Qt::MouseButton button )
{
	I_iterator it = imagePlanes.begin();
	for ( ; it != imagePlanes.end(); ++it )
	{
		ImagePtr img = it->second;
		if ( img->rect().contains( pos ) )
		{
			img->mousePressEvent( pos, button );
			return;
		}
	}
}

void FeaturesScene::mouseCtrlClicked( const QPointF& pos )
{
	if(lastNView == NULL || lastNView->getNumViews() < 2 || lastNView->getNumViews() > 3)
		return;
	if(lastNView->isNormalized())
		return;

	ImagePtr clickedImg;
	QPointF correctedPoint;
	I_iterator it = imagePlanes.begin();
	for ( ; it != imagePlanes.end(); ++it )
	{
		ImagePtr img = it->second;
		if ( img->rect().contains( pos ) )
		{
			clickedImg = img;
			correctedPoint = pos - img->getPosition();
			break;
		}
	}
	if(clickedImg == NULL) return;

	ViewT * clickedView = NULL;
	for (NView::ViewIterator vit = lastNView->getViewsBegin(); vit != lastNView->getViewsEnd(); ++vit )
	{
		if(clickedImg->getView() == *vit)
		{
			clickedView = clickedImg->getView();
			break;
		}
	}
	if(clickedView == NULL) return;

	if(firstView == NULL) firstView = clickedView;
	else if(firstView != clickedView && secondView == NULL) secondView = clickedView;

	if(firstView == clickedView)
		ptInFirstView = correctedPoint;
	else if(secondView == clickedView)
		ptInSecondView = correctedPoint;

	updateEpipolarLines();
}

void swapMat(AMatrix & m1, AMatrix& m2, AMatrix& temp)
{
	temp = m1; m1 = m2; m2 = temp;
}

void FeaturesScene::updateEpipolarLines()
{
	if(lastNView == NULL || lastNView->getNumViews() < 2 || lastNView->getNumViews() > 3 || firstView == NULL)
	{
		removeEpipolarLines();
		return;
	}

	uint numViews = lastNView->getNumViews();
	AMatrix fmat12(3, 3), fmat13(3, 3), fmat23(3, 3), tempMat(3, 3);
	ATensor3 ttensor(3, 3, 3);
	ViewBase * v1 = firstView, * v2 = secondView, * v3 = NULL;

	NViewInfo *info = lastNView->getInfo();
	if(info == NULL) {
		removeEpipolarLines();
		return;
	}

	if(numViews == 2) {
		if ( isNullArray( info->getFMatrix() ) )
		{
			removeEpipolarLines();
			return;
		}
		fmat12 = info->getFMatrix();
		if(*lastNView->getViewsBegin() != firstView)
		{
			std::swap(v1, v2);
			v2 = *lastNView->getViewsBegin();
			transInPlace(fmat12);
		} else {
			v2 = *++lastNView->getViewsBegin();
		}
	}
	else {
		if ( isNullArray( info->getTTensor() ) )
		{
			removeEpipolarLines();
			return;
		}
		ttensor.resize( 3, 3, 3 );
		ttensor = info->getTTensor();

		ViewBase* testV1 = *lastNView->getViewsBegin();
		ViewBase* testV2 = *++lastNView->getViewsBegin();
		ViewBase* testV3 = *++(++lastNView->getViewsBegin());

		TTensorCalculator calc;
		AVector epi_12( calc.calculateEpipole( ttensor, true ) );
		AVector epi_13( calc.calculateEpipole( ttensor, false ) );

		fmat12 = calc.calculateFundamentalMatrix( ttensor, epi_12, epi_13, TTensorCalculator::FMATRIX_F21 );
		fmat13 = calc.calculateFundamentalMatrix( ttensor, epi_12, epi_13, TTensorCalculator::FMATRIX_F31 );
		fmat23 = calc.calculateFundamentalMatrix( ttensor, epi_12, epi_13, TTensorCalculator::FMATRIX_F32 );


		if(v1 == testV1) {
			if(v2 == NULL) v2 = testV2;

			if(v2 == testV2) {
				v3 = testV3;
			} else {
				v3 = testV2;
				transInPlace(fmat23);
				swapMat(fmat12, fmat13, tempMat);
			}
		} else if(v1 == testV2) {
			if(v2 == NULL) v2 = testV3;

			transInPlace(fmat12);
			if(v2 == testV1) {
				v3 = testV3;
				swapMat(fmat13, fmat23, tempMat);
			} else {
				v3 = testV1;
				transInPlace(fmat13);
				swapMat(fmat12, fmat23, tempMat);
				swapMat(fmat13, fmat23, tempMat);
			}
		} else {
			if(v2 == NULL) v2 = testV2;

			transInPlace(fmat23);
			transInPlace(fmat13);
			if(v2 == testV1) {
				v3 = testV2;
				swapMat(fmat12, fmat13, tempMat);
				swapMat(fmat13, fmat23, tempMat);
			} else {
				v3 = testV1;
				transInPlace(fmat12);
				swapMat(fmat12, fmat23, tempMat);
			}
		}
	}

	AVector pt(3);
	pt = ptInFirstView.x(), ptInFirstView.y(), 1;
	AVector line = matMul( fmat12, pt );
	imagePlanes[v2]->setEpiline1(line);
	if(numViews == 3)
	{
		AVector line = matMul( fmat13, pt );
		imagePlanes[v3]->setEpiline1(line);

		if(secondView == NULL) return;

		pt = ptInSecondView.x(), ptInSecondView.y(), 1;
		line = matMul( fmat23, pt );
		imagePlanes[v3]->setEpiline2(line);
	}
}

void FeaturesScene::removeEpipolarLines()
{
	firstView = secondView = NULL;
	I_iterator iit = imagePlanes.begin();
	for ( ; iit != imagePlanes.end(); ++iit )
		iit->second->removeEpilines();
}

void FeaturesScene::featuresChanged()
{
	FL_const_iterator fi = featureLayers.begin();
	if ( fi == featureLayers.end() )
	{
		I_const_iterator ii = imagePlanes.begin();
		while ( ii != imagePlanes.end() )
		{
			ImagePtr img = ii->second;
			FeaturesPtr fLayer( new FeaturesLayer( img->getView(), true ) );
			fLayer->setPosition( img->getPosition() );
			fLayer->setFeaturesVisible( featsShown );
			fLayer->setLinesVisible( linesShown );
			featureLayers[img->getView()] = fLayer;
			++ii;
		}
		return;
	}
	while ( fi != featureLayers.end() )
	{
		FeaturesPtr fLayer = fi->second;
		fLayer->updateFeatures();
		++fi;
	}
}

void FeaturesScene::nViewSelected ( NView * nv )
{
	assert ( nv );

	/// remove epipolar lines
	firstView = NULL;
	updateEpipolarLines();

	I_iterator iit = imagePlanes.begin();
	for ( ; iit != imagePlanes.end(); ++iit )
		selectView ( iit->first, false );

	NView::ViewIterator vit = nv->getViewsBegin();
	for ( ; vit != nv->getViewsEnd(); ++vit )
		selectView ( *vit, true );

	lastSelectedCorr.reset();
	lastNView = nv;
	ViewBase* view = * ( nv->getViewsBegin() );
	if ( view && 0 != nv->getNumCameras() )
		focusView ( view );
	else if ( 0 == nv->getNumCameras() )
		focusView ( 0 );

	bool skip = corrLayers.find( nv ) != corrLayers.end();
	if ( !skip )
	{
		vector<ImagePtr> images;
		NView::ViewIterator vit = nv->getViewsBegin();
		NView::ViewIterator end = nv->getViewsEnd();
		for ( ; vit != end; ++vit )
		{
			ImagePtr img;
			for(OrderedImagesT::iterator iit = orderedImages.begin(); iit != orderedImages.end(); ++iit)
			{
				if((*iit)->getView() == *vit) {
					img = *iit;
					break;
				}
			}
			if(img != NULL)
				images.push_back( img );
		}
		if ( images.size() < 2 )
			return;
		CorrsPtr corrs( new CorrLayerItem( nv, images ) );
		corrs->setPointCorrsVisible( pCorrsShown );
		corrs->setLineCorrsVisible( lCorrsShown );
		corrLayers[nv] = corrs;
	}
}

void FeaturesScene::highlightPointCorr ( PointCorr * c )
{
	lastSelectedCorr.reset();

	if( !lastNView ) return;
	if ( !c ) return;

	vector<ImagePtr> images;
	NView::ViewIterator vit = lastNView->getViewsBegin();
	NView::ViewIterator end = lastNView->getViewsEnd();
	for ( ; vit != end; ++vit )
	{
		ImagePtr img;
		for(OrderedImagesT::iterator iit = orderedImages.begin(); iit != orderedImages.end(); ++iit)
		{
			if((*iit)->getView() == *vit) {
				img = *iit;
				break;
			}
		}
		if(img != NULL)
			images.push_back( img );
	}
	if ( images.size() < 2 )
		return;

	lastSelectedCorr.reset( new CorrGL2D( c,images ) );
}

void FeaturesScene::setShowFeats( bool on )
{
	FL_iterator fit = featureLayers.begin();
	for ( ; fit != featureLayers.end(); ++fit )
		fit->second->setFeaturesVisible( on );
	featsShown = on;
}

void FeaturesScene::setShowLines( bool on )
{
	FL_iterator fit = featureLayers.begin();
	for ( ; fit != featureLayers.end(); ++fit )
		fit->second->setLinesVisible( on );
	linesShown = on;
}

void FeaturesScene::setShowPointCorrs ( bool on )
{
	CL_iterator cit = corrLayers.begin();
	for ( ; cit != corrLayers.end(); ++cit )
		cit->second->setPointCorrsVisible( on );
	pCorrsShown = on;
}

void FeaturesScene::setShowLineCorrs ( bool on )
{
	CL_iterator cit = corrLayers.begin();
	for ( ; cit != corrLayers.end(); ++cit )
		cit->second->setLineCorrsVisible( on );
	lCorrsShown = on;
}
