

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <QtGui/QColor>
#include <QtGui/QMouseEvent>
#include <QtGui/QImage>

#include "pixmapplane.h"
#include "qglhelper.h"

#include <adrcore/util/mathutil.h>

#include <GL/glu.h>

#include <boost/shared_array.hpp>

PixmapPlane::PixmapPlane( ViewT* v ) :
		view( v ),
		texId( 0 ),
		iconId( 0 ),
		selected( false ),
		focused( false ),
		highlighted( false )
{
	loadView();
	if ( !texId )
	{
		double width = v->getWidth();
		double height = v->getHeight();
		bool norm = v->getNormalized();
		if ( norm )
		{
			double scale = v->getInvTrafo()( 0, 0 );
			width = ceil( scale * width );
			height = ceil( scale * height );
		}
		size = QSizeF( width, height );
	}
	generateFocusIcon();
}

PixmapPlane::~PixmapPlane()
{
	if ( 0 != texId )
		glDeleteTextures( 1, &texId );
	if ( 0 != iconId )
		glDeleteTextures( 1, &iconId );
}

void PixmapPlane::loadView()
{
	ImagePlaneT* img = view->getImage();
	uint numChannels = img->getNumChannels();

	int width = img->getWidth();
	width += width % 2;
	int height = img->getHeight();
	height += height % 2;

	if ( 1 != numChannels && 3 != numChannels )
		return;
	if ( width <= 0 || height <= 0 )
		return;

	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );

	int texW = width, texH = height;
	boost::shared_array<imgTypeT> deleter;

	GLenum format = numChannels == 1 ? GL_LUMINANCE : GL_RGB;

	imgTypeT * imgPtr = img->getPointer(0, 0);

	if(texW > MaxTextureExtent || texH > MaxTextureExtent)
	{
		if(texW > texH) {
			texH = int( round( double(MaxTextureExtent) * (double(texH) / texW) ) );
			texW = MaxTextureExtent;
		} else {
			texW = int( round( double(MaxTextureExtent) * (double(texH) / texW) ) );
			texH = MaxTextureExtent;
		}
		int pixelSize = format == GL_LUMINANCE ? 1 : 4;
		deleter.reset(new imgTypeT[texW * texH * pixelSize]);
		GLint err = gluScaleImage(format, width, height, GL_UNSIGNED_BYTE, img->getPointer( 0, 0),
					  texW, texH, GL_UNSIGNED_BYTE, deleter.get());
		if(err != 0) {
			size = QSizeF( width, height );
			return;
		}
		imgPtr = deleter.get();
	}

	glGenTextures( 1, &texId );
	glBindTexture( GL_TEXTURE_2D, texId );
	if ( numChannels == 1 )
	{
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, texW, texH, 0,
					  GL_LUMINANCE, GL_UNSIGNED_BYTE, imgPtr );
	}
	else
	{
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, texW, texH, 0,
					  GL_RGB, GL_UNSIGNED_BYTE, imgPtr );
	}
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glBindTexture( GL_TEXTURE_2D, 0 );
	size = QSizeF( width, height );
}

void PixmapPlane::generateFocusIcon()
{
	QImage img(":/images/eye.png");

	glGenTextures( 1, &iconId );
//	GLenum err = glGetError();
	glBindTexture( GL_TEXTURE_2D, iconId );
//	err = glGetError();
//	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, img.width(), img.height(), 0,
//				  GL_RGBA, GL_UNSIGNED_BYTE, img.bits() );
//	err = glGetError();
	/*err = */gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGBA, img.width(), img.height(), GL_BGRA, GL_UNSIGNED_BYTE, img.bits());
//	err = glGetError();
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
//	err = glGetError();

	glBindTexture( GL_TEXTURE_2D, 0 );
}

void PixmapPlane::paint() const
{
	if ( 0 == texId ) {
		glColor4f( 0.7, 0.7, 0.7, 0.0 );
	} else {
		glColor4f(1.0, 1.0, 1.0, 0.0);
		glBindTexture( GL_TEXTURE_2D, texId );
	}

	glDisable(GL_BLEND);

	glBegin( GL_QUADS );
	glTexCoord2f( 0.0, 0.0 ); glVertex3f( x(), y(), 0.0 );
	glTexCoord2f( 1.0, 0.0 ); glVertex3f( x() + width(), y(), 0.0 );
	glTexCoord2f( 1.0, 1.0 ); glVertex3f( x() + width(), y() + height(), 0.0 );
	glTexCoord2f( 0.0, 1.0 ); glVertex3f( x(), y() + height(), 0.0 );
	glEnd();

	if(0 != texId)
		glBindTexture( GL_TEXTURE_2D, 0 );

	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	QRectF imgRect( position, size );
	drawSelectionMark(imgRect);
	drawHighlightMark(imgRect);
	drawFocusMark(imgRect);
	drawEpilines();
}

void PixmapPlane::drawSelectionMark(const QRectF& imgRect) const
{
	if ( selected ) {
		static QColor green(0, 255, 0, 100);
		drawFrame( imgRect, green, frameW );
	}
}

void PixmapPlane::drawHighlightMark(const QRectF& imgRect) const
{
	if ( highlighted )
	{
		QRectF frame(imgRect);
		if ( selected )
			frame.adjust( frameW, frameW, -frameW, -frameW );
		if ( !frame.isValid() || 1 > frame.width() || 1 > frame.height() )
		{
			return;
		}
		drawFrame( frame, QColor( 255, 0, 255, 255 ), frameW );
	}
}

void PixmapPlane::drawFocusMark(const QRectF& imgRect) const
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	if(!focused) glColor4ub( 0, 0, 255, 100 );
	else glColor4ub( 255, 255, 255, 255 );
	glBindTexture( GL_TEXTURE_2D, iconId );
	QRectF fMark(calculateFocusMark(imgRect));

	glBegin( GL_QUADS );
	glTexCoord2f( 0.0, 0.0 ); glVertex2f( fMark.x(), fMark.y());
	glTexCoord2f( 1.0, 0.0 ); glVertex2f( fMark.x() + fMark.width(), fMark.y() );
	glTexCoord2f( 1.0, 1.0 ); glVertex2f( fMark.x() + fMark.width(), fMark.y() + fMark.height() );
	glTexCoord2f( 0.0, 1.0 ); glVertex2f( fMark.x(), fMark.y() + fMark.height() );
	glEnd();
	glBindTexture( GL_TEXTURE_2D, 0 );
}

QRectF PixmapPlane::calculateFocusMark(const QRectF& imgRect) const
{
	double w = imgRect.width() * 0.1;
	double h = imgRect.height() * 0.1;
	double s = std::min(w, h);
	QRectF fMark(imgRect.x() + frameW * 2, imgRect.y() + frameW * 2, s, s);
	if(!fMark.isValid()) fMark = imgRect;
	return fMark;
}

void PixmapPlane::drawEpilines() const
{
	if(epilinePoints.empty() || epilinePoints.size() < 2) return;
	const QPointF& p1 = epilinePoints[0];
	const QPointF& p2 = epilinePoints[1];
	glColor3ub(0, 255, 0);
	drawLineGl(p1, p2, 3.0);
	if(epilinePoints.size() < 4) return;
	const QPointF& p3 = epilinePoints[2];
	const QPointF& p4 = epilinePoints[3];
	drawLineGl(p3, p4, 3.0);
}

void PixmapPlane::drawFrame( const QRectF& rect, const QColor& color, double width ) const
{
	double w = qMin( rect.width(), rect.height() );
	w /= 10; w = qMin( w, width );

	glColor4ub( color.red(), color.green(), color.blue(), color.alpha() );
	drawRect( rect.x(), rect.y(), rect.width(), w );
	drawRect( rect.right() - w, rect.y(), w, rect.height() );
	drawRect( rect.x(), rect.bottom() - w, rect.width(), w );
	drawRect( rect.x(), rect.y(), w, rect.height() );
}

void PixmapPlane::drawRect( double x, double y, double w, double h ) const
{
	glBegin( GL_QUADS );
	glVertex3d( x, y, 0.1 );
	glVertex3d( x, y + h, 0.1 );
	glVertex3d( x + w, y + h, 0.1 );
	glVertex3d( x + w, y, 0.1 );
	glEnd();
}

void PixmapPlane::setSelected( bool on )
{
	if ( selected == on ) return;
	selected = on;
}

void PixmapPlane::setFocused( bool on )
{
	if ( focused == on ) return;
	focused = on;
}

void PixmapPlane::setHighlighted( bool on )
{
	if ( highlighted == on ) return;
	highlighted = on;
}

void PixmapPlane::mousePressEvent( const QPointF& p, Qt::MouseButton button )
{
	if ( Qt::LeftButton == button ) {
		QRectF imgRect( position, size );
		QRectF fRect( calculateFocusMark(imgRect) );
		if(fRect.contains(p)) {
			emit signalFocused( view );
			if(fRect == imgRect)
				emit signalSelected( view, selected );
		}
		else emit signalSelected( view, selected );
	}
	else if ( Qt::RightButton ==  button )
		emit signalFocused( view );
}

void PixmapPlane::setEpiline1( const adore::core::AVector& line )
{
	double x1, x2, y1, y2;
	double w = width();
	double h = height();
	bool ok = adore::core::pointAsLine2( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w, h );
	if ( !ok ) return;
	if( epilinePoints.empty() )
		epilinePoints.resize(2);
	epilinePoints[0] = QPointF(x1, y1) + position;
	epilinePoints[1] = QPointF(x2, y2) + position;
}

void PixmapPlane::setEpiline2( const adore::core::AVector& line )
{
	double x1, x2, y1, y2;
	double w = width();
	double h = height();
	bool ok = adore::core::pointAsLine2( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w, h );
	if ( !ok ) return;
	assert(!epilinePoints.empty());
	epilinePoints.resize(4);
	epilinePoints[2] = QPointF(x1, y1) + position;
	epilinePoints[3] = QPointF(x2, y2) + position;
}

void PixmapPlane::removeEpilines()
{
	epilinePoints.clear();
}

void PixmapPlane::drawLineGl( const QPointF & p0, const QPointF & p1, double thickness ) const
{
	if(!QGLHelper::isMultisamplingAvailable()) {
		boost::shared_ptr<GLEnabler> enabler;
		glDisable(GL_DEPTH_TEST);
		enabler.reset(new GLEnabler(GL_BLEND));
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		QGLHelper::aaGlLineWidth(thickness);
		QGLHelper::aaGlBeginLines();
		QGLHelper::aaGlVertex2(p0.x(), p0.y());
		QGLHelper::aaGlVertex2(p1.x(), p1.y());
		QGLHelper::aaGlEndLines();
		return;
	}
	double thOver2 = thickness * 0.5;
	glBegin( GL_QUADS );
	glVertex3d( p0.x() - thOver2, p0.y() - thOver2, 0.0 );
	glVertex3d( p1.x() + thOver2, p1.y() - thOver2, 0.0 );
	glVertex3d( p1.x() + thOver2, p1.y() + thOver2, 0.0 );
	glVertex3d( p0.x() - thOver2, p0.y() + thOver2, 0.0 );
	glEnd();
}
