

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <cassert>
using namespace std;

#include <boost/shared_array.hpp>
using namespace boost;

#include <adrcore/types/aarray.h>
#include <adrcore/util/mathutil.h>
using namespace adore::core;

#include "qglhelper.h"

QGLHelper::QGLHelper() :
		m_current_context_holder(NULL),
		multisamplingAvailable(false),
		lineWidth(1.0),
		started(false)
{}

void QGLHelper::checkMultisampling()
{
	bool multisamplingOK = true;
	boost::shared_ptr<GLEnabler> enabler( new GLEnabler( GL_MULTISAMPLE ) );

	GLboolean hasMultiS;
	glGetBooleanv(GL_MULTISAMPLE, &hasMultiS);
	GLenum glErr = glGetError();
	if(GL_NO_ERROR != glErr || GL_FALSE == hasMultiS)
		multisamplingOK = false;

	if(multisamplingOK) {
		GLint numSamples;
		glGetIntegerv(GL_SAMPLES, &numSamples);
		glErr = glGetError();
		if(GL_NO_ERROR != glErr || 4 > numSamples)
			multisamplingOK = false;
	}

	QGLHelper& helper = Holder::instance();
	helper.multisamplingAvailable = multisamplingOK;
}

bool QGLHelper::isMultisamplingAvailable()
{
	return Holder::const_instance().multisamplingAvailable;
}

void QGLHelper::prepareAATextures(QGLWidget * contextholder)
{
	QGLHelper& helper = Holder::instance();
	if(helper.m_textures.find(contextholder) != helper.m_textures.end())
		return;

	shared_ptr<GLEnabler> enabler;
	if(GL_FALSE == glIsEnabled( GL_TEXTURE_2D ) )
		enabler.reset(new GLEnabler( GL_TEXTURE_2D ) );

	GLuint t0 = makeCircleTexture(0.6, 0.85);
	GLuint t1 = makeCircleTexture(0.1, 0.2);
	helper.m_textures.insert(std::make_pair(contextholder, std::make_pair(t0, t1)));
}

void QGLHelper::setCurrentContext(QGLWidget * contextholder)
{
	QGLHelper& helper = Holder::instance();
	helper.m_current_context_holder = contextholder;
}

GLuint QGLHelper::makeCircleTexture(double stop1, double stop2)
{
	const int radius = 32;
	const int diameter = 64;
	const double sqStop1 = (stop1 * radius) * (stop1 * radius);
	const double sqStop2 = (stop2 * radius) * (stop2 * radius);
	double sqDiff = sqStop2 - sqStop1;

	shared_array<unsigned char> data( new unsigned char[diameter * diameter] );
	double y = -0.5 * (diameter - 1);
	for(int i = 0; i < radius; ++i, y += 1.0)
	{
		double x = -0.5 * (diameter - 1);
		unsigned char * lineP = data.get() + i * diameter;
		for(int j = 0; j < radius; ++j, x += 1.0)
		{
			double sqDist = y * y + x * x;
			if(sqStop2 <= sqDist)
				*(lineP + j) = 0;
			else if(sqStop1 <= sqDist) {
				double f = sqDist - sqStop1;
				f /= sqDiff;
				double falloff = std::min(255.0, f * 255);
				*(lineP + j) = 255 - (unsigned char) falloff;
			}
			else
				*(lineP + j) = 255;
		}
	}

	for(int i = 0; i < radius; ++i) {
		unsigned char * lineP1 = data.get() + i * diameter;
		unsigned char * lineP2 = data.get() + (i + 1) * diameter - 1;
		for(int j = 0; j < radius; ++j) {
			*(lineP2 - j) = *(lineP1 + j);
		}
	}

	for(int i = 0; i < radius; ++i) {
		unsigned char * lineP1 = data.get() + (radius - i - 1) * diameter;
		unsigned char * lineP2 = data.get() + (i + radius) * diameter;
		for(int j = 0; j < diameter; ++j) {
			*(lineP2 + j) = *(lineP1 + j);
		}
	}

	GLuint texId = 0;
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(1, &texId);
	glBindTexture(GL_TEXTURE_2D, texId);

	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, diameter, diameter, GL_ALPHA, GL_UNSIGNED_BYTE, data.get());
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glError();

	glBindTexture( GL_TEXTURE_2D, 0 );

	return texId;
}

void QGLHelper::aaGlLineWidth(double width)
{
	QGLHelper& helper = Holder::instance();
	if(helper.started) return;
	helper.lineWidth = std::max(0.001, width);
}

void QGLHelper::aaGlBeginLines()
{
	QGLHelper& helper = Holder::instance();
	std::vector<double>().swap(helper.coordsCache);
	helper.coordsCache.reserve(1000);
	helper.started = true;
}

void QGLHelper::aaGlEndLines()
{
	QGLHelper& helper = Holder::instance();
	if(!helper.started) return;
	const std::vector<double>& cache = helper.coordsCache;
	for(size_t i = 3; i < cache.size(); i += 4) {
		aaGlLine(cache[i - 3], cache[i - 2], cache[i - 1], cache[i]);
	}
	std::vector<double>().swap(helper.coordsCache);
	helper.started = false;
}

void QGLHelper::aaGlVertex2(double x, double y)
{
	QGLHelper& helper = Holder::instance();
	if(!helper.started) return;
	helper.coordsCache.push_back(x);
	helper.coordsCache.push_back(y);
}

void QGLHelper::aaGlLine( double x1, double y1, double x2, double y2 )
{
	QGLHelper& helper = Holder::instance();
	double thickness = std::max(0.5, helper.lineWidth);
	double capF = 1.0;
	if(thickness <= 8) {
		glBindTexture( GL_TEXTURE_2D, helper.m_textures[helper.m_current_context_holder].second );
		glError();
		thickness *= 8.0;
		capF = 0.3333;
	}
	else {
		glError();
		glBindTexture( GL_TEXTURE_2D, helper.m_textures[helper.m_current_context_holder].first );
		glError();
		thickness *= 1.5;
	}

	double thOver2 = thickness * 0.5;
	AVector2D p0(2), p1(2), cap(2), ortho(2);
	p0 = x1, y1; p1 = x2, y2;
	cap = p1 - p0;
	normalize(cap, 1.0);
	cap *= thOver2;
	ortho = -cap(1), cap(0);

	p0 = p0 + cap * capF;
	p1 = p1 - cap * capF;
	glBegin( GL_TRIANGLE_STRIP );
		glTexCoord2d( 0.0, 0.0 ); glVertex2d( p0(0) - cap(0) + ortho(0), p0(1) - cap(1) + ortho(1) );
		glTexCoord2d( 0.0, 1.0 ); glVertex2d( p0(0) - cap(0) - ortho(0), p0(1) - cap(1) - ortho(1) );
		glTexCoord2d( 0.5, 0.0 ); glVertex2d( p0(0) + ortho(0), p0(1) + ortho(1) );
		glTexCoord2d( 0.5, 1.0 ); glVertex2d( p0(0) - ortho(0), p0(1) - ortho(1) );
		glTexCoord2d( 0.5, 0.0 ); glVertex2d( p1(0) + ortho(0), p1(1) + ortho(1) );
		glTexCoord2d( 0.5, 1.0 ); glVertex2d( p1(0) - ortho(0), p1(1) - ortho(1) );
		glTexCoord2d( 1.0, 0.0 ); glVertex2d( p1(0) + cap(0) + ortho(0), p1(1) + cap(1) + ortho(1) );
		glTexCoord2d( 1.0, 1.0 ); glVertex2d( p1(0) + cap(0) - ortho(0), p1(1) + cap(1) - ortho(1) );
	glEnd();
	glError();
}
