

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "scrollbargl.h"

#include "qglhelper.h"

#include <cmath>

ScrollBarGL::ScrollBarGL(Type t) :
		m_type(t),
		m_content_min(0),
		m_max(0),
		m_pos(0),
		m_factor(1),
		m_slider_length(0)
{
}

void ScrollBarGL::initialize(QGLWidget* )
{
	if(m_type == Horizontal)
		size.setHeight(Thickness);
	else
		size.setWidth(Thickness);
}

void ScrollBarGL::paint() const
{
	double extent = m_type == Horizontal ? width() : height();
	if(m_max == 0 || (extent - m_slider_length) < 0.5)
		return;

	glTranslated(position.x(), position.y(), 0.0);

	glEnable( GL_TEXTURE_2D );
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	double thickness(Thickness);
	double x1 = m_type == Horizontal ? 0 : thickness * 0.5;
	double y1 = m_type == Horizontal ? thickness * 0.5 : 0;
	double x2 = m_type == Horizontal ? width() : thickness * 0.5;
	double y2 = m_type == Horizontal ? thickness * 0.5 : height();

	glColor4ub(0, 127, 0, 128);
	QGLHelper::aaGlLineWidth(thickness);
	QGLHelper::aaGlBeginLines();
	QGLHelper::aaGlVertex2(x1, y1);
	QGLHelper::aaGlVertex2(x2, y2);
	QGLHelper::aaGlEndLines();

	x1 = m_type == Horizontal ? m_pos: x1;
	y1 = m_type == Horizontal ? y1 : m_pos;
	x2 = m_type == Horizontal ? x1 + m_slider_length : x1;
	y2 = m_type == Horizontal ? y1 : y1 + m_slider_length;

	glColor4ub(0, 255, 0, 200);
	QGLHelper::aaGlLineWidth(SliderWidth);
	QGLHelper::aaGlBeginLines();
	QGLHelper::aaGlVertex2(x1, y1);
	QGLHelper::aaGlVertex2(x2, y2);
	QGLHelper::aaGlEndLines();

	glTranslated(-position.x(), -position.y(), 0.0);
}

void ScrollBarGL::click(double x, double y)
{
	if(m_type == Horizontal) {
		m_pos = std::max(0.0, x - 0.5 * m_slider_length);
		if(m_pos > m_max) m_pos = m_max;
	} else {
		m_pos = std::max(0.0, y - 0.5 * m_slider_length);
		if(m_pos > m_max) m_pos = m_max;
	}
	emit positionChanged(getPos());
}

void ScrollBarGL::beginDragging(double x, double y)
{
	if(m_type == Horizontal) {
		if(x < m_pos || x > m_pos + m_slider_length)
			click(x, y);
	} else {
		if(y < m_pos || y > m_pos + m_slider_length)
			click(x, y);
	}
}

void ScrollBarGL::drag(double dx, double dy)
{
	if(m_type == Horizontal) {
		click(m_pos + dx + 0.5 * m_slider_length, 0);
	} else {
		click(0, m_pos + dy + 0.5 * m_slider_length);
	}
}

void ScrollBarGL::resize( const QRectF& viewport, const QRectF& content )
{
	if ( m_type == Horizontal )
	{
		position = QPointF( Border, viewport.height() - Thickness - Border );
		double pathLen = viewport.width() - Border * 2;
		size = QSizeF( pathLen, Thickness );

		if(content.isEmpty()) {
			m_factor = 1.0;
			m_slider_length = pathLen;
			m_pos = m_content_min = m_max = 0;
			return;
		}

		m_factor = content.width() / pathLen;
		double factor = content.width() / viewport.width();
		m_slider_length = pathLen / factor;
		m_slider_length = std::min(pathLen, m_slider_length);
		m_max = pathLen - m_slider_length;
		m_max = std::max(0.0, m_max);
		m_content_min = content.x();
		m_pos = std::max(m_pos, 0.0);
		m_pos = std::min(m_pos, m_max);
	}
	else
	{
		position = QPointF( viewport.width() - Thickness - Border, Border );
		double pathLen = viewport.height() - Thickness - Border * 3;
		size = QSizeF( Thickness, pathLen );

		if(content.isEmpty()) {
			m_factor = 1.0;
			m_slider_length = pathLen;
			m_pos = m_content_min = m_max = 0;
			return;
		}

		m_factor = content.height() / pathLen;
		double factor = content.height() / viewport.height();
		m_slider_length = pathLen / factor;
		m_slider_length = std::min(pathLen, m_slider_length);
		m_max = pathLen - m_slider_length;
		m_max = std::max(0.0, m_max);
		m_content_min = content.y();
		m_pos = std::max(m_pos, 0.0);
		m_pos = std::min(m_pos, m_max);
	}
}

void ScrollBarGL::move( double delta )
{
	double extent = m_type == Horizontal ? width() : height();
	if(m_slider_length == extent || m_max == 0.0) return;
	double old_pos(m_pos);
	m_pos += delta;
	m_pos = std::max(m_pos, 0.0);
	m_pos = std::min(m_pos, m_max);
	if(old_pos != m_pos)
		emit positionChanged(getPos());
}
