

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin
/// Author: Klaus Haeming


#include <iostream>
#include <string>

using namespace std;

/// this can test if the file at path an accepted image : isImage(sring)
#include <adrcore/persistence/imgloader.h>

/// this class represents images
#include <adrcore/imgprocessing/imageplane.h>

/// this class can have associated image and feature points
#include <adrcore/types/view.h>

/// this one manages views, correspondences, cameras, 3d points, etc.
#include <adrcore/types/nview.h>

/// implementation of the SURF algorithm
#include <adrcore/surf/surffinder.h>

/// finds harris features in a scale space environment
#include <adrcore/viewmatch/scaleharrisfinder.h>

/// the descriptor of SURF
#include <adrcore/surf/surfdescriptor.h>

/// persistence class, can handle NView and NViewMap classes
#include <adrcore/persistence/persist.h>

/// persistence backend, based on boost's serialization library
#include <adrcore/persistence/persist_boost.h>

/// a collection of good parameters, e.g. for feature finding
#include <adrcore/util/recommendations.h>

/// A Filelocator assists in finding files in the file system
#include <adrcore/util/filelocator.h>

/// functions for scale and shift transforations, needed for image coordinate normalization
#include <adrcore/util/scaleshifttrafo.h>

/// matches features by constructing a kd-tree
#include <adrcore/viewmatch/kdtreematcher.h>

/// a special coordinate transformation to ensure numerical stability
#include <adrcore/nview/normalizer.h>

/// An NViewOperator for RANSAC filtering
#include <adrcore/nview/featurepointfilter2.h>

/// RANSAC filtering of feature matches
#include <adrcore/mestimate/filterutil.h>

/** An NViewOperator for matching based on fundamental matrix transfer (point in one image to line in another one)
  * This one make usage of storing of feature points in tree structure (currently kd-tree).
  */
#include <adrcore/viewmatch/epilinematcher2.h>

using namespace adore::core;
using namespace persist;

/// define the types of Views and Images. These are simply abbreviations
typedef ImagePlane<uint8_t> ImageT;
typedef View<uint8_t> ViewT;

/// These parameters can be modified.
/** InitialPointsNum: initial number of feature points. A bigger number can give more
 * precise fundamental matrix, but it takes more time to match and can make epipolarmatching
 * redundant. A smaller number can lead to the situation without correct correspondences at all or
 * the fundamental matrix can be very imprecise.
 * This situation makes usage of epipolarmatching redundant too.
 */
enum { InitialPointsNum = 300 };

/** addMore: Add more featured points after the first matching. Prefferably the featurefinder must be fast and
  * generate many feature points.
  */
const bool addMore = true;

/**
  * @param fileName is a path to some image
  * @returns ViewT with image loaded and maximum InitialPointsNum feature points found.
  */
ViewT* makeView(const string& fileName)
{
	cout << "loading image : " << fileName << endl;
	/// load image
	ImageT * img = new ImageT ( fileName.c_str() );

	/// create a View, that holds the image. "true" means, that it owns the image and has to destroy it upon deletion
	ViewT * view = new ViewT ( img, true );

	/// let the view know, where the image has come from.
	/// this is used, when it comes to saving your results
	view->setImagePath ( fileName );

	/// initialize the transformations and the feature space
	view->initialize();

	/// since our reconstruction is based on found features, we need to find features
	SurfFinder finder ( 0 /** id */,
						4 /** number of octaves (each octave means downscaling minimum x2) */,
						4 /** number of intervals (downscaled images) in each octave */,
						( int )getGoodMinStrength<SurfFinder>() /** surf strength **/,
						false /** if the finder has to double image before processing */,
						InitialPointsNum /** max number of points **/ );

	SurfDescriptor descr(0 /** feature id */, 0 /** descriptor id */);

	/// tell the feature finder to use SurfDescriptor.
	/// it is also possible to use the some descriptor after the feature finding process.
	/// letting the feature finder invoke the descriptor creator ensures the descriptor
	/// size matches the feature scale
	finder.setDescriptorCreator(&descr, false);

	finder.findFeatures(view);

	return view;
}

/** for numerical stability, we need to transform the image space into the intervall [-1, 1][-1, 1]
 * since the images may have different sizes, we use the following W and H to find out what sizes to use for the transformation
 * the initialization is arbitrary, but large.
 */
void getMaxWidthAndHeight ( const NView& nview, int& W, int& H )
{
	W = INT_MAX >> 1;
	H = INT_MAX >> 1;
	for(NView::ViewIterator vit = nview.getViewsBegin(); vit != nview.getViewsEnd(); ++vit)
	{
		ViewT* view = static_cast<ViewT*> ( *vit );
		double width = view->getWidth();
		double height = view->getHeight();

		/// if this view is already normalized, we must take this fact in consideration.
		bool norm = view->getNormalized();
		if ( norm )
		{
			double scale = view->getInvTrafo() ( 0, 0 );
			width = ceil ( scale * width );
			height = ceil ( scale * height );
		}

		/// decide, which W and H to use for the normalizing trafo
		if ( ( ( int ) width + ( int ) height ) < ( W + H ) )
		{
			W = ( int ) width;
			H = ( int ) height;
		}
	}
}

/** Input: NView must contain 2 views with feature points.
  * NView and Views assumed to be not normalized
  */
void makeInitialGuess(NView& nview)
{
	cout << "making initial guess" << endl;

	/// create a matcher to match features, which is an NViewOperator
	/// the parameter tells the matcher to return those matches that are the best and
	/// where the ratio of the best to the second best is larger than a given threshold
	KDTreeMatcher matcher( KDTREE_SINGLE_BEST );

	/// some values that seems to be good for various images
	matcher.setMaxBBFNodes( 500 );
	matcher.setDistThresh ( 0.15f );
	matcher.setGoodRatio ( 0.8f );

	/// match feature points between first and second view
	matcher.modify(nview);

	/// this moves the coordinates of the features into the range [-1, 1][-1, 1]
	/// this is necessary because the linear algorithm for fundamental matrix computation
	/// and trifocal tensor computation are unstable otherwise
	Normalizer normer;
	normer.modify(nview);

	///
	/// 2-view feature filtering
	///
	/// filter those matches that do not comply with the epipolar geometry
	/// This is done by a feature point filter
	/// its parameters are filters for pairs and tripplets. These are set a few lines below
	FeaturePointFilter2 fpf ( NULL, NULL );

	/// compute a threshold that requires featues to be as close as 1.5 pixel to their epipolar line
	int W, H;
	getMaxWidthAndHeight(nview, W, H);
	float opt = getNormalizedPixel( 1.5, W, H );

	/// create a filter that tests matches if they obey the 2-view epipolar constraint imposed by a fundametal matrix
	FilterFMat ranser2 ( opt );

	/// once a good subset of matches is found, they may be used to improve the result further
	/// this is specified by the optMethod
	ranser2.setOptMethod ( INNER_RANSAC );

	/// set the parameters of the feature point filter that we omitted above
	fpf.setRanser2 ( &ranser2 );
	fpf.setRanser3 ( NULL );

	/// apply this filter as an NViewOperator
	fpf.modify(nview);
}

/**
 * it takes two paths to images in png or jpeg format as argument
 */
int main(int argc, char* argv[])
{
	if ( argc < 3 ) {
		cerr <<  "You need to specify two image paths." << endl;
		return -1;
	}

	string file1 = argv[1];
	string file2 = argv[2];

	/// are our two files accepted images?
	if ( !isImage ( file1 ) || !isImage ( file2 ) ) {
		cerr <<  "You need to specify two image paths (png or jpeg format)." << endl;
		return -1;
	}

	/// load images and find feature points
	ViewT* view1 = makeView(file1);
	ViewT* view2 = makeView(file2);

	/// create NView class and add two views in it.
	NView nview;
	nview.addView(view1);
	nview.addView(view2);

	/// find correspondences between first and second views;
	/// filter correspondences and find some fundamental matrix, that describes these
	/// correspondences acceptable.
	makeInitialGuess(nview);
	cout << "number of correspondences found after first matching and filtering: " << nview.getNumPointCorrs() << endl;

	string name = extractName( file1 ) + ".che";
	{
		/// save the result
		Persist<PersistBoost> database( name );
		database.saveNView(nview, "epilinematching_first_matching", false, Options(true));
	}

	/// after makeInitialGuess we have normalized NView. For feature finding and
	/// matching we need not normalized one
	nview.denormalize();

	/// add more feature points in views.
	if(addMore) {
		ScaleHarrisFinder finder ( 0, 1, 1, ( int )getGoodMinStrength<ScaleHarrisFinder>(), INT_MAX );
		SurfDescriptor descr(0, 0);
		finder.setDescriptorCreator(&descr, false);
		finder.findFeatures(view1);
		finder.findFeatures(view2);
	}

	/// retrieve the fundamental matrix wrapper from nview
	NViewInfoFM* fmatInfo = static_cast<NViewInfoFM*>(nview.getInfo());

	/// initialize epiline matcher with fundamental matrix.
	EpiLineMatcherKDTree epimatcher(fmatInfo->getFMatrix(),
		3 /** maximum distance from epipolar line in second image (that corresponds to point in first view) to point in second view*/,
		1 /** maximum difference of feature descriptors of the feature points assuming to correspond */);

	/// match feature points
	epimatcher.modify(nview);
	cout << "number of correspondences after epipolar matching: " << nview.getNumPointCorrs() << endl;

	/// save the result
	Persist<PersistBoost> database( name );
	database.saveNView(nview, "epilinematching", false, Options(true));

	return 0;
}
