

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <iostream>
#include <cstdlib>
#include <vector>
#include <cfloat>
#include <cmath>
/// A Filelocator assists in finding files in the file system
#include <adrcore/util/filelocator.h>

/// An Imagecache assists in managing different versions of an image, e.g. color and grayscale
#include <adrcore/types/imagecache.h>

/// The images need to be related to know which one are similar to which
/// the sequenceviewrelater simply assumes that the images next to each other
/// in a sequence are also similar to each other
#include <adrcore/viewmatch/sequenceviewrelater.h>

/// matches features by constructing a kd-tree
#include <adrcore/viewmatch/kdtreematcher.h>

/// finds harris features in a scale space environment
#include <adrcore/viewmatch/scaleharrisfinder.h>

/// the descriptor of SIFT
#include <adrcore/viewmatch/siftdescriptor.h>

/// splits sequences of images into pairs and triplets
#include <adrcore/viewmatch/splitter.h>

/// a collection of good parameters, e.g. for feature finding
#include <adrcore/util/recommendations.h>

/// RANSAC filtering of feature matches
#include <adrcore/mestimate/filterutil.h>

/// conversion of color to grey-scale images
#include <adrcore/imgprocessing/rgbtogray.h>

/// functions for scale and shift transforations, needed for image coordinate normalization
#include <adrcore/util/scaleshifttrafo.h>

/// a special coordinate transformation to ensure numerical stability
#include <adrcore/nview/normalizer.h>

/// applies operators to NViews
#include <adrcore/nview/modifier.h>

/// An NViewOperator for RANSAC filtering
#include <adrcore/nview/featurepointfilter2.h>

/// merges short reconstructed sequences into longer ones
#include <adrcore/nview/merger.h>

/// an NViewOperator that re-estimates the fundamental matrix or the trifocal tensor
#include <adrcore/nview/tensorupdater.h>

/// an NViewOperator that triangulates 3D points
#include <adrcore/nview/point3dcalculator.h>

/// an NViewOperator that upgrades projective reonstruction to metric
#include <adrcore/nview/autocalibrator2.h>

/// choose how to treat different camera parameters during calibration
#include <adrcore/parameterization/camparamchoice.h>

/// parameterization of a camera
#include <adrcore/parameterization/calibrationparamversatile.h>

/// soft constraints, that is a value and a "comfort zone"
#include <adrcore/parameterization/softconstraints.h>

/// an NViewOperator that applies bundle adjustment
#include <adrcore/nview/baprojective.h>

/// general loading and saving of reconstructed data
#include <adrcore/persistence/persist.h>
/// loading and saving using boost persistence
#include <adrcore/persistence/persist_boost.h>

using namespace std;
using namespace adore::core;

/**
 * a complete reconstruction from three images
 * it takes a directory with images as the argument
 */
int main ( int argc, char *argv[] )
{
	if ( argc < 2 )
	{
		fprintf ( stderr, "You need to specify a directory.\n" );
		return -1;
	}

	/// the path to the images
	string path = argv[1];
	
	/// the name of the directory is used to derive filenames
	string name = extractName( path );

	/// read the contets of the directory
	vector<string> entries; entries.reserve ( 50 );
	getDirEntries ( path.c_str(), entries );
	
	/// make sure that we have some files
	if ( !entries.size() ) exit ( 1 );
	
	/// sort the files, because digicams usually name the images using increasing numbers
	std::sort ( entries.begin(), entries.end() );

	/// define the types of Views and Images. These are simply abbreviations
	typedef ImagePlane<uint8_t> ImageT;
	/// a view is essentially an image with an extra space to store features
	typedef View<uint8_t> ViewT;

	/// An NViewMap manages the relations between Views and NViews
	/// It can be used to create NViews
	/// An NView is a part of a reconstruction containing View, Cameras and 3D-points
	NViewMap map;
	
	/// The image cache is used to convert conveniently to gray-scale images
	/// it can exchange the image a view refers to
	ImageCache<uint8_t> cache;
	
	/// And this is the converter that does the gray-scale conversion
	RGBtoGray<uint8_t> convgray;

	/// for numerical stability, we need to transform the image space into the intervall [-1, 1][-1, 1]
	/// since the images may have different sizes, we use the following W and H to find out what sizes to use for the transformation
	/// the initialization is arbitrary, but large
	int W = INT_MAX >> 1;
	int H = INT_MAX >> 1;

	/// we do this for three images
	const int maxImages = 3;

	/// to count the number of images we have loaded
	int count = 0;
	
	/// load the images by iterating over the directory entries
	for ( uint c = 0; ( c < entries.size() ) && ( count < maxImages ); ++c )
	{
		/// construct the image path
		string file = path + entries[c];
		
		/// test, if file is an image we can handle
		if ( !isImage ( file ) ) continue;

		/// print, that we have successfully found an image
		printf ( "loading image %s\n", file.c_str() );
		
		/// and load it
		ImagePlane<uint8_t> *img = new ImagePlane<uint8_t> ( file.c_str() );
		
		/// create a View, that holds the image. "true" means, that it owns the image and has to destroy it upon deletion
		ViewT *view = new View<uint8_t> ( img, true );
		
		/// give an id to the view
		view->setId ( count );
		
		/// let the view know, where the image has come from.
		/// this is used, when it comes to saving your results
		view->setImagePath ( file );
		
		/// initialize the transformations and the feature space
		view->initialize();
		
		/// print information about the constructed view
		printf ( "view %d: (%f, %f)\n", count, view->getWidth(), view->getHeight() );
		
		/// decide, which W and H to use for the normalizing trafo
		if ( ( ( int )view->getWidth() + ( int )view->getHeight() ) < ( W + H ) )
		{
			W = ( int )view->getWidth();
			H = ( int )view->getHeight();
		}

		/// let the NViewMap be aware of the view
		map.addView ( view );
		
		/// tell the image cache the original image belongs to the view and has index 0
		/// the index can later be used to switch the images a view points to
		cache.add( view, img, 0 );
		
		/// check, if we have color images.
		/// if so, convert them to gray-scale
		if ( img->getNumChans() == 3 )
		{
			ImagePlane<uint8_t> *gray = new ImagePlane<uint8_t> ( ( int )view->getWidth(), ( int )view->getHeight(), 1 );
			convgray.process( img, gray );
			
			/// add the gray-scale image to the image cache with index 1
			cache.add( view, gray, 1 );
		}
		else if ( img->getNumChans() == 1 )
		{
			/// add the original image as the gray-scale image to the image cache with index 1
			cache.add( view, img, 1 );
		}
		else
		{
			/// bail out
			printf("wrong number of channels\n");
			exit(1);
		}

		/// we found an image, count it!
		++count;
	}
	
	/// this is a global value (nearly the only one), which affects nonlinear minimization
	/// it essentially determins the rtransition of the least-squares to the linear error function
	/// the smaller it is, the more tolerant we are with outliers and the more inaccurate we get
	HUBER_VALUE = getNormalizedPixel( 12, W, H );
	
	/// inform us about the chosen values
	printf( "W is %d, H is %d, huber is %f\n", W, H, HUBER_VALUE );

	/// since our reconstruction is based on found features, we need to find features
	ScaleHarrisFinder featfind ( 0 /** id **/, 
				     5 /** number of images before subsampling **/, 
				     14 /** total number of images **/, 
				     ( int )getGoodMinStrength<ScaleHarrisFinder>() /** harris strength **/, 
				     INT_MAX /** max number of points **/ );
	/// use SIFT feature descriptor
	SiftDescriptor sifter;
	
	/// tell the feature finder to use SIFT.
	/// it is also possible to use the SIFT descriptor after the feature finding process.
	/// letting the feature finder invoke the descriptor creator ensures the descriptor
	/// size matches the feature scale
	featfind.setDescriptorCreator( &sifter );

	/// switch to gray-scale images
	cache.switchTo( 1 );

	/// find features in all views of the NViewMap
	featfind.findFeatures ( map );

	/// create pairs according to the view relater
	SequenceViewRelater relater( false /** no closed sequence **/ );
	relater.ViewRelater::createRelations( map );
	ViewKnitter knit;
	knit.createPairs( map );
	knit.createPlan( 2/*overlap*/ );

	//create 2-NViews
	Splitter splitter;
	splitter.split ( map, knit );

	/// A modifier applies an NViewOperator on all the NViews in an NViewMap
	/// the patter is:
	/// 1) instatiate an NViewOperator
	/// 2) call modifier.modifyAll(map, operator);
	Modifier modifier;
		
	/// create a matcher to match features, which is an NViewOperator
	/// the parameter tells the matcher to return those matches that are the best and
	/// where the ratio of the best to the second best is larger than a given threshold
	KDTreeMatcher matcher( KDTREE_RATIO_BEST );
	
	/// match the view pairs
	modifier.modifyAll( map, matcher );

	///
	/// normalize
	///
	/// this moves the coordinates of the features into the range [-1, 1][-1, 1]
	/// this is necessary because the linear algorithm for fundamental matrix computation
	/// and trifocal tensor computation are unstable otherwise
	Normalizer normer;
	modifier.modifyAll ( map, normer );

	/// 
	/// 2-view feature filtering
	///
	/// filter those matches that do not comply with the epipolar geometry
	/// This is done by a feature point filter
	/// its parameters are filters for pairs and tripplets. These are set a few lines below
	FeaturePointFilter2 fpf ( NULL, NULL );

	/// compute a threshold that requires featues to be as close as 1.5 pixel to their epipolar line
	float opt = getNormalizedPixel( 1.5, W, H );
	
	/// create a filter that tests matches if they obey the 2-view epipolar constraint imposed by a fundametal matrix
	FilterFMat ranser2 ( opt );
	/// once a good subset of matches is found, they may be used to improve the result further
	/// this is specified by the optMethod
	ranser2.setOptMethod ( INNER_RANSAC );
	
	/// set the parameters of the feature point filter that we omitted above
	fpf.setRanser2 ( &ranser2 );
	fpf.setRanser3 ( NULL );
	/// and apply the filter as an NViewOperator
	modifier.modifyAll ( map, fpf );
	
	/// 
	/// from 2-views to 3-views
	///
	/// create a Merger that will create Tripplets from the pairs following the plan 
	/// devised by the knitter
	Merger merger;
	/// looks like an NViewOperator, but it is not, because it creates new NViews
	merger.merge ( map, knit );

	/// 
	/// 3-view feature filtering
	///
	/// matches over three views must fulfill another epipolar constraint which is
	/// given by the trifocal tensor. This constraint is now imposed upon the matches
	/// get another threshold which is less strict
	opt = getNormalizedPixel( 12, W, H );
	/// create a filter that tests matches if they obey the 3-view epipolar constraint imposed by a trifocal tensor
	FilterTTensor ranser3 ( opt );
	/// no further improvements once a good match is found. These improvements are rather costly in the three-view case
	/// we will polish the results later
	ranser3.setOptMethod ( NONE );

	/// re-set the parameters of the feature point filter that we omitted above
	fpf.setRanser2 ( NULL );
	fpf.setRanser3 ( &ranser3 );
	/// and apply the filter as an NViewOperator
	modifier.modifyAll ( map, fpf );
	
	///
	/// polish the results
	///
	/// the tensorupdater re-estimates the fundametal matrix or trifocal tensor from the matches
	TensorUpdater tup;
	/// minimize an algebraic error
	tup.setMinAlgebraic( true );
	/// and apply as an NViewOperator
	modifier.modifyNumViews ( map, 3 /** num Views **/, tup );

	///
	/// structure computation
	///
	/// create an object that calculates cameras from multi-view tensors
	CameraCalculator calcCam;
	/// and apply as an NViewOperator
	modifier.modifyNumViews ( map, 3 /** num Views **/, calcCam );
	/// create an object that calculates 3D-points by triangulation using the cameras and 2D-points
	Point3DCalculator calc3D;
	/// let us minimize the reprojection error to improve accuracy (of the 3D-points only)
	calc3D.setMinimizeError ( true );
	/// and apply as an NViewOperator
	modifier.modifyNumViews ( map, 3 /** num Views **/, calc3D );

	///
	/// projective bundle adjustment
	///
	/// this minimizes the reprojection error by simultaneously modifying cameras and 3D-points
	/// this polishes the result further
	BAProjectiveSimple BA;
	modifier.modifyNumViews ( map, 3 /** num Views **/, BA );

	///
	/// Autocalibration
	///
	/// upgrade the projective reconstruction to metric
	/// create an Autocalibrator with a "versatile" parameterisation
	/// this parameterization allows for a fine-granular camera description
	Autocalibrator2<CalibrationParamVersatile> calibrator;
	/// get a reference to the Parameters
	CalibrationParamVersatile &calibParam = calibrator.getCalibrationParam();
	/// Softconstraints are used to tell the minimizer which values would be nice
	SoftConstraintSet &constSoftConstraints = calibParam.getCONSTSoftConstraints();
	/// a choice defines how to treat the different parameters
	CamParamChoice &metricChoice = calibParam.getChoice();

	/// a FREE_CONST parameter can assume any value but is constant for all cameras
	metricChoice.setFLyMode ( FREE_CONST );
	metricChoice.setARMode ( FREE_CONST );
	metricChoice.setPPxMode ( FREE_CONST );
	metricChoice.setPPyMode ( FREE_CONST );
	/// a HARD_CONST has a fixed value which is not altered by the minimization
	metricChoice.setSkewMode ( HARD_CONST );
	/// we need to tell the minimizer which value this actually is
	metricChoice.setSkew ( 0.0 );

	/// we now specify the soft constraints the parameters focal length,
	/// aspect ratio, and the x and y coordinate of the principal point
	/// the constraint has the form (what parameter, what value, comfort zone)
	/// the last parameter of addConstraint (which is 0), specifies the camera
	/// for which the constraint is meant (0 = all of them)
	constSoftConstraints.addConstraint ( SoftConstraint ( FL, 1.0, 3.0 ), 0 );
	constSoftConstraints.addConstraint ( SoftConstraint ( AR, 1.0, 0.1 ), 0 );
	constSoftConstraints.addConstraint ( SoftConstraint ( PPX, 0.0, 0.1 ), 0 );
	constSoftConstraints.addConstraint ( SoftConstraint ( PPY, 0.0, 0.1 ), 0 );

	/// apply as NViewOperator
	modifier.modifyAll ( map, calibrator );

	/// save the result
	persist::Persist<persist::PersistBoost> database( "threeimages.che" );
	database.saveNViewMap( map, name );

	return EXIT_SUCCESS;
}


